/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "local.h"

/*
 * Create and initialize the first area of all volumes.
 */
Dx_rc
Area_create(Dx *db)
{
  Physvolnum i;
  Dx_rc rc;

  for (i = 0; i < db->nvolumes; i++) {
	if ((rc = Area_create_volume(db, i, i, NULL)) != DX_OK)
	  return(rc);
  }
  return(DX_OK);
}

/*
 * Create and initialize the given area of the given volume.
 */
Dx_rc
Area_create_volume(Dx *db, Physvolnum volnum, Dx_ui32 areanum, Area **areap)
{
  Dx_ui32 rel_areanum;
  Area *ap, *area;
  Dx_rc rc;

  area = (Area *) Xmalloc(sizeof(Area));
  area->areanum = areanum;
  area->logmap = NULL;
  area->physmap.map = NULL;
  area->next = NULL;

  if ((ap = db->volume[volnum]->area) == NULL)
	db->volume[volnum]->area = area;
  else {
	while (ap->next != NULL)
	  ap = ap->next;
	ap->next = area;
  }

  rel_areanum = REL_AREANUM(db, areanum);
  rc = Logmap_create(db, db->volume[volnum], rel_areanum, &area->logmap);
  if (rc != DX_OK)
	return(rc);

  rc = Physmap_create(db, db->volume[volnum], rel_areanum, &area->physmap);
  if (areap != NULL && rc == DX_OK)
	*areap = area;

  return(rc);
}

/*
 * Map each area of a pre-existing db and initialize data structures.
 */
Dx_rc
Area_init(Dx *db)
{
  Physvolnum i;
  Dx_ui32 j, nareas;
  struct stat sbuf;
  Area *ap, *area;

  for (i = 0; i < db->nvolumes; i++) {
	if (fstat(db->volume[i]->fd, &sbuf) == -1) {
	  db->last_errno = errno;
	  perror("Area_init: fstat");
	  exit(1);
	}
	nareas = (((Dx_ui64) sbuf.st_size - META_HEADER_NBYTES)
			  + AREA_SIZE(db) - 1) / AREA_SIZE(db);
	for (j = 0; j < nareas; j++) {
	  area = (Area *) Xmalloc(sizeof(Area));
	  area->areanum = i + (j * db->nvolumes);
	  area->logmap = NULL;
	  area->physmap.map = NULL;
	  area->next = NULL;
	  if ((ap = db->volume[i]->area) == NULL)
		db->volume[i]->area = area;
	  else {
		while (ap->next != NULL)
		  ap = ap->next;
		ap->next = area;
	  }
	}
  }

  return(DX_OK);
}

/*
 * Write each area's meta data.
 */
Dx_rc
Area_write(Dx *db)
{
  Physvolnum i;
  Area *area;

  for (i = 0; i < db->nvolumes; i++) {
	area = db->volume[i]->area;
	while (area != NULL) {
#ifdef __FreeBSD__
	  if (fsync(db->volume[i]->fd) == -1) {
		db->last_errno = errno;
		perror("Area_write: fsync");
		return(DX_OS_ERROR);
	  }
#else
	  if (msync((char *) area->logmap, LOGMAP_MAP_NBYTES, MS_SYNC) == -1) {
		db->last_errno = errno;
		perror("Area_write: msync");
		return(DX_OS_ERROR);
	  }
	  if (msync((char *) area->physmap.map, PHYSMAP_MAP_NBYTES, MS_SYNC)
		  == -1) {
		db->last_errno = errno;
		perror("Area_write: msync");
		return(DX_OS_ERROR);
	  }
#endif
	  area = area->next;
	}
  }

  return(DX_OK);
}

void
Area_show(Dx *db, Dx_ui32 areanum)
{
  Dx_ui32 i, j, nmaps;
  Dx_ui32 *p, rel_areanum, used;
  Area *ap;
  Physoff *map;
  Volume *vol;

  fprintf(stderr, "Area: %lu\n", areanum);
  vol = AREANUM_TO_VOLUME(db, areanum);
  rel_areanum = REL_AREANUM(db, areanum);
  for (ap = vol->area; ap != NULL && ap->areanum != areanum; ap = ap->next)
	;
  if (ap != NULL) {
	map = (Physoff *) ap->logmap;
	if (map == NULL) {
	  fprintf(stderr, "logmap is NULL\n");
	  return;
	}
	nmaps = 0;
	for (i = 0; i < LOGMAP_PAGES_MAPPED(); i++) {
	  if (map[i] != 0) {
		fprintf(stderr, "(%5u %10llu)", (unsigned int) i, (Dx_ui64) map[i]);
		if ((++nmaps % 5) == 0)
		  fprintf(stderr, "\n");
		else
		  fprintf(stderr, " ");
	  }
	}
	if (nmaps % 5)
	  fprintf(stderr, "\n");

	fprintf(stderr, "nmaps=%u\n", (unsigned int) nmaps);
	used = (PHYSMAP_MAP_NBYTES * BITS_PER_BYTE) - ap->physmap.avail;
	fprintf(stderr, "pages: avail=%lu, used=%lu\n",
			ap->physmap.avail, used);
			
	if (used != 0 && ap->physmap.map != NULL) {
	  p = (Dx_ui32 *) ap->physmap.map;
	  for (i = 0; i < (PHYSMAP_MAP_NBYTES / sizeof(Dx_ui32)) / 8; i++) {
		for (j = 0; j < 8; j++) {
		  if (p[j] != 0)
			break;
		}
		if (j != 8) {
		  for (j = 7; j > 0; j--)
			fprintf(stderr, "%08lx ", p[j]);
		  fprintf(stderr, "%08lx\n", p[j]);
		}
		p += 8;
	  }
	}
	fprintf(stderr, "\n");
  }
}
