/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * These are the user-visible functions.
 */

#include "local.h"

static void
handle_segv(int signum)
{

  fprintf(stderr, "Dx: a SIGSEGV signal has occurred... exiting\n");
  abort();
}

Dx_rc
Dx_open(char **volume_paths, Dx_config *config, Dx **dbp)
{
  int flock_flags, last_errno, open_flags, persistent;
  Dx_ui32 i, is_new;
  char **paths;
  struct stat sbuf;
  struct timeval tv;
  Dx *db;
  Dx_config *conf, configuration;
  Dx_rc rc;
  static int first = 1;

  if (config == NULL) {
	conf = &configuration;
	Dx_get_configuration(NULL, conf);
  }
  else
	conf = config;

  db = *dbp = (Dx *) Xmalloc(sizeof(Dx));

  /* Might use sysconf() here, where supported. */
  db->system_page_size = getpagesize();
  db->dx_bits = DX_BITS;
  db->last_errno = 0;
  db->iter.b = 0;
  db->iter.ind = 0;
  db->keyspace = NULL;
  db->volume_weights.weights = db->volume_weights.cw = NULL;
  db->nlogical_volumes = 0;

  db->flags = conf->flags;
  db->open_mode = conf->open_mode;
  if (conf->hashfunc == NULL)
	db->hashfunc = hash_bj;
  else
	db->hashfunc = conf->hashfunc;

  is_new = 1;
  paths = volume_paths;

  if (paths == NULL || paths[0] == NULL || paths[0][0] == '\0'
	  || (conf->flags & DX_CONFIG_PERSISTENT) == 0) {
	/* It doesn't make sense to open a volatile db read-only. */
	if ((conf->flags & DX_CONFIG_READ_ONLY)
		|| conf->flags & DX_CONFIG_MUST_EXIST)
	  return(DX_BADCONFIGARG);

	/* Not specifying any paths can't be persistent, even if requested. */
	db->flags &= ~DX_CONFIG_PERSISTENT;
	persistent = 0;

	for (i = 0; paths != NULL && paths[i] != NULL && paths[i][0] != '\0'; i++)
	  ;
	if (i == 0)
	  db->nvolumes = 1;
	else
	  db->nvolumes = i;

	db->volume = (Volume **) Xmalloc(sizeof(Volume *) * db->nvolumes);
	for (i = 0; i < db->nvolumes; i++)
	  db->volume[i] = NULL;

	for (i = 0; i < db->nvolumes; i++) {
	  db->volume[i] = (Volume *) Xmalloc(sizeof(Volume));
	  db->volume[i]->volnum = i;
	  if (paths == NULL || paths[0] == NULL || paths[0][0] == '\0')
		db->volume[i]->path = Xstrdup("");
	  else
		db->volume[i]->path = Xstrdup(paths[i]);
	  db->volume[i]->meta = NULL;
	  db->volume[i]->mmap_size = 0;
	  db->volume[i]->area = NULL;

	  if ((db->volume[i]->fd = open("/dev/zero", O_RDWR)) == -1) {
		db->last_errno = errno;
		rc = DX_OPENFAILED;
		goto err;
	  }

	  if (fstat(db->volume[i]->fd, &sbuf) == -1) {
		db->last_errno = errno;
		rc = DX_FSTATFAILED;
		goto err;
	  }

	  db->volume[i]->st_blksize = sbuf.st_blksize;
	  db->volume[i]->mmap_size = 0;
	}
  }
  else {
	open_flags = O_CREAT
	  | ((conf->flags & DX_CONFIG_READ_ONLY) ? O_RDONLY : O_RDWR);
	if (conf->flags & DX_CONFIG_TRUNCATE_ON_OPEN)
	  open_flags |= O_TRUNC;
	if (conf->flags & DX_CONFIG_MUST_EXIST)
	  open_flags |= O_EXCL;
	persistent = 1;

	for (i = 0; paths[i] != NULL && paths[i][0] != '\0'; i++)
	  ;
	db->nvolumes = i;
	db->volume = (Volume **) Xmalloc(sizeof(Volume *) * db->nvolumes);
	for (i = 0; i < db->nvolumes; i++)
	  db->volume[i] = NULL;

	for (i = 0; i < db->nvolumes; i++) {
	  db->volume[i] = (Volume *) Xmalloc(sizeof(Volume));
	  db->volume[i]->volnum = i;
	  db->volume[i]->path = Xstrdup(paths[i]);
	  db->volume[i]->meta = NULL;
	  db->volume[i]->mmap_size = 0;
	  db->volume[i]->area = NULL;

	  /*
	   * If in read-only mode or when requested, the db files must already
	   * exist.
	   */
	  if (((db->flags & DX_CONFIG_READ_ONLY)
		   || (db->flags & DX_CONFIG_MUST_EXIST))
		  && access(paths[i], F_OK) == -1) {
		db->last_errno = errno;
		rc = DX_ACCESSFAILED;
		goto err;
	  }
	  if ((db->volume[i]->fd
		   = open(paths[i], open_flags, db->open_mode)) == -1) {
		db->last_errno = errno;
		rc = DX_OPENFAILED;
		goto err;
	  }

	  if (fstat(db->volume[i]->fd, &sbuf) == -1) {
		db->last_errno = errno;
		rc = DX_FSTATFAILED;
		goto err;
	  }

	  if (i == 0 && sbuf.st_size != 0)
		is_new = 0;
	  db->volume[i]->st_blksize = sbuf.st_blksize;
	  db->volume[i]->mmap_size = sbuf.st_size;

	  if (db->flags & DX_CONFIG_READ_ONLY)
		flock_flags = LOCK_SH | LOCK_NB;
	  else
		flock_flags = LOCK_EX | LOCK_NB;
	  if (flock(db->volume[i]->fd, flock_flags) == -1) {
		db->last_errno = errno;
		rc = DX_FLOCKFAILED;
		goto err;
	  }
	}
  }

  db->volume_weights.weights =
	(Dx_ui32 *) Xmalloc(db->nvolumes * sizeof(Dx_ui32));
  db->volume_weights.cw =
	(Dx_ui32 *) Xmalloc(db->nvolumes * sizeof(Dx_ui32));

  if (conf->volume_weights != NULL) {
	for (i = 0; i < db->nvolumes; i++)
	  db->volume_weights.weights[i] = conf->volume_weights[i];
  }
  else {
	for (i = 0; i < db->nvolumes; i++)
	  db->volume_weights.weights[i] = 1;
  }

  for (i = 0; i < db->nvolumes; i++) {
	if (i == 0)
	  db->volume_weights.cw[i] = db->volume_weights.weights[0];
	else
	  db->volume_weights.cw[i] = db->volume_weights.weights[i]
		+ db->volume_weights.cw[i - 1];
  }
  db->nlogical_volumes = db->volume_weights.cw[db->nvolumes - 1];

  db->last_segment_lookup
	= (SegmentMap **) Xmalloc(db->nvolumes * sizeof(SegmentMap *));
  db->active_list = NULL;

  if (db->flags & DX_CONFIG_PERSISTENT) {
	db->mmap_flags = MAP_SHARED;
	if (db->flags & DX_CONFIG_READ_ONLY)
	  db->mmap_prot = PROT_READ;
	else
	  db->mmap_prot = PROT_READ | PROT_WRITE;
  } else {
	db->mmap_flags = MAP_PRIVATE;
	db->mmap_prot = PROT_READ | PROT_WRITE;
  }

  if (is_new) {
	if (conf->page_size == 0)
	  db->page_size = DEFAULT_PAGE_SIZE;
	else
	  db->page_size = conf->page_size;

	if (conf->endian == DX_LITTLE_ENDIAN || conf->endian == DX_BIG_ENDIAN)
	  db->endian = conf->endian;
	else if (conf->endian == DX_NATIVE_ENDIAN) {
	  if (DX_HOST_BIG_ENDIAN)
		db->endian = DX_BIG_ENDIAN;
	  else
		db->endian = DX_LITTLE_ENDIAN;
	}
	else if (conf->endian == DX_PORTABLE_ENDIAN)
	  db->endian = DX_BIG_ENDIAN;
	else {
	  rc = DX_INVALID_ENDIAN;
	  goto err;
	}

	if (db->page_size < 4096 || db->page_size > 65536) {
	  rc = DX_INVALID_PAGESIZE;
	  goto err;
	}
	if ((db->page_size % db->system_page_size) != 0) {
	  rc = DX_INVALID_PAGESIZE;
	  goto err;
	}
	  
	gettimeofday(&tv, NULL);
	/* XXX Assumes that it's ok to convert a long to Dx_ui32 */
	db->creation_stamp.high = (Dx_ui32) tv.tv_sec;
	db->creation_stamp.low = tv.tv_usec;

	/* XXX I'm no longer so sure that using the time is necessary. */
	db->current_uid.tv_sec = (Dx_ui32) tv.tv_sec;
	db->current_uid.count = 0;

	rc = Meta_create(db);
  }
  else
	rc = Meta_init(db, sbuf.st_size);

  if (rc != DX_OK)
	goto err;

  if (conf->max_stripe_npages == 0)
	db->max_stripe_npages = DEFAULT_MAX_STRIPE_NPAGES;
  else
	db->max_stripe_npages = conf->max_stripe_npages;
  if (db->max_stripe_npages == 0) {
	rc = DX_INVALID_MAX_STRIPE_NPAGES;
	goto err;
  }

  /*
   * The maximum stripe size can't be too big to be handled by size_t,
   * which is the limit on many library and system calls, and it can't be
   * too big to be reasonably allocated from one area.
   * Making it really huge is probably a stupid idea anyway.
   */
  if ((db->max_stripe_npages * db->page_size) > MAXIMUM_MAX_STRIPE_NBYTES) {
	rc = DX_INVALID_MAX_STRIPE_NPAGES;
	goto err;
  }

  if (first) {
	signal(SIGSEGV, handle_segv);
	Xsrand((unsigned int) getpid());
	first = 0;
  }

  Stats_init(db);

  return(rc);

 err:

  /* Clean up after an error! */
  Page_release(db);
  for (i = 0; i < db->nvolumes; i++) {
	if (persistent) {
	  flock(db->volume[i]->fd, LOCK_UN);
	  if (is_new)
		unlink(db->volume[i]->path);
	}
	close(db->volume[i]->fd);		/* ignores error code */
	Xfree(db->volume[i]->path);
	Xfree(db->volume[i]);
  }

  if (db->volume_weights.weights != NULL)
	Xfree(db->volume_weights.weights);
  if (db->volume_weights.cw != NULL)
	Xfree(db->volume_weights.cw);

  Xfree(db->volume);
  *dbp = NULL;

  last_errno = db->last_errno;
  Xfree(db);

  errno = last_errno;
  return(rc);
}

static const Dx_config default_config = {
  DX_CONFIG_DEFAULT_FLAGS,
  DX_DEFAULT_OPEN_MODE,
  DEFAULT_PAGE_SIZE,
  DEFAULT_MAX_STRIPE_NPAGES,
  DX_NATIVE_ENDIAN,
  NULL,							/* Volume weights */
  NULL							/* Hash function */
};

void
Dx_get_configuration(Dx *db, Dx_config *config)
{

  if (db == NULL) {
	config->flags = default_config.flags;
	config->open_mode = default_config.open_mode;
	config->page_size = default_config.page_size;
	config->max_stripe_npages = default_config.max_stripe_npages;
	config->endian = default_config.endian;
	config->volume_weights = default_config.volume_weights;
	config->hashfunc = default_config.hashfunc;
  }
  else {
	config->flags = db->flags;
	config->open_mode = db->open_mode;
	config->page_size = db->page_size;
	config->max_stripe_npages = db->max_stripe_npages;
	config->endian = db->endian;
	config->volume_weights = db->volume_weights.weights;
	config->hashfunc = db->hashfunc;
  }
}

Dx_rc
Dx_sync(Dx *db)
{
  Dx_rc rc;

  rc = DX_OK;
  db->last_errno = 0;

  if (db->flags & DX_CONFIG_PERSISTENT) {
	/* XXX Only need to do this if the db has been modified. */
	rc = Meta_write(db);
	if (rc != DX_OK)
	  return(rc);

	rc = Page_sync(db);
  }
   
  return(rc);
}

/*
 * Create a persistent version of the given volatile db.
 * If any of the new volumes (VOLUME_PATHS) exist, they are truncated.
 * The file mode for the volumes is taken from OLD_DB.
 */
Dx_rc
Dx_create_persistent_copy(Dx *old_db, char **volume_paths)
{
  Dx_ui32 i, n;
  Dx *new_db;
  Dx_config config;
  Dx_rc rc;

  if (old_db->flags & DX_CONFIG_PERSISTENT)
	return(DX_BADARG);
  if (volume_paths == NULL)
	return(DX_BADARG);

  for (n = 0; volume_paths[n] != NULL && volume_paths[n][0] != '\0'; n++)
	;
  if (n != old_db->nvolumes)
	return(DX_BADARG);

  Dx_get_configuration(old_db, &config);
  config.flags |= (DX_CONFIG_PERSISTENT | DX_CONFIG_TRUNCATE_ON_OPEN);

  if ((rc = Dx_open(volume_paths, &config, &new_db)) != DX_OK)
	return(rc);

  if ((rc = Page_duplicate(new_db, old_db)) != DX_OK)
	return(rc);

  if ((rc = Meta_duplicate(new_db, old_db)) != DX_OK)
	return(rc);

  rc = Page_release(new_db);
  if (rc != DX_OK)
	return(rc);

  for (i = 0; i < new_db->nvolumes; i++) {
	if (close(new_db->volume[i]->fd) == -1) {
	  old_db->last_errno = errno;
	  /* XXX */
	  fprintf(stderr, "Dx_duplicate: close error\n");
	  rc = DX_OS_ERROR;
	}
	Xfree(new_db->volume[i]->path);
	Xfree(new_db->volume[i]);
  }

  Xfree(new_db->volume);
  /* Nuke it, just to be sure. */
  Dx_memset(new_db, 0, sizeof(Dx));

  Xfree(new_db->last_segment_lookup);

  Xfree(new_db);

  return(rc);
}

Dx_rc
Dx_close(Dx *db)
{
  Dx_ui32 i;
  Dx_rc rc;

  db->last_errno = 0;
  if (db == NULL)
	return(DX_OK);

  rc = DX_OK;
  if (db->flags & DX_CONFIG_PERSISTENT) {
	/* XXX Only need to do this if the db has been modified. */
	rc = Meta_write(db);
	if (rc != DX_OK)
	  return(rc);
  }

#ifdef NOTDEF
  /* This doesn't seem to be necessary. */
  rc = Page_sync(db);
  if (rc != DX_OK)
	return(rc);
#endif

  rc = Page_release(db);
  if (rc != DX_OK)
	return(rc);

  for (i = 0; i < db->nvolumes; i++) {
	/* close() also unlocks, if necessary. */
	if (close(db->volume[i]->fd) == -1) {
	  db->last_errno = errno;
	  /* XXX */
	  fprintf(stderr, "Dx_close: close error\n");
	  rc = DX_OS_ERROR;
	}
	Xfree(db->volume[i]->path);
	Xfree(db->volume[i]);
  }

  if (db->volume_weights.weights != NULL)
	Xfree(db->volume_weights.weights);
  if (db->volume_weights.cw != NULL)
	Xfree(db->volume_weights.cw);

  Xfree(db->volume);
  /* Nuke it, just to be sure. */
  Dx_memset(db, 0, sizeof(Dx));

  if (db->last_segment_lookup != NULL)
	Xfree(db->last_segment_lookup);

  Xfree(db);

  return(rc);
}

/*
 * Store an item, inserting or replacing according to MODE, from memory.
 */
Dx_rc
Dx_store(Dx *db, Dx_datum *key, Dx_opmode mode, Dx_datum *value)
{

  if ((mode & DX_OVERWRITE) || (mode & DX_REGION) || (mode & DX_UIO))
	return(DX_BADARG);

  if (DX_OPMODE(mode) != DX_INSERT && DX_OPMODE(mode) != DX_REPLACE
	  && DX_OPMODE(mode) != DX_INSERT_UNIQUE)
	return(DX_BADACCESSMODE);

  return(Dx_storex(db, key, mode, value));
}

/*
 * Extended store.
 * In addition to the simple Dx_store() semantics, this provides an interface
 * to:
 * a) delete an arbitrary chunk of the value, and/or
 * b) 1. insert new data at an arbitrary position or
 *    2. replace (overwrite) old data with new data at an arbitrary
 *       position, and/or
 * c) give a user-specified get-function that will be called to provide
 *    the new value data.
 *
 * The mode argument is the bitwise-or of one of the following operation flags:
 *   DX_STORE   - no deletion; a position arg is expected
 *   DX_DEFAULT - equivalent to DX_STORE
 *   DX_INSERT  - add this item without deleting a previous version
 *   DX_INSERT_UNIQUE - add this item only if it doesn't already exist
 *   DX_REPLACE - replace an existing item with this one
 *   DX_EDIT    - edit an existing item; a region arg is expected
 *
 * The fourth argument is either an edit spec or a position spec.
 * The fifth argument is either a DX_user_io or a Dx_datum.
 *
 * with zero or more of the following operation modifiers:
 *   DX_REGION    - a region to delete
 *   DX_UIO       - a get-function is used instead of a value Dx_datum
 *   DX_OVERWRITE - overwrite rather than insert
 *   DX_ADD_FIRST - if an identical key exists, add this one first
 *   DX_ADD_LAST  - if an identical key exists, add this one last
 *   DX_ADD_INDEX - if an identical key exists, use the 'ind' field
 *
 * The DX_EDIT flag specifies a region to be deleted and establishes an
 * insertion point for new data.  If no DX_EDIT flag is given, then the
 * parameter only establishes the insertion point (POS or END) and if the
 * parameter is NULL, the insertion point's value is taken to be zero.
 *
 * If given DX_EDIT, starting at offset POS within the value, if END is zero,
 * DLEN bytes will be deleted; if END is non-zero, DLEN is ignored and all
 * bytes (starting at POS) will be deleted.  If both EOF and LEN are zero,
 * the insertion point is established.
 * If not given DX_EDIT, if END is zero, POS establishes the insertion point;
 * if END is non-zero, the insertion point is set to the end of the Dx_datum.
 * In both cases, DLEN is not used.
 *
 * Following deletion, the value bytes, whether given explicitly or via the
 * get-function, are inserted or old bytes are overwritten starting at the
 * insertion point.
 * The insertion point may be beyond the current length of the value (zero
 * fill occurs, but only if bytes are actually inserted/overwritten) and the
 * value may end up bigger or smaller than its current length
 * when the operation is completed.
 */
Dx_rc
Dx_storex(Dx *db, Dx_datum *key, Dx_opmode mode, ...)
{
  Dx_ui32 f, m;
  Dx_datum value;
  Item item;
  Ehash_xarg arg;
  va_list ap;

  db->last_errno = 0;
  if (DX_OPMODE(mode) != DX_DEFAULT && DX_OPMODE(mode) != DX_EDIT
	  && DX_OPMODE(mode) != DX_STORE
	  && DX_OPMODE(mode) != DX_INSERT && DX_OPMODE(mode) != DX_REPLACE
	  && DX_OPMODE(mode) != DX_INSERT_UNIQUE)
	return(DX_BADARG);

  va_start(ap, mode);
  arg.region = NULL;
  arg.pos = NULL;
  if (DX_OPMODE(mode) == DX_EDIT) {
	arg.region = va_arg(ap, Dx_region *);
	arg.pos = NULL;
  }
  else if (DX_OPMODE(mode) == DX_STORE || DX_OPMODE(mode) == DX_DEFAULT) {
	arg.region = NULL;
	arg.pos = va_arg(ap, Dx_point *);
  }

  if ((mode & DX_UIO)) {
	arg.io.uio = va_arg(ap, Dx_user_io *);
	value.dptr = NULL;
	value.dlen = 0;
	value.ind = value.count = 0;
	value.flags = DX_DATUM_ALIGN0;
	value.keyspace = NULL;
	item.value = &value;
  }
  else {
	arg.io.uio = NULL;
	item.value = va_arg(ap, Dx_datum *);
  }

  arg.flags = (mode & (DX_OVERWRITE | DX_MAY_CREATE));
  f = (mode & (DX_ADD_FIRST | DX_ADD_LAST | DX_ADD_INDEX));
  if (f)
	arg.flags |= f;
  else
	arg.flags |= DX_ADD_FIRST;		/* XXX default behaviour */

  va_end(ap);

  if (key->dlen == 0 || key->dlen > DX_MAX_UI32)
	return(DX_BAD_KEY);
  item.key = key;
  item.keyhash = db->hashfunc((Dx_ui8 *) key->dptr, key->dlen, 0);

  if (DX_DBMCOMPAT) {
	key->flags = 0;
	key->ind = 0;
  }

  /* Decide if this is really a simple store. */
  if (arg.pos == NULL && arg.region == NULL) {
	m = (((mode & DX_ADD_FIRST) >> 8) + ((mode & DX_ADD_LAST) >> 9)
		 + ((mode & DX_ADD_INDEX) >> 10));
	if (m == 0)
	  arg.flags |= DX_ADD_FIRST;
	else if (m != 1)
	  return(DX_BADARG);
	else if (DX_OPMODE(mode) == DX_REPLACE)
	  return(DX_BADARG);
	return(Ehash_add(db, &item, mode, &arg));
  }

  if (DX_OPMODE(mode) == DX_REPLACE || DX_OPMODE(mode) == DX_INSERT
	  || DX_OPMODE(mode) == DX_INSERT_UNIQUE)
	return(DX_BADARG);

  return(Ehash_add(db, &item, DX_EDIT, &arg));
}

/*
 * Fetch the value for the given key.
 * The returned Dx_datum has its DX_DATUM_MALLOC flag set if it is an allocated
 * copy, otherwise it is a pointer to the read-only data.
 */
Dx_rc
Dx_fetch(Dx *db, Dx_datum *key, Dx_datum *value)
{

  return(Dx_fetchx(db, key, DX_FETCH, value));
}

/*
 * Fetch with extended functionality.
 * In addition to the simple Dx_fetch() semantics, this provides an interface
 * to:
 *  a) read an arbitrary chunk of the value, and/or
 *  b) give a user-specified put-function that will be called to accept
 *     the value/chunk instead of setting the Dx_datum pointer.
 * Note that if a value Dx_datum is given, value.dlen is the number of bytes
 * returned, not the actual size of the Dx_datum.
 * If a put-function is given, it is called as many times as necessary
 * and must return non-zero iff it wants to report a fatal error for the
 * operation.
 */
Dx_rc
Dx_fetchx(Dx *db, Dx_datum *key, Dx_opmode mode, ...)
{
  Item item;
  Ehash_xarg arg;
  Dx_region chunk;
  va_list ap;

  db->last_errno = 0;

  if (DX_OPMODE(mode) != DX_DEFAULT && DX_OPMODE(mode) != DX_FETCH
	  && DX_OPMODE(mode) != DX_PROBE && DX_OPMODE(mode) != DX_EXISTS)
	return(DX_BADARG);

  arg.flags = mode;
  arg.io.uio = NULL;
  item.value = NULL;
  arg.pos = NULL;

  va_start(ap, mode);
  if ((mode & DX_REGION)) {
	arg.region = va_arg(ap, Dx_region *);
	if (arg.region == NULL) {
	  chunk.mode = DX_REGION_ALL;
	  arg.region = &chunk;
	}
  }
  else {
	chunk.mode = DX_REGION_ALL;
	arg.region = &chunk;
  }

  if ((mode & DX_UIO)) {
	if (DX_OPMODE(mode) == DX_PROBE || DX_OPMODE(mode) == DX_EXISTS)
	  return(DX_BADARG);
	arg.io.uio = va_arg(ap, Dx_user_io *);
  }
  else if (DX_OPMODE(mode) != DX_EXISTS)
	item.value = va_arg(ap, Dx_datum *);
  va_end(ap);

  if (key->dlen == 0 || key->dlen > DX_MAX_UI32)
	return(DX_BAD_KEY);
  item.key = key;
  item.keyhash = db->hashfunc((Dx_ui8 *) key->dptr, key->dlen, 0);

  if (DX_DBMCOMPAT) {
	key->flags = 0;
	key->ind = 0;
	if (item.value != NULL) {
	  item.value->flags = 0;
	  item.value->ind = 0;
	}
  }

  if (DX_OPMODE(mode) == DX_DEFAULT)
	return(Ehash_get(db, &item, DX_FETCH, &arg));
  return(Ehash_get(db, &item, DX_OPMODE(mode), &arg));
}

/*
 * Remove the item from the database.
 */
Dx_rc
Dx_delete(Dx *db, Dx_datum *key)
{

  return(Dx_deletex(db, key, DX_DELETE));
}

/*
 * Delete with extended functionality.
 * Actually, this really exists only as a convenient alternative when
 * deleting a region from a value.
 */
Dx_rc
Dx_deletex(Dx *db, Dx_datum *key, Dx_opmode mode, ...)
{
  Dx_region *chunk;
  Dx_datum value;
  Item item;
  va_list ap;

  db->last_errno = 0;

  if (DX_OPMODE(mode) != DX_DELETE && DX_OPMODE(mode) != DX_DEFAULT)
	return(DX_BADARG);

  va_start(ap, mode);
  if ((mode & DX_REGION))
	chunk = va_arg(ap, Dx_region *);
  else
	chunk = NULL;
  va_end(ap);

  if (chunk == NULL) {
	if (key->dlen == 0 || key->dlen > DX_MAX_UI32)
	  return(DX_BAD_KEY);
	item.key = key;
	item.keyhash = db->hashfunc((Dx_ui8 *) key->dptr, key->dlen, 0);

	if (DX_DBMCOMPAT) {
	  key->flags = 0;
	  key->ind = 0;
	}

	return(Ehash_get(db, &item, DX_DELETE, NULL));
  }

  value.dlen = 0;
  value.dptr = NULL;
  value.keyspace = NULL;
  return(Dx_storex(db, key, DX_EDIT, chunk, &value));
}

Dx_rc
Dx_firstkey(Dx *db, Dx_datum *key)
{
  Dx_datum value;

  return(Dx_firstkeyx(db, key, DX_PROBE, &value));
}

/*
 * Initialize to begin iterating through all keys in the db and return the
 * first one.  If mode is:
 *   DX_EXISTS, simply check if there's a key.
 *   DX_PROBE, return the key, its meta data, and meta data for the value.
 *   DX_FETCH, also do a full fetch on the value.
 * To iterate through a particular key, use Dx_fetch() with the 'ind' field.
 */
Dx_rc
Dx_firstkeyx(Dx *db, Dx_datum *key, Dx_opmode mode, ...)
{
  Item item;
  Ehash_xarg arg;
  Dx_region chunk;
  va_list ap;

  db->last_errno = 0;

  if (DX_OPMODE(mode) != DX_FETCH && DX_OPMODE(mode) != DX_PROBE
	  && DX_OPMODE(mode) != DX_EXISTS)
	return(DX_BADACCESSMODE);

  arg.flags = mode;
  arg.io.uio = NULL;
  item.value = NULL;

  va_start(ap, mode);
  if ((mode & DX_REGION))
	arg.region = va_arg(ap, Dx_region *);
  else {
	chunk.mode = DX_REGION_ALL;
	arg.region = &chunk;
  }

  if ((mode & DX_UIO) && DX_OPMODE(mode) != DX_FETCH)
	return(DX_BADACCESSMODE);

  if ((mode & DX_UIO))
	arg.io.uio = va_arg(ap, Dx_user_io *);
  else
	item.value = va_arg(ap, Dx_datum *);
  va_end(ap);

  item.key = key;
  key->flags = 0;
  key->ind = 0;

  Ehash_iter_init(db);

  return(Ehash_iter(db, &item, mode, &arg));
}

Dx_rc
Dx_nextkey(Dx *db, Dx_datum *key)
{
  Dx_datum value;

  return(Dx_nextkeyx(db, key, DX_PROBE, &value));
}

/*
 * Like Dx_firstkeyx(), except continue to the next key in the sequence.
 * Assumes Dx_firstkeyx() has already been called to initialize state.
 */
Dx_rc
Dx_nextkeyx(Dx *db, Dx_datum *key, Dx_opmode mode, ...)
{
  Item item;
  Dx_region chunk;
  Ehash_xarg arg;
  va_list ap;

  db->last_errno = 0;

  if (DX_OPMODE(mode) != DX_FETCH && DX_OPMODE(mode) != DX_PROBE
	  && DX_OPMODE(mode) != DX_EXISTS)
	return(DX_BADACCESSMODE);

  arg.flags = mode;
  arg.io.uio = NULL;
  item.value = NULL;

  va_start(ap, mode);
  if ((mode & DX_REGION))
	arg.region = va_arg(ap, Dx_region *);
  else {
	chunk.mode = DX_REGION_ALL;
	arg.region = &chunk;
  }

  if ((mode & DX_UIO) && DX_OPMODE(mode) != DX_FETCH)
	return(DX_BADACCESSMODE);

  if ((mode & DX_UIO))
	arg.io.uio = va_arg(ap, Dx_user_io *);
  else
	item.value = va_arg(ap, Dx_datum *);
  va_end(ap);

  item.key = key;
  key->flags = 0;
  key->ind = 0;

  return(Ehash_iter(db, &item, mode, &arg));
}

/*
 * Initialize for iterating through identical keys.
 */
void
Dx_key_first(Dx_datum *keyp)
{

  keyp->ind = 0;
}

/*
 * Point to the next identical key for iteration.
 */
void
Dx_key_next(Dx_datum *keyp)
{

  keyp->ind++;
}

void
Dx_key(Dx_datum *keyp, void *keyval, Dx_len keylen)
{

	keyp->dptr = keyval;
	keyp->dlen = keylen;
	keyp->flags = DX_DATUM_ALIGN0;
	keyp->ind = 0;
	keyp->count = 0;
	keyp->keyspace = NULL;
}

void
Dx_value(Dx_datum *valuep, void *valueval, Dx_len valuelen)
{
  
	valuep->dptr = valueval;
	valuep->dlen = valuelen;
	valuep->flags = DX_DATUM_ALIGN0;
	valuep->ind = 0;
	valuep->count = 0;
	valuep->keyspace = NULL;
}

Dx_rc
Dx_reorganize(Dx *db)
{

  db->last_errno = 0;

  return(Page_reorg(db));
}

int
Dx_last_errno(Dx *db)
{

  return(db->last_errno);
}

void
Dx_stats_reset(Dx *db)
{

  Stats_init(db);
}

Dx_stats_counters *
Dx_stats_get(Dx *db)
{

  return(&db->stats);
}

void
Dx_show(Dx *db)
{
  unsigned int b;
  Page page;
  Dx_rc rc;

  for (b = 0; b < db->npages; b++) {
	if ((rc = Page_get(db, db->volume[0], b, PAGE_MUSTEXIST, &page)) == DX_OK
		&& Item_count(db, page, db->page_size)) {
	  fprintf(stderr, "\nBucket %u:\n", b);
	  Item_show(db, page, db->page_size);
	}
  }
}

void
Dx_check(Dx *db)
{
  unsigned int b;

  for (b = 0; b < db->npages; b++)
	page_check(db, b);
}

void
Dx_set_point(Dx_point *point, Dx_len start_val)
{

  point->mode = DX_POINT_ABSOLUTE;
  point->u.pos = start_val;
}

void
Dx_set_point_at_start(Dx_point *point)
{
  
	Dx_set_point(point, 0);
}

void
Dx_set_point_from_end(Dx_point *point, Dx_len end_val)
{

  point->mode = DX_POINT_RELATIVE;
  point->u.offset_from_end = end_val;
}

void
Dx_set_point_at_end(Dx_point *point)
{
  
    Dx_set_point_from_end(point, 0);
}

void
Dx_set_region_from_start(Dx_region *region)
{
  
  Dx_set_point_at_start(&region->start);
}

void
Dx_set_region_start(Dx_region *region, Dx_len start_val)
{
  
  Dx_set_point(&region->start, start_val);
}

void
Dx_set_region_end(Dx_region *region, Dx_len end_val)
{
  
  Dx_set_point(&region->u.end, end_val);
  region->mode = DX_REGION_PTOP;
}

void
Dx_set_region_to_end(Dx_region *region)
{
  
  Dx_set_point_at_end(&region->u.end);
  region->mode = DX_REGION_PTOP;
}

void
Dx_set_region_from_end(Dx_region *region, Dx_len end_val)
{
  
  region->mode = DX_REGION_PTOP;
  Dx_set_point_from_end(&region->u.end, end_val);
}

void
Dx_set_region_len(Dx_region *region, Dx_len dlen)
{
  
  region->mode = DX_REGION_LEN;
  region->u.dlen = dlen;
}

void
Dx_set_region_all(Dx_region *region)
{

  Dx_set_region_start(region, 0);
  Dx_set_region_to_end(region);
}
