/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _DX_ITEM_H_
#define _DX_ITEM_H_

/*
 * There's a tension here between making this very efficient, portable, and
 * easy to debug, understand, and change.  Changing the layout of the various
 * meta data or individual data types shouldn't be too difficult, but
 * unfortunately isn't as simple as just changing a typedef.
 *
 * XXX Shouldn't the offsets be configurable to be bigger than Dx_ui32?
 */

static INLINE Dx_ui16
item_get_ui16(Dx *db, Page page, Dx_ui32 offset)
{
  Dx_ui16 *ptr, val;

  ptr = (Dx_ui16 *) ((Dx_ui8 *) page + offset);
  val = Dx_import_ui16(db, ptr);

  return(val);
}

static INLINE void
item_put_ui16(Dx *db, Page page, Dx_ui32 offset, Dx_ui16 val)
{
  Dx_ui16 *ptr;

  ptr = (Dx_ui16 *) ((Dx_ui8 *) page + offset);
  *ptr = Dx_export_ui16(db, &val);
}

static INLINE Dx_ui32
item_get_ui32(Dx *db, Page page, Dx_ui32 offset)
{
  Dx_ui32 *ptr, val;

  ptr = (Dx_ui32 *) ((Dx_ui8 *) page + offset);
  val = Dx_import_ui32(db, ptr);

  return(val);
}

static INLINE void
item_put_ui32(Dx *db, Page page, Dx_ui32 offset, Dx_ui32 val)
{
  Dx_ui32 *ptr;

  ptr = (Dx_ui32 *) ((Dx_ui8 *) page + offset);
  *ptr = Dx_export_ui32(db, &val);
}

static INLINE Dx_ui64
item_get_ui64(Dx *db, Page page, Dx_ui32 offset)
{
  Dx_ui64 *ptr, val;

  ptr = (Dx_ui64 *) ((Dx_ui8 *) page + offset);
  val = Dx_import_ui64(db, ptr);

  return(val);
}

static INLINE void
item_put_ui64(Dx *db, Page page, Dx_ui32 offset, Dx_ui64 val)
{
  Dx_ui64 *ptr;

  ptr = (Dx_ui64 *) ((Dx_ui8 *) page + offset);
  *ptr = Dx_export_ui64(db, &val);
}

#if defined(Dx_hash32)
#define item_get_hash(DB, P, O)		item_get_ui32(DB, P, O)
#define item_put_hash(DB, P, O, V)	item_put_ui32(DB, P, O, V)
#elif defined(Dx_hash64)
#define item_get_hash(DB, P, O)		item_get_ui64(DB, P, O)
#define item_put_hash(DB, P, O, V)	item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Dx_hash"
#endif

#if defined(Physvolnum32)
#define item_get_physvolnum(DB, P, O)		item_get_ui32(DB, P, O)
#define item_put_physvolnum(DB, P, O, V)	item_put_ui32(DB, P, O, V)
#elif defined(Physvolnum64)
#define item_get_physvolnum(DB, P, O)		item_get_ui64(DB, P, O)
#define item_put_physvolnum(DB, P, O, V)	item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Physvolnum"
#endif

#if defined(Dx_len32)
#define item_get_dxlen(DB, P, O)		item_get_ui32(DB, P, O)
#define item_put_dxlen(DB, P, O, V)		item_put_ui32(DB, P, O, V)
#elif defined(Dx_len64)
#define item_get_dxlen(DB, P, O)		item_get_ui64(DB, P, O)
#define item_put_dxlen(DB, P, O, V)		item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Dx_len"
#endif

#if defined(Physoff32)
#define item_get_physoff(DB, P, O)		item_get_ui32(DB, P, O)
#define item_put_physoff(DB, P, O, V)	item_put_ui32(DB, P, O, V)
#elif defined(Physoff64)
#define item_get_physoff(DB, P, O)		item_get_ui64(DB, P, O)
#define item_put_physoff(DB, P, O, V)	item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Physoff"
#endif

#if defined(Seglen32)
#define item_get_seglen(DB, P, O)		item_get_ui32(DB, P, O)
#define item_put_seglen(DB, P, O, V)	item_put_ui32(DB, P, O, V)
#elif defined(Seglen64)
#define item_get_seglen(DB, P, O)		item_get_ui64(DB, P, O)
#define item_put_seglen(DB, P, O, V)	item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Seglen"
#endif

/* A page's meta data layout is as follows: */
typedef Dx_ui16 Nheaders_t;
typedef Dx_ui16 Metaflags_t;
typedef Dx_ui16 Endfree_t;
typedef Dx_ui16 Avail_t;
typedef Dx_ui16 Hashtab_bucket_t;

enum {
  Nheaders_s       = sizeof(Nheaders_t),
  Metaflags_s      = sizeof(Metaflags_t),
  Endfree_s        = sizeof(Endfree_t),
  Avail_s          = sizeof(Avail_t),
  Hashtab_bucket_s = sizeof(Hashtab_bucket_t)
};

static INLINE Dx_ui32
ITEM_MAX_NHEADERS(void)
{
  return((1 << (Nheaders_s * BITS_PER_BYTE)) - 1);
}

#define ITEM_REORG_FLAG_BITS	((Metaflags_t) (1 << 2))
#define ITEM_REORG_FLAG(FLAGS)	((Dx_ui32) (FLAGS & 04))

enum {
  META_NHEADERS_OFFSET  = 0,
  META_METAFLAGS_OFFSET = (META_NHEADERS_OFFSET + Nheaders_s),
  META_ENDFREE_OFFSET   = (META_METAFLAGS_OFFSET + Metaflags_s),
  META_AVAIL_OFFSET     = (META_ENDFREE_OFFSET + Endfree_s),
  META_OFFSET_END       = (META_AVAIL_OFFSET + Avail_s),
  META_HEADER_SIZE      = META_OFFSET_END,

  ITEM_HASHTAB_OFFSET    = META_HEADER_SIZE
};

#define ITEM_GET_NHEADERS(DX, PAGE) \
  (Nheaders_t) item_get_ui16(DX, PAGE, META_NHEADERS_OFFSET)
#define ITEM_PUT_NHEADERS(DX, PAGE, VAL) \
  item_put_ui16(DX, PAGE, META_NHEADERS_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_METAFLAGS(DX, PAGE) \
  (Metaflags_t) item_get_ui16(DX, PAGE, META_METAFLAGS_OFFSET)
#define ITEM_PUT_METAFLAGS(DX, PAGE, VAL) \
  item_put_ui16(DX, PAGE, META_METAFLAGS_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_ENDFREE(DX, PAGE) \
  (Endfree_t) item_get_ui16(DX, PAGE, META_ENDFREE_OFFSET)
#define ITEM_PUT_ENDFREE(DX, PAGE, VAL) \
  item_put_ui16(DX, PAGE, META_ENDFREE_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_AVAIL(DX, PAGE) \
  (Avail_t) item_get_ui16(DX, PAGE, META_AVAIL_OFFSET)
#define ITEM_PUT_AVAIL(DX, PAGE, VAL) \
  item_put_ui16(DX, PAGE, META_AVAIL_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_HASHTAB_VAL(DX, PAGE, IND) \
  (Hashtab_bucket_t) item_get_ui16(DX, PAGE, \
    ITEM_HASHTAB_OFFSET + (IND * Hashtab_bucket_s))

#define ITEM_PUT_HASHTAB_VAL(DX, PAGE, IND, VAL) \
  item_put_ui16(DX, PAGE, \
    ITEM_HASHTAB_OFFSET + (IND * Hashtab_bucket_s), (Dx_ui16) VAL)

/* An item's meta data layout */
typedef Dx_ui16 Keyp_t;
typedef Dx_ui16 Valuep_t;
typedef Dx_ui16 Keylen_t;
typedef Dx_ui16 Valuelen_t;
typedef Dx_ui16 Chainp_t;
typedef Dx_ui16 Item_flags_t;

enum {
  Keyp_s       = sizeof(Keyp_t),
  Valuep_s     = sizeof(Valuep_t),
  Keylen_s     = sizeof(Keylen_t),
  Valuelen_s   = sizeof(Valuelen_t),
  Chainp_s     = sizeof(Chainp_t),
  Item_flags_s = sizeof(Item_flags_t)
};

enum {
  ITEM_STARTP_OFFSET   = 0,
  ITEM_KEYP_OFFSET     = 0,
  ITEM_VALUEP_OFFSET   = (ITEM_KEYP_OFFSET + Keyp_s),
  ITEM_KEYLEN_OFFSET   = (ITEM_VALUEP_OFFSET + Valuep_s),
  ITEM_VALUELEN_OFFSET = (ITEM_KEYLEN_OFFSET + Keylen_s),
  ITEM_UIDHI_OFFSET    = (ITEM_VALUELEN_OFFSET + Valuelen_s),
  ITEM_UIDLO_OFFSET    = (ITEM_UIDHI_OFFSET + sizeof(Dx_ui32)),
  ITEM_HASH_OFFSET     = (ITEM_UIDLO_OFFSET + sizeof(Dx_ui32)),
  ITEM_CHAINP_OFFSET   = (ITEM_HASH_OFFSET + Dx_hash_s),
  ITEM_FLAGS_OFFSET    = (ITEM_CHAINP_OFFSET + Chainp_s),
  ITEM_HEADER_SIZE     = (ITEM_FLAGS_OFFSET + Item_flags_s)
};

#define ITEM_GET_KEYP(DX, HEADERP) \
  (Keyp_t) item_get_ui16(DX, HEADERP, ITEM_KEYP_OFFSET)
#define ITEM_PUT_KEYP(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_KEYP_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_VALUEP(DX, HEADERP) \
  (Valuep_t) item_get_ui16(DX, HEADERP, ITEM_VALUEP_OFFSET)
#define ITEM_PUT_VALUEP(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_VALUEP_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_KEYLEN(DX, HEADERP) \
  (Keylen_t) item_get_ui16(DX, HEADERP, ITEM_KEYLEN_OFFSET)
#define ITEM_PUT_KEYLEN(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_KEYLEN_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_VALUELEN(DX, HEADERP) \
  (Valuelen_t) item_get_ui16(DX, HEADERP, ITEM_VALUELEN_OFFSET)
#define ITEM_PUT_VALUELEN(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_VALUELEN_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_UIDHI(DX, HEADERP) \
  (Dx_ui32) item_get_ui32(DX, HEADERP, ITEM_UIDHI_OFFSET)
#define ITEM_PUT_UIDHI(DX, HEADERP, VAL) \
  item_put_ui32(DX, HEADERP, ITEM_UIDHI_OFFSET, (Dx_ui32) VAL)

#define ITEM_GET_UIDLO(DX, HEADERP) \
  (Dx_ui32) item_get_ui32(DX, HEADERP, ITEM_UIDLO_OFFSET)
#define ITEM_PUT_UIDLO(DX, HEADERP, VAL) \
  item_put_ui32(DX, HEADERP, ITEM_UIDLO_OFFSET, (Dx_ui32) VAL)

#define ITEM_GET_HASH(DX, HEADERP) \
  (Dx_hash) item_get_hash(DX, HEADERP, ITEM_HASH_OFFSET)
#define ITEM_PUT_HASH(DX, HEADERP, VAL) \
  item_put_hash(DX, HEADERP, ITEM_HASH_OFFSET, (Dx_hash) VAL)

#define ITEM_GET_CHAINP(DX, HEADERP) \
  (Chainp_t) item_get_ui16(DX, HEADERP, ITEM_CHAINP_OFFSET)
#define ITEM_PUT_CHAINP(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_CHAINP_OFFSET, (Dx_ui16) VAL)

#define ITEM_GET_FLAGS(DX, HEADERP) \
  (Item_flags_t) item_get_ui16(DX, HEADERP, ITEM_FLAGS_OFFSET)
#define ITEM_PUT_FLAGS(DX, HEADERP, VAL) \
  item_put_ui16(DX, HEADERP, ITEM_FLAGS_OFFSET, (Dx_ui16) VAL)


#define ITEM_KEY_INDIRECT(FLAGS)		(FLAGS & 001)
#define ITEM_KEY_INDIRECT_FLAG_BITS		(01 << 0)
#define ITEM_VALUE_INDIRECT(FLAGS)		(FLAGS & 002)
#define ITEM_VALUE_INDIRECT_FLAG_BITS	(01 << 1)

#define ITEM_KEY_ALIGNMENT_VALUE(FLAGS) ((FLAGS & 014) >> 2)
#define ITEM_KEY_ALIGNMENT(FLAGS)	\
		alignments[ITEM_KEY_ALIGNMENT_VALUE(FLAGS)]
#define ITEM_KEY_ALIGNMENT_FLAG_BITS(BITS)	(((BITS) & 02) << 2)
#define ITEM_VALUE_ALIGNMENT_VALUE(FLAGS)	((FLAGS & 060) >> 4)
#define ITEM_VALUE_ALIGNMENT(FLAGS)		\
		alignments[ITEM_VALUE_ALIGNMENT_VALUE(FLAGS)]
#define ITEM_VALUE_ALIGNMENT_FLAG_BITS(BITS)	(((BITS) & 02) << 4)

#define ITEM_DELETED(FLAGS)				(FLAGS & 0100)
#define ITEM_DELETED_BITS				(1 << 6)

/* The maximum number of bytes allocated for a stripe. */
static INLINE Dx_len
ITEM_INDIRECT_STRIPE_NBYTES(Dx *db)
{
  Dx_len val;

  val = (Dx_len) (NPAGES_TO_NBYTES(db, db->max_stripe_npages));
  return(val);
}

/* The number of maximum length stripes need for NPAGES. */
static INLINE Dx_ui32
ITEM_INDIRECT_NPAGES_TO_NSTRIPES(Dx *db, Dx_ui32 npages)
{
  Dx_ui32 val;

  val = (npages + (db->max_stripe_npages - 1)) / db->max_stripe_npages;
  return(val);
}

/* The number of stripes needed for a datum of DLEN bytes.  */
static INLINE Dx_ui32
ITEM_INDIRECT_NBYTES_TO_NSTRIPES(Dx *db, Dx_len dlen)
{
  Dx_ui32 val;

  val = ITEM_INDIRECT_NPAGES_TO_NSTRIPES(db, NBYTES_TO_NPAGES(db, dlen));
  return(val);
}

#ifdef NOTDEF
typedef struct {
  Dx_hash hash;
  Dx_uid uid;
} Ilocator;
#endif

#ifdef NOTDEF
#define IINDEX_OFFSET0			0
#define IINDEX_STARTP(ID)		((u_int8 *)       ((ID) + IINDEX_OFFSET0))
#define IINDEX_HASH(ID)			(*((Dx_hash *)     ((ID) + IINDEX_OFFSET0)))
#define IINDEX_OFFSET1			(IINDEX_OFFSET0 + Dx_hash_s)
#define IINDEX_UID_HI(ID)		(*((long *)       ((ID) + IINDEX_OFFSET1)))
#define IINDEX_OFFSET2			(IINDEX_OFFSET1 + sizeof(long))
#define IINDEX_UID_LO(ID)		(*((Dx_ui32 *)     ((ID) + IINDEX_OFFSET2)))
#define IINDEX_OFFSET3			(IINDEX_OFFSET2 + sizeof(Dx_ui32))
#define IINDEX_ENDP(ID)			((u_int8    *)    ((ID) + IINDEX_OFFSET3))

enum { IINDEX_HEADER_SIZE = IINDEX_OFFSET3 };

typedef struct {
  Physvolnum vol;
  Dx_len dlen;
  Physoff offset;
  Seglen seglen;
} Idesc;
#endif

enum {
  IDESC_VOLUME_OFFSET =	0,
  IDESC_DLEN_OFFSET   = (IDESC_VOLUME_OFFSET + Physvolnum_s),
  IDESC_OFFSET_OFFSET = (IDESC_DLEN_OFFSET + Dx_len_s),
  IDESC_SEGLEN_OFFSET = (IDESC_OFFSET_OFFSET + Physoff_s),
  IDESC_SIZE          = (IDESC_SEGLEN_OFFSET + Seglen_s)
};

#define IDESC_GET_VOLUME(DB, ID) \
  (Physvolnum) item_get_physvolnum(DB, ID, IDESC_VOLUME_OFFSET)
#define IDESC_PUT_VOLUME(DB, ID, VAL) \
  item_put_physvolnum(DB, ID, IDESC_VOLUME_OFFSET, (Physvolnum) VAL)

#define IDESC_GET_DLEN(DB, ID) \
  (Dx_len) item_get_dxlen(DB, ID, IDESC_DLEN_OFFSET)
#define IDESC_PUT_DLEN(DB, ID, VAL) \
  item_put_dxlen(DB, ID, IDESC_DLEN_OFFSET, (Dx_len) VAL)

#define IDESC_GET_OFFSET(DB, ID) \
  (Physoff) item_get_physoff(DB, ID, IDESC_OFFSET_OFFSET)
#define IDESC_PUT_OFFSET(DB, ID, VAL) \
  item_put_physoff(DB, ID, IDESC_OFFSET_OFFSET, (Physoff) VAL)

#define IDESC_GET_SEGLEN(DB, ID) \
  (Seglen) item_get_seglen(DB, ID, IDESC_SEGLEN_OFFSET)
#define IDESC_PUT_SEGLEN(DB, ID, VAL) \
  item_put_seglen(DB, ID, IDESC_SEGLEN_OFFSET, (Seglen) VAL)


typedef Dx_ui32 Istripe_count_t;
#define Istripe_count32

enum {
  Istripe_count_s = sizeof(Istripe_count_t)
};

#ifdef Istripe_count32
#define item_get_istripe_count(DB, P, O)	item_get_ui32(DB, P, O)
#define item_put_istripe_count(DB, P, O, V)	item_put_ui32(DB, P, O, V)
#elif Istripe_count64
#define item_get_istripe_count(DB, P, O)	item_get_ui64(DB, P, O)
#define item_put_istripe_count(DB, P, O, V)	item_put_ui64(DB, P, O, V)
#else
#error "Unsupported length for Istripe_count_t"
#endif

typedef struct {
  Istripe_count_t count;	/* Number of descriptors in this segment */
  Physvolnum vol;			/* Next stripe descriptor segment location */
  Physoff offset;
  Seglen seglen;
  Segment seg;				/* This segment in memory; not stored on disk. */
} Istripe_hdr;

enum {
  ISTRIPE_HDR_COUNT_OFFSET  = 0,
  ISTRIPE_HDR_VOLUME_OFFSET = (ISTRIPE_HDR_COUNT_OFFSET + Istripe_count_s),
  ISTRIPE_HDR_OFFSET_OFFSET = (ISTRIPE_HDR_VOLUME_OFFSET + Physvolnum_s),
  ISTRIPE_HDR_SEGLEN_OFFSET = (ISTRIPE_HDR_OFFSET_OFFSET + Physoff_s),
  ISTRIPE_HDR_SIZE          = (ISTRIPE_HDR_SEGLEN_OFFSET + Seglen_s)
};

#define ISTRIPE_HDR_GET_COUNT(DB, ID) \
  (Istripe_count_t) item_get_istripe_count(DB, ID, ISTRIPE_HDR_COUNT_OFFSET)
#define ISTRIPE_HDR_PUT_COUNT(DB, ID, VAL) \
  item_put_istripe_count(DB, ID, ISTRIPE_HDR_COUNT_OFFSET, \
    (Istripe_count_t) VAL)

#define ISTRIPE_HDR_GET_VOLUME(DB, ID) \
  (Physvolnum) item_get_physvolnum(DB, ID, ISTRIPE_HDR_VOLUME_OFFSET)
#define ISTRIPE_HDR_PUT_VOLUME(DB, ID, VAL) \
  item_put_physvolnum(DB, ID, ISTRIPE_HDR_VOLUME_OFFSET, (Physvolnum) VAL)

#define ISTRIPE_HDR_GET_OFFSET(DB, ID) \
  (Physoff) item_get_physoff(DB, ID, ISTRIPE_HDR_OFFSET_OFFSET)
#define ISTRIPE_HDR_PUT_OFFSET(DB, ID, VAL) \
  item_put_physoff(DB, ID, ISTRIPE_HDR_OFFSET_OFFSET, (Physoff) VAL)

#define ISTRIPE_HDR_GET_SEGLEN(DB, ID) \
  (Seglen) item_get_seglen(DB, ID, ISTRIPE_HDR_SEGLEN_OFFSET)
#define ISTRIPE_HDR_PUT_SEGLEN(DB, ID, VAL) \
  item_put_seglen(DB, ID, ISTRIPE_HDR_SEGLEN_OFFSET, (Seglen) VAL)


enum {
  ISTRIPE_VOLUME_OFFSET = 0,
  ISTRIPE_OFFSET_OFFSET = (ISTRIPE_VOLUME_OFFSET + Physvolnum_s),
  ISTRIPE_SEGLEN_OFFSET = (ISTRIPE_OFFSET_OFFSET + Physoff_s),
  ISTRIPE_DLEN_OFFSET   = (ISTRIPE_SEGLEN_OFFSET + Seglen_s),
  ISTRIPE_SIZE          = (ISTRIPE_DLEN_OFFSET + Dx_len_s)
};

#define ISTRIPE_GET_VOLUME(DB, ID) \
  (Physvolnum) item_get_physvolnum(DB, ID, ISTRIPE_VOLUME_OFFSET)
#define ISTRIPE_PUT_VOLUME(DB, ID, VAL) \
  item_put_physvolnum(DB, ID, ISTRIPE_VOLUME_OFFSET, (Physvolnum) VAL)

#define ISTRIPE_GET_OFFSET(DB, ID) \
  (Physoff) item_get_physoff(DB, ID, ISTRIPE_OFFSET_OFFSET)
#define ISTRIPE_PUT_OFFSET(DB, ID, VAL) \
  item_put_physoff(DB, ID, ISTRIPE_OFFSET_OFFSET, (Physoff) VAL)

#define ISTRIPE_GET_SEGLEN(DB, ID) \
  (Seglen) item_get_seglen(DB, ID, ISTRIPE_SEGLEN_OFFSET)
#define ISTRIPE_PUT_SEGLEN(DB, ID, VAL) \
  item_put_seglen(DB, ID, ISTRIPE_SEGLEN_OFFSET, (Seglen) VAL)

#define ISTRIPE_GET_DLEN(DB, ID) \
  (Dx_len) item_get_dxlen(DB, ID, ISTRIPE_DLEN_OFFSET)
#define ISTRIPE_PUT_DLEN(DB, ID, VAL) \
  item_put_dxlen(DB, ID, ISTRIPE_DLEN_OFFSET, (Dx_len) VAL)

static INLINE Dx_ui8 *
ISTRIPE_PTR(void *id, Dx_ui32 ind)
{
  Dx_ui8 *val;

  val = (Dx_ui8 *) (((Dx_ui8 *) id) + ISTRIPE_HDR_SIZE + (ind * ISTRIPE_SIZE));
  return(val);
}
	
static INLINE Istripe_count_t
ISTRIPE_ISTRIPES_ALLOC(Dx_len seglen)
{
  Istripe_count_t val;

  val = (Istripe_count_t) (seglen - ISTRIPE_HDR_SIZE) / ISTRIPE_SIZE;
  return(val);
}

/*
 * The "indirect descriptor" points to the start of the meta data for a
 * datum.
 */
static INLINE void
pack_idescriptor_fast(Dx *db, Dx_ui8 *id, Dx_len *dlen, Physoff *offset,
					  Seglen *seglen, Physvolnum *volume)
{

  if (dlen != NULL)
	IDESC_PUT_DLEN(db, id, *dlen);
  if (offset != NULL)
	IDESC_PUT_OFFSET(db, id, *offset);
  if (seglen != NULL)
	IDESC_PUT_SEGLEN(db, id, *seglen);
  if (volume != NULL)
	IDESC_PUT_VOLUME(db, id, *volume);
}

static INLINE void
unpack_idescriptor_fast(Dx *db, Dx_ui8 *id, Dx_len *dlen, Physoff *offset,
						Seglen *seglen, Physvolnum *volume)
{

  if (dlen != NULL)
	*dlen = IDESC_GET_DLEN(db, id);
  if (offset != NULL)
	*offset = IDESC_GET_OFFSET(db, id);
  if (seglen != NULL)
	*seglen = IDESC_GET_SEGLEN(db, id);
  if (volume != NULL)
	*volume = IDESC_GET_VOLUME(db, id);
}

static INLINE void
pack_istripe(Dx *db, Dx_ui8 *id, Stripe *stripe)
{

  ISTRIPE_PUT_VOLUME(db, id, stripe->volnum);
  ISTRIPE_PUT_OFFSET(db, id, stripe->offset);
  ISTRIPE_PUT_SEGLEN(db, id, stripe->seglen);
  ISTRIPE_PUT_DLEN(db, id, stripe->dlen);
}

static INLINE void
unpack_istripe(Dx *db, Dx_ui8 *id, Stripe *stripe)
{

  stripe->volnum = ISTRIPE_GET_VOLUME(db, id);
  stripe->offset = ISTRIPE_GET_OFFSET(db, id);
  stripe->seglen = ISTRIPE_GET_SEGLEN(db, id);
  stripe->dlen = ISTRIPE_GET_DLEN(db, id);
  stripe->segment = NULL;
}

extern void pack_istripe_hdr(Dx *db, Dx_ui8 *id, char *fmt, ...);
extern void unpack_istripe_hdr(Dx *db, Dx_ui8 *id, char *fmt, ...);

extern Dx_rc Stripe_allocate(Dx *db, Dx_len nbytes, Istripe_count_t nstripes,
							 Stripe *stripes);
extern Dx_rc Stripe_contig(Dx *db, Physvolnum initial_volume,
						   Physoff offset, Seglen seglen, Dx_len dlen,
						   Dx_region *chunk, Ehash_io *io);
extern Dx_rc Stripe_delete_data(Dx *db, Physvolnum vol, Physoff offset,
								Seglen seglen, Dx_len dlen, Dx_len start_pos,
								Dx_len delete_length, Dx_desc *desc,
								Ehash_io *io);
extern Dx_rc Stripe_free(Dx *db, Physvolnum vol, Physoff offset,
						 Seglen seglen);
extern Dx_rc Stripe_free_set(Dx *db, Stripe *stripes, Istripe_count_t count);
extern Dx_rc Stripe_insert_data(Dx *db, Physvolnum vol, Dx_ui8 *id,
								Physoff offset,
								Seglen seglen, Dx_datum *datum, Dx_len dlen,
								Dx_len start_pos, Ehash_xarg *arg);
extern Dx_rc Item_stripe_copy(Dx *db, Volume *v, Dx_datum *datum,
							  Dx_len zero_fill,
							  Segment *istripe_seg, Physoff *istripe_offset,
							  Seglen *istripe_seglen, Ehash_xarg *arg);
extern Dx_rc item_verify_region(Dx_len datum_dlen, Dx_region *chunk,
								Dx_len *start_pos, Dx_len *end_pos,
								Dx_len *lenp);
extern Dx_rc item_verify_position(Dx_len dlen, Dx_point *r, Dx_len *pos);
#endif
