/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Keyspace management
 *
 * A keyspace is a way for the user to segregate keys that are related in
 * some way into their own namespace.  The user gives an arbitrary character
 * string name to a keyspace and this simple management layer maintains
 * tem.
 * The keyspace name's hash value is used to modify a key's normal hash value.
 *
 * The keyspace directory is a vector of the names for the benefit of
 * the user.
 */

#include "local.h"

#define KSNAME			"*keyspaces*"
#define DIR_KEY			"keyspace_dir"

static Dx_keyspace ksname;

static INLINE void
keyspace_init(Dx *db)
{
  static int done_init = 0;

  if (!done_init) {
	ksname.name = KSNAME;
	ksname.hash = db->hashfunc((Dx_ui8 *) KSNAME, strlen(KSNAME), 0);
	done_init = 1;
  }
}

/*
 * The directory is a vector of keyspace names.
 */
static INLINE Dx_rc
keyspace_dir_add(Dx *db, char *ksn)
{
  Dx_datum key, value;
  Dx_point pos;
  Dx_rc rc;

  keyspace_init(db);
  Dx_key(&key, DIR_KEY, strlen(DIR_KEY) + 1);
  key.keyspace = &ksname;
  Dx_value(&value, ksn, strlen(ksn) + 1);

  Dx_set_point_at_end(&pos);
  rc = Dx_storex(db, &key, DX_STORE | DX_MAY_CREATE, &pos, &value);

  return(rc);
}

/*
 * Check if KSN is a keyspace name that already exists.
 * Also check if KSN has the same hash value as another keyspace name.
 */
static INLINE Dx_rc
keyspace_name_find(Dx *db, char *ksn)
{
  int len, nlen, olen;
  char *p, *q;
  Dx_datum key, value;
  Dx_hash nhval, ohval;
  Dx_rc rc;

  keyspace_init(db);
  len = strlen(DIR_KEY);
  Dx_key(&key, DIR_KEY, len + 1);
  key.keyspace = &ksname;

  rc = Dx_fetchx(db, &key, DX_FETCH | DX_ALWAYS_COPY, &value);
  if (rc == DX_KEY_NOT_FOUND)
	return(DX_KEYSPACE_NOTFOUND);
  if (rc != DX_OK)
	return(rc);

  /* The hash of a keyspace name doesn't include the null. */
  nlen = strlen(ksn);
  nhval = db->hashfunc((Dx_ui8 *) ksn, (Dx_len) nlen, 0);

  q = p = (char *) value.dptr;
  while (q < (p + value.dlen)) {
	olen = strlen(q);
	if (strncmp(q, ksn, olen + 1) == 0) {
	  Xfree(value.dptr);
	  return(DX_OK);
	}
	ohval = db->hashfunc((Dx_ui8 *) ksn, (Dx_len) nlen, 0);
	if (ohval == nhval) {
	  Xfree(value.dptr);
	  return(DX_KEYSPACE_EXISTS);
	}
	q += olen + 1;
  }

  Xfree(value.dptr);
  return(DX_KEYSPACE_NOTFOUND);
}

static INLINE Dx_rc
keyspace_list_fetch(Dx *db, char **vec, Dx_ui32 *veclen, void **mem)
{
  Dx_ui32 n;
  char *p, *q;
  Dx_datum key, value;
  Dx_rc rc;

  keyspace_init(db);
  Dx_key(&key, DIR_KEY, strlen(DIR_KEY) + 1);
  key.keyspace = &ksname;

  rc = Dx_fetchx(db, &key, DX_FETCH | DX_ALWAYS_COPY, &value);
  if (rc != DX_OK)
	return(rc);

  p = q = *vec = (char *) value.dptr;
  n = 0;
  while (q < (p + value.dlen)) {
	q += strlen(q) + 1;
	n++;
  }

  *veclen = n;
  *mem = value.dptr;
  return(DX_OK);
}

static INLINE Dx_rc
keyspace_list_delete(Dx *db, char *ksn)
{

  return(DX_NOTIMPLEMENTED);

}

/*
 * Create a new keyspace called KSN.
 * Don't allow two keyspace names that have the same hash value.
 * Add the keyspace name to the keyspace directory.
 */
Dx_rc
Dx_create_keyspace(Dx *db, char *ksn)
{
  Dx_rc rc;

  if (ksn == NULL || *ksn == '\0')
	return(DX_INVALID_KEYSPACE);
  if (*ksn == '*')
	return(DX_KEYSPACE_RESERVED);

  rc = keyspace_name_find(db, ksn);
  if (rc != DX_KEYSPACE_NOTFOUND)
	return(DX_KEYSPACE_EXISTS);

  rc = keyspace_dir_add(db, ksn);
  return(rc);
}

/*
 * Return the current keyspace name, or NULL.
 */
char *
Dx_get_keyspace(Dx *db, Dx_datum *key)
{

  if (key != NULL && key->keyspace != NULL)
	return(key->keyspace->name);
  if (db->keyspace == NULL)
	return(NULL);
  return(db->keyspace->name);
}

Dx_rc
Dx_set_keyspace(Dx *db, Dx_datum *key, char *ksn)
{
  Dx_keyspace *ks;
  Dx_rc rc;

  if (ksn == NULL || *ksn == '\0') {
	if (key != NULL)
	  key->keyspace = NULL;
	else
	  db->keyspace = NULL;
	return(DX_OK);
  }
  else if (*ksn == '*')
	return(DX_KEYSPACE_RESERVED);

  /* XXX verify ks exists */
  rc = keyspace_name_find(db, ksn);
  if (rc != DX_OK)
	return(rc);

  ks = (Dx_keyspace *) Xmalloc(sizeof(Dx_keyspace));
  ks->name = Xstrdup(ksn);
  ks->hash = db->hashfunc((Dx_ui8 *) ksn, (Dx_len) strlen(ksn), 0);

  if (key == NULL) {
	if (db->keyspace != NULL) {
	  Xfree(db->keyspace->name);
	  Xfree(db->keyspace);
	}
	db->keyspace = ks;
  }
  else {
	if (key->keyspace != NULL) {
	  Xfree(key->keyspace->name);
	  Xfree(key->keyspace);
	}
	key->keyspace = ks;
  }

  return(DX_OK);
}

Dx_rc
Dx_get_keyspace_list(Dx *db, char ***ksn_vec, void **mem)
{
  Dx_ui32 i, n;
  char *p, *q, **vec;
  Dx_rc rc;

  rc = keyspace_list_fetch(db, &p, &n, mem);
  if (rc != DX_OK)
	return(rc);

  vec = *ksn_vec = (char **) Xmalloc(sizeof(char *) * (n + 1));
  if (vec == NULL) {
	Xfree(*mem);
	return(DX_MALLOC_FAILED);
  }

  q = p;
  for (i = 0; i < n; i++) {
	vec[i] = p;
	p += strlen(p) + 1;
  }
  vec[i] = NULL;

  return(DX_OK);
}

Dx_rc
Dx_delete_keyspace(Dx *db, char *ksn)
{
  Dx_rc rc;

  if (ksn == NULL || *ksn == '\0' || *ksn == '*')
	return(DX_BADARG);

  rc = keyspace_list_delete(db, ksn);

  return(rc);
}
