/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Initialize and maintain the meta information in the database header area.
 * The header area is statically allocated, with its size determined at compile
 * time.
 *
 * The header is a single, fixed-length segment that contains:
 *   o configuration constants for this db
 *   o dynamic state variables
 *
 * The information is replicated in each volume, except that each volume
 * has its own volume number.
 *
 * The header is only accessed when the db is opened and again when it is
 * closed.
 *
 * XXX In theory, we could handle misordered but otherwise correct volumes
 * by putting them in the proper order.
 */

#include "local.h"

enum {
  HEADER_MAGIC             =  0,	/* Dx_ui32 */
  HEADER_IS_BIG_ENDIAN     =  4,	/* Dx_ui32 */
  HEADER_NPAGES            =  8,	/* Dx_ui64 */
  HEADER_BOUNDARY          = 16,	/* Dx_ui64 */
  HEADER_NEXPANSIONS       = 24,	/* Dx_ui64 */
  HEADER_PAGE_SIZE         = 32,	/* Dx_ui32 */
  HEADER_NVOLUMES          = 36,	/* Dx_ui32 */
  HEADER_STAMP_HI          = 40,	/* Dx_ui32 */
  HEADER_STAMP_LO          = 44,	/* Dx_ui32 */
  HEADER_VOLUMENUM         = 48,	/* Dx_ui32 */
  HEADER_UID_TV_SEC        = 52,	/* Dx_ui32 */
  HEADER_DX_BITS           = 56,    /* Dx_ui32 */
  HEADER_VOLUME_WEIGHT     = 60     /* Dx_ui32 */
};

/*
 * Create and initialize meta data for a new db.
 */
Dx_rc
Meta_create(Dx *db)
{
  Physvolnum volnum;
  Dx_ui64 need_len;
  Dx_rc rc;

  need_len = META_HEADER_NBYTES;

  Page_init(db);

  for (volnum = 0; volnum < db->nvolumes; volnum++) {
	rc = Segment_map(db, db->volume[volnum], 0, META_HEADER_NBYTES, 0, 0,
					 SEGMENT_META | SEGMENT_FIXED,
					 &db->volume[volnum]->meta);
	if (rc != DX_OK)
	  return(rc);
  }

  if ((rc = Area_create(db)) != DX_OK)
	  return(rc);

  if ((rc = Ehash_create(db)) != DX_OK)
	return(rc);

  if ((rc = Meta_write(db)) != DX_OK)
	return(rc);

  rc = Area_write(db);

  return(rc);
}

/*
 * Obtain meta data from an existing db.
 */
Dx_rc
Meta_init(Dx *db, off_t st_size)
{
  Physvolnum volnum;
  Dx_rc rc;

  Page_init(db);

  for (volnum = 0; volnum < db->nvolumes; volnum++) {
	rc = Segment_map(db, db->volume[volnum], 0, META_HEADER_NBYTES, 0, 0,
					 SEGMENT_META | SEGMENT_FIXED, &db->volume[volnum]->meta);
	if (rc != DX_OK)
	  return(rc);
  }

  if ((rc = Meta_read(db)) != DX_OK)
	return(rc);

  if ((rc = Area_init(db)) != DX_OK)
	return(rc);

  return(rc);
}

/*
 * Read meta data for DB, check it, and initialize the Dx structure.
 */
Dx_rc
Meta_read(Dx *db)
{
  Dx_ui8 *header;
  Dx_ui32 magic;
  Physvolnum volnum;

#ifdef NOTDEF
  struct timeval tv;
  gettimeofday(&tv, NULL);
#endif

  for (volnum = 0; volnum < db->nvolumes; volnum++) {
	header = (Dx_ui8 *) db->volume[volnum]->meta;

	/* This is needed before the rest can be decoded. */
	db->endian = *(Dx_ui32 *) (header + HEADER_IS_BIG_ENDIAN)
	  ? DX_BIG_ENDIAN : DX_LITTLE_ENDIAN;

	magic = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_MAGIC));
	if (magic != META_MAGIC) {
	  fprintf(stderr, "Meta_read: bad magic number in file header.\n");
	  return(DX_BADMAGIC);
	}

	if (volnum == 0) {
	  db->npages
		= Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NPAGES));
	  db->boundary
		= Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_BOUNDARY));
	  db->nexpansions
		= Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NEXPANSIONS));
	  db->page_size
		= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_PAGE_SIZE));
	  db->creation_stamp.high
		= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_HI));
	  db->creation_stamp.low
		= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_LO));
	  if (Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_VOLUMENUM)) != 0)
		return(DX_HEADER_MISMATCH);

	  db->current_uid.tv_sec
		= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_UID_TV_SEC));
	  db->current_uid.count = 0;

	  if (db->nvolumes != Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_NVOLUMES)))
		return(DX_VOLUME_MISMATCH);
#ifdef NOTDEF
	  if (db->current_uid.tv_sec >= (Dx_ui32) tv.tv_sec)
		return(DX_TIME_MISMATCH);
#endif
	  if (Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_DX_BITS)) != DX_BITS)
		return(DX_HEADER_MISMATCH);
	}
	else {
	  if (Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NPAGES)) != db->npages
		  || Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_BOUNDARY))
		    != db->boundary
		  || Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NEXPANSIONS))
		    != db->nexpansions
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_PAGE_SIZE))
		    != db->page_size
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_NVOLUMES))
		    != db->nvolumes
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_HI))
		    != db->creation_stamp.high
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_LO))
		    != db->creation_stamp.low
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_VOLUMENUM)) != volnum
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_UID_TV_SEC))
		      != (Dx_ui32) db->current_uid.tv_sec
		  || Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_DX_BITS))
              != (Dx_ui32) DX_BITS
		  )
		return(DX_HEADER_MISMATCH);
	}
	db->volume_weights.weights[volnum]
	  = Dx_import_ui32(db, (Dx_ui32 *) (header + HEADER_VOLUME_WEIGHT));
	if (volnum == 0)
	  db->volume_weights.cw[volnum] = db->volume_weights.weights[0];
	else
	  db->volume_weights.cw[volnum] = db->volume_weights.weights[volnum]
		+ db->volume_weights.cw[volnum - 1];
  }

  db->nlogical_volumes = db->volume_weights.cw[db->nvolumes - 1];

  return(DX_OK);
}

/*
 * Write header meta data back to persistent storage.
 */
Dx_rc
Meta_write(Dx *db)
{
  Physvolnum volnum;
  Dx_ui8 *header;
  static Dx_ui32 meta_magic = META_MAGIC;
  static Dx_ui32 dx_bits = DX_BITS;

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(DX_OK);

  for (volnum = 0; volnum < db->nvolumes; volnum++) {
	header = (Dx_ui8 *) db->volume[volnum]->meta;
	*(Dx_ui32 *) (header+HEADER_MAGIC)
	  = Dx_export_ui32(db, &meta_magic);
	*(Dx_ui64 *) (header+HEADER_NPAGES)
	  = Dx_export_ui64(db, &db->npages);
	*(Dx_ui64 *) (header+HEADER_BOUNDARY)
	  = Dx_export_ui64(db, &db->boundary);
	*(Dx_ui64 *) (header+HEADER_NEXPANSIONS)
	  = Dx_export_ui64(db, &db->nexpansions);
	*(Dx_ui32 *) (header+HEADER_PAGE_SIZE)
	  = Dx_export_ui32(db, &db->page_size);
	*(Dx_ui32 *) (header+HEADER_IS_BIG_ENDIAN)
	  = (db->endian == DX_BIG_ENDIAN) ? 1 : 0;
	*(Dx_ui32 *) (header+HEADER_NVOLUMES)
	  = Dx_export_ui32(db, &db->nvolumes);
	*(Dx_ui32 *) (header+HEADER_STAMP_HI)
	  = Dx_export_ui32(db, &db->creation_stamp.high);
	*(Dx_ui32 *) (header+HEADER_STAMP_LO)
	  = Dx_export_ui32(db, &db->creation_stamp.low);
	*(Dx_ui32 *) (header+HEADER_VOLUMENUM)
	  = Dx_export_ui32(db, &volnum);
	*(Dx_ui32 *) (header+HEADER_UID_TV_SEC)
	  = Dx_export_ui32(db, &db->current_uid.tv_sec);
	*(Dx_ui32 *) (header+HEADER_DX_BITS)
	  = Dx_export_ui32(db, &dx_bits);
	*(Dx_ui32 *) (header+HEADER_VOLUME_WEIGHT)
	  = Dx_export_ui32(db, &db->volume_weights.weights[volnum]);
  }

  return(DX_OK);
}

/*
 * Copy the current (in-memory) header meta data for OLD_DB to NEW_DB.
 * Store the meta data in NEW_DB's byte ordering.
 */
Dx_rc
Meta_duplicate(Dx *new_db, Dx *old_db)
{
  Physvolnum volnum;
  Dx_ui8 *header;
  static Dx_ui32 meta_magic = META_MAGIC;
  static Dx_ui32 dx_bits = DX_BITS;

  for (volnum = 0; volnum < new_db->nvolumes; volnum++) {
	header = (Dx_ui8 *) new_db->volume[volnum]->meta;

	*(Dx_ui32 *) (header+HEADER_MAGIC)
	  = Dx_export_ui32(new_db, &meta_magic);
	*(Dx_ui64 *) (header+HEADER_NPAGES)
	  = Dx_export_ui64(new_db, &old_db->npages);
	*(Dx_ui64 *) (header+HEADER_BOUNDARY)
	  = Dx_export_ui64(new_db, &old_db->boundary);
	*(Dx_ui64 *) (header+HEADER_NEXPANSIONS)
	  = Dx_export_ui64(new_db, &old_db->nexpansions);
	*(Dx_ui32 *) (header+HEADER_PAGE_SIZE)
	  = Dx_export_ui32(new_db, &old_db->page_size);
	*(Dx_ui32 *) (header+HEADER_IS_BIG_ENDIAN)
	  = (new_db->endian == DX_BIG_ENDIAN) ? 1 : 0;
	*(Dx_ui32 *) (header+HEADER_NVOLUMES)
	  = Dx_export_ui32(new_db, &old_db->nvolumes);
	*(Dx_ui32 *) (header+HEADER_STAMP_HI)
	  = Dx_export_ui32(new_db, &old_db->creation_stamp.high);
	*(Dx_ui32 *) (header+HEADER_STAMP_LO)
	  = Dx_export_ui32(new_db, &old_db->creation_stamp.low);
	*(Dx_ui32 *) (header+HEADER_VOLUMENUM)
	  = Dx_export_ui32(new_db, &volnum);
	*(Dx_ui32 *) (header+HEADER_UID_TV_SEC)
	  = Dx_export_ui32(new_db, &old_db->current_uid.tv_sec);
	*(Dx_ui32 *) (header+HEADER_DX_BITS)
	  = Dx_export_ui32(new_db, &dx_bits);
	*(Dx_ui32 *) (header+HEADER_VOLUME_WEIGHT)
	  = Dx_export_ui32(new_db, &old_db->volume_weights.weights[volnum]);
  }

  return(DX_OK);
}

void
meta_show(Dx *db)
{
  Physvolnum pvn;
  Dx_ui8 *header;
  Dx_ui32 dx_bits, endian, magic, nvolumes, page_size, volnum;
  Dx_ui64 boundary, nexpansions, npages;
  Dx_stamp creation_stamp;
  Dx_uid current_uid;

  header = (Dx_ui8 *) db->volume[0]->meta;

  /* This is needed before the rest can be decoded. */
  endian = *(Dx_ui32 *) (header + HEADER_IS_BIG_ENDIAN)
	? DX_BIG_ENDIAN : DX_LITTLE_ENDIAN;

  magic = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_MAGIC));
  npages = Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NPAGES));
  boundary = Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_BOUNDARY));
  nexpansions = Dx_import_ui64(db, (Dx_ui64 *) (header+HEADER_NEXPANSIONS));
  page_size = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_PAGE_SIZE));
  creation_stamp.high
	= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_HI));
  creation_stamp.low
	= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_STAMP_LO));
  volnum = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_VOLUMENUM));
  current_uid.tv_sec
	= Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_UID_TV_SEC));
  current_uid.count = db->current_uid.count;

  nvolumes = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_NVOLUMES));
  dx_bits = Dx_import_ui32(db, (Dx_ui32 *) (header+HEADER_DX_BITS));

  fprintf(stderr, "magic=%lu, npages=%llu, boundary=%llu, ",
		  magic, npages, boundary);
  fprintf(stderr, "nexpansions=%llu, page_size=%lu\n",
		  nexpansions, page_size);
  fprintf(stderr, "creation_stamp=%lu/%lu, volnum=%lu\n",
		  creation_stamp.high, creation_stamp.low, volnum);
  fprintf(stderr, "current_uid=%lu/%lu\n",
		  current_uid.tv_sec, current_uid.count);
  fprintf(stderr, "endian=\"%s\", nvolumes=%lu, dx_bits=%lu\n",
		  endian == DX_BIG_ENDIAN ? "big endian" : "little endian",
		  nvolumes, dx_bits);
  fprintf(stderr, "DX_DEBUG=%d, DX_CHECK=%d, DX_STATS=%d\n",
		  DX_DEBUG, DX_CHECK, DX_STATS);
  fprintf(stderr, "Volume weights:");
  for (pvn = 0; pvn < db->nvolumes; pvn++)
	fprintf(stderr, " %lu", db->volume_weights.weights[pvn]);
  fprintf(stderr, "\n");
}

void
dx_show(Dx *db)
{

  fprintf(stderr, "npages=%llu, boundary=%llu, ",
		  db->npages, db->boundary);
  fprintf(stderr, "nexpansions=%llu, page_size=%lu\n",
		  db->nexpansions, db->page_size);
  fprintf(stderr, "creation_stamp=%lu/%lu",
		  db->creation_stamp.high, db->creation_stamp.low);
  fprintf(stderr, "current_uid=%lu/%lu\n",
		  db->current_uid.tv_sec, db->current_uid.count);
  fprintf(stderr, "nvolumes=%lu\n",
		  db->nvolumes);
}
