/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "local.h"

static const struct {
  Dx_rc rc;
  char *msg;
} messages[] = {
  { DX_OK,               "No error (DX_OK)" },
  { DX_HASH_OVERFLOW,    "Too many identical keys/hash values" },
  { ITEM_NOSPACE,        "ITEM_NOSPACE" },
  { ITEM_NOTFOUND,       "Key not found" },
  { DX_BAD_POSITION,     "DX_BAD_POSITION" },
  { DX_NOTIMPLEMENTED,   "Unimplemented feature" },
  { DX_BADARG,           "Bad argument" },
  { DX_USERIO_ERROR,     "User I/O callback function failed" },
  { ITEM_NOMATCH,        "ITEM_NOMATCH" },
  { PAGE_NOTFOUND,       "PAGE_NOTFOUND" },
  { DX_BAD_REGION,		 "Invalid region specification" },
  { DX_BADMAGIC,         "Database has bad magic number" },
  { DX_INITFAILED,       "DX_INITFAILED" },
  { DX_PHYSMAP_FAILED,   "DX_PHYSMAP_FAILED" },
  { DX_INVALID_PAGESIZE, "DX_INVALID_PAGESIZE" },
  { DX_BADACCESSMODE,    "Invalid mode flag or combination of flags" },
  { DX_OS_ERROR,         "DX_OS_ERROR" },
  { DX_INVALID_NVOLUMES, "DX_INVALID_NVOLUMES" },
  { DX_INTERNAL_ERROR,   "DX_INTERNAL_ERROR" },
  { DX_MALLOC_FAILED,    "DX_MALLOC_FAILED" },
  { DX_BADEDITARG,       "Bad DX_EDIT specification" },
  { DX_VOLUME_MISMATCH,  "Number of volumes is incorrect" },
  { DX_HEADER_MISMATCH,  "Incorrect volume configuration/corrupted database" },
  { DX_INVALID_MAX_STRIPE_NPAGES, "Invalid max_stripe_npages value" },
  { DX_TIME_MISMATCH,    "System time has been turned back" },
  { DX_BADCONFIGARG,     "Invalid configuration argument to Dx_open" },
  { DX_OPENFAILED,       "Could not open a volume - open() failed" },
  { DX_READONLY,         "Attempt to modify a read-only db" },
  { DX_INVALID_ENDIAN,   "Invalid endian value" },
  { DX_INVALID_KEYSPACE, "Invalid keyspace name" },
  { DX_KEY_NOT_FOUND,    "Key not found" },
  { DX_KEY_EXISTS,       "Key exists" },
  { DX_FRAME_NOT_POPPED, "Frame must be popped for this operation" },
  { DX_FRAME_NO_FRAME,   "There is no valid frame for this function" },
  { DX_FRAME_BAD_POINTER, "A pointer argument does not point to frame memory"},
  { DX_KEYSPACE_NOTFOUND, "Keyspace doesn't exist" },
  { DX_KEYSPACE_EXISTS,  "Keyspace already exists" },
  { DX_KEYSPACE_RESERVED, "Keyspace name is reserved" },
  { SEGMENT_NOTFOUND,     "Segment not found" },
  { DX_BAD_KEY,           "Key is invalid" },
  { DX_FSTATFAILED,       "fstat(2) failed" },
  { DX_ACCESSFAILED,      "A volume does not exist - access(2) failed" },
  { DX_FLOCKFAILED,       "Could not lock a volume - flock(2) failed" },
  { DX_INVALID_VOLUME_WEIGHTS, "Incorrect number of volume weights" },
  { DX_OK, NULL }
};

char *
Dx_strerror(Dx_rc rc)
{
  int i;

  for (i = 0; messages[i].msg != NULL; i++) {
	if (messages[i].rc == rc)
	  return(messages[i].msg);
  }

  return("Unknown result code - no error message.");
}

void *
Xmalloc(size_t size)
{
  void *ptr;

  if ((ptr = malloc(size)) == NULL)
	Fatal(DX_MALLOC_FAILED);

  return(ptr);
}

char *
Xstrdup(const char *s)
{
  char *ptr;

  if ((ptr = (char *) Xmalloc(strlen(s) + 1)) == NULL)
	Fatal(DX_MALLOC_FAILED);

  strcpy(ptr, s);
  return(ptr);
}

void *
Xrealloc(void *ptr, size_t size)
{
  void *nptr;

  if ((nptr = realloc(ptr, size)) == NULL)
	Fatal(DX_MALLOC_FAILED);

  return(nptr);
}

void
Xfree(void *ptr)
{

  free(ptr);
}

void
Fatal(Dx_rc rc)
{

  fprintf(stderr, "%s\n", Dx_strerror(rc));
  abort();

  /*NOTREACHED*/
  exit(1);
}

/*
 * Return a random integer between LOW and HIGH, inclusive.
 * As per rand(3), quoting from "Numerical Recipes in C: The Art of Scientific
 * Computing", William  H.  Press, Brian P. Flannery, Saul A. Teukolsky,
 * William T.  Vetterling;  New  York:  Cambridge  University Press,
 * 1990 (1st ed, p. 207)
 */
int
Xrand(int low, int high)
{
  int j;

  if (DX_DEBUG && low > high) {
	fprintf(stderr, "xrand1");
	Fatal(DX_INTERNAL_ERROR);
  }

  if (low == high)
	return(low);

  j = low
	+ (int) (((double) rand() / ((double) RAND_MAX + 1)) * (high - low + 1));

  if (DX_DEBUG && (j < low || j > high)) {
	fprintf(stderr, "xrand2");
	Fatal(DX_INTERNAL_ERROR);
  }

  return(j);
}

void
Xsrand(unsigned int seed)
{

  srand(seed);
}

void
Stats_init(Dx *db)
{
  Dx_stats_counters *s;

  s = &db->stats;
  s->nsplits = s->nadds = s->nmoves
    = s->nsearches = s->nprobes = s->nreorgs = 0;
}

void
Stats_show(Dx *db, FILE *fp)
{
  Dx_stats_counters *s;

  s = &db->stats;
  fprintf(fp, "nsplits = %d, ", s->nsplits);
  fprintf(fp, "nadds   = %d, ", s->nadds);
  fprintf(fp, "nmoves  = %d, ", s->nmoves);
  fprintf(fp, "nprobes = %d, ", s->nprobes);
  fprintf(fp, "nreorgs = %d\n", s->nreorgs);
}
