/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

static void err(char *s, ...);

/*
 * mkargvec: scan a string and split up into argv
 * argc contains maximum number of argv elements upon entry
 * and is set to the actual number of elements on exit
 * Argv list terminated by entry of 0
 * If ifs is 0 or the empty string, then white space (tabs and blanks)
 * are used as field separators; otherwise each character in ifs
 * is considered to be a field separator.
 * Leading field separators in string are skipped.
 * Returns -1 if not enough argv elements, 0 oth.
 *
 * 22-Jun-86	Checking for no more argv pointers first caused problems when
 *		there were trailing blanks.  Delay the check until necessary.
 */

#ifndef NULL
#define NULL	0
#endif

static char *default_ifs = "\t ";

int
mkargvec(char *string, int *argc, char *argv[], char *space, char *ifs)
{
	register int i;
	register char *p, *sp;
	int max, count;

	p = string;
	sp = space;
	count = 0;
	max = *argc - 1;
	if (ifs == NULL || ifs[0] == '\0')
		ifs = default_ifs;

	/* For each arg, copy and set argv pointer */
	i = 0;
	while (1) {
		while (*p != '\0' && strchr(ifs, *p) != NULL)
			p++;
		if (*p == '\0')
			break;
		if (i == max)		/* no more pointers */
			break;
		argv[i++] = sp;
		count++;
		while (*p != '\0' && strchr(ifs, *p) == NULL)
			*sp++ = *p++;
		*sp++ = '\0';
	}
	*argc = count;
	argv[count] = 0;
	if (*p != '\0')			/* only if i == max */
		return(-1);
	else
		return(0);
}


/*
 * mkargvecq: modified mkargvec
 * scan a string and split up into argv
 * argc contains maximum number of argv elements upon entry
 * and is set to the actual number of elements on exit
 * Argv list terminated by entry of 0
 * If ifs is 0 or the empty string, then white space (tabs and blanks)
 * are used as field separators; otherwise each character in ifs
 * is considered to be a field separator.
 * Leading field separators in string are skipped.
 *
 * This version knows about quotes - single and double quotes may be used
 * to delimit words
 * Two consecutive quotes makes a null word
 * If the 'keepq' arg is non-zero, then the quote marks delimiting words are
 * kept
 *
 * Also knows about the escape character (backslash) which may be used to
 * prevent interpretation of ifs characters, quotes, or the escape char itself
 * Returns -1 if not enough argv elements, -2 if quotes screwed,
 * -3 if backslash quoting problem, 0 oth.
 */
int
mkargvecq(char *string, int *argc, char *argv[], char *space, char *ifs,
		  int keepq)
{
	register int i;
	register char *p, *sp;
	int count, max;
	char *q, quote[10];

	p = string;
	sp = space;
	count = 0;
	max = *argc - 1;
	if (ifs == NULL || ifs[0] == '\0')
		ifs = default_ifs;

	quote[0] = '\0';
	if (strchr(ifs, '\'') == NULL)
		strcat(quote, "'");
	if (strchr(ifs, '"') == NULL)
		strcat(quote, "\"");

	if (quote[0] == '\0')
		err("Quote mode but quotes are all ifs");
	if (strchr(ifs, '\\') != NULL)
		err("Escape character (backslash) in ifs");

	q = NULL;
	/* For each arg, copy and set argv pointer */
	i = 0;
	while (1) {
		while (*p != '\0' && strchr(ifs, *p) != NULL)
			p++;
		if (*p == '\0')
			break;
		if (i == max)
			break;
		argv[i++] = sp;
		if ((q = strchr(quote, *p)) != NULL) {
			if (keepq)
				*sp++ = *q;
			if (*(++p) == '\0')
				break;
		}
		count++;
		while (*p != '\0') {
			if (q != NULL && *p == *q) {
				if (keepq)
					*sp++ = *q;
				q = NULL;
				p++;
				break;
			}
			else if (q == NULL && (q = strchr(quote, *p)) != NULL) {
				if (!keepq) {
					p++;
					continue;
				}
			}
			else if (*p == '\\') {
				if (*(++p) == '\0') {
					*argc = count;
					argv[count] = 0;
					err("Misquote");
					return(-3);
				}
				*sp++ = *p++;
				continue;
			}
			else if (q == NULL && strchr(ifs, *p) != NULL)
				break;
			*sp++ = *p++;
		}
		*sp++ = '\0';
	}
	*argc = count;
	argv[count] = 0;
	if (q != NULL) {
		err("Unmatched quote: %c", *q);
		return(-2);
	}

	if (*p != '\0') {		/* only if i == max */
		err("Not enough pointers");
		return(-1);
	}
	else
		return(0);
}

#include <stdarg.h>

static void
err(char *s, ...)
{
  va_list ap;

  va_start(ap, s);

#ifdef DEBUG
	vfprintf(stderr, s, ap);
	fprintf(stderr, "\n");
#endif

	va_end(ap);
}

/*
 * Breakup a string into argc/argv format, allocating all the space
 * Assumes worst-case for number of pointers; i.e., an alternating
 * character/ifs sequence
 * Caller must free the space pool and the vector of pointers
 */
int
breakup(char *buf, char **argv[], char **sp, int *n, char *ifs)
{
	register int len, rc;
	char **ptr;

	len = strlen(buf);
	if ((*sp = (char *) malloc((unsigned) (len + 1))) == NULL) {
		*n = 0;
		return(-4);
	}
	*n = (len + 1) / 2 + 1;
	ptr = (char **) malloc((unsigned) (sizeof(char *) * *n));
	if (ptr == NULL) {
		free(*sp);
		*n = 0;
		return(-4);
	}
	if ((rc = mkargvec(buf, n, ptr, *sp, ifs)) < 0) {
		free(*sp);
		free(ptr);
		*n = 0;
		return(rc);
	}
	*argv = ptr;
	return(0);
}

int
breakupq(char *buf, char **argv[], char **sp, int *n, char *ifs, int keepq)
{
	register int len, rc;
	char **ptr;

	len = strlen(buf);
	if ((*sp = (char *) malloc((unsigned) (len + 1))) == NULL) {
		*n = 0;
		return(-4);
	}
	*n = (len + 1) / 2 + 1;
	ptr = (char **) malloc((unsigned) (sizeof(char *) * *n));
	if (ptr == NULL) {
		free(*sp);
		*n = 0;
		return(-4);
	}
	if ((rc = mkargvecq(buf, n, ptr, *sp, ifs, keepq)) < 0) {
		free(*sp);
		free(ptr);
		*n = 0;
		return(rc);
	}
	*argv = ptr;
	return(0);
}

