/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Manage dynamically allocated/released memory space using bitmap blocks.
 * Each bit represents a one page long segment, which may be used to hold
 * items or an indirect key/data.
 * Each bitmap block is a vector of bits that indicates whether a particular
 * page is available for allocation or in use.
 *
 * A page is available if its corresponding bit is ZERO.
 * The bitmap is accessed byte-wise, where bits are numbered little-endian:
 *   bit0 -> block0, bit1->block1, ..., bit7->block7
 *
 * Iblocks are used to link together a sequence of bitmap blocks.
 */

#include "local.h"

#ifdef __cplusplus
#include "physbits.cc"
#else
#include "physbits.c"
#endif

/*
 * Some precomputed mask values.
 * byte_mask[X][Y] gives the bitmask corresponding to a bit pattern
 * where X is the number of consecutive set bits that follow and Y is the
 * index (0-7) of the first set bit.
 * Example:
 * byte_mask[3][4] is 0xf0 (1111 0000)
 * byte_mask[5][1] is 0x7e (0111 1110)
 * 
 */
static const Dx_ui8 byte_mask[8][8] = {
  /* Starting bit number */
  /* 0,    1,    2,    3,    4,    5,    6,    7 */
   { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 },	/* 0 */
   { 0x03, 0x06, 0x0c, 0x18, 0x30, 0x60, 0xc0, 0x80 },	/* 1 */
   { 0x07, 0x0e, 0x1c, 0x38, 0x70, 0xe0, 0xc0, 0x80 },	/* 2 */
   { 0x0f, 0x1e, 0x3c, 0x78, 0xf0, 0xe0, 0xc0, 0x80 },	/* 3 */
   { 0x1f, 0x3e, 0x7c, 0xf8, 0xf0, 0xe0, 0xc0, 0x80 },	/* 4 */
   { 0x3f, 0x7e, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80 },	/* 5 */
   { 0x7f, 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80 },	/* 6 */
   { 0xff, 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80 }	/* 7 */
};

static Dx_rc find_possible_map(Dx *db, Volume *volume, Dx_ui32 npages,
							   Area **areap);

static Dx_rc compute_starting_location(Dx *db, Physvolnum volnum,
									   Physoff offset,
									   Dx_ui32 rel_npages, Area **areap,
									   Dx_ui32 *start_bytenump,
									   Dx_ui32 *start_bitnump);

/*
 * Create the physical map for the given relative area number and volume.
 */
Dx_rc
Physmap_create(Dx *db, Volume *volume, Dx_ui32 rel_areanum, Physmap *physmap)
{
  Physoff offset;
  Dx_rc rc;

  offset = PHYSMAP_OFFSET(db, rel_areanum);
  rc = Segment_map(db, volume, offset, PHYSMAP_MAP_NBYTES, 0, 0,
				   SEGMENT_META | SEGMENT_FIXED, &physmap->map);

  Dx_memset((void *) physmap->map, 0, (size_t) PHYSMAP_MAP_NBYTES);
  physmap->avail = PHYSMAP_MAP_NBYTES * BITS_PER_BYTE;

  return(DX_OK);
}

/*
 * Map a bitmap for the given area number.
 */
Dx_rc
Physmap_init(Dx *db, Dx_ui32 areanum, Physmap *physmap)
{
  int i;
  Dx_ui8 *p;
  Dx_rc rc;

  rc = Segment_map(db, AREANUM_TO_VOLUME(db, areanum),
				   PHYSMAP_OFFSET(db, REL_AREANUM(db, areanum)),
				   PHYSMAP_MAP_NBYTES, 0, 0,
				   SEGMENT_META | SEGMENT_FIXED, &physmap->map);
  if (rc != DX_OK)
	return(rc);

  physmap->avail = 0;
  p = (Dx_ui8 *) physmap->map;
  for (i = 0; i < PHYSMAP_MAP_NBYTES; i++)
	physmap->avail += byte_info[*p++].t_bits;

  return(DX_OK);
}

/*
 * Using the given bitmap SEG, find a free block and return its Physoff
 * in the db.  SEG_SIZE is the size of an allocation map, in bytes.
 */
Dx_rc
Physmap_allocate_page(Dx *db, Volume *volume, Seglen seg_len, Physoff *offset)
{

  return(Physmap_allocate_segment(db, volume, seg_len, 1, offset));

#ifdef NOTDEF
  /* XXX It may pay to optimize allocation of 1 page as a special case. */
  Dx_ui32 bn, bytenum;
  Dx_ui8 *p;
  Area *area;
  const Byteinfo *bi;
  Dx_rc rc;


  area = NULL;
  if ((rc = find_possible_map(db, volume, 1, &area)) != DX_OK)
	return(rc);

  p = (Dx_ui8 *) area->physmap.map;

  for (bytenum = 0; bytenum < seg_size; bytenum++) {
	if (*p != 0xff)
	  break;
	p++;
  }

  if (bytenum == seg_size) {
	fprintf(stderr, "Physmap_allocate_page: avail count inconsistency!\n");
	return(DX_PHYSMAP_FAILED);
  }

  bi = &byte_info[*p];
  if (bi->i_bit_index == -1) {
	fprintf(stderr, "Physmap_allocate_page: Byteinfo bug!\n");
	return(DX_PHYSMAP_FAILED);
  }

  /* *p |= (1 << bi->i_bit_index); */
  *p |= byte_mask[0][bi->i_bit_index];

  bn = (bytenum * BITS_PER_BYTE) + bi->i_bit_index;
  *offset = USERPAGE_OFFSET_START(db, REL_AREANUM(db, area->areanum))
	+ (bn * db->page_size);
  area->physmap.avail--;

  return(DX_OK);
#endif
}

/*
 * Scan through the bitmap, starting at byte index IND, for NFULL available
 * bytes and PARTIAL available bits, all contiguous.  If successful, return
 * 1; if the entire map has been scanned, return -1; otherwise return 0
 * and set *INDP to the index at which the scan stopped; the caller may
 * resume searching from that byte.
 */
static int inline
super_peek(Dx_ui8 *base, Dx_ui32 seg_size, Dx_ui32 ind, Dx_ui32 nfull,
		   Dx_ui32 partial, Dx_ui32 *indp)
{
  Dx_ui32 i, n;
  Dx_ui8 *p;

  /* XXX This could be accelerated by looking at 16/32/64 bits at a time. */
  p = base + ind;
  n = nfull;
  for (i = ind; nfull != 0 && i < seg_size; i++) {
	if (*p++ != 0x00) {
	  *indp = i;
	  return(0);
	}
	if (--n == 0)
	  break;
  }

  /* If we're at the end of the map, we lose. */
  if (i == seg_size)
	return(-1);

  /* If there's no remainder to find, we're done. */
  if (n == 0 && partial == 0)
	return(1);

  if (byte_info[*p].r_bits >= partial)
	return(1);

  *indp = ind;
  return(0);
}

/*
 * Allocate NPAGES contiguous pages and return the physical offset of the
 * first.
 * Note that a segment may not span bitmaps, so the biggest possible allocation
 * from one bitmap is the number of bits in PHYSMAP_MAP_NBYTES.
 *
 * XXX Should cache information about freed/truncated segments to make their
 * reallocation much faster by sometimes avoiding a search through bitmaps.
 *
 * XXX This might take a flag argument to hint that this segment is likely
 * to expand.  It would then prefer an allocation with spare contiguous space.
 */
Dx_rc
Physmap_allocate_segment(Dx *db, Volume *volume, Seglen seg_len,
						 Dx_ui32 npages, Physoff *offset)
{
  Dx_ui32 bitnum, nfull, bytenum, need, partial, *pp, start_bitnum;
  Dx_ui8 *p;
  Dx_ui8 *start_p, start_mask;
  Dx_ui8 *end_p, end_mask;
  Area *area;
  const Byteinfo *b;
  Dx_rc rc;

  /* Shuts up compiler. */
  nfull = 0;
  start_bitnum = 0;
  start_p = NULL;
  start_mask = 0;
  end_p = NULL;
  end_mask = 0;

  if (npages > (PHYSMAP_MAP_NBYTES * BITS_PER_BYTE)) {
	fprintf(stderr, "Physmap_allocate_segment: npages too big!\n");
	return(DX_INTERNAL_ERROR);
  }

  area = NULL;

 again:

  if ((rc = find_possible_map(db, volume, npages, &area)) != DX_OK)
	return(rc);

  /*
   * This bitmap page has enough free space, but it may not be contiguous
   * space. If it's not, we need to get another bitmap page and repeat.
   *
   * Here's the algorithm for finding the contiguous pages.  We use the
   * precomputed information to examine the bitmap page one byte at a time.
   */
  p = (Dx_ui8 *) area->physmap.map;

  bytenum = 0;
  /* Accelerate past an allocated block at the start. */
  for (pp = (Dx_ui32 *) p; *pp == 0xffffffff
		 && bytenum < seg_len; bytenum += 4)
	pp++;
  p = (Dx_ui8 *) pp;

  need = npages;
  while (bytenum < seg_len && need) {
	b = &byte_info[*p];
	if (b->r_bits >= need) {
	  start_p = p;
	  start_bitnum = 0;
	  start_mask = byte_mask[npages - 1][start_bitnum];
	  end_p = NULL;
	  need = 0;
	}
	else if (b->max_bits >= need) {
	  start_p = p;
	  start_bitnum = b->max_bits_start;
	  start_mask = byte_mask[npages - 1][start_bitnum];
	  end_p = NULL;
	  need = 0;
	}
	else if (b->l_bits >= need) {
	  start_p = p;
	  start_bitnum = BITS_PER_BYTE - b->l_bits;
	  start_mask = byte_mask[npages - 1][start_bitnum];
	  end_p = NULL;
	  need = 0;
	}
	else if (b->l_bits) {
	  int r;

	  nfull = (need - b->l_bits) / BITS_PER_BYTE;
	  partial = (need - b->l_bits) % BITS_PER_BYTE;
	  r = super_peek((Dx_ui8 *) area->physmap.map, seg_len, bytenum + 1,
					 nfull, partial, &bytenum);
	  if (r == 0)
		continue;

	  if (r == -1)
		break;
	  /* Found enough. */
	  start_p = p;
	  start_bitnum = BITS_PER_BYTE - b->l_bits;
	  start_mask = byte_mask[b->l_bits - 1][start_bitnum];
	  end_p = p + nfull + (partial > 0);
	  end_mask = 0;
	  if (partial)
		end_mask = byte_mask[partial - 1][0];
	  need = 0;
	} else {
	  p++;
	  bytenum++;
	}
  }

  if (need) {
	if (area->next != NULL)
	  area = area->next;
	else {
	  rc = Area_create_volume(db, volume->volnum,
							  area->areanum + db->nvolumes, &area);
	  if (rc != DX_OK)
		return(rc);
	}
	goto again;
  }

  /* Set the select bits to indicate that they're allocated. */

  if (end_p != NULL) {
	unsigned int i;

	*start_p |= start_mask;
	for (i = 1; i <= nfull; i++)
	  start_p[i] = 0xff;
	*end_p |= end_mask;
  }
  else
	*start_p |= start_mask;

  bitnum = (bytenum * BITS_PER_BYTE) + start_bitnum;
  *offset = USERPAGE_OFFSET_START(db, REL_AREANUM(db, area->areanum))
	+ (bitnum * db->page_size);
  area->physmap.avail -= npages;

  if (DX_DEBUG > 1)
	fprintf(stderr,
			"Alloc: %3lu npage%s %5lu bitnum (byte %lu, sbitnum %lu), %llu\n",
			npages, npages == 1 ? ", " : "s,", bitnum, bytenum, start_bitnum,
			(unsigned long long) *offset);

  return(DX_OK);
}

static Dx_rc
compute_starting_location(Dx *db, Physvolnum volnum, Physoff offset,
						  Dx_ui32 rel_npages, Area **areap,
						  Dx_ui32 *start_bytenump, Dx_ui32 *start_bitnump)
{
  Dx_ui32 start_bitoffset;
  Area *area;
  Physoff start_offset;
  Dx_ui32 volrelnum;
  Dx_rc rc;

  if ((area = AREA_PHYSOFFSET_TO_AREA(db, volnum, offset)) == NULL) {
	fprintf(stderr, "physmap_compute_starting_location: lookup failed!\n");
	return(DX_INTERNAL_ERROR);
  }

  if (area->physmap.map == NULL) {
	if ((rc = Physmap_init(db, area->areanum, &area->physmap)) != DX_OK)
	  return(rc);
  }

  /* Compute the location of the starting bit within the bitmap. */
  volrelnum = REL_AREANUM(db, area->areanum);
  start_offset = USERPAGE_OFFSET_START(db, volrelnum);
  start_bitoffset = (offset - start_offset + NPAGES_TO_NBYTES(db, rel_npages))
	/ db->page_size;

  *start_bytenump = start_bitoffset / 8;
  *start_bitnump = start_bitoffset % 8;
  *areap = area;

  return(DX_OK);
}

/*
 * Attempt to extend the segment that starts at OFFSET and is OLD_NPAGES long
 * to be NEW_NPAGES long.
 */
Dx_rc
Physmap_extend_segment(Dx *db, Volume *volume, Seglen seg_len, Physoff offset,
					   Dx_ui32 old_npages, Dx_ui32 new_npages)
{
  Dx_ui32 i, ind, need, nleft, tail_count;
  Dx_ui32 start_bitnum, start_bytenum;
  Dx_ui8 *p;
  Dx_ui8 *start_p, start_mask;
  Dx_ui8 end_mask;
  Area *area;
  const Byteinfo *b;
  Dx_rc rc;

  /* Shuts up compiler. */
  start_p = NULL;
  start_mask = 0;
  end_mask = 0;

  if (new_npages == old_npages)
	return(DX_OK);

  if (new_npages < old_npages) {
	fprintf(stderr, "Physmap_extend_segment: new_npages too small!\n");
	return(DX_INTERNAL_ERROR);
  }

  if (new_npages > (PHYSMAP_MAP_NBYTES * BITS_PER_BYTE)) {
	fprintf(stderr, "Physmap_extend_segment: new_npages too big!\n");
	return(DX_INTERNAL_ERROR);
  }

  /* Locate the bit that follows the last bit of the allocation. */
  rc = compute_starting_location(db, volume->volnum, offset, old_npages,
								 &area, &start_bytenum, &start_bitnum);
  if (rc != DX_OK)
	return(rc);

  need = new_npages - old_npages;

  /* Check if this can't possibly work. */
  if (area->physmap.avail < need)
	  return(DX_PHYSMAP_FAILED);

  /* This bitmap page has enough free space, but it may not be contiguous. */
  p = (Dx_ui8 *) area->physmap.map + start_bytenum;

  /* Find out how many available pages abut in the mapping byte. */
  tail_count = byte_info[*p].c_bits[start_bitnum] & BYTEINFO_C_BITS_COUNT;
  if (tail_count >= need) {
	/* Found enough in this mapping byte. */
	*p |= byte_mask[need - 1][start_bitnum];
	area->physmap.avail -= need;
	return(DX_OK);
  }

  if (!byte_info[*p].c_bits[start_bitnum] & BYTEINFO_C_BITS_ENDRUN)
	return(DX_PHYSMAP_FAILED);

  /* We must examine at least one more mapping byte. */
  nleft = need;
  start_mask = byte_mask[need - 1][start_bitnum];
  start_p = p++;
  ind = start_bytenum + 1;

  while (ind < seg_len && nleft) {
	b = &byte_info[*p];
	if (b->r_bits >= nleft) {
	  end_mask = byte_mask[nleft - 1][0];
	  nleft = 0;
	}
	else if (b->r_bits == 8)
	  nleft -= 8;
	else
	  break;
	p++;
	ind++;
  }

  if (nleft != 0)
	return(DX_PHYSMAP_FAILED);

  /* Set the select bits to indicate that they're allocated. */
  *start_p |= start_mask;
  for (i = 1; i < (Dx_ui32) (p - start_p); i++)
	start_p[i] = 0xff;
  *p |= end_mask;

  area->physmap.avail -= need;

  return(DX_OK);
}

/*
 * Deallocate a user data segment of length NPAGES on the given volume.
 * The given offset is page aligned.
 */
Dx_rc
Physmap_free_segment(Dx *db, Physvolnum volnum, Physoff offset, Dx_ui32 npages)
{

  return(Physmap_truncate_segment(db, volnum, offset, npages, 0));
}

/*
 * Truncate the user data segment that starts at OFFSET of length OLD_NPAGES
 * on the given volume to NEW_NPAGES.
 * If NEW_NPAGES is zero, this is equivalent to freeing the entire segment.
 */
Dx_rc
Physmap_truncate_segment(Dx *db, Physvolnum volnum, Physoff offset,
						 Dx_ui32 old_npages, Dx_ui32 new_npages)
{
  Dx_ui32 n, nfreed, nleft, start_bitnum, start_bytenum;
  Dx_ui8 bm, *p;
  Area *area;
  Dx_rc rc;

  if (old_npages < new_npages) {
	fprintf(stderr, "Physmap_truncate_segment: old_npages < new_npages!\n");
	return(DX_INTERNAL_ERROR);
  }

  if (old_npages == new_npages)
	return(DX_OK);

  rc = compute_starting_location(db, volnum, offset, new_npages, &area,
								 &start_bytenum, &start_bitnum);
  if (rc != DX_OK)
	return(rc);

  p = (Dx_ui8 *) area->physmap.map + start_bytenum;
  nleft = old_npages - new_npages;
  while (nleft) {
	if (start_bitnum == 0 && nleft >= BITS_PER_BYTE) {
	  if (DX_DEBUG && *p != 0377) {
		fprintf(stderr, "Physmap_truncate_segment: pages already free!\n");
		return(DX_INTERNAL_ERROR);
	  }
	  *p++ = 0;
	  nfreed = 8;
	}
	else {
	  n = (nleft >= BITS_PER_BYTE) ? BITS_PER_BYTE : nleft;
	  bm = byte_mask[n - 1][start_bitnum];
	  if (DX_DEBUG && (*p & bm) != bm) {
		fprintf(stderr, "Physmap_truncate_segment: bug2!\n");
		return(DX_INTERNAL_ERROR);
	  }
	  *p++ &= (~bm);
	  nfreed = Dx_minimum_ui32(BITS_PER_BYTE - start_bitnum, nleft);
	}

	nleft -= nfreed;
	start_bitnum = 0;
  }

  area->physmap.avail += (old_npages - new_npages);

  return(DX_OK);
}

static Dx_rc
find_possible_map(Dx *db, Volume *volume, Dx_ui32 npages, Area **areap)
{
  Dx_ui32 rel_areanum;
  Area *area;
  Dx_rc rc;

  /* Find a map with sufficient space. */
  rel_areanum = 0;

  /* Start from head of list or resume from a previous point. */
  if (*areap == NULL)
	area = volume->area;
  else
	area = *areap;

  while (area != NULL) {
	if (area->physmap.map == NULL) {
	  if ((rc = Physmap_init(db, area->areanum, &area->physmap)) != DX_OK)
		return(rc);
	}
	if (area->physmap.avail >= npages)
	  break;
	rel_areanum++;
	area = area->next;
  }

  if (area == NULL) {
	/* There's not enough space available in any map, so allocate a new one. */
	rc = Area_create_volume(db, volume->volnum,
							rel_areanum * db->nvolumes + volume->volnum,
							&area);
	if (rc != DX_OK)
	  return(rc);
  }

  if (area->physmap.map == NULL) {
	if ((rc = Physmap_init(db, area->areanum, &area->physmap)) != DX_OK)
	  return(rc);
  }

  *areap = area;

  return(DX_OK);
}
