/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * An ordered set of segments, each of which is called a stripe, make up
 * an indirect Dx_datum.
 * The first stripe is allocated on volume N, the second on
 * ((N + 1) mod nvolumes), the third on ((N + 2) mod nvolumes), etc.
 * Dx_datum editing may change that ordering though.
 *
 * The indirect descriptor header in an item's home page entry points to
 * first in a singly-linked chain of stripe descriptor segments (Istripes).
 * Each stripe descriptor segment starts with a header (Istripe_hdr) that
 * has a count of the number of stripe descriptors in this segment (count)
 * and the location of the next stripe descriptor segment in the chain
 * (seglen == 0 flags the last stripe descriptor segment).
 * Following the header is a set of descriptors that point to data:
 * (Physvolnum, Physoff, Seglen, Dx_len), where the Seglen indicates the
 * allocated size of the data segment and Dx_len indicates how much data
 * is actually in that data segment.
 * Editing may leave an empty descriptor slot (Seglen == Dx_len == 0), which
 * could (theoretically anyway) be reused.
 *
 * Zero filling, which happens when a hole is created by data inserted beyond
 * the end of a value, is represented by a stripe descriptor that has
 * a non-zero Dx_len but a Seglen of zero.  It is logically equivalent
 * to a segment at least Dx_len bytes in length filled with zeroes.
 *
 * XXX The stripe handling code can certainly be improved (e.g., to reuse
 * segments etc.) but more experience and measurements are needed before
 * worthwhile improvements can be made.
 */

#include "local.h"
#include "item.h"

/*
 * Locate a starting position START_POS within a datum.
 * If the position is beyond the end of the datum, return DX_BAD_POSITION
 * and information about the last stripe for the datum.
 * If either DX_BAD_REGION or DX_OK are returned, FIRST_SEG will be
 * mapped upon return.
 *
 * Return a description of the stripe descriptor segment containing START_POS
 * (HDR), the index within that stripe descriptor segment of the stripe
 * for the START_POS (FIRST_STRIPE_INDEX), and the byte index within
 * the data segment of START_POS (FIRST_STRIPE_OFFSET).
 */
static INLINE Dx_rc
stripe_locate_start(Dx *db, Physvolnum vol, Dx_len datum_dlen,
					Physoff offset, Seglen seglen,
					Dx_len start_pos, Istripe_hdr *hdr,
					Dx_len *first_stripe_offset, Dx_len *first_stripe_index)
{
  int found_start;
  Dx_ui8 *id;
  Dx_len i, pos;
  Istripe_count_t count;
  Seglen next_seglen;
  Physoff next_offset;
  Stripe stripe;
  Physvolnum next_volume;
  Physoff off;
  Segment istripe_seg;
  Seglen slen;
  Physvolnum v;
  Dx_rc rc;

  pos = 0;
  v = vol;
  off = offset;
  slen = seglen;

  i = 0;
  found_start = 0;

  while (1) {
	/* Get a stripe descriptor page. */
	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &istripe_seg);
	if (rc != DX_OK)
	  return(rc);

	id = (Dx_ui8 *) istripe_seg;
	unpack_istripe_hdr(db, id, "cosv", &count,
					   &next_offset, &next_seglen, &next_volume);

	/*
	 * Scan the stripe descriptor segment, looking for the descriptor
	 * belonging to the requested start position.
	 */
	for (; i < count; i++) {
	  unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	  if ((pos + stripe.dlen) > start_pos)
		break;
	  pos += stripe.dlen;
	}

	if (i < count) {
	  found_start = 1;
	  break;
	}

	if (next_offset == 0)
	  break;

	/* XXX unmap the segment before moving to the next one. */

	/* Continue with the next segment. */
	off = next_offset;
	slen = next_seglen;
	v = next_volume;
	i = 0;
  }

  /* If we reached the end, return the last descriptor. */
  hdr->seg = istripe_seg;
  hdr->count = count;
  hdr->vol = v;
  hdr->offset = off;
  hdr->seglen = slen;
  *first_stripe_offset = start_pos - pos;
  *first_stripe_index = i;

  if (!found_start)
	return(DX_BAD_POSITION);

  return(DX_OK);
}

/*
 * Identify the descriptor for the first stripe descriptor page for the datum,
 * taking a chunk specification into account.
 * VOL is the initial volume, OFFSET and SEGLEN identify the first stripe
 * descriptor segment, and DATUM_DLEN is the total length of the datum.
 * START_POS and END_POS identify the region of interest.
 *
 * Return the first stripe desciptor segment (FIRST_SEG), the total number of
 * segment descriptor pages needed (NSEGS),
 * the index within the first stripe descriptor page of the stripe for the
 * first byte of the chunk (FIRST_STRIPE_INDEX), and the byte index within the
 * first segment of the starting byte of the chunk (FIRST_STRIPE_OFFSET).
 * Assumes that item_verify_region() has already validated the parameters.
 */
static INLINE Dx_rc
stripe_get_stripes(Dx *db, Physvolnum vol, Dx_len datum_dlen,
				   Physoff offset, Seglen seglen,
				   Dx_len start_pos, Dx_len end_pos,
				   Dx_len *nsegs, Segment *first_seg,
				   Dx_len *first_stripe_offset, Dx_len *first_stripe_index)
{
  Dx_ui8 *id;
  Dx_len i, n, pos;
  Istripe_count_t count;
  Seglen next_seglen;
  Physoff next_offset;
  Stripe stripe;
  Physvolnum next_volume;
  Istripe_hdr hdr;
  Physoff off;
  Segment istripe_seg;
  Seglen slen;
  Physvolnum v;
  Dx_rc rc;

  rc = stripe_locate_start(db, vol, datum_dlen, offset, seglen,
						   start_pos, &hdr,
						   first_stripe_offset, first_stripe_index);
  if (rc != DX_OK)
	return(rc);

  pos = start_pos;
  *first_seg = hdr.seg;
  off = hdr.offset;
  slen = hdr.seglen;
  i = *first_stripe_index;
  n = 0;
  v = hdr.vol;

  while (off != 0) {
	/* Get a stripe descriptor page. */
	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &istripe_seg);
	if (rc != DX_OK)
	  return(rc);

	id = (Dx_ui8 *) istripe_seg;
	unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
					   &next_volume);

	n++;

	for (; i < count; i++) {
	  unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	  if ((pos + stripe.dlen) >= end_pos)
		break;
	  pos += stripe.dlen;
	}

	if (i != count)
	  break;

	off = next_offset;
	slen = next_seglen;
	v = next_volume;
	i = 0;
  }

  *nsegs = n;
  return(DX_OK);
}

/*
 * Given OLD_STRIPE, break it into two stripes.
 * Allocate a new stripe on NEW_VOL, copy bytes starting at SPLIT_OFFSET
 * until the end of the original stripe, and truncate the original stripe.
 * PTR is the old stripe's entry in the descriptor segment.
 * Verify that SPLIT_OFFSET isn't zero or beyond the length of the old
 * stripe, as that doesn't make much sense.
 *
 * NEW_STRIPE describes the allocated second part of OLD_STRIPE.
 * Returns a description of the new stripe and updates the description of
 * OLD_STRIPE.
 */
static INLINE Dx_rc
split_stripe(Dx *db, Dx_ui8 *ptr, Stripe *old_stripe, Dx_len split_offset,
			 Physvolnum new_vol, Stripe *new_stripe)
{
  Dx_ui8 *dst, *src;
  Dx_len need_bytes;
  Segment new_seg;
  Seglen new_seglen;
  Physoff new_offset;
  Dx_rc rc;

  if (split_offset == 0 || old_stripe->dlen <= split_offset)
	return(DX_INTERNAL_ERROR);

  need_bytes = old_stripe->dlen - split_offset;
  new_seglen = NBYTES_TO_PAGE_BYTES(db, need_bytes);

  rc = Segment_allocate_segment(db, db->volume[new_vol], new_seglen, 0,
								&new_seg, &new_offset);
  if (rc != DX_OK)
	return(rc);

  dst = (Dx_ui8 *) new_seg;
  src = (Dx_ui8 *) old_stripe->segment + split_offset;
  /* Safe and faster not to use memmove(). */
  Dx_memcpy(dst, src, (size_t) need_bytes);

  new_stripe->volnum = new_vol;
  new_stripe->offset = new_offset;
  new_stripe->seglen = new_seglen;
  new_stripe->dlen = need_bytes;
  new_stripe->segment = new_seg;

  /* Truncate the old segment. */
  rc = Segment_truncate(db, db->volume[old_stripe->volnum],
						old_stripe->offset,
						NBYTES_TO_NPAGES(db, old_stripe->seglen),
						NBYTES_TO_NPAGES(db, split_offset));
  if (rc != DX_OK)
	return(rc);

  old_stripe->dlen = split_offset;
  old_stripe->seglen = NBYTES_TO_PAGE_BYTES(db, old_stripe->dlen);

  pack_istripe(db, ptr, old_stripe);

  return(DX_OK);
}

/*
 * Split a descriptor segment (HDR) into two descriptor segments
 * (the new one is NEW_OFFSET, NEW_SEG, NEW_SEGLEN), where SPLIT_STRIPE_INDEX
 * is the descriptor for the stripe being split and SPLIT_OFFSET is the byte
 * offset within that stripe of the first byte being assigned to the new
 * descriptor.
 *
 * The original descriptor segment will point at the new one and the new
 * one will have a null next pointer.
 * NEW_COUNT is set to the number of descriptors in the new descriptor segment.
 */
static INLINE Dx_rc
split_descriptor_segment(Dx *db, Istripe_hdr *hdr, Dx_len split_stripe_offset,
						 Dx_len split_stripe_index, Segment *new_seg,
						 Seglen *new_seglen, Physoff *new_offset,
						 Physvolnum *new_volume, Dx_len *new_countp)
{
  Dx_ui32 i;
  Dx_ui8 *new_id, *old_id;
  Dx_len first_full_stripe_index, need_bytes, new_count, old_count;
  Physvolnum v;
  Stripe new_stripe, old_stripe, stripe;
  Dx_rc rc;

  if (hdr->count == 0 || hdr->count <= split_stripe_index)
	return(DX_INTERNAL_ERROR);

  /*
   * How many stripe descriptors will be in the new descriptor segment?
   * If the split point is within a stripe, a new stripe descriptor is
   * needed for its second part.
   */
  *new_countp = split_stripe_index;
  first_full_stripe_index = split_stripe_index;
  old_count = hdr->count - split_stripe_index - 1;

  if (split_stripe_offset != 0) {
	(*new_countp)++;
	first_full_stripe_index++;
	old_count++;
  }

  if (first_full_stripe_index == 0) {
	*new_seg = NULL;
	*new_seglen = 0;
	return(DX_OK);
  }

  /* Allocate the new descriptor segment. */
  need_bytes = ISTRIPE_HDR_SIZE + (*new_countp * ISTRIPE_SIZE);
  *new_seglen = NBYTES_TO_PAGE_BYTES(db, need_bytes);

  /* XXX On which volume should this new segment be allocated? */
  v = Xrand(0, db->nvolumes - 1);
  rc = Segment_allocate_segment(db, db->volume[v], *new_seglen, 0,
								new_seg, new_offset);
  if (rc != DX_OK)
	return(rc);

  /* Initialize its header. */
  old_id = (Dx_ui8 *) hdr->seg;
  new_id = (Dx_ui8 *) (*new_seg);
  pack_istripe_hdr(db, new_id, "cosv", *new_countp, (Physoff) 0,
				   (Seglen) 0, (Physvolnum) 0);

  /* If necessary, split the stripe at the insertion point. */
  if (split_stripe_offset != 0) {
	unpack_istripe(db, ISTRIPE_PTR(old_id, split_stripe_index), &old_stripe);
	rc = Segment_map(db, db->volume[old_stripe.volnum], old_stripe.offset,
					 old_stripe.seglen, 0, 0, 0, &old_stripe.segment);
	if (rc != DX_OK)
	  return(rc);
	rc = split_stripe(db, ISTRIPE_PTR(old_id, split_stripe_index), &old_stripe,
					  split_stripe_offset, v, &new_stripe);
	if (rc != DX_OK)
	  return(rc);
	pack_istripe(db, ISTRIPE_PTR(new_id, 0), &new_stripe);
	new_count = 1;
  }
  else
	new_count = 0;

  /* Copy stripe descriptors from the old descriptor segment to the new one. */
  for (i = first_full_stripe_index; i < hdr->count; i++) {
	unpack_istripe(db, ISTRIPE_PTR(old_id, i), &stripe);
	pack_istripe(db, ISTRIPE_PTR(new_id, new_count), &stripe);
	new_count++;
  }

  pack_istripe_hdr(db, old_id, "cosv", old_count, *new_offset, *new_seglen, v);
  *new_volume = v;

  return(DX_OK);
}

/*
 * Allocate a sufficient number of stripes that total NBYTES bytes, starting
 * with the given volume.
 * Return the number of stripes and an initialized vector of stripe
 * descriptors.
 */
Dx_rc
Stripe_allocate(Dx *db, Dx_len nbytes, Istripe_count_t nstripes,
				Stripe *stripes)
{
  Dx_ui32 bytes_left, i, stripe_nbytes, stripe_npages;
  Physvolnum v;
  Volume *vol;
  Dx_rc rc;

  bytes_left = nbytes;

  i = 0;
  while (bytes_left > 0) {
	stripe_nbytes = (bytes_left <= ITEM_INDIRECT_STRIPE_NBYTES(db))
	  ? bytes_left : ITEM_INDIRECT_STRIPE_NBYTES(db);
	stripe_npages = NBYTES_TO_NPAGES(db, stripe_nbytes);

	if (i == nstripes) {
	  fprintf(stderr, "Stripe_allocate: istripes[] overflow!\n");
	  return(DX_INTERNAL_ERROR);
	}

	vol = Random_volume(db);
	v = vol->volnum;

	stripes[i].volnum = v;
	stripes[i].seglen = NPAGES_TO_NBYTES(db, stripe_npages);
	stripes[i].dlen = stripe_nbytes;

	rc = Segment_allocate_segment(db, db->volume[v], stripe_nbytes,
								  SEGMENT_FIXED,
								  &stripes[i].segment, &stripes[i].offset);
	if (rc != DX_OK) {
	  fprintf(stderr, "Stripe_allocate: alloc failed!\n");
	  /* XXX Free allocations and return */
	  return(rc);
	}
	bytes_left -= stripe_nbytes;
	i++;
  }

  return(DX_OK);
}

/*
 * Stream a contiguous region of the stripes for a datum.
 * VOL is the starting volume, OFFSET and SEGLEN describe the first
 * stripe descriptor segment, DLEN is the total length of the datum, and
 * CHUNKP describes the region to be extracted.
 */
Dx_rc
Stripe_contig(Dx *db, Physvolnum vol, Physoff offset,
			  Seglen seglen, Dx_len dlen, Dx_region *chunkp, Ehash_io *io)
{
  Dx_ui8 *ptr;
  Physvolnum next_volume;
  Dx_len nleft;
  Seglen next_seglen;
  Physoff next_offset;
  Istripe_count_t count;
  Segment seg;
  Segment first_seg;
  Stripe stripe;
  Dx_desc *desc;
  Dx_len chunk_dlen, end_pos, start_pos;
  Dx_len first_stripe_index, fsi, first_stripe_offset, fso;
  Dx_len i, len, nsegs, st;
  Dx_ui8 *id;
  Dx_rc rc;

  if (chunkp == NULL) {
	start_pos = 0;
	end_pos = dlen - 1;
	chunk_dlen = dlen;
  }
  else {
	rc = item_verify_region(dlen, chunkp, &start_pos, &end_pos, &chunk_dlen);
	if (rc != DX_OK)
	  return(rc);
  }

  rc = stripe_get_stripes(db, vol, dlen, offset, seglen,
						  start_pos, end_pos, &nsegs, &first_seg,
						  &first_stripe_offset, &first_stripe_index);
  if (rc != DX_OK)
	return(rc);

  seg = first_seg;
  fso = first_stripe_offset;
  fsi = first_stripe_index;
  nleft = chunk_dlen;
  desc = &io->desc;

  while (nleft) {
	/*
	 * Look at each of the nsegs consecutive descriptors that point to data
	 * that make up the chunk.
	 * The first one (index fsi within the first segment of descriptors) is
	 * special because the first byte of the chunk is byte fso within that
	 * descriptor.
	 */
	id = (Dx_ui8 *) seg;
	unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
					   &next_volume);

	for (i = fsi; nleft && i < count; i++) {
	  fsi = 0;
	  unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	  if (stripe.seglen == 0)
		rc = Segment_map_zeroes(db, stripe.dlen, PROT_READ, &seg);
	  else
		rc = Segment_map(db, db->volume[stripe.volnum], stripe.offset,
						 stripe.seglen, 0, 0, 0, &seg);
	  if (rc != DX_OK)
		return(rc);

	  if (fso) {
		ptr = (Dx_ui8 *) seg + fso;
		len = stripe.dlen - fso;
		fso = 0;
	  } else {
		ptr = (Dx_ui8 *) seg;
		len = stripe.dlen;
	  }

	  len = (len <= nleft) ? len : nleft;
	  if (len == 0)
		continue;

	  /* Call the put-function (flags and dlen already set). */
	  desc->error_occurred = 0;
	  desc->last_errno = 0;
	  st = (io->uio->func)(io->uio->arg, (void *) ptr, len, desc);

	  /* Release the segment, then check for errors. */
	  if (stripe.seglen == 0)
		rc = Segment_unmap_zeroes(db, stripe.dlen);
	  else
		rc = Segment_unmap(db, db->volume[stripe.volnum], stripe.offset,
						   stripe.segment, stripe.seglen, 1);

	  if (st != len || desc->error_occurred) {
		db->last_errno = desc->last_errno;
		return(DX_USERIO_ERROR);
	  }
	  if (rc != DX_OK)
		return(rc);

	  nleft -= len;
	}

	if (nleft) {
	  if (next_offset && next_seglen) {
		rc = Segment_map(db, db->volume[next_volume], next_offset, next_seglen,
						 0, 0, 0, &seg);
		if (rc != DX_OK)
		  return(rc);
	  }
	}
  }

  /* The actual number of bytes retrieved from the Dx_datum. */
  io->dlen = chunk_dlen;

  return(DX_OK);
}

Dx_rc
Stripe_append_data(Dx *db, Physvolnum vol, Physoff offset,
				   Seglen seglen, Dx_datum *datum, Dx_len dlen,
				   Dx_desc *desc, Ehash_xarg *arg)
{

  return(DX_NOTIMPLEMENTED);
}

/*
 * Insert data within a datum.
 * Given a starting position within a datum (START_POS), insert the new data.
 * Insertion is within or beyond the end of the datum.
 * ID is the datum's entry meta data.
 *
 * This is an important routine because it affects stripe balancing
 * and, consequently, the performance of any future changes to the datum.
 *
 * Caller updates size of datum.
 */
Dx_rc
Stripe_insert_data(Dx *db, Physvolnum vol, Dx_ui8 *id, Physoff offset,
				   Seglen seglen, Dx_datum *datum, Dx_len dlen,
				   Dx_len start_pos, Ehash_xarg *arg)
{
  Dx_ui8 *new_id, *old_id;
  Dx_len first_stripe_offset, first_stripe_index, zero_fill;
  Istripe_hdr hdr;
  Segment new_seg;
  Seglen istripe_seglen;
  Physoff istripe_offset, split_offset;
  Physvolnum split_volume;
  Dx_rc rc;

  if (datum->dptr == NULL && (arg == NULL || arg->io.uio == NULL))
	return(DX_OK);

  /* Find out where the insertion begins. */
  rc = stripe_locate_start(db, vol, dlen, offset, seglen, start_pos,
						   &hdr, &first_stripe_offset, &first_stripe_index);
  if (rc != DX_OK && rc != DX_BAD_POSITION)
	return(rc);

  if (start_pos > dlen) {
	/*
	 * This is an insertion that requires zero filling.
	 */
	zero_fill = start_pos - dlen;
  }
  else
	zero_fill = 0;

  /* Acquire the new data and allocate a new set of stripes. */
  rc = Item_stripe_copy(db, db->volume[vol], datum, zero_fill, &new_seg,
						&istripe_offset, &istripe_seglen, arg);
  if (rc != DX_OK)
	return(rc);

  if (datum->dlen == dlen) {
	/* Nothing was inserted. */
	return(DX_OK);
  }

  /*
   * There are several cases to consider:
   *   1. Insertion is at the start of the datum (start_pos == 0)
   *   2. Insertion is at the end of the datum (start_pos == dlen)
   *   3. Insertion is at the start of a stripe (first_stripe_offset == 0)
   *   4. Insertion is within a stripe (first_stripe_offset != 0)
   *   5. Insertion is beyond the end of the datum (start_pos > dlen)
   */

  new_id = (Dx_ui8 *) new_seg;
  old_id = (Dx_ui8 *) hdr.seg;

  /* Case 1 */
  if (start_pos == 0) {
	Physoff orig_off;
	Seglen orig_seglen;
	Physvolnum orig_initial_volume;

	unpack_idescriptor_fast(db, id, NULL, &orig_off, &orig_seglen,
							&orig_initial_volume);
	pack_idescriptor_fast(db, id, NULL, &istripe_offset, &istripe_seglen, &vol);
	pack_istripe_hdr(db, new_id, "osv", orig_off, orig_seglen,
					 orig_initial_volume);
	return(DX_OK);
  }

  /* Cases 2 and 5 */
  if (start_pos >= dlen) {
	/* The new segment descriptor will already have a NULL "next" pointer. */
	pack_istripe_hdr(db, old_id, "osv", istripe_offset, istripe_seglen, vol);
	return(DX_OK);
  }

  /* Cases 3 and 4 */
  if (start_pos < dlen) {
	/*
	 * The new stripes are being inserted "in between" existing stripes.
	 * We could grow the stripe descriptor segment or we could split it
	 * in two, chaining the new stripe descriptor segment between the two
	 * parts.
	 * We choose to split it.
	 */
	Segment split_seg;
	Seglen split_seglen;
	Dx_len split_count;

	rc = split_descriptor_segment(db, &hdr, first_stripe_offset,
								  first_stripe_index, &split_seg,
								  &split_seglen, &split_offset, &split_volume,
								  &split_count);
	if (rc != DX_OK)
	  return(rc);

	/* Do the insertion. */
	pack_istripe_hdr(db, new_id, "osv", split_offset, split_seglen, split_volume);
	pack_istripe_hdr(db, old_id, "osv", istripe_offset, istripe_seglen, vol);
	return(DX_OK);
  }

  /* Shouldn't reach here. */
  return(DX_NOTIMPLEMENTED);
}

/*
 * Overwrite data within a datum.
 * Given a starting position within a datum (START_POS), overwrite the
 * new data.
 */
Dx_rc
Stripe_overwrite_data(Dx *db, Physvolnum vol, Physoff offset,
					  Seglen seglen, Dx_len dlen, Dx_len start_pos,
					  Dx_desc *desc, Ehash_io *io)
{

  return(DX_NOTIMPLEMENTED);
}

/*
 * Delete a region within a datum.
 * Given a starting position (START_POS), delete DELETE_LENGTH bytes.
 * Caller updates size of datum.
 *
 * XXX stripe descriptor segments are not deleted.
 */
Dx_rc
Stripe_delete_data(Dx *db, Physvolnum vol, Physoff offset,
				   Seglen seglen, Dx_len dlen, Dx_len start_pos,
				   Dx_len delete_length, Dx_desc *desc, Ehash_io *io)
{
  Dx_ui8 *dst, *id, *src;
  Physvolnum next_volume;
  Seglen next_seglen;
  Physoff next_offset;
  Istripe_count_t count, i, j;
  Dx_len first_stripe_offset, first_stripe_index;
  Dx_len amount_deleted, nsegs, tail_length;
  Segment first_seg, seg;
  Stripe stripe;
  Dx_rc rc;

  if (start_pos >= dlen)
	return(DX_INTERNAL_ERROR);
  if (delete_length == 0)
	return(DX_OK);

  rc = stripe_get_stripes(db, vol, dlen, offset, seglen, start_pos,
						  start_pos + delete_length, &nsegs, &first_seg,
						  &first_stripe_offset, &first_stripe_index);
  if (rc != DX_OK)
	return(rc);

  /*
   *  first_seg
   *     |
   *     V
   * stripe(0) ----->
   * stripe(1) ----->
   *  ...                        --- first_stripe_offset (start_pos)
   *                            /
   *                     -------------------------------------
   * stripe(fsi)  ----->| .....|  <- delete_length  -> | .... |
   *                     -------------------------------------
   *  ...
   * stripe(fsi + nsegs - 1) ----->
   */

  id = (Dx_ui8 *) first_seg;
  unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
					 &next_volume);
  unpack_istripe(db, ISTRIPE_PTR(id, first_stripe_index), &stripe);

  /*
   * If the entire first stripe will be deleted through to its end, just
   * truncate it.  Otherwise, part of the first stripe needs to be kept.
   * If the "middle" part of a stripe is being kept, the second part needs
   * to be moved to abut the first part.
   */
  if (delete_length >= stripe.dlen) {
	/* Truncate the segment.  */
	amount_deleted = stripe.dlen - first_stripe_offset;
	stripe.dlen = first_stripe_offset;
	rc = Segment_truncate(db, db->volume[stripe.volnum], stripe.offset,
						  NBYTES_TO_NPAGES(db, stripe.seglen),
						  NBYTES_TO_NPAGES(db, stripe.dlen));
	if (rc != DX_OK)
	  return(rc);
	stripe.seglen = NBYTES_TO_PAGE_BYTES(db, stripe.dlen);
	if (stripe.dlen == 0)
	  stripe.segment = NULL;
	pack_istripe(db, ISTRIPE_PTR(id, first_stripe_index), &stripe);
  }
  else {
	/* Less than the remainder of the entire segment is being deleted. */
	rc = Segment_map(db, db->volume[stripe.volnum], stripe.offset,
					 stripe.seglen, 0, 0, 0, &stripe.segment);
	if (rc != DX_OK)
	  return(rc);

	amount_deleted = delete_length;
	dst = (Dx_ui8 *) stripe.segment + first_stripe_offset;
	src = (Dx_ui8 *) stripe.segment + first_stripe_offset + delete_length;
	tail_length = stripe.dlen - first_stripe_offset - delete_length;
	memmove(dst, src, (size_t) tail_length);
	stripe.dlen = first_stripe_offset + tail_length;
	rc = Segment_truncate(db, db->volume[stripe.volnum], stripe.offset,
						  NBYTES_TO_NPAGES(db, stripe.seglen),
						  NBYTES_TO_NPAGES(db, stripe.dlen));
	if (rc != DX_OK)
	  return(rc);
	stripe.seglen = NBYTES_TO_PAGE_BYTES(db, stripe.dlen);
	pack_istripe(db, ISTRIPE_PTR(id, first_stripe_index), &stripe);
  }

  /* Truncate all fully deleted stripes.  */
  seg = first_seg;
  j = first_stripe_index + 1;
  for (i = 1; i < nsegs - 1; i++) {
	if (j == count) {
	  /* Next descriptor page. */
	again:
	  if (next_offset == 0) {
		/* XXX ??? */
		return(DX_INTERNAL_ERROR);
	  }
	  rc = Segment_map(db, db->volume[next_volume], next_offset, next_seglen,
					   0, 0, 0, &seg);
	  if (rc != DX_OK)
		return(rc);
	  id = (Dx_ui8 *) seg;
	  unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
						 &next_volume);
	  if (count == 0)
		goto again;
	  j = 0;
	}

	unpack_istripe(db, ISTRIPE_PTR(id, j), &stripe);
	rc = Segment_free(db, db->volume[stripe.volnum],
					  stripe.offset, NBYTES_TO_NPAGES(db, stripe.seglen));
	if (rc != DX_OK)
	  return(rc);
	amount_deleted += stripe.dlen;
	stripe.dlen = 0;
	stripe.seglen = 0;
	stripe.segment = NULL;
	pack_istripe(db, ISTRIPE_PTR(id, j), &stripe);
	j++;
  }

  /* Handle the last partial stripe, if necessary.  */
  if (amount_deleted < delete_length) {
	if (j == count) {
	  /* Next descriptor page. */
	again2:
	  if (next_offset == 0) {
		/* XXX ??? */
		return(DX_INTERNAL_ERROR);
	  }
	  rc = Segment_map(db, db->volume[next_volume], next_offset, next_seglen,
					   0, 0, 0, &seg);
	  if (rc != DX_OK)
		return(rc);
	  id = (Dx_ui8 *) seg;
	  unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
						 &next_volume);
	  if (count == 0)
		goto again2;
	  j = 0;
	}
	unpack_istripe(db, ISTRIPE_PTR(id, j), &stripe);
	rc = Segment_map(db, db->volume[stripe.volnum], stripe.offset,
					 stripe.seglen, 0, 0, 0, &stripe.segment);
	if (rc != DX_OK)
	  return(rc);

	tail_length = stripe.dlen - (delete_length - amount_deleted);
	dst = (Dx_ui8 *) stripe.segment;
	src = (Dx_ui8 *) stripe.segment + (stripe.dlen - tail_length);
	memmove(dst, src, (size_t) tail_length);
	stripe.dlen = tail_length;
	rc = Segment_truncate(db, db->volume[stripe.volnum], stripe.offset,
						  NBYTES_TO_NPAGES(db, stripe.seglen),
						  NBYTES_TO_NPAGES(db, stripe.dlen));
	if (rc != DX_OK)
	  return(rc);
	stripe.seglen = NBYTES_TO_PAGE_BYTES(db, stripe.dlen);
	pack_istripe(db, ISTRIPE_PTR(id, j), &stripe);
	amount_deleted += (delete_length - amount_deleted);
  }

  return(DX_OK);
}

/*
 * Remove null (unused) stripe descriptors and return the new count.
 */
static INLINE Istripe_count_t
stripe_compact_descriptor_segment(Dx *db, Dx_ui8 *id,
								  Istripe_count_t count)
{
  int do_copy;
  Istripe_count_t i, low_index, new_count;
  Stripe stripe;

  if (count == 0)
	return(0);

  low_index = 0;

  new_count = 0;
  do_copy = 0;
  for (i = 0; i < count; i++) {
	unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	if (stripe.offset != 0) {
	  if (do_copy) {
		pack_istripe(db, ISTRIPE_PTR(id, low_index), &stripe);
		low_index++;
	  }
	  new_count++;
	  continue;
	}
	else if (do_copy == 0) {
	  low_index = i;
	  do_copy = 1;
	}
  }

  return(new_count);
}

/*
 * Copy DST_COUNT stripes, starting at zero, from DST_ID to
 * SRC_ID, starting at SRC_COUNT.
 */
static INLINE void
stripe_compact_copy_stripes(Dx *db, Dx_ui8 *dst_id, Istripe_count_t dst_count,
							Dx_ui8 *src_id, Istripe_count_t src_index)
{
  Istripe_count_t d, s;
  Stripe stripe;

  for (d = dst_count, s = 0; s < src_index; s++, d++) {
	unpack_istripe(db, ISTRIPE_PTR(src_id, s), &stripe);
	pack_istripe(db, ISTRIPE_PTR(dst_id, d), &stripe);
  }
}

/*
 * Consolidate the stripe descriptor segments for a datum.
 * The first such segment (VOL, OFFSET, SEGLEN) is described.
 * Editing can cause a list of stripe descriptor segments to become
 * fragmented.
 *
 * This isn't the most efficient algorithm, but I'm not entirely
 * sure about which approach is best right now.  In the meantime, this
 * algorithm is simple and most important, correct.
 */
Dx_rc
Stripe_compact(Dx *db, Physvolnum vol, Physoff offset, Seglen seglen)
{
  Dx_ui8 *id, *next_id;
  Physvolnum next_volume;
  Seglen next_seglen, slen;
  Physoff next_offset, off;
  Istripe_count_t avail, c, count, new_count, next_count;
  Physvolnum v;
  Segment next_seg, seg;
  Dx_rc rc;

  if (seglen == 0 || offset == 0)
	return(DX_OK);

  off = offset;
  slen = seglen;
  v = vol;
  while (off != 0) {
	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &seg);
	if (rc != DX_OK)
	  return(rc);

	id = (Dx_ui8 *) seg;
	unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
					   &next_volume);
	new_count = stripe_compact_descriptor_segment(db, id, count);
	if (new_count != count) {
	  pack_istripe_hdr(db, id, "c", &new_count);
	  count = new_count;
	}

	avail = ISTRIPE_ISTRIPES_ALLOC(slen) - count;
	while (avail != 0 && next_offset != 0) {
	  rc = Segment_map(db, db->volume[next_volume], next_offset, next_seglen,
					   0, 0, 0, &next_seg);
	  if (rc != DX_OK)
		return(rc);

	  next_id = (Dx_ui8 *) next_seg;
	  unpack_istripe_hdr(db, next_id, "cosv", &next_count, &next_offset,
						 &next_seglen, &next_volume);

	  new_count = stripe_compact_descriptor_segment(db, next_id, next_count);
	  if (new_count != next_count) {
		pack_istripe_hdr(db, id, "c", &new_count);
		next_count = new_count;
	  }
	  if (new_count == 0)
		break;

	  if (next_count >= avail)
		c = avail;
	  else
		c = next_count;
	  stripe_compact_copy_stripes(db, id, count, next_id, c);
	  count += c;
	  avail -= c;
	  pack_istripe_hdr(db, id, "c", count);
	  next_count -= c;
	  pack_istripe_hdr(db, next_id, "c", next_count);  
	}

	off = next_offset;
	slen = next_seglen;
	v = next_volume;
  }

  /*
   * There may be stripe descriptor segments with no stripes at the end.
   * Unlink and free them.
   */
  rc = Segment_map(db, db->volume[vol], offset, seglen, 0, 0, 0, &seg);
  if (rc != DX_OK)
	return(rc);

  id = (Dx_ui8 *) seg;
  unpack_istripe_hdr(db, id, "osv", &off, &slen, &v);
  while (off != 0) {
	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &seg);
	if (rc != DX_OK)
	  return(rc);

	next_id = (Dx_ui8 *) seg;
	unpack_istripe_hdr(db, next_id, "cosv", &next_count, &next_offset,
					   &next_seglen, &next_volume);
	if (next_count == 0) {
	  pack_istripe_hdr(db, id, "osv", next_offset, next_seglen, next_volume);
	  rc = Segment_free(db, db->volume[v], off, NBYTES_TO_NPAGES(db, slen));
	  if (rc != DX_OK)
		return(rc);
	}
	off = next_offset;
	slen = next_seglen;
	v = next_volume;
  }

  return(DX_OK);
}

/*
 * Deallocate all the stripes for a datum.
 */
Dx_rc
Stripe_free(Dx *db, Physvolnum vol, Physoff offset, Seglen seglen)
{
  Physvolnum next_volume;
  Seglen next_seglen, slen;
  Physoff next_offset, off;
  Istripe_count_t count, i;
  Physvolnum v;
  Segment seg;
  Stripe stripe;
  Dx_ui8 *id;
  Dx_rc rc;

  if (seglen == 0)
	return(DX_OK);

  off = offset;
  slen = seglen;
  v = vol;

  while (off != 0) {
	/* Get the stripe descriptor page. */
	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &seg);
	if (rc != DX_OK)
	  return(rc);

	id = (Dx_ui8 *) seg;
	unpack_istripe_hdr(db, id, "cosv", &count, &next_offset, &next_seglen,
					   &next_volume);

	for (i = 0; i < count; i++) {
	  unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	  if (stripe.seglen != 0) {
		rc = Segment_free(db, db->volume[stripe.volnum],
						  stripe.offset, NBYTES_TO_NPAGES(db, stripe.seglen));
		if (rc != DX_OK)
		  return(rc);
	  }
	}

	rc = Segment_free(db, db->volume[v], off, NBYTES_TO_NPAGES(db, slen));
	if (rc != DX_OK)
	  return(rc);

	off = next_offset;
	slen = next_seglen;
	v = next_volume;
  }

  return(DX_OK);
}

/*
 * Deallocate a given set of stripes within a single istripe segment.
 */
Dx_rc
Stripe_free_set(Dx *db, Stripe *stripes, Istripe_count_t count)
{
  Istripe_count_t i;
  Stripe *stripe;
  Dx_rc rc;

  for (i = 0; i < count; i++) {
	stripe = &stripes[i];
	if ((rc = Segment_free(db, db->volume[stripe->volnum], stripe->offset,
						   NBYTES_TO_NPAGES(db, stripe->seglen))) != DX_OK)
	  return(rc);
  }

  return(DX_OK);
}

static void
stripe_segment_show(Dx *db, Segment seg)
{
  Dx_ui8 *id;
  Dx_len i;
  Istripe_count_t count;
  Stripe stripe;
  Seglen next_seglen;
  Physoff next_offset;
  Physvolnum next_volume;

  id = (Dx_ui8 *) seg;
  unpack_istripe_hdr(db, id, "cosv", &count,
					 &next_offset, &next_seglen, &next_volume);
  fprintf(stderr,
		  " count=%lu next_offset=%llu next_seglen=%llu next_volume=%lu\n",
		  count, (unsigned long long) next_offset,
		  (unsigned long long) next_seglen, next_volume);

  /*
   * Scan the stripe descriptor segment, looking for the descriptor
   * belonging to the requested start position.
   */
  for (i = 0; i < count; i++) {
	unpack_istripe(db, ISTRIPE_PTR(id, i), &stripe);
	fprintf(stderr, "  %llu: vol=%lu offset=%llu seglen=%llu dlen=%llu\n",
			(unsigned long long) i,
			stripe.volnum, (unsigned long long) stripe.offset,
			(unsigned long long) stripe.seglen,
			(unsigned long long) stripe.dlen);
  }

}

void
stripe_show(Dx *db, Dx_ui8 *datum_id)
{
  Dx_ui8 *id;
  Dx_len dlen, n;
  Physoff off, offset;
  Seglen seglen, slen;
  Segment istripe_seg;
  Physvolnum initial_volume, v;
  Seglen next_seglen;
  Physoff next_offset;
  Physvolnum next_volume;
  Dx_rc rc;

  unpack_idescriptor_fast(db, datum_id, &dlen, &offset, &seglen, &initial_volume);
  off = offset;
  slen = seglen;
  v = initial_volume;
  n = 0;
  while (1) {
	fprintf(stderr, "Stripe descriptor segment %llu (%llu %llu %llu %lu)\n",
			(unsigned long long) n, (unsigned long long) dlen,
			(unsigned long long) off, (unsigned long long) slen, v);

	rc = Segment_map(db, db->volume[v], off, slen, 0, 0, 0, &istripe_seg);
	if (rc != DX_OK)
	  return;

	stripe_segment_show(db, istripe_seg);

	id = (Dx_ui8 *) istripe_seg;
	unpack_istripe_hdr(db, id, "osv", &next_offset, &next_seglen, &next_volume);

	if (next_offset == 0)
	  break;

	off = next_offset;
	slen = next_seglen;
	v = next_volume;
	n++;
  }
}
