/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Copy a file that may have holes, preserving them in the copy.
 * When present, the holes are assumed to be aligned to PAGE_SIZE.
 *
 * Usage: dxcp inputfile outputfile
 */

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/file.h>
#include <stdio.h>

/*
 * PAGE_SIZE should not be bigger than the minimum hole size.
 * For a Dxstore db, that is 4096.
 * It must be a multiple of sizeof(unsigned long).
 */
enum { PAGE_SIZE = 4096 };

char buf[PAGE_SIZE];

/* Copies of argv[0] ... for error messages. */
char *arg0, *arg1, *arg2;

void Lseek(int fd, off_t offset, int whence);
void Write(int fd, const void *bufp, size_t count);
int Is_hole(void);

int
main(int argc, char **argv)
{
	register int rd, fdin, fdout, nblks, nholes, prev_was_hole;

	arg0 = argv[0];
	arg1 = argv[1];
	arg2 = argv[2];

	if (argc != 3) {
		fprintf(stderr, "Usage: dxcp inputfile outputfile\n");
		exit(1);
	}

	if ((fdin = open(argv[1], O_RDONLY, 0)) == -1) {
		fprintf(stderr, "%s: %s: %s\n", arg0, arg1, strerror(errno));
		exit(1);
	}

	if ((fdout = creat(argv[2], 0644)) == -1) {
		fprintf(stderr, "%s: %s: %s\n", arg0, arg2, strerror(errno));
		exit(1);
	}

	nblks = nholes = 0;
	prev_was_hole = 0;
	while ((rd = read(fdin, buf, PAGE_SIZE)) == PAGE_SIZE) {
		if (Is_hole()) {
			Lseek(fdout, PAGE_SIZE, SEEK_CUR);
			nholes++;
			prev_was_hole = 1;
		}
		else {
			Write(fdout, buf, PAGE_SIZE);
			prev_was_hole = 0;
		}
		nblks++;
	}

	if (rd == 0) {
		if (prev_was_hole && nblks) {
			/* Fill in the last chunk to make the file size correct. */
			Lseek(fdout, -PAGE_SIZE, SEEK_CUR);
			memset(buf, 0, PAGE_SIZE);
			Write(fdout, buf, PAGE_SIZE);
		}
	}
	else if (rd == -1) {
		fprintf(stderr, "%s: %s: %s\n", arg0, arg1, strerror(errno));
		exit(1);
	}
	else {
		fprintf(stderr,
			"Warning: input file size is not a multiple of PAGE_SIZE\n");
		Write(fdout, buf, rd);
	}

	fprintf(stderr,
		"%d blocks written (%d blocks are holes, %d bytes per block)\n",
		nblks, nholes, PAGE_SIZE);

	close(fdin);
	close(fdout);

	exit(0);
}

int
Is_hole(void)
{
	register unsigned long *p;

	p = (unsigned long *) buf;
	while (p < (unsigned long *) &buf[PAGE_SIZE]) {
		if (*p++ != 0)
			return(0);
	}

	return(1);
}

void
Write(int fd, const void *bufp, size_t count)
{
	int wr;

	wr = write(fd, bufp, count);

	if (wr != count) {
		if (wr == -1)
			fprintf(stderr, "%s: %s: %s\n", arg0, arg2, strerror(errno));
		else
			fprintf(stderr, "%s: short write\n", arg0);
		exit(1);
	}

}

void
Lseek(int fd, off_t offset, int whence)
{

	if (lseek(fd, offset, whence) == -1) {
		fprintf(stderr, "%s: %s: %s\n", arg0, arg2, strerror(errno));
		exit(1);
	}
}

