/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Store items, taking each value from a file.  File names are read from stdin.
 * Usage: dxstore [-t] volume [volume ...]
 *
 * Where:
 *   the optional -t flag causes the volumes to be truncated if they exist
 *   volume [volume ...] are the volumes of the db
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>

#include "dx.h"

char buf[102400];

#define MAX_VOLUMES		100

static void add(Dx *dbm);
static int add_file(Dx *dbm, char *filename, int len);
Dx_len io_getfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc);

int
main(int argc, char **argv)
{
    int i;
	Dx *dbm;
	Dx_config config;
	Dx_rc rc;
	char *volumes[MAX_VOLUMES];

    if (argc < 2) {
        fprintf(stderr, "Usage: dxstore [-t] volume [volume ...]\n");
        exit(1);
    }

    Dx_get_configuration(NULL, &config);
    config.flags = DX_CONFIG_DEFAULT_FLAGS;
	config.max_stripe_npages = 512;
    if (strcmp(argv[1], "-t") == 0) {
      argv++;
      argc--;
      config.flags |= DX_CONFIG_TRUNCATE_ON_OPEN;
    }

    if (argc == 1) {
        fprintf(stderr, "Usage: dxstore [-t] volume [volume ...]\n");
        exit(1);
    }

   if ((argc - 1) >= MAX_VOLUMES) {
     fprintf(stderr, "Too many volumes\n");
     exit(1);
   }

    for (i = 1; i < argc; i++)
      volumes[i-1] = argv[i];
    volumes[i-1] = NULL;

	if ((rc = Dx_open(volumes, &config, &dbm)) != DX_OK) {
	  fprintf(stderr, "dxstore: %s\n", Dx_strerror(rc));
	  exit(1);
	}

    add(dbm);

	if ((rc = Dx_close(dbm)) != DX_OK) {
	  fprintf(stderr, "dxstore: %s\n", Dx_strerror(rc));
	  exit(1);
	}

	exit(0);
}

int nwords;

static void
add(Dx *dbm)
{
  int len;

  nwords = 0;
  while (fgets(buf, sizeof(buf), stdin) != NULL) {
    len = strlen(buf);
    len--;
    buf[len] = '\0';
    if (add_file(dbm, buf, len) != -1)
		nwords++;
  }
}

struct io_info {
  int fd;
  Dx_len count;
};

static int
add_file(Dx *dbm, char *filename, int len)
{
  int fd;
  Dx_datum key;
  Dx_user_io uio;
  struct io_info info;
  Dx_rc rc;

  fprintf(stderr, "%s: ", filename);
  if ((fd = open(filename, O_RDONLY)) == -1) {
	perror("open");
	return(-1);
  }

  Dx_key(&key, filename, len + 1);

  info.count = 0;
  info.fd = fd;
  uio.func = io_getfunc;
  uio.arg = (void *) &info;

  rc = Dx_storex(dbm, &key, DX_REPLACE | DX_UIO, &uio);
  if (rc != DX_OK) {
	  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  exit(1);
  }

  close(fd);
  fd = -1;

  fprintf(stderr, "%lu bytes\n", info.count);
  return(0);
}

Dx_len
io_getfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc)
{
  int fd, n;
  struct io_info *info;
  struct stat sbuf;

  info = (struct io_info *) arg;
  fd = info->fd;

  if (fd == -1)
    return(0);

  if (info->count == 0) {
	if (fstat(fd, &sbuf) == -1) {
	  desc->error_occurred = 1;
	  desc->last_errno = errno;
	  return(0);
	}
	desc->flags = DX_DATUM_ALIGN0;
	desc->dlen = sbuf.st_size;
  }

  /* XXX Assume n == dlen unless EOF, not necessarily true. */
  n = read(fd, dptr, (size_t) dlen);

  if (n == -1) {
	desc->error_occurred = 1;
	desc->last_errno = errno;
  }
  else
	info->count += n;

  return((Dx_len) n);
}
