/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>

#include "dx.h"

char buf[102400];

static void add_words(Dx *dbm, int flag, int do_read);
static int add_file(Dx *dbm, char *filename, int len, int do_read);
Dx_len io_getfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc);

int
main(int argc, char **argv)
{
	Dx *dbm;
	Dx_config config;
	Dx_rc rc;
	char *volumes[2];

    if (argc != 3 && argc != 4) {
        fprintf(stderr, "Usage: mkdx -w[kvf=] dx-file [doread]\n");
        exit(1);
    }

    Dx_get_configuration(NULL, &config);
    config.flags = DX_CONFIG_DEFAULT_FLAGS | DX_CONFIG_TRUNCATE_ON_OPEN;
	/*config.page_size = 32768;*/
    volumes[0] = argv[2];
    volumes[1] = NULL;
	if ((rc = Dx_open(volumes, &config, &dbm)) != DX_OK) {
	  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  exit(1);
	}

    add_words(dbm, argv[1][2], argv[3] != NULL);

	Dx_close(dbm);

	exit(0);
}

int nwords;

static void
add_words(Dx *dbm, int flag, int do_read)
{
  int len, rc;
  char v[10];
  Dx_datum key, value;

  nwords = 0;
  while (fgets(buf, sizeof(buf), stdin) != NULL) {
    len = strlen(buf);
    len--;
    buf[len] = '\0';
    sprintf(v, "%d", nwords);
    if (flag == '\0' || flag == 'k') {
	  Dx_key(&key, buf, len + 1);
	  Dx_value(&value, v, strlen(v) + 1);
    }
    else if (flag == '=') {
	  Dx_key(&key, buf, len + 1);
	  Dx_value(&value, buf, len + 1);
    }
	else if (flag == 'f') {
	  if (add_file(dbm, buf, len, do_read) != -1)
		nwords++;
	  continue;
	}
    else {
	  Dx_key(&key, v, strlen(v) + 1);
	  Dx_value(&value, buf, len + 1);
    }
    rc = Dx_store(dbm, &key, DX_INSERT, &value);
    if (rc != DX_OK) {
      fprintf(stderr, "Failed\n");
      fprintf(stderr, "Key: %s\n", buf);
    }
    else
      nwords++;
  }
  fprintf(stderr, "Added %d items\n", nwords);
}

struct io_info {
  int fd;
  Dx_len count;
};

static int
add_file(Dx *dbm, char *filename, int len, int do_read)
{
  int fd;
  Dx_datum key, value;
  Dx_user_io uio;
  struct io_info info;
  struct stat sbuf;
  Dx_rc rc;

  fprintf(stderr, "%s: ", filename);
  if ((fd = open(filename, O_RDONLY)) == -1) {
	perror("open");
	return(-1);
  }

  Dx_key(&key, filename, len + 1);

  info.count = 0;
  if (do_read) {
	int n;
	char *buf;

	if (fstat(fd, &sbuf) == -1) {
	  perror("fstat");
	  close(fd);
	  return(-1);
	}
	buf = (char *) alloca(sbuf.st_size);
	n = read(fd, buf, sbuf.st_size);
	if (n != sbuf.st_size) {
	  if (n == -1)
		perror("read");
	  else
		fprintf(stderr, "Short read!\n");
	  return(-1);
	}
	Dx_value(&value, buf, n);
	rc = Dx_store(dbm, &key, DX_INSERT, &value);
	if (rc != DX_OK) {
	  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  exit(1);
	}
	info.count += n;
  }
  else {
	info.fd = fd;
	uio.func = io_getfunc;
	uio.arg = (void *) &info;

	rc = Dx_storex(dbm, &key, DX_INSERT | DX_UIO, &uio);
	if (rc != DX_OK) {
	  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  exit(1);
	}
  }

  close(fd);
  fd = -1;

  fprintf(stderr, "%lu bytes\n", info.count);
  return(0);
}

Dx_len
io_getfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc)
{
  int fd, n;
  struct io_info *info;
  struct stat sbuf;

  info = (struct io_info *) arg;
  fd = info->fd;

  if (fd == -1)
    return(0);

  if (info->count == 0) {
	if (fstat(fd, &sbuf) == -1) {
	  desc->error_occurred = 1;
	  desc->last_errno = errno;
	  return(0);
	}
	desc->flags = DX_DATUM_ALIGN0;
	desc->dlen = sbuf.st_size;
  }

  /* XXX Assume n == dlen unless EOF, not necessarily true. */
  n = read(fd, dptr, (size_t) dlen);

  if (n == -1) {
	desc->error_occurred = 1;
	desc->last_errno = errno;
  }
  else
	info->count += n;

  return((Dx_len) n);
}
