/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _DX_H_
#define _DX_H_

#include <sys/types.h>

#include "dxtypes.h"

#if DX_BITS == 64
typedef Dx_ui64 Dx_hash;
typedef Dx_ui64 Dx_len;
typedef Dx_i64 Dx_off;

#define Dx_hash64
#define Dx_len64
#define Dx_off64

#define DX_MAX_UINT DX_MAX_UI64

#elif DX_BITS == 32
typedef Dx_ui32 Dx_hash;
typedef Dx_ui32 Dx_len;
typedef Dx_i32 Dx_off;

#define Dx_hash32
#define Dx_len32
#define Dx_off32

#define DX_MAX_UINT DX_MAX_UI32

#else
#error "DX_BITS must be defined to be 32 or 64!"
#endif

typedef Dx_ui8 Dx_datum_flags;
typedef Dx_ui32 Dx_opmode;

typedef enum {
  DX_LITTLE_ENDIAN   = 0,
  DX_BIG_ENDIAN      = 1,
  DX_NATIVE_ENDIAN   = 2,
  DX_PORTABLE_ENDIAN = DX_BIG_ENDIAN	/* Network byte order is big endian */
} Dx_endian;

typedef struct Dx_stats_counters {
  int nsplits;
  int nadds;
  int nmoves;
  int nsearches;
  int nprobes;
  int nreorgs;
} Dx_stats_counters;

typedef struct Dx_config {
  Dx_ui32 flags;
  mode_t open_mode;
  Dx_ui32 page_size;
  Dx_ui32 max_stripe_npages;		/* Max number of pages in a stripe. */
  Dx_endian endian;
  Dx_ui32 *volume_weights;
  Dx_hash (*hashfunc)(Dx_ui8 *str, Dx_len len, Dx_hash init);
} Dx_config;

enum {
  DX_CONFIG_TRUNCATE_ON_OPEN =  1,
  DX_CONFIG_ERASE_ON_DELETE  =  2,
  DX_CONFIG_READ_ONLY        =  4,
  DX_CONFIG_PERSISTENT       =  8,
  DX_CONFIG_MUST_EXIST       = 16,
  DX_CONFIG_DEFAULT_FLAGS    = (DX_CONFIG_PERSISTENT)
};

typedef struct Dx_keyspace {
  char *name;
  Dx_hash hash;				/* Used to modify a key's hash value. */
} Dx_keyspace;

typedef struct Dx Dx;

/* The low-order 3 bits of a Dx_datum_flags */
enum {
  DX_DATUM_ALIGN0  = 0,                     /* Any alignment */
  DX_DATUM_ALIGN2  = 1,                     /* Evenly divisible by 2 */
  DX_DATUM_ALIGN4  = 2,                     /* Evenly divisible by 4 */
  DX_DATUM_ALIGN8  = 3,                     /* Evenly divisible by 8 */
  DX_DATUM_ALIGN16 = 4,                     /* Evenly divisible by 16 */
  DX_DATUM_ALIGNXX = 5,                     /* Unused */
  DX_DATUM_ALIGNYY = 6,                     /* Unused */
  DX_DATUM_ALIGNZZ = 7,                     /* Unused */
  DX_DATUM_ALIGN_MAX = DX_DATUM_ALIGN16     /* Maximum supported alignment */
};

/* Dx_datum has been malloced by db (a bit flag in Dx_datum_flags) */
enum { DX_DATUM_MALLOC = 0x08 };

/*
 * For a put-function call, if DESC is not NULL it points to the total
 * size of the value and its alignment.  This approach is taken mainly
 * because the size is most useful when known as soon as possible; if it was
 * returned in the value Dx_datum (or another regular parameter), the size
 * wouldn't be known until after the operation completed.  Or a DX_PROBE
 * would need to be done in a preceding operation.
 *
 * For a get-function call, if DESC is not NULL the caller may optionally
 * set it to the total size of the value and its alignment.  Th size is taken
 * as an efficiency hint and needn't be the actual size; the closer it is to
 * the actual size, however, the more efficient the operation will be.  If
 * it is much different than the actual size, the operation may be less
 * efficient than it would otherwise be.
 *
 * If DESC is provided, the caller is expected to indicate an error
 * by setting ERROR_OCCURRED to a non-zero value and may optionally set
 * LAST_ERRNO to an appropriate "errno" value (both ERROR_OCCURRED and
 * LAST_ERRNO are zero when the function is called).  LAST_ERRNO will
 * be made available when the operation returns to the user.
 */
typedef struct Dx_desc {
  Dx_len dlen;
  Dx_datum_flags flags;
  int error_occurred;
  int last_errno;
} Dx_desc;

typedef struct Dx_user_io {
  Dx_len (*func)(void *arg, void *dptr, Dx_len io_dlen, Dx_desc *desc);
  void *arg;
} Dx_user_io;

typedef enum { DX_POINT_ABSOLUTE = 0, DX_POINT_RELATIVE = 1 } Dx_point_mode;

typedef struct Dx_point {
  Dx_point_mode mode;				/* Absolute or relative to the end */
  union {
	Dx_len pos;						/* Starts at zero */
	Dx_off offset_from_end;  		/* Signed offset from the end */
  } u;
} Dx_point;

/*
 * A Dx_region is used to specify a region to fetch (Dx_fetchx(),
 * Dx_firstkeyx(), or Dx_nextkeyx()) or a region to delete (Dx_storex()).
 */
typedef enum {
  DX_REGION_PTOP = 0, DX_REGION_LEN = 1, DX_REGION_ALL
} Dx_region_mode;

typedef struct Dx_region {
  Dx_region_mode mode;
  Dx_point start;		/* Starting position for the region */
  union {
	Dx_len dlen;		/* Length of the region */
	Dx_point end;		/* Ending position for the region */
  } u;
} Dx_region;

typedef struct Dx_datum {
  void *dptr;
  Dx_len dlen;			  /* Maximum of (2^32)-1 for keylen. */
  Dx_ui32 ind;			  /* Used to select among equal keys only for keys. */
  Dx_ui32 count;		  /* Used by DX_PROBE only for keys. */
  Dx_datum_flags flags;
  Dx_keyspace *keyspace;
} Dx_datum;

/* Operations */
typedef enum {
  DX_DEFAULT       = 0x000,
  DX_PROBE         = 0x001,
  DX_FETCH         = 0x002,
  DX_EXISTS        = 0x003,
  DX_REPLACE       = 0x004,
  DX_INSERT        = 0x005,
  DX_EDIT          = 0x006,
  DX_DELETE        = 0x007,
  DX_STORE         = 0x008,
  DX_INSERT_UNIQUE = 0x009,
  DX_OP_NOP3       = 0x00a,
  DX_OP_NOP4       = 0x00b,
  DX_OP_NOP5       = 0x00c,
  DX_OP_NOP6       = 0x00d,
  DX_OP_NOP7       = 0x00e,
  DX_OP_NOP8       = 0x00f,
  DX_OPMODE_MASK   = 0x00f
} Dx_optype;

/* Operation modifiers */
enum {
  DX_OVERWRITE    = 0x0010,
  DX_REGION       = 0x0020,
  DX_UIO          = 0x0040,
  DX_ALWAYS_COPY  = 0x0080,
  DX_ADD_FIRST    = 0x0100,
  DX_ADD_LAST     = 0x0200,
  DX_ADD_INDEX    = 0x0400,
  DX_ADDMODE_MASK =	0x0700,
  DX_MAY_CREATE   = 0x1000
};

typedef enum {
  DX_OK                        =  0,
  DX_HASH_OVERFLOW             =  1,
  ITEM_NOSPACE                 =  2,
  ITEM_NOTFOUND                =  3,
  DX_BAD_POSITION              =  4,
  DX_NOTIMPLEMENTED            =  5,
  DX_BADARG                    =  6,
  DX_USERIO_ERROR              =  7,
  ITEM_NOMATCH                 =  8,
  DX_BAD_REGION                =  9,
  PAGE_NOTFOUND                = 10,
  DX_BADMAGIC                  = 11,
  DX_INITFAILED                = 12,
  DX_PHYSMAP_FAILED            = 13,
  DX_INVALID_PAGESIZE          = 14,
  DX_BADACCESSMODE             = 15,
  DX_OS_ERROR                  = 16,
  DX_INVALID_NVOLUMES          = 17,
  DX_INTERNAL_ERROR            = 18,
  DX_MALLOC_FAILED             = 19,
  DX_BADEDITARG                = 20,
  DX_VOLUME_MISMATCH           = 21,
  DX_HEADER_MISMATCH           = 22,
  DX_INVALID_MAX_STRIPE_NPAGES = 23,
  DX_TIME_MISMATCH             = 24,
  DX_BADCONFIGARG              = 25,
  DX_OPENFAILED                = 26,
  DX_READONLY                  = 27,
  DX_INVALID_ENDIAN            = 28,
  DX_INVALID_KEYSPACE          = 29,
  DX_KEY_NOT_FOUND             = 30,
  DX_KEY_EXISTS                = 31,
  DX_FRAME_NOT_POPPED          = 32,
  DX_FRAME_NO_FRAME            = 33,
  DX_FRAME_BAD_POINTER         = 34,
  DX_KEYSPACE_NOTFOUND         = 35,
  DX_KEYSPACE_EXISTS           = 36,
  DX_KEYSPACE_RESERVED         = 37,
  SEGMENT_NOTFOUND             = 38,
  DX_BAD_KEY                   = 39,
  DX_FSTATFAILED               = 40,
  DX_ACCESSFAILED              = 41,
  DX_FLOCKFAILED               = 42,
  DX_INVALID_VOLUME_WEIGHTS    = 43
} Dx_rc;

extern Dx_rc Dx_close(Dx *db);
extern void Dx_get_configuration(Dx *db, Dx_config *config);
extern Dx_rc Dx_delete(Dx *db, Dx_datum *key);
extern Dx_rc Dx_deletex(Dx *db, Dx_datum *key, Dx_opmode mode, ...);
extern Dx_rc Dx_fetch(Dx *db, Dx_datum *key, Dx_datum *value);
extern Dx_rc Dx_fetchx(Dx *db, Dx_datum *key, Dx_opmode mode, ...);
extern Dx_rc Dx_firstkey(Dx *db, Dx_datum *key);
extern Dx_rc Dx_firstkeyx(Dx *db, Dx_datum *key, Dx_opmode mode, ...);
extern Dx_rc Dx_nextkey(Dx *db, Dx_datum *key);
extern Dx_rc Dx_nextkeyx(Dx *db, Dx_datum *key, Dx_opmode mode, ...);
extern Dx_rc Dx_open(char **volume_paths, Dx_config *config, Dx **db);
extern Dx_rc Dx_create_persistent_copy(Dx *db, char **volume_paths);
extern Dx_rc Dx_reorganize(Dx *db);
extern Dx_rc Dx_store(Dx *db, Dx_datum *key, Dx_opmode mode, Dx_datum *value);
extern Dx_rc Dx_storex(Dx *db, Dx_datum *key, Dx_opmode mode, ...);
extern char *Dx_strerror(Dx_rc rc);
extern Dx_rc Dx_sync(Dx *db);

extern void Dx_key(Dx_datum *keyp, void *keyval, Dx_len keylen);
extern void Dx_key_first(Dx_datum *keyp);
extern void Dx_key_next(Dx_datum *keyp);
extern void Dx_value(Dx_datum *valuep, void *valueval, Dx_len valuelen);

extern void Dx_stats_reset(Dx *db);
extern Dx_stats_counters *Dx_stats_get(Dx *db);

extern void Dx_check(Dx *db);
extern void Dx_show(Dx *db);

extern void Dx_set_point(Dx_point *point, Dx_len start_val);
extern void Dx_set_point_at_start(Dx_point *point);
extern void Dx_set_point_at_start(Dx_point *point);
extern void Dx_set_point_from_end(Dx_point *point, Dx_len end_val);
extern void Dx_set_point_at_end(Dx_point *point);
extern void Dx_set_region_from_start(Dx_region *region);
extern void Dx_set_region_start(Dx_region *region, Dx_len start_val);
extern void Dx_set_region_end(Dx_region *region, Dx_len end_val);
extern void Dx_set_region_to_end(Dx_region *region);
extern void Dx_set_region_from_end(Dx_region *region, Dx_len end_val);
extern void Dx_set_region_len(Dx_region *region, Dx_len dlen);
extern void Dx_set_region_all(Dx_region *region);

extern const char *Dx_version;

extern Dx_rc Dx_create_keyspace(Dx *db, char *ksn);
extern Dx_rc Dx_set_keyspace(Dx *db, Dx_datum *key, char *ksn);
extern char *Dx_get_keyspace(Dx *db, Dx_datum *key);
extern Dx_rc Dx_get_keyspace_list(Dx *db, char ***ksn_vec, void **mem);
extern Dx_rc Dx_delete_keyspace(Dx *db, char *ksn);

extern Dx_rc Dx_last_rc;
extern int Dx_last_errno(Dx *db);

extern int Dx_new_frame(Dx *db, char *name);
extern int Dx_push_frame(void *ptr);
extern void *Dx_pop_frame(void);
extern void Dx_exch_frame(void);
extern int Dx_unite_frames(void *frame1, void *frame2);

extern int Dx_free(void);
extern int Dx_free_frame(void *ptr);
extern int Dx_free_frames(void);
extern int Dx_free_all_frames(void);
extern void *Dx_malloc(size_t size);
extern void *Dx_malloc_src(Dx *db, size_t size);
extern void *Dx_frame_malloc(void *ptr, size_t size);
extern void *Dx_frame_malloc_src(void *ptr, Dx *db, size_t size);
extern void *Dx_calloc(size_t nmemb, size_t size);
extern void *Dx_calloc_src(Dx *db, size_t nmemb, size_t size);
extern void *Dx_frame_calloc(void *ptr, size_t nmemb, size_t size);
extern void *Dx_frame_calloc_src(void *ptr, Dx *db, size_t nmemb, size_t size);
extern void *Dx_realloc(void *ptr, size_t size);
extern void *Dx_realloc_src(Dx *dbsrc, void *ptr, size_t size);
extern void *Dx_frame_realloc(void *frame, void *ptr, size_t size);
extern void *Dx_frame_realloc_src(void *frame, Dx *db, void *ptr, size_t size);

extern int Dx_frame_import(Dx *db, void *buf, size_t buflen, void **ptr);
extern void *Dx_frame_import_frame(Dx *db, void *buf, size_t buflen,
								   void **ptr);
extern int Dx_frame_export(Dx *db, void *ptr, void **buf, size_t *buflen);
extern int Dx_frame_export_frame(Dx *db, void *frame, void *ptr, void **buf,
								  size_t *buflen);
extern Dx_rc Dx_swizzle(Dx *db, void **frame_mem_ptr);
extern Dx_rc Dx_frame_swizzle(Dx *db, void *frame, void **frame_mem_ptr);
extern Dx_rc Dx_unswizzle(Dx *db, void **swizzledp);
extern Dx_rc Dx_frame_unswizzle(Dx *db, void *frame, void **swizzledp);

#endif
