/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * dxtest - Dxstore database utility
 *
 * Command line flags:
 * -w[=kv]   create a new 3 volume database (default: db[123])
 *           -w or -wk reads keys, one per line, from stdin and successive
 *           integers as values
 *           -w= reads keys, one per line, from stdin and also uses the key
 *           as the value
 *           -wv reads values, one per line, from stdin and uses successive
 *           integers as keys
 * -ro       open an existing database (default: db[123]) read only
 * -v#       use # volumes, (db1, db2, ...)
 * -volatile create a volatile database instead of persistent
 * -b        batch mode.  After loading with -w, exit.  The default is to
 *           enter interactive mode.
 * -d[kvc]   After opening a db, dump keys (k) or values (v) to stdout
 *           -dc just prints a count of the number of items.
 *
 * In interactive mode, type '?' for a simple usage message.
 */

#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "dx.h"
#include "local.h"

extern int breakup(char *buf, char **argv[], char **sp, int *n, char *ifs);

void usage(void);

FILE *fp;

static char buf[102400];

static int dumpem(Dx *db, int flag);
static void add_words(Dx *db, int flag);
static void load_rand_words(Dx *db, int, int);

struct io_arg
{
  FILE *fp;
  int count;
};

Dx_len
io_getfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc)
{
  struct io_arg *io_arg;
  size_t n;

  io_arg = (struct io_arg *) arg;

  if (io_arg->fp == NULL)
    return(0);

  if (feof(io_arg->fp))
	return(0);

  n = fread(dptr, sizeof(char), (size_t) dlen, io_arg->fp);

  io_arg->count += n;

  if (n != dlen) {
	if (ferror(io_arg->fp)) {
	  desc->error_occurred = 1;
	  desc->last_errno = errno;
	}
  }

  return((Dx_len) n);
}

Dx_len
io_putfunc(void *arg, void *dptr, Dx_len dlen, Dx_desc *desc)
{
  struct io_arg *io_arg;
  size_t n;

  io_arg = (struct io_arg *) arg;

  if (io_arg->fp == NULL)
    return(0);

  if (feof(io_arg->fp))
    return(0);

  n = fwrite(dptr, sizeof(char), (size_t) dlen, io_arg->fp);

  io_arg->count += n;

  if (n != dlen) {
	if (ferror(io_arg->fp)) {
	  desc->error_occurred = 1;
	  desc->last_errno = errno;
	}
  }

  return((Dx_len) n);
}

int
main(int argc, char **argv)
{
  int count, do_add, do_batch, do_dump, do_dump_flag, i, len, use_file;
  int n, next, nvolumes;
  int add_flag;
  Dx_ui32 ind;
  Dx_ui32 flags;
  int rand, bigkeys, bigvalues;
  char **mkargv, *p, *space;
  char cmd[100], prevcmd[100], *volumes[10];
  struct io_arg io_arg;
  Dx_datum key, value;
  Dx *db;
  Dx_config config;
  Dx_opmode addmode, mode;
  Dx_region region;
  Dx_user_io uio;
  Dx_ui32 volweights[100];
  Dx_rc rc;
  unsigned int seed = 0;

  Dx_get_configuration(NULL, &config);

  do_add = 0;
  add_flag = 0;
  do_batch = 0;
  do_dump = do_dump_flag = 0;
  nvolumes = 3;
  volweights[0] = 1;
  volweights[1] = 3;
  volweights[2] = 6;
  volweights[3] = 0;
  config.volume_weights = volweights;
  flags = DX_CONFIG_DEFAULT_FLAGS;

  while (argv[1] != NULL && argv[1][0] == '-') {
	if (strncmp(argv[1], "-w", 2) == 0) {
	  do_add = 1;
	  add_flag = argv[1][2];
	  flags |= DX_CONFIG_TRUNCATE_ON_OPEN;
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "-ro") == 0) {
	  flags |= DX_CONFIG_READ_ONLY;
	  argc--;
	  argv++;
	}
	else if (strncmp(argv[1], "-v", 2) == 0 && isdigit(argv[1][2])) {
	  nvolumes = atoi(&argv[1][2]);
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "-volatile") == 0) {
	  flags &= (~DX_CONFIG_PERSISTENT);
	  argc--;
	  argv++;
	}
	else if (strcmp(argv[1], "-b") == 0) {
	  do_batch = 1;
	  argc--;
	  argv++;
	}
	else if (strncmp(argv[1], "-d", 2) == 0) {
	  do_dump = 1;
	  do_dump_flag = argv[1][2];
	  argc--;
	  argv++;
	}
  }

  if (nvolumes == 0)
	exit(1);
  for (i = 0; i < nvolumes; i++) {
	volumes[i] = (char *) malloc(8);
	sprintf(volumes[i], "db%d", i + 1);
  }
  volumes[i] = NULL;

  config.flags = flags;
#ifndef ENDIAN
#define ENDIAN	DX_PORTABLE_ENDIAN
#endif
  config.endian = ENDIAN;

  if ((rc = Dx_open(volumes, &config, &db)) != DX_OK) {
	fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	exit(1);
  }

  Stats_init(db);

  if (do_dump) {
	count = dumpem(db, do_dump_flag);
	if (do_dump_flag == 'c')
	  printf("%d\n", count);
	Dx_close(db);
	exit(0);
  }

  addmode = DX_ADD_FIRST;

  if (do_add) {
	add_words(db, add_flag);
	if (do_batch) {
	  rc = Dx_close(db);
	  if (rc != DX_OK) {
		fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		exit(1);
	  }
	  exit(0);
	}

	db_check(db);
	if (freopen("/dev/tty", "r+", stdin) == NULL) {
	  perror("freopen");
	  exit(1);
	}
  }

  ind = 0;
  while (1) {
	fprintf(stderr, "> ");
	fflush(stderr);
	if (fgets(cmd, sizeof(cmd), stdin) == NULL)
	  break;

	len = strlen(cmd);
	cmd[len - 1] = '\0';

	if (cmd[0] == '\0')
	  strcpy(cmd, prevcmd);

	breakup(cmd, &mkargv, &space, &n, " ");
	if (n == 0)
	  continue;

	io_arg.fp = NULL;

	if (db == NULL) {
	  fprintf(stderr, "The db is closed\n");
	  exit(0);
	}

	switch (mkargv[0][0]) {
	case 'a':
	case 'r':
	  if (strcmp(mkargv[0], "reorg") == 0) {
		Dx_reorganize(db);
		break;
	  }

	  Dx_key(&key, mkargv[1], strlen(mkargv[1]) + 1);
	  if (mkargv[2][0] == '<') {
		if ((fp = fopen(&mkargv[2][1], "r")) == NULL) {
		  perror("fopen");
		  break;
		}
		io_arg.count = 0;
		io_arg.fp = fp;
		uio.func = io_getfunc;
		uio.arg = &io_arg;
	  }
	  else {
		Dx_value(&value, mkargv[2], strlen(mkargv[2]) + 1);
		fp = NULL;
	  }

	  if (mkargv[0][0] == 'a') {
		if (fp == NULL)
		  rc = Dx_storex(db, &key, DX_INSERT | addmode, &value);
		else
		  rc = Dx_storex(db, &key, DX_INSERT | addmode | DX_UIO, &uio);
	  }
	  else {
		if (mkargv[0][1] == 'o') {
		  Dx_set_region_all(&region);
		  if (fp == NULL)
			rc = Dx_storex(db, &key, DX_EDIT | DX_OVERWRITE, &region, &value);
		  else
			rc = Dx_storex(db, &key, DX_EDIT | DX_OVERWRITE | DX_UIO,
						   &region, &uio);
		}
		else {
		  if (fp == NULL)
			rc = Dx_storex(db, &key, DX_REPLACE, &value);
		  else
			rc = Dx_storex(db, &key, DX_REPLACE | DX_UIO, &uio);
		}
	  }

	  if (rc != DX_OK)
		fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  else if (fp != NULL)
		fprintf(stderr, "%d bytes\n", io_arg.count);

	  addmode = DX_ADD_FIRST;
	  ind++;

	  break;

	case 'c':
	  if (strcmp(mkargv[0], "check") == 0)
		db_check(db);
	  else if (strcmp(mkargv[0], "close") == 0) {
		Dx_close(db);
		db = NULL;
	  }
	  else if (strcmp(mkargv[0], "complete") == 0)
		Ehash_complete_expansion(db);
	  else
		fprintf(stderr, "?\n");
	  break;

	case 'd':
	  if (n != 2 && n != 4) {
		usage();
		break;
	  }
	  Dx_key(&key, mkargv[1], strlen(mkargv[1]) + 1);
	  value.dptr = NULL;
	  value.dlen = 0;
	  if (n == 2)
		rc = Dx_delete(db, &key);
	  else {
		Dx_set_region_start(&region, atol(mkargv[2]));
		Dx_set_region_end(&region, atol(mkargv[3]));
		rc = Dx_deletex(db, &key, DX_REGION, &region);
	  }
	  if (rc != DX_OK)
		fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  break;

	case 'f':
	  if (mkargv[0][1] == 'n')
		Dx_key_next(&key);
	  else if (mkargv[0][1] != '\0') {
		fprintf(stderr, "?\n");
		break;
	  }
	  else {
		Dx_key(&key, mkargv[1], strlen(mkargv[1]) + 1);
		if (mkargv[2] != NULL && mkargv[2][0] == '>') {
		  if ((fp = fopen(&mkargv[2][1], "w")) == NULL) {
			perror("fopen");
			break;
		  }
		}
	  }
	  value.dptr = NULL;
	  value.dlen = 0;
	  Stats_init(db);
	  if (fp == NULL) {
		if ((rc = Dx_fetch(db, &key, &value)) != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		else
		  fprintf(stderr, "`%s' (%lu bytes, desc=%u)\n",
				  (char *) value.dptr, (unsigned long) value.dlen,
				  value.flags);
		fprintf(stderr, "%d probe%s\n",
				db->stats.nprobes, db->stats.nprobes == 1 ? "" : "s");
	  }
	  else {
		io_arg.count = 0;
		io_arg.fp = fp;
		uio.func = io_putfunc;
		uio.arg = (void *) &io_arg;
		if ((rc = Dx_fetchx(db, &key, DX_FETCH | DX_UIO, &uio)) != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		else
		  fprintf(stderr, "%d bytes\n", io_arg.count);
	  }
	  break;

	case 'i':
	  mode = DX_FETCH;
	  p = mkargv[0] + 1;
	  if (*p == 'n') {
		next = 1;
		p++;
	  }
	  else
		next = 0;
	  if (*p == 'f' || *p == '\0')
		mode = DX_FETCH;
	  else if (*p == 'p')
		mode = DX_PROBE;

	  key.keyspace = NULL;
	  if (next)
		rc = Dx_nextkeyx(db, &key, mode, &value);
	  else
		rc = Dx_firstkeyx(db, &key, mode, &value);

	  if (rc == DX_OK) {
		fprintf(stderr, "key=`%s'\n", (char *) key.dptr);
		if (mode == DX_PROBE)
		  fprintf(stderr, "value=(%lu bytes)\n", (unsigned long) value.dlen);
		else
		  fprintf(stderr, "value=`%s' (%lu bytes)\n", (char *) value.dptr,
				  (unsigned long) value.dlen);
	  }
	  else
		fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  break;

	case 'k':
	  if (strcmp(mkargv[0], "ks") == 0) {
		char *keyspace;

		rc = DX_OK;
		if (n == 1) {
		  keyspace = Dx_get_keyspace(db, NULL);
		  if (keyspace != NULL)
			fprintf(stderr, "'%s'\n", keyspace);
		}
		else if (n == 2) {
		  if (mkargv[1][0] == '-') {
			if (mkargv[1][1] == '\0')
			  rc = Dx_set_keyspace(db, NULL, NULL);
			else
			  rc = Dx_delete_keyspace(db, &mkargv[1][1]);
		  }
		  else if (mkargv[1][0] == '+')
			rc = Dx_create_keyspace(db, &mkargv[1][1]);
		  else
			rc = Dx_set_keyspace(db, NULL, mkargv[1]);
		}
		else
		  fprintf(stderr, "?\n");

		if (rc != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		break;
	  }
	  else if (strcmp(mkargv[0], "ksl") == 0) {
		char **vec;
		void *mem;

		if ((rc = Dx_get_keyspace_list(db, &vec, &mem)) == DX_OK) {
		  for (i = 0; vec[i] != NULL; i++)
			fprintf(stderr, "'%s'\n", vec[i]);
		  Xfree(mem);
		}
		break;
	  }

	  p = mkargv[1];
	  if (n == 2 && isalpha(*p)) {
		if (*p == 'f')
		  addmode = DX_ADD_FIRST;
		else if (*p == 'l')
		  addmode = DX_ADD_LAST;
		else if (*p == 'i')
		  addmode = DX_ADD_INDEX;
		else
		  break;
		ind = 0;
	  }
	  else if (n == 2 && isdigit(*p))
		ind = atoi(p);

	  break;
	case 'l':
	  rand = 0;
	  bigkeys = 0;
	  bigvalues = 0;
	  p = &mkargv[0][1];
	  if (isdigit(*p)) {
		seed = atoi(p);
		break;
	  }
	  while (*p != '\0') {
		switch (*p) {
		case 'r':
		  rand = 1;
		  break;
		case 'k':
		  bigkeys = 1;
		  break;
		case 'v':
		  bigvalues = 1;
		  break;
		default:
		  break;
		}
		p++;
	  }
	  if (rand) {
		time_t now;

		time(&now);
		seed = now;
	  }
	  srandom(seed);
	  fprintf(stderr, "rand = %u\n", seed);

	  load_rand_words(db, bigkeys, bigvalues);
	  break;

	case 'p':
	  if (strcmp(mkargv[0], "persist") == 0) {
		rc = Dx_create_persistent_copy(db, volumes);
		if (rc != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		break;
	  }
	  if (mkargv[0][1] == 'n')
		Dx_key_next(&key);
	  else
		Dx_key(&key, mkargv[1], strlen(mkargv[1]) + 1);
	  value.dptr = NULL;
	  value.dlen = 0;
	  if ((rc = Dx_fetchx(db, &key, DX_PROBE, &value)) != DX_OK)
		fprintf(stderr, "Not found\n");
	  else
		fprintf(stderr, "%lu bytes (%lu of %lu)\n",
				(unsigned long) value.dlen, key.ind + 1, key.count);
	  break;

	case 's':
	  if (strcmp(mkargv[0], "sync") == 0) {
		rc = Dx_sync(db);
		if (rc != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  }
	  else if (strcmp(mkargv[0], "sa") == 0) {
		rc = Segment_uncache(db, 1);
		if (rc != DX_OK)
		  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  }
	  else if (strcmp(mkargv[0], "sb") == 0)
		db_dump(db);
	  else if (strcmp(mkargv[0], "sdx") == 0)
		dx_show(db);
	  else if (strcmp(mkargv[0], "sm") == 0)
		meta_show(db);
	  else if (strcmp(mkargv[0], "ss") == 0)
		Segment_mapshow(db);
	  else
		Dx_show(db);
	  break;

	case 'q':
	  if (mkargv[0][1] == 'f')
		exit(0);
	  if (mkargv[0][1] == 'd')
		unlink("db");

	  rc = Dx_close(db);
	  if (rc != DX_OK) {
		fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
		exit(1);
	  }
	  exit(0);
	  /*NOTREACHED*/

	case 'v':
	  fprintf(stderr, "%s\n", Dx_version);
	  break;


	default:
	  usage();
	  break;
	}

	strcpy(prevcmd, cmd);
	if (io_arg.fp != NULL) {
	  fclose(io_arg.fp);
	  io_arg.fp = NULL;
	}

	free(space);
	free(mkargv);
  }

  exit(0);
}

int nwords;

static void
add_words(Dx *db, int flag)
{
  int len;
  char v[10];
  Dx_datum key, value;
  Dx_rc rc;

  nwords = 0;
  while (fgets(buf, sizeof(buf), stdin) != NULL) {
	len = strlen(buf);
	len--;
	buf[len] = '\0';
	sprintf(v, "%d", nwords);
	if (flag == '\0' || flag == 'k') {
	  Dx_key(&key, buf, len + 1);
	  Dx_value(&value, (void *) v, strlen(v) + 1);
	}
	else if (flag == '=') {
	  Dx_key(&key, buf, len + 1);
	  Dx_value(&value, buf, len + 1);
	}
	else {
	  Dx_key(&key, v, strlen(v) + 1);
	  Dx_value(&value, (void *) buf, len + 1);
	}
	rc = Dx_store(db, &key, DX_INSERT, &value);
	/* Dx_check(); */
	if (rc != DX_OK) {
	  fprintf(stderr, "Failed: %s\n", Dx_strerror(rc));
	  fprintf(stderr, "Key: %s\n", buf);
	}
	else
	  nwords++;
  }
  fprintf(stderr, "Added %d items\n", nwords);
}

char *alph = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

int
rand_char(void)
{
  unsigned int r;
  static int alen = 0;

  if (alen == 0)
	alen = strlen(alph);

  r = (unsigned int) ((unsigned long) random() % alen);
  return((int) alph[r]);
}

char *
rand_word(char *word, int len)
{
  int i;

  for (i = 0; i < len; i++)
	word[i] = rand_char();

  word[len] = '\0';
  return(word);
}

static void
load_rand_words(Dx *db, int bigkeys, int bigvalues)
{
  int i, klen, vlen;
  char *kw = NULL, *vw = NULL;
  int kwlen = 0, vwlen = 0;
  Dx_datum key, value;
  Dx_rc rc;

  nwords = 0;
  for (i = 0; i < 200; i++) {
	if (bigkeys)
	  klen = (unsigned long) random() % 8000 + 1;
	else
	  klen = (unsigned long) random() % 100 + 1;

	if (kwlen == 0) {
	  kwlen = klen + 1;
	  kw = (char *) malloc(kwlen);
	}
	else if ((klen + 1) > kwlen) {
	  kwlen = klen + 1;
	  kw = (char *) realloc(kw, kwlen);
	}
	rand_word(kw, klen);

	if (bigvalues)
	  vlen = (unsigned long) random() % 8000 + 1;
	else
	  vlen = (unsigned long) random() % 100 + 1;
	if (vwlen == 0) {
	  vwlen = vlen + 1;
	  vw = (char *) malloc(vwlen);
	}
	else if ((vlen + 1) > vwlen) {
	  vwlen = vlen + 1;
	  vw = (char *) realloc(vw, vwlen);
	}
	rand_word(vw, vlen);

	Dx_key(&key, kw, klen + 1);
	Dx_value(&value, vw, vlen + 1);

	rc = Dx_store(db, &key, DX_INSERT, &value);
	if (rc != DX_OK)
	  fprintf(stderr, "Failed: %d. %s/%s\n", i, kw, vw);
	else
	  nwords++;
  }

  fprintf(stderr, "Added %d items\n", nwords);
}

static int
dumpem(Dx *db, int flag)
{
  int count;
  Dx_datum key, value;
  Dx_rc rc;

  count = 0;

  Dx_key(&key, NULL, 0);
  rc = Dx_firstkeyx(db, &key, DX_FETCH, &value);
  if (rc == DX_KEY_NOT_FOUND)
	return(0);
  if (rc != DX_OK) {
	fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	exit(1);
  }
  if (flag == 'k' || flag == '\0')
	printf("\"%s\"", (char *) key.dptr);
  if (flag == '\0')
	printf(" ");
  if (flag == 'v' || flag == '\0')
	printf("\"%s\"", (char *) value.dptr);
  if (flag != 'c')
	printf("\n");

  count = 1;

  while (1) {
	rc = Dx_nextkeyx(db, &key, DX_FETCH, &value);
	if (rc == DX_KEY_NOT_FOUND)
	  return(count);
	if (rc != DX_OK) {
	  fprintf(stderr, "Dx: %s\n", Dx_strerror(rc));
	  exit(1);
	}
	if (flag == 'k' || flag == '\0')
	  printf("\"%s\"", (char *) key.dptr);
	if (flag == '\0')
	  printf(" ");
	if (flag == 'v' || flag == '\0')
	  printf("\"%s\"", (char *) value.dptr);
	if (flag != 'c')
	printf("\n");

	count++;
  }
}

void
usage(void)
{

  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "a key value    Add\n");
  fprintf(stderr, "check          Check db\n");
  fprintf(stderr, "close          Close db\n");
  fprintf(stderr, "complete       Reduce access times by expanding db\n");
  fprintf(stderr, "d key          Delete\n");
  fprintf(stderr, "f key [>file]  Fetch [store value in file]\n");
  fprintf(stderr, "fn             Fetch next key while iterating\n");
  fprintf(stderr, "i[n][f|p]      Iterate [next] fetch/probe)\n");
  fprintf(stderr, "l#             Set random load seed\n");
  fprintf(stderr, "k [fil#]       Set key index (first, index, last, #)\n");
  fprintf(stderr, "ks [name|+name|-name|-] Print/Set/Remove current keyspace\n");
  fprintf(stderr, "ksl            List all keyspaces\n");
  fprintf(stderr, "l[rkv]...      Load with new random seed/big key/big value\n");
  fprintf(stderr, "p[n] key       Probe [next]\n");
  fprintf(stderr, "persist        Make a volatile db persistent\n");
  fprintf(stderr, "q[f|d]         Quit [fast|delete]\n");
  fprintf(stderr, "r[o] key value Replace [overwrite]\n");
  fprintf(stderr, "reorg          Perform delayed cleanups\n");
  fprintf(stderr, "s[a|b|dx|m|s]  Show [adjust|dump|dx|meta|segments]\n");
  fprintf(stderr, "sync           Sync\n");
  fprintf(stderr, "v              Version\n");

  fprintf(stderr, "\n");
  fprintf(stderr, "When adding/replacing, '<filename' will read it from 'filename'\n");
  fprintf(stderr, "When fetching, the value can optionally be stored in a file.\n");
  fprintf(stderr, "Enter a newline to repeat the previous command.\n");
}
