/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Extensible hashing, without a page-split directory and with "in place"
 * hash overflow.
 */

#include "local.h"

static int bucket_create(Dx *db, PageNum b);
static PageNum compute_bucket(Dx *db, Dx_hash h, Dx_ui32 exp);
static void expand(Dx *db, PageNum b);
static PageNum next_bucket(Dx *db, PageNum b);
static int next_bucket_exists(Dx *db, PageNum b);

static int do_controlled_split(Dx *db, Page page);

static INLINE Dx_ui32
compute_next_expansion(Dx *db, Dx_ui32 b)
{
  int i;
  Dx_ui32 next_size;

  next_size = db->npages;
  for (i = 1; next_size <= (b + db->npages); i++)
	next_size = Compute_expansion_buffers(db, db->nexpansions + i);

  return(next_size);
}

/*
 * If necessary, expand the number of buckets to create the split bucket
 * corresponding to the given bucket.
 */
static INLINE void
expand(Dx *db, PageNum b)
{

  if (next_bucket_exists(db, b))
	return;

  if (DX_DEBUG > 1)
	fprintf(stderr, "expand=%lu\n", (unsigned long) b);

  db->npages = compute_next_expansion(db, b);
}

/*
 * Also optionally indicate how many changes were made.
 * Recursive.
 */
static INLINE Dx_rc
ehash_split(Dx *db, PageNum b, Dx_ui32 base_level, int *nchanges,
			Dx_hash *check_hash)
{
  int changed, first, nitems, nident;
  Dx_ui32 ind, level;
  PageNum newb;
  Dx_hash h, hident;
  Page newpage, page;
  Dx_rc rc;

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(DX_READONLY);

  if ((db->nexpansions + base_level) >= MAX_LEVELS)
	return(DX_HASH_OVERFLOW);

  expand(db, b);
  rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
  if (rc != DX_OK) {
	if (rc == PAGE_NOTFOUND)
	  page = Page_alloc(db, b);
	else
	  return(rc);
  }

  changed = 0;
  level = base_level;

  first = 1;
  nitems = 0;
  nident = 0;
  hident = 0;

  /*
   * Examine each item in the page; with luck, about half will stay
   * where they are (in page b) and about half will move to newb.
   */
  while (Item_iterate(db, page, first, &ind, &h) == DX_OK) {

	if (nitems == 0)
	  hident = h;
	else if (hident == h)
	  nident += (nitems == 1) ? 2 : 1;
	nitems++;

	first = 0;
	newb = compute_bucket(db, h, db->nexpansions + level);
	expand(db, newb);

	if (newb != b) {
	  /*
	   * Item rehashes to a different bucket.
	   * If the item's new bucket is full, that bucket must be
	   * split before we can continue.
	   */
	  bucket_create(db, newb);
	  rc = Page_get(db, PAGENUM_TO_VOLUME(db, newb), newb, PAGE_MAYCREATE,
					&newpage);
	  if (rc != DX_OK)
		return(rc);

	  /*
	   * If the new page is already full, it must be split to make room.
	   * Then we must recurse and continue with this page.
	   */
	  if ((rc = Item_move(db, page, ind, newpage)) == ITEM_NOSPACE) {
		rc = ehash_split(db, newb, level + 1, NULL, NULL);
		if (DX_CHECK)
		  page_check(db, newb);
		rc = ehash_split(db, b, level + 1, nchanges, NULL);
		if (DX_CHECK)
		  page_check(db, b);
		return(rc);
	  }
	  else if (rc != DX_OK) {
		return(rc);
	  }
	  changed++;
	}
  }

  if (b == (PageNum) db->boundary) {
	/*
	 * If the boundary bucket was just split, advance the boundary pointer.
	 * If a full expansion has been completed, start the next one.
	 */
	if (++db->boundary == Compute_expansion_buffers(db, db->nexpansions)) {
	  if (DX_DEBUG > 1)
		fprintf(stderr, "Boundary split %lu\n",
				(unsigned long) db->nexpansions);
	  db->nexpansions++;
	  db->boundary = 0;
	}
  }
  else {
	if (DX_DEBUG > 1)
	  fprintf(stderr, "Uncontrolled split %lu\n",
			  (unsigned long) db->nexpansions);
	ehash_split(db, (PageNum) db->boundary, level + 1, NULL, NULL);
	if (DX_CHECK)
	  page_check(db, (PageNum) db->boundary);
  }

  if (changed)
	Stats_count(db, SPLIT);

  if (nchanges != NULL)
	*nchanges += changed;

  if (check_hash != NULL && nitems > 0 && nitems == nident
	  && *check_hash == hident)
	return(DX_HASH_OVERFLOW);

  if (DX_CHECK)
	page_check(db, b);

  return(DX_OK);
}

Dx_rc
Ehash_add(Dx *db, Item *item, Dx_opmode mode, Ehash_xarg *arg)
{
  int nsplit_moves;
  Dx_ui64 level;
  PageNum b, prev;
  Dx_rc rc;
  Page page;
  Volume *v;

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(DX_READONLY);

  Segment_uncache(db, 0);

  if (item->key->keyspace != NULL) {
	item->ks = item->key->keyspace;
	item->hash = item->keyhash ^ item->ks->hash;
  }
  else if (db->keyspace != NULL) {
	item->ks = db->keyspace;
	item->hash = item->keyhash ^ item->ks->hash;
  }
  else {
	item->hash = item->keyhash;
	item->ks = NULL;
  }
#ifdef NOTDEF
  item->hash = item->keyhash;
  item->ks = NULL;
#endif

  if (arg != NULL) {
	arg->io.dlen = 0;
	arg->io.uio_buf = NULL;
	arg->io.bytes_buffered = 0;
  }

  prev = 0;		/* Shuts up compiler about its uninitialized use. */

  for (level = 0; (db->nexpansions + level) < MAX_LEVELS; level++) {

	b = compute_bucket(db, item->hash, db->nexpansions + level);
	if (level && b == prev)
	  continue;

	bucket_create(db, b);

	v = PAGENUM_TO_VOLUME(db, b);
	if ((rc = Page_get(db, v, b, PAGE_MAYCREATE, &page)) != DX_OK)
	  return(rc);

	/*
	 * Any of these calls may return ITEM_NOSPACE to indicate that there is
	 * insufficient free space in this page to make necessary changes, so it
	 * must be split.
	 */
	if (DX_OPMODE(mode) == DX_INSERT || DX_OPMODE(mode) == DX_INSERT_UNIQUE)
	  rc = Item_add(db, v, page, DX_OPMODE(mode), item, arg);
	else if (DX_OPMODE(mode) == DX_REPLACE)
	  rc = Item_replace(db, v, page, item, arg);
	else if (DX_OPMODE(mode) == DX_EDIT)
	  rc = Item_edit(db, v, page, item, arg);
	else
	  rc = DX_BADARG;

	if (rc == DX_OK) {
	  if (DX_DEBUG > 1)
		fprintf(stderr, "key=%s, b=%lu\n",
				(char *) item->key->dptr, (unsigned long) b);
	  Stats_count(db, ADD);
	  if (DX_CHECK)
		page_check(db, b);

	  /*
	   * Do a controlled split.
	   * XXX A DX_EDIT/DX_STORE may do a delete/add, which may be relevant.
	   */
	  if (DX_OPMODE(mode) == DX_INSERT
		  || DX_OPMODE(mode) == DX_INSERT_UNIQUE) {
		if (do_controlled_split(db, page))
		  rc = ehash_split(db, (PageNum) db->boundary, 1, NULL, NULL);
	  }
	  /*Segment_clear_fixed_page_active(db, b);*/
	  return(rc);
	}

	if (rc != ITEM_NOSPACE) {
	  /*Segment_clear_fixed_page_active(db, b);*/
	  return(rc);
	}

	/* There's no room in this bucket.  Let's split! */
	nsplit_moves = 0;
	rc = ehash_split(db, b, 1, &nsplit_moves, &item->hash);
	if (rc != DX_OK)
	  return(rc);
	prev = b;
  }

  return(DX_HASH_OVERFLOW);
}

/*
 * Search for a page that may contain the given item and perform the
 * specified function on it.  This function follows the implicit overflow
 * chain when looking for the item.  The item might not exist.
 */
Dx_rc
Ehash_get(Dx *db, Item *item, Dx_opmode mode, Ehash_xarg *arg)
{
  PageNum b, prev;
  Dx_ui64 level;
  Dx_rc rc;
  Page page;

  Segment_uncache(db, 0);

  prev = 0;		/* Shuts up compiler about its uninitialized use. */

  if (item->key->keyspace != NULL) {
	item->ks = item->key->keyspace;
	item->hash = item->keyhash ^ item->ks->hash;
  }
  else if (db->keyspace != NULL) {
	item->ks = db->keyspace;
	item->hash = item->keyhash ^ item->ks->hash;
  }
  else {
	item->hash = item->keyhash;
	item->ks = NULL;
  }
#ifdef NOTDEF
  item->hash = item->keyhash;
  item->ks = NULL;
#endif

  if (arg != NULL) {
	arg->io.dlen = 0;
	arg->io.uio_buf = NULL;
	arg->io.bytes_buffered = 0;
  }

  for (level = 0; (db->nexpansions + level) < MAX_LEVELS; level++) {

    b = compute_bucket(db, item->hash, db->nexpansions + level);

    if (level && b == prev)
      continue;

	if (! Ehash_bucket_exists(db, b))
	  return(DX_KEY_NOT_FOUND);

	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc != DX_OK) {
	  prev = b;
	  continue;
	}

	Stats_count(db, PROBE);
	if (DX_OPMODE(mode) == DX_PROBE || DX_OPMODE(mode) == DX_EXISTS)
	  rc = Item_probe(db, page, mode, item, arg);
	else if (DX_OPMODE(mode) == DX_DELETE) {
	  if (db->flags & DX_CONFIG_READ_ONLY) {
		/*Segment_clear_fixed_page_active(db, b);*/
		return(DX_READONLY);
	  }
	  rc = Item_delete(db, page, item);
	}
	else if (DX_OPMODE(mode) == DX_FETCH)
	  rc = Item_fetchx(db, page, item, arg);
	else
	  rc = DX_INTERNAL_ERROR;

	if (rc != ITEM_NOTFOUND) {
	  /*Segment_clear_fixed_page_active(db, b);*/
	  return(rc);
	}

    prev = b;
  }

  return(DX_KEY_NOT_FOUND);
}

/*
 * Initialize for iterating through all items in the db.
 */
void
Ehash_iter_init(Dx *db)
{

  db->iter.b = 0;
  db->iter.ind = 0;
}

/*
 * The guts of Dx_firstkeyx()/Dx_nextkeyx() db iteration.
 * Examine each page in the database, iterating through each key in a page.
 */
Dx_rc
Ehash_iter(Dx *db, Item *item, Dx_opmode mode, Ehash_xarg *arg)
{
  Dx_ui32 ind;
  Dx_keyspace ks;
  PageNum b;
  Page page;
  Dx_rc rc;

  Segment_uncache(db, 0);

  if (item->key->keyspace != NULL) {
	item->ks = item->key->keyspace;
	item->hash = 0;
  }
  else if (db->keyspace != NULL) {
	item->ks = db->keyspace;
	item->hash = 0;
  }
  else {
	item->hash = item->keyhash;
	ks.name = NULL;
	ks.hash = 0;
	item->ks = &ks;
  }
#ifdef NOTDEF
  item->hash = item->keyhash;
  item->ks = NULL;
#endif

  ind = db->iter.ind;
  b = db->iter.b;

  rc = ITEM_NOTFOUND;
  while (Ehash_bucket_exists(db, b)) {
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc == DX_OK) {
	  rc = Item_index(db, page, ind, mode, item, arg);
	  if (rc == DX_OK) {
		db->iter.b = b;
		db->iter.ind = ind + 1;
		break;
	  }
	  else if (rc != ITEM_NOTFOUND)
		break;
	}
	else if (rc != PAGE_NOTFOUND)
	  return(rc);

	ind = 0;
	b++;
	rc = ITEM_NOTFOUND;
  }

  /*Segment_clear_fixed_page_active(db, b);*/

  if (rc == ITEM_NOTFOUND)
	return(DX_KEY_NOT_FOUND);

  return(rc);
}

Dx_rc
Ehash_complete_expansion(Dx *db)
{
  PageNum b, limit;
  Dx_rc rc;

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(DX_READONLY);

  limit = (PageNum) Compute_expansion_buffers(db, db->nexpansions);

  for (b = (PageNum) db->boundary; b < limit; b++) {
	/*
	 * Note that a split may cause a cascade of splits, so the boundary
	 * variable may be updated several times by one call to ehash_split().
	 * In fact, the expansion may have been completed, so check.
	 */
	if (db->boundary == 0 && b != 0)
	  break;
	if (b < (PageNum) db->boundary)
	  b = (PageNum) db->boundary;

	rc = Page_exists(db, b);
	if (rc == DX_OK) {
	  if ((rc = ehash_split(db, b, 1, NULL, NULL)) != DX_OK)
		return(rc);
	}
	else if (rc != PAGE_NOTFOUND)
	  return(rc);
  }

  /*
   * The loop above will end without having done the bookkeeping to complete
   * the expansion if the last page didn't exist.
   */
  if (db->boundary) {
	db->nexpansions++;
	db->boundary = 0;
  }

  return(DX_OK);
}

/*
 * Initialize the extensible hashing data structures.
 *
 * Note: might initialize several pages here to reduce startup overhead
 * due to rehashing.
 */
Dx_rc
Ehash_create(Dx *db)
{

  if (Page_alloc(db, 0) == NULL)
	return(DX_INITFAILED);

  db->npages = 1;
  db->boundary = 0;
  db->nexpansions = 0;

  return(DX_OK);
}

static PageNum
next_bucket(Dx *db, PageNum b)
{
  PageNum last_page;

  last_page = (PageNum) Compute_expansion_buffers(db, db->nexpansions);
  return(b + last_page);
}

int
Ehash_bucket_exists(Dx *db, PageNum b)
{

  return(b < db->npages);
}

static int
next_bucket_exists(Dx *db, PageNum b)
{

  return(Ehash_bucket_exists(db, next_bucket(db, b)));
}

static PageNum
compute_bucket(Dx *db, Dx_hash h, Dx_ui32 exp)
{
  PageNum b;

  b = ((u_int) h) & (Compute_expansion_buffers(db, exp) - 1);

  if (exp == db->nexpansions && b < (PageNum) db->boundary)
	b = ((u_int) h) & (Compute_expansion_buffers(db, exp + 1) - 1);
	
  return(b);
}

static int
bucket_create(Dx *db, PageNum b)
{
  Dx_rc rc;

  if (Ehash_bucket_exists(db, b)) {
	if ((rc = Page_exists(db, b)) == DX_OK) {
	  if (DX_CHECK)
		page_check(db, b);
	  return(0);
	}
	if (rc != PAGE_NOTFOUND)
	  return(-1);
  }

  if (DX_DEBUG > 1)
	fprintf(stderr, "create=%lu\n", (unsigned long) b);

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(-1);

  expand(db, b);
  if (Page_alloc(db, b) == NULL)
	return(-1);

  return(0);
}

/*
 * Return 1 if a split should be done, 0 otherwise.
 */
static int
do_controlled_split(Dx *db, Page page)
{
  double real_usage;

#define DX_CONTROLLED_SPLIT_CHECK	5
  /*
   *
   */
  if ((++db->controlled_split_check % DX_CONTROLLED_SPLIT_CHECK) != 0)
	return(0);

#ifdef NOTDEF
  Page_load pl;
  Item_load(page, db->page_size, &pl, 0);
  real_usage = pl.actual_meta + pl.used;

  if (DX_CHECK && (pl.actual_meta + pl.used) > db->page_size) {
	fprintf(stderr, "do_controlled_split!\n");
	exit(1);
  }
#endif

  real_usage = Item_page_usage(db, page, db->page_size);

  if (real_usage >= EHASH_CONTROLLED_SPLIT_FACTOR)
	return(1);

  return(0);
}
