/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "local.h"

/*
 * From sdbm, an ndbm work-alike hashed database library
 * Author: Ozan Yigit <oz@nexus.yorku.ca>
 * Status: public domain.
 *
 * polynomial conversion ignoring overflows
 * [this seems to work remarkably well, in fact better
 * then the ndbm hash function. Replace at your own risk]
 * use: 65599   nice.
 *      65587   even better.
 *
 * [In one experiment, this function hashed 84165 symbols (English words
 * plus symbol table values) with no collisions. -bjb]
 */
inline Dx_hash
hash_sdbm32(Dx_ui8 *str, Dx_len xlen, Dx_ui32 level)
{
    register Dx_ui32 len, n;

	len = xlen;
    n = level;

#define HASHC   n = *str++ + 65587 * n

#ifdef NODUFF
    while (len--)
        n = HASHC;
#else

    if (len > 0) {
        Dx_ui32 loop;

        loop = (len + 8 - 1) >> 3;
        switch (len & (8 - 1)) {
        case 0:
            do {
                HASHC;
            case 7: HASHC;
            case 6: HASHC;
            case 5: HASHC;
            case 4: HASHC;
            case 3: HASHC;
            case 2: HASHC;
            case 1: HASHC;
            } while (--loop);
        }
    }
#endif
    return(n);
}

Dx_hash
hash_sdbm(Dx_ui8 *str, Dx_len xlen)
{

  return(hash_sdbm32(str, xlen, 0));
}

/*
 * lookup8.c, by Bob Jenkins, January 4 1997
 * hash(), hash2(), hash3, and mix() are externally useful functions.
 * Routines to test the hash are included if SELF_TEST is defined.
 * You can use this free for any purpose.  It has no warranty.
 */

#define hashsize(n) ((Dx_ui64)1<<(n))
#define hashmask(n) (hashsize(n)-1)

/*
 * mix -- mix 3 64-bit values reversibly.
 * mix() takes 48 machine instructions, but only 24 cycles on a superscalar
 *   machine (like Intel's new MMX architecture).  It requires 4 64-bit
 *   registers for 4::2 parallelism.
 * All 1-bit deltas, all 2-bit deltas, all deltas composed of top bits of
 *   (a,b,c), and all deltas of bottom bits were tested.  All deltas were
 *   tested both on random keys and on keys that were nearly all zero.
 *   These deltas all cause every bit of c to change between 1/3 and 2/3
 *   of the time (well, only 113/400 to 287/400 of the time for some
 *   2-bit delta).  These deltas all cause at least 80 bits to change
 *   among (a,b,c) when the mix is run either forward or backward (yes it
 *   is reversible).
 * This implies that a hash using mix64 has no funnels.  There may be
 *   characteristics with 3-bit deltas or bigger, I didn't test for
 *   those.
 */

#define mix64(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>43); \
  b -= c; b -= a; b ^= (a<<9); \
  c -= a; c -= b; c ^= (b>>8); \
  a -= b; a -= c; a ^= (c>>38); \
  b -= c; b -= a; b ^= (a<<23); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>35); \
  b -= c; b -= a; b ^= (a<<49); \
  c -= a; c -= b; c ^= (b>>11); \
  a -= b; a -= c; a ^= (c>>12); \
  b -= c; b -= a; b ^= (a<<18); \
  c -= a; c -= b; c ^= (b>>22); \
}

/*
 * hash() -- hash a variable-length key into a 64-bit value
 *   k     : the key (the unaligned variable-length array of bytes)
 *   len   : the length of the key, counting by bytes
 *   level : can be any 8-byte value
 * Returns a 64-bit value.  Every bit of the key affects every bit of
 * the return value.  No funnels.  Every 1-bit and 2-bit delta achieves
 * avalanche.  About 41+5len instructions.
 * 
 * The best hash table sizes are powers of 2.  There is no need to do
 * mod a prime (mod is sooo slow!).  If you need less than 64 bits,
 * use a bitmask.  For example, if you need only 10 bits, do
 *   h = (h & hashmask(10));
 * In which case, the hash table should have hashsize(10) elements.
 * 
 * If you are hashing n strings (u_int8 **)k, do it like this:
 *   for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);
 * 
 * By Bob Jenkins, Jan 4 1997.  bob_jenkins@compuserve.com.  You may
 * use this code any way you wish, private, educational, or commercial,
 * as long as this whole comment accompanies it.
 * 
 * See http://ourworld.compuserve.com/homepages/bob_jenkins/evahash.htm
 * Use for hash table lookup, or anything where one collision in 2^^64
 * is acceptable.  Do NOT use for cryptographic purposes.
 */

/*
 * The key, the length of the key, the previous hash, or an arbitrary value.
 */
inline Hash64
hash_bj64(Dx_ui8 *k, Dx_ui32 length, Dx_ui64 level)
{
  register Hash64 a,b,c;
  Dx_ui32 len;

  /* Set up the internal state */
  len = length;
  a = b = level;                         /* the previous hash value */
  c = 0x9e3779b97f4a7c13LL; /* the golden ratio; an arbitrary value */

  /*---------------------------------------- handle most of the key */
  while (len >= 24)
  {
    a += (k[0]+((Dx_ui64)k[ 1]<< 8)
		  +((Dx_ui64)k[ 2]<<16)+((Dx_ui64)k[ 3]<<24)
		  +((Dx_ui64)k[4 ]<<32)+((Dx_ui64)k[ 5]<<40)
		  +((Dx_ui64)k[ 6]<<48)+((Dx_ui64)k[ 7]<<56));
    b += (k[8]+((Dx_ui64)k[ 9]<< 8)+((Dx_ui64)k[10]<<16)
		  +((Dx_ui64)k[11]<<24) +((Dx_ui64)k[12]<<32)
		  +((Dx_ui64)k[13]<<40)+((Dx_ui64)k[14]<<48)
		  +((Dx_ui64)k[15]<<56));
    c += (k[16] +((Dx_ui64)k[17]<< 8)+((Dx_ui64)k[18]<<16)
		  +((Dx_ui64)k[19]<<24) +((Dx_ui64)k[20]<<32)
		  +((Dx_ui64)k[21]<<40)+((Dx_ui64)k[22]<<48)
		  +((Dx_ui64)k[23]<<56));
    mix64(a,b,c);
    k += 24; len -= 24;
  }

  /*------------------------------------- handle the last 63 bytes */
  c += length;
  switch(len)              /* all the case statements fall through */
  {
  case 23: c+=((Dx_ui64)k[22]<<56);
  case 22: c+=((Dx_ui64)k[21]<<48);
  case 21: c+=((Dx_ui64)k[20]<<40);
  case 20: c+=((Dx_ui64)k[19]<<32);
  case 19: c+=((Dx_ui64)k[18]<<24);
  case 18: c+=((Dx_ui64)k[17]<<16);
  case 17: c+=((Dx_ui64)k[16]<<8);
    /* the first byte of c is reserved for the length */
  case 16: b+=((Dx_ui64)k[15]<<56);
  case 15: b+=((Dx_ui64)k[14]<<48);
  case 14: b+=((Dx_ui64)k[13]<<40);
  case 13: b+=((Dx_ui64)k[12]<<32);
  case 12: b+=((Dx_ui64)k[11]<<24);
  case 11: b+=((Dx_ui64)k[10]<<16);
  case 10: b+=((Dx_ui64)k[ 9]<<8);
  case  9: b+=((Dx_ui64)k[ 8]);
  case  8: a+=((Dx_ui64)k[ 7]<<56);
  case  7: a+=((Dx_ui64)k[ 6]<<48);
  case  6: a+=((Dx_ui64)k[ 5]<<40);
  case  5: a+=((Dx_ui64)k[ 4]<<32);
  case  4: a+=((Dx_ui64)k[ 3]<<24);
  case  3: a+=((Dx_ui64)k[ 2]<<16);
  case  2: a+=((Dx_ui64)k[ 1]<<8);
  case  1: a+=((Dx_ui64)k[ 0]);
    /* case 0: nothing left to add */
  }
  mix64(a,b,c);
  /*-------------------------------------------- report the result */
  return c;
}

Dx_hash
hash_bj(Dx_ui8 *str, Dx_len len, Dx_hash init)
{

#if DX_BITS == 64
  return(hash_bj64((Dx_ui8 *) str, (Dx_ui32) len, init));
#else
  return(hash_sdbm32((Dx_ui8 *) str, (Dx_ui32) len, init));
#endif
}
