/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "local.h"
#include "item.h"

/*
 * The database stores key/value pairs called items.
 * Basically, an item is a pair of Dx_datums plus associated meta-data.
 *
 * Items are generally packed into a page.  All pages are the same size.
 * A page consists of a fixed-length page header, which comprises
 * some general meta-data for the page, followed by a fixed-length vector of
 * hash table buckets, and then zero or more fixed-length item headers,
 * each of which describes one item stored in the page.  Space for key and
 * value data is allocated from the end of the page towards the start so
 * that the fixed-length item headers can be contiguous without any space
 * wasted due to alignment requirements.
 *
 * The page meta-data is as follows:
 * Field offset     Field purpose           Field type/size
 * +0               Number of headers        Dx_ui16
 * +2               Flags                    Dx_ui16
 *                    bit 0/1:  Hash table size (00=16, 01=32, 10=64, 11=128)
 *                    bit 2:    Reorg
 * +4               Free space pointer       Dx_ui16
 * +6               Total bytes free         Dx_ui16
 *
 * +8               Hash table starts        Dx_ui16 (* 8, 16, 32, 64)
 *
 * (Followed by start of vector of headers)
 *
 * The total number of headers (NHEADERS) includes those of deleted items.
 * The Reorg flag is set if the page is "dirty" (i.e., when items have been
 * deleted but the space has not been reclaimed yet).
 * The free space pointer (ENDFREE) is the index of the first unavailable
 * byte in the page for allocating data.
 * The free count (AVAIL) is actually a lower bound estimate
 * of the number of bytes available in the page.  It may not be accurate but
 * there are at least this many bytes free, possibly after a reorganization.
 *
 * Each item header has the following layout:
 *  +0              Key offset               Dx_ui16
 *  +2              Value offset             Dx_ui16
 *  +4              Key length               Dx_ui16
 *  +6              Value length             Dx_ui16
 *  +8              UID                      Dx_uid
 *  +16             Key hash                 Dx_hash
 * +20/24           Hash chain pointer       Dx_ui16
 * +22/26           Flags                    Dx_ui16
 *                    bit 0:   Indirect key flag
 *                    bit 1:   Indirect value flag
 *                    bit 2/3: Key alignment   (00=none, 01=2, 10=4, 11=8)
 *                    bit 4/5: Value alignment (00=none, 01=2, 10=4, 11=8)
 *                    bit 6:   Deleted flag
 * +24/28
 *
 * Every item has a unique id (uid).  No two items in the db have the same uid
 * and a uid can never be reused in a particular database.  A DX_REPLACE
 * operation actually creates a new item and it will have a different uid.
 * Uids are purely for internal use and should not be visible to the user.
 *
 * If a non-aligned value (or key) is bigger than ITEM_INDIRECT_THRESHOLD,
 * the value (or key) is too big to be stored in any page and it is stored
 * separately in an indirect segment, but the meta-data for the indirect
 * value (or key) is stored in this page.  Additionally, if the sum of the
 * lengths of the key and value is bigger than ITEM_INDIRECT_THRESHOLD,
 * the value is stored indirectly.
 *
 * NB: The value of ITEM_INDIRECT_THRESHOLD must not exceed the maximum
 * value that can fit in the data type used for the key and value lengths.
 * Also, ITEM_INDIRECT_THRESHOLD must be sufficiently smaller than the page
 * size lest it become impossible for some keys/values to be stored at all.
 *
 * An "indirect descriptor header" is created where the value (or key) would
 * ordinarily go.  This header tells where the data actually is.
 * For an indirect value (or key), the value (key) length in the item header is
 * the size of the indirect descriptor header; the size of the value (key)
 * itself is part of the indirect descriptor header.
 * An indirect key or value is assumed to be page aligned, meaning it
 * satisfies the worst case alignment constraint.
 * It is possible for the length of an indirect value to not exceed
 * ITEM_INDIRECT_THRESHOLD.  In fact, it may be zero, in which case
 * the IDESC_SEGLEN will also be zero.
 *
 * XXX Look at reducing the number of endian conversions.
 */

/*
 * Because items that hash to the same page have key hashes that are similar
 * in their least significant bits, a different hash value is needed to hash
 * within a page lest they all map to the same bucket.
 * Note that this should jibe with the maximum hashtab_sizes[] value.
 */
enum { HASH_SHIFT = (Dx_hash_nbits - 7) };

static INLINE Dx_hash
H(Dx_hash hash)
{

  return((Dx_hash) (hash ^ (hash >> HASH_SHIFT)));
}

static const Dx_ui32 hashtab_sizes[4] = { 8, 16, 32, 64 };

static INLINE Dx_ui32
HASHTAB_SIZE_OF_PAGE_SIZE(Dx_ui32 page_size)
{

  if (page_size <= 2048)
	return(hashtab_sizes[0]);
  else if (page_size <= 4096)
	return(hashtab_sizes[1]);
  else if (page_size <= 16384)
	return(hashtab_sizes[2]);
  else
	return(hashtab_sizes[3]);
}

static INLINE Dx_ui32
ITEM_HASHTAB_NBUCKETS(Dx_ui32 flags)
{
  Dx_ui32 val;

  val = hashtab_sizes[flags & 03];
  return(val);
}

static INLINE Dx_ui32
ITEM_HASHTAB_SIZE(Dx_ui32 flags)
{
  Dx_ui32 val;

  val = Hashtab_bucket_s * ITEM_HASHTAB_NBUCKETS(flags);
  return(val);
}

static INLINE Dx_ui16
ITEM_START_HEADERS(Dx_ui32 flags)
{
  Dx_ui16 val;

  val = ITEM_HASHTAB_OFFSET + ITEM_HASHTAB_SIZE(flags);
  return(val);
}

static INLINE Dx_ui16
ITEM_HEADER(Page pagep, Dx_ui32 n, Dx_ui32 flags)
{
  Dx_ui16 val;

  val = (n * ITEM_HEADER_SIZE) + ITEM_START_HEADERS(flags);
  return(val);
}

static INLINE Dx_ui16
ITEM_CONTIG(Dx *db, Page pagep, Dx_ui32 nheaders, Dx_ui32 flags)
{
  Dx_ui16 val;

  val = (Dx_ui16) (ITEM_GET_ENDFREE(db, pagep)
				   - ITEM_HEADER(pagep, nheaders, flags));
  return(val);
}
		
static INLINE Dx_ui16
ITEM_TOTAL_META_SIZE(Dx *db, Page pagep, Dx_ui32 flags)
{
  Dx_ui16 val;

  val = META_HEADER_SIZE
	+ ITEM_HASHTAB_SIZE(flags)
	+ (ITEM_GET_NHEADERS(db, pagep) * ITEM_HEADER_SIZE);
  return(val);
}

static INLINE Hashtab_bucket_t
pack_hashtab_val(Dx *db, Page page, Dx_ui32 page_size, Dx_ui16 ind,
				 Hashtab_bucket_t val)
{

  ITEM_PUT_HASHTAB_VAL(db, page, ind, val);
  return(val);
}

static INLINE Hashtab_bucket_t
unpack_hashtab_val(Dx *db, Page page, Dx_ui32 page_size, Dx_ui16 ind)
{

  return(ITEM_GET_HASHTAB_VAL(db, page, ind));
}

static INLINE Dx_len
ITEM_INDIRECT_THRESHOLD(Dx_ui32 page_size)
{
  Dx_len val;

  val = (Dx_len) (ITEM_INDIRECT_THRESHOLD_FACTOR * page_size);
  return(val);
}

static const Dx_ui32 alignments[8] = {
  1, 2, 4, 8, 16, 16, 16, 16	/* Subscripted by DX_DATUM_ALIGN0, ... */
};

/*
 * XXX DX_DATUM_ALIGN0, ALIGN1, ..., must have values that agree with the
 * way the bits are used here.
 */
static INLINE Dx_ui32
DATUM_ALIGNMENT(Dx_datum *datum)
{

  return(alignments[datum->flags & 03]);
}

typedef struct {
  void *dptr;
  Dx_len offset;
} Contig_callback_state;

static Dx_rc item_add(Dx *db, Volume *v, Page page, Item *item,
					  Ehash_xarg *arg);
static Dx_rc item_delete(Dx *db, Page page, Dx_ui8 *hptr);
static Dx_rc item_edit_direct(Dx *db, Page page, Dx_ui8 *hptr,
							  Dx_ui16 valuep, Dx_ui16 valuelen,
							  Dx_len start_pos, Dx_len requested_delete_length,
							  Dx_len add_length, Dx_datum *value,
							  Ehash_xarg *arg);
static Dx_rc item_edit_indirect(Dx *db, Volume *, Page page,
								Dx_ui8 *hptr, Dx_len start_pos,
								Dx_len requested_delete_length,
								Dx_len add_length, Dx_datum *value,
								Ehash_xarg *arg);
static int item_show_bucket(Dx *db, Page page, Dx_ui32 page_size,
							Dx_ui32 hindex);
static Dx_rc item_substitute(Dx *, Page page, Dx_ui8 *hptr,
							 Item *item);
static Dx_rc item_force_indirect_value(Dx *db, Volume *volume, Page page,
									   Dx_ui8 *hptr);
static Dx_rc item_replace_direct_value(Dx *db, Volume *volume,
									   Page page,
									   Dx_ui8 *hptr, Dx_datum *value);
static Dx_rc item_replace_indirect_value(Dx *db, Volume *volume, Page page,
										 Dx_ui8 *hptr,
										 Dx_datum *value);
static Dx_rc item_replace(Dx *db, Volume *v, Page page,
						  Dx_ui8 *hptr, Item_flags_t hflags, Dx_len valuelen,
						  Item *item, Ehash_xarg *arg);
static Dx_len item_stripe_contig_callback(void *arg, void *dptr,
										  Dx_len io_dlen, Dx_desc *desc);


static Dx_rc item_stripe_contig(Dx *db, Dx_ui8 *id, Ehash_xarg *arg);
static Dx_rc item_stripe_copy_uio(Dx *db, Volume *v, Dx_datum *datum,
								  Dx_len zero_fill, Segment *istripe_seg,
								  Physoff *istripe_offset,
								  Seglen *istripe_seglen, Ehash_xarg *arg);
static Dx_rc item_fetch_indirect_datum(Dx *db, Dx_ui8 *id, Ehash_xarg *arg);
static Chainp_t chain_zero = 0;
static Item_flags_t flags_zero = 0;
static Nheaders_t nheaders_zero = 0;

typedef enum {
  PACK_NOP, PACK_OR, PACK_ADD, PACK_SUB, PACK_NOT
} pack_op;

/* XXX GCC won't INLINE varargs functions. */

static INLINE void
pack_meta_fast(Dx *db, Page page, Nheaders_t *nheaders, Metaflags_t *flags,
			   Endfree_t *end_free, Avail_t *avail)
{

  if (nheaders != NULL)
	ITEM_PUT_NHEADERS(db, page, *nheaders);
  if (flags != NULL)
	ITEM_PUT_METAFLAGS(db, page, *flags);
  if (end_free != NULL)
	ITEM_PUT_ENDFREE(db, page, *end_free);
  if (avail != NULL)
	ITEM_PUT_AVAIL(db, page, *avail);
}

static INLINE void
pack_meta_add_avail(Dx *db, Page page, int avail)
{
  int new_avail;

  new_avail = ITEM_GET_AVAIL(db, page) + avail;
  ITEM_PUT_AVAIL(db, page, (Avail_t) new_avail);
}

static INLINE void
pack_meta_or_flags(Dx *db, Page page, Metaflags_t flags)
{

  ITEM_PUT_METAFLAGS(db, page, ITEM_GET_METAFLAGS(db, page) | flags);
}

#ifdef NOTDEF
static void
pack_meta(Page page, Dx_ui32 page_size, char *fmt, ...)
{
  char *s;
  pack_op op;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;
  op = PACK_NOP;

  while (*s != '\0') {
	switch (*s) {
	case '|':
	  op = PACK_OR;
	  break;
	case '+':
	  op = PACK_ADD;
	  break;
	case '-':
	  op = PACK_SUB;
	  break;
	case '!':
	  op = PACK_NOT;
	  break;

	case 'a':
	  if (op == PACK_ADD)
		ITEM_AVAIL(page) += va_arg(ap, Avail_t);
	  else if (op == PACK_SUB)
		ITEM_AVAIL(page) -= va_arg(ap, Avail_t);
	  else
		ITEM_AVAIL(page) = va_arg(ap, Avail_t);
	  op = PACK_NOP;
	  break;
	case 'e':
	  ITEM_END_FREE(page) = va_arg(ap, End_free_t);
	  op = PACK_NOP;
	  break;
	case 'f':
	  if (op == PACK_OR)
		ITEM_META_FLAGS(page) |= va_arg(ap, Meta_flags_t);
	  else if (op == PACK_NOT)
		ITEM_META_FLAGS(page) &= ~(va_arg(ap, Meta_flags_t));
	  else
		ITEM_META_FLAGS(page) = va_arg(ap, Meta_flags_t);
	  op = PACK_NOP;
	  break;
	case 'n':
	  ITEM_NHEADERS(page) = va_arg(ap, Nheaders_t);
	  op = PACK_NOP;
	  break;
	default:
	  fprintf(stderr, "pack_meta: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}

	s++;
  }

  va_end(ap);
}
#endif

static INLINE void
unpack_meta(Dx *db, Page page, Avail_t *avail, Endfree_t *end_free,
			Metaflags_t *flags, Nheaders_t *nheaders)
{

  if (avail != NULL)
	*avail = ITEM_GET_AVAIL(db, page);
  if (end_free != NULL)
	*end_free = ITEM_GET_ENDFREE(db, page);
  if (flags != NULL)
	*flags = ITEM_GET_METAFLAGS(db, page);
  if (nheaders != NULL)
	*nheaders = ITEM_GET_NHEADERS(db, page);
}

#ifdef NOTDEF
unpack_meta(Page page, Dx_ui32 page_size, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'a':
	  {
		Avail_t *v;
		v = va_arg(ap, Avail_t *);
		*v = ITEM_AVAIL(page);
		break;
	  }
	case 'e':
	  {
		End_free_t *v;
		v = va_arg(ap, End_free_t *);
		*v = ITEM_END_FREE(page);
		break;
	  }
	case 'f':
	  {
		Meta_flags_t *v;
		v = va_arg(ap, Meta_flags_t *);
		*v = ITEM_META_FLAGS(page);
		break;
	  }
	case 'n':
	  {
		Nheaders_t *v;
		v = va_arg(ap, Nheaders_t *);
		*v = ITEM_NHEADERS(page);
		break;
	  }
	default:
	  fprintf(stderr, "unpack_meta: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	s++;
  }

  va_end(ap);
}
#endif

#ifdef NOTDEF
static void
pack_idescriptor(Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'd':
	  IDESC_DLEN(id) = va_arg(ap, Dx_len);
	  break;
	case 'o':
	  IDESC_OFFSET(id) = va_arg(ap, Physoff);
	  break;
#ifdef NOTDEF
	case 's':
	  {
		int c, i;
		Stripe *s;

		/* Two arguments: a count followed by a pointer to a vector. */
		c = va_arg(ap, Idesc_count_t);
		IDESC_COUNT(id) = c;
		s = va_arg(ap, Stripe *);
		for (i = 0; i < c; i++)
		  IDESC_OFFSET_VAL(id, i) = s[i].offset;
		break;
	  }
#endif
	case 's':
	  IDESC_SEGLEN(id) = va_arg(ap, Seglen);
	  break;
	case 'v':
	  IDESC_VOLUME(id) = va_arg(ap, Physvolnum);
	  break;
	default:
	  fprintf(stderr, "pack_idescriptor: bad flag `%c'\n", *s);
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}

	s++;
  }

  va_end(ap);
}
#endif

#ifdef NOTDEF
static void
unpack_idescriptor(Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'd':
	  {
		Dx_len *v;

		v = va_arg(ap, Dx_len *);
		*v = IDESC_DLEN(id);
		break;
	  }
	case 'o':
	  {
		Physoff *v;

		v = va_arg(ap, Physoff *);
		*v = IDESC_OFFSET(id);
		break;
	  }
	case 's':
	  {
		Seglen *v;

		v = va_arg(ap, Seglen *);
		*v = IDESC_SEGLEN(id);
		break;
	  }
	case 'v':
	  {
		Physvolnum *v;

		v = va_arg(ap, Physvolnum *);
		*v = IDESC_VOLUME(id);
		break;
	  }
	default:
	  fprintf(stderr, "unpack_idescriptor: bad flag `%c'\n", *s);
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	s++;
  }

  va_end(ap);
}
#endif

/*
 * The stripe header contains a count of the number of entries (volume,
 * offset, segment length, data length) in this segment plus a pointer
 * (volume, offset, segment length) to the next such stripe
 * descriptor in the chain.
 */
void
pack_istripe_hdr(Dx *db, Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'c':
	  ISTRIPE_HDR_PUT_COUNT(db, id, va_arg(ap, Istripe_count_t));
	  break;
	case 'o':
	  ISTRIPE_HDR_PUT_OFFSET(db, id, va_arg(ap, Physoff));
	  break;
	case 's':
	  ISTRIPE_HDR_PUT_SEGLEN(db, id, va_arg(ap, Seglen));
	  break;
	case 'v':
	  ISTRIPE_HDR_PUT_VOLUME(db, id, va_arg(ap, Physvolnum));
	  break;
	default:
	  fprintf(stderr, "pack_istripe_hdr: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}

	s++;
  }

  va_end(ap);
}

void
unpack_istripe_hdr(Dx *db, Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'c':
	  {
		Istripe_count_t *v;

		v = va_arg(ap, Istripe_count_t *);
		*v = ISTRIPE_HDR_GET_COUNT(db, id);
		break;
	  }
	case 'o':
	  {
		Physoff *v;

		v = va_arg(ap, Physoff *);
		*v = ISTRIPE_HDR_GET_OFFSET(db, id);
		break;
	  }
	case 's':
	  {
		Seglen *v;

		v = va_arg(ap, Seglen *);
		*v = ISTRIPE_HDR_GET_SEGLEN(db, id);
		break;
	  }
	case 'v':
	  {
		Physvolnum *v;

		v = va_arg(ap, Physvolnum *);
		*v = ISTRIPE_HDR_GET_VOLUME(db, id);
		break;
	  }
	default:
	  fprintf(stderr, "unpack_istripe_hdr: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	s++;
  }

  va_end(ap);
}

#ifdef NOTDEF
/*
 * Format codes:
 *    c - chain field (an index)
 *    f - flags
 *    h - hash
 *    k - key or indirect key structure (an index)
 *    K - size of the key or indirect key structure, in bytes
 *    u - UID
 *    v - value or indirect value structure (an index)
 *
 * Code prefix modifiers:
 *    | - Logical or value with existing value
 *    + - Add value to existing value
 */
static void
pack_header(Dx_ui8 *header, char *fmt, ...)
{
  pack_op op;
  char *s;
  Dx_uid uid;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  op = PACK_NOP;
  while (*s != '\0') {
	if (*s == '|') {
	  op = PACK_OR;
	  s++;
	  continue;
	} else if (*s == '+') {
	  op = PACK_ADD;
	  s++;
	  continue;
	}

	switch (*s) {
	case 'c':
	  ITEM_CHAINP(header) = va_arg(ap, Chainp_t);
	  break;
	case 'f':
	  if (op == PACK_OR)
		ITEM_FLAGS(header) |= va_arg(ap, Item_flags_t);
	  else
		ITEM_FLAGS(header) = va_arg(ap, Item_flags_t);
	  break;
	case 'h':
	  ITEM_HASH(header) = va_arg(ap, Dx_hash);
	  break;
	case 'k':
	  ITEM_KEYP(header) = va_arg(ap, Keyp_t);
	  break;
	case 'K':
	  ITEM_KEYLEN(header) = va_arg(ap, Keylen_t);
	  break;
	case 'u':
	  uid = va_arg(ap, Dx_uid);
	  ITEM_UID_HI(header) = uid.tv_sec;
	  ITEM_UID_LO(header) = uid.count;
	  break;
	case 'v':
	  ITEM_VALUEP(header) = va_arg(ap, Valuep_t);
	  break;
	case 'V':
	  ITEM_VALUELEN(header) = va_arg(ap, Valuelen_t);
	  break;
	default:
	  fprintf(stderr, "pack_header: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}

	op = PACK_NOP;
	s++;
  }
}
#endif

static INLINE void
pack_header_or_flags(Dx *db, Dx_ui8 *header, Item_flags_t flags)
{

  ITEM_PUT_FLAGS(db, header, ITEM_GET_FLAGS(db, header) | flags);
}

static INLINE void
pack_header_meta(Dx *db, Dx_ui8 *header, Chainp_t *chainp, Item_flags_t *flags,
				 Dx_hash *hash, Dx_uid *uid)
{

  if (chainp != NULL)
	ITEM_PUT_CHAINP(db, header, *chainp);
  if (flags != NULL)
	ITEM_PUT_FLAGS(db, header, *flags);
  if (hash != NULL)
	ITEM_PUT_HASH(db, header, *hash);
  if (uid != NULL) {
	ITEM_PUT_UIDHI(db, header, uid->tv_sec);
	ITEM_PUT_UIDLO(db, header, uid->count);
  }

}

static INLINE void
pack_header_kv(Dx *db, Dx_ui8 *header, Dx_ui16 *keyp, Dx_ui16 *keylen,
			   Dx_ui16 *valuep, Dx_ui16 *valuelen)
{

  if (keyp != NULL)
	ITEM_PUT_KEYP(db, header, *keyp);
  if (keylen != NULL)
	ITEM_PUT_KEYLEN(db, header, *keylen);
  if (valuep != NULL)
	ITEM_PUT_VALUEP(db, header, *valuep);
  if (valuelen != NULL)
	ITEM_PUT_VALUELEN(db, header, *valuelen);
}

#ifdef NOTDEF
static void
unpack_header(Dx_ui8 *header, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'c':
	  {
		Chainp_t *v;
		v = va_arg(ap, Chainp_t *);
		*v = ITEM_CHAINP(header);
		break;
	  }
	case 'f':
	  {
		Item_flags_t *v;
		v = va_arg(ap, Item_flags_t *);
		*v = ITEM_FLAGS(header);
		break;
	  }
	case 'h':
	  {
		Dx_hash *v;
		v = va_arg(ap, Dx_hash *);
		*v = ITEM_HASH(header);
		break;
	  }
	case 'k':
	  {
		Keyp_t *v;
		v = va_arg(ap, Keyp_t *);
		*v = ITEM_KEYP(header);
		break;
	  }
	case 'K':
	  {
		Keylen_t *v;
		v = va_arg(ap, Keylen_t *);
		*v = ITEM_KEYLEN(header);
		break;
	  }
	case 'u':
	  {
		Dx_uid *v;
		v = va_arg(ap, Dx_uid *);
		v->tv_sec = ITEM_UID_HI(header);
		v->count = ITEM_UID_LO(header);
		break;
	  }
	case 'v':
	  {
		Valuep_t *v;
		v = va_arg(ap, Valuep_t *);
		*v = ITEM_VALUEP(header);
		break;
	  }
	case 'V':
	  {
		Valuelen_t *v;
		v = va_arg(ap, Valuelen_t *);
		*v = ITEM_VALUELEN(header);
		break;
	  }
	default:
	  fprintf(stderr, "unpack_header: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	s++;
  }
}
#endif

static INLINE void
unpack_header_kv(Dx *db, Dx_ui8 *header, Dx_ui16 *keyp, Dx_ui16 *keylen,
				 Dx_ui16 *valuep, Dx_ui16 *valuelen)
{

  if (keyp != NULL)
	*keyp = ITEM_GET_KEYP(db, header);
  if (keylen != NULL)
	*keylen = ITEM_GET_KEYLEN(db, header);
  if (valuep != NULL)
	*valuep = ITEM_GET_VALUEP(db, header);
  if (valuelen != NULL)
	*valuelen = ITEM_GET_VALUELEN(db, header);
}

static INLINE void
unpack_header_meta(Dx *db, Dx_ui8 *header, Chainp_t *chainp,
				   Item_flags_t *flags, Dx_hash *hash, Dx_uid *uid)
{

  if (chainp != NULL)
	*chainp = ITEM_GET_CHAINP(db, header);
  if (flags != NULL)
	*flags = ITEM_GET_FLAGS(db, header);
  if (hash != NULL)
	*hash = ITEM_GET_HASH(db, header);
  if (uid != NULL) {
	uid->tv_sec = ITEM_GET_UIDHI(db, header);
	uid->count = ITEM_GET_UIDLO(db, header);
  }
}

#ifdef NOTDEF
void
pack_index(Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  Dx_uid uid;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'h':
	  IINDEX_HASH(id) = va_arg(ap, Dx_hash);
	  break;
	case 'u':
	  uid = va_arg(ap, Dx_uid);
	  IINDEX_UID_HI(id) = uid.tv_sec;
	  IINDEX_UID_LO(id) = uid.count;
	  break;
	default:
	  fprintf(stderr, "pack_index: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}

	s++;
  }

  va_end(ap);
}

void
unpack_index(Dx_ui8 *id, char *fmt, ...)
{
  char *s;
  va_list ap;

  va_start(ap, fmt);
  s = fmt;

  while (*s != '\0') {
	switch (*s) {
	case 'h':
	  {
		Dx_hash *v;
		v = va_arg(ap, Dx_hash *);
		*v = IINDEX_HASH(id);
		break;
	  }
	case 'u':
	  {
		Dx_uid *v;
		v = va_arg(ap, Dx_uid *);
		v->tv_sec = IINDEX_UID_HI(id);
		v->count = IINDEX_UID_LO(id);
		break;
	  }
	default:
	  fprintf(stderr, "unpack_index: bad flag\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	s++;
  }

  va_end(ap);
}
#endif

static INLINE void
item_make_new_uid(Dx_uid *uid)
{

  if (++uid->count == 0)
	uid->tv_sec++;
}

static INLINE void
item_advance_uid(Dx *db)
{

  item_make_new_uid(&db->current_uid);
}

/*
 * Compute the total amount of space needed to store ITEM, taking
 * alignment into account.  If there's insufficient space available, return
 * ITEM_NOSPACE.
 * If the key or value are indirect, they stay indirect even if they
 * could fit directly.
 */
static INLINE Dx_rc
item_need(Dx_len thresh, Item *item, Dx_ui16 avail, Dx_ui16 end_free,
		  Item_flags_t hflags, int *indirect_key, int *indirect_value,
		  Dx_ui32 *key_alignment_bytes, Dx_ui32 *value_alignment_bytes,
		  Dx_ui32 *needp)
{
  Dx_ui32 base_need, need;

  if (ITEM_KEY_INDIRECT(hflags))
	*indirect_key = 1;
  else
	*indirect_key = (item->key->dlen > thresh);
  if (ITEM_VALUE_INDIRECT(hflags))
	*indirect_value = 1;
  else
	*indirect_value = (item->value->dlen > thresh);

  if (!(*indirect_key) && !(*indirect_value)
	  && (item->key->dlen + item->value->dlen) > thresh)
	*indirect_value = 1;

  if (*indirect_key)
	base_need = IDESC_SIZE;
  else
	base_need = (Dx_ui32) item->key->dlen;

  if (*indirect_value)
	base_need += IDESC_SIZE;
  else
	base_need += (Dx_ui32) item->value->dlen;

  need = ITEM_HEADER_SIZE + base_need;

  if (need > avail)
	return(ITEM_NOSPACE);

  if (*indirect_key) {
	if (end_free < IDESC_SIZE)
	  return(ITEM_NOSPACE);
	*key_alignment_bytes = (end_free - IDESC_SIZE)
	  % ALIGNMENT32_CONSTRAINT;
  }
  else {
	if (end_free < (Dx_ui16) item->key->dlen)
	  return(ITEM_NOSPACE);
	*key_alignment_bytes = (end_free - (Dx_ui16) item->key->dlen)
	  % DATUM_ALIGNMENT(item->key);
  }

  if (*indirect_value) {
	if (end_free < IDESC_SIZE)
	  return(ITEM_NOSPACE);
	*value_alignment_bytes = (end_free - IDESC_SIZE)
	  % ALIGNMENT32_CONSTRAINT;
  }
  else {
	if (end_free < (Dx_ui16) item->value->dlen)
	  return(ITEM_NOSPACE);
	*value_alignment_bytes = (end_free - (Dx_ui16) item->value->dlen)
	  % DATUM_ALIGNMENT(item->value);
  }

  if (!(*indirect_key) && !(*indirect_value)) {
	if (*key_alignment_bytes <= *value_alignment_bytes) {
	  need += *key_alignment_bytes;
	  if ((int) end_free - (int) item->key->dlen - (int) *key_alignment_bytes
		  - (int) item->value->dlen < 0)
		return(ITEM_NOSPACE);
	  *value_alignment_bytes = ((end_free - (Dx_ui16) item->key->dlen
								 - *key_alignment_bytes)
								- (Dx_ui16) item->value->dlen)
		% DATUM_ALIGNMENT(item->value);
	  need += *value_alignment_bytes;
	} else {
	  need += *value_alignment_bytes;
	  if ((int) end_free - (int) item->value->dlen
		  - (int) *value_alignment_bytes - (int) item->key->dlen < 0)
		return(ITEM_NOSPACE);
	  *key_alignment_bytes = ((end_free - (Dx_ui16) item->value->dlen
							   - *value_alignment_bytes)
							  - (Dx_ui16) item->key->dlen)
		% DATUM_ALIGNMENT(item->key);
	  need += *key_alignment_bytes;
	}
  }
  else {
	need += *key_alignment_bytes;
	need += *value_alignment_bytes;
  }

  if (need > avail)
	return(ITEM_NOSPACE);

  *needp = need;
  return(DX_OK);
}

static INLINE void
item_add_key(Dx *db, Page page, Dx_ui32 page_size, Dx_ui8 *hptr, Dx_datum *key,
			 Dx_ui32 alignment)
{
  Dx_ui32 need;
  Dx_ui16 end_free, iptr, len;
  Dx_ui8 *ptr;

  /* Copy the new key. */
  need = (Dx_ui32) key->dlen + alignment;
  unpack_meta(db, page, NULL, &end_free, NULL, NULL);
  iptr = end_free - need;
  ptr = (Dx_ui8 *) page + iptr;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(ptr, key->dptr, (size_t) key->dlen);

  /* Fill in a new header entry. */
  pack_header_meta(db, hptr, &chain_zero, NULL, NULL, NULL);
  len = (Dx_ui16) key->dlen;
  pack_header_kv(db, hptr, &iptr, &len, NULL, NULL);
  pack_header_or_flags(db, hptr, ITEM_KEY_ALIGNMENT_FLAG_BITS(key->flags));

  pack_meta_fast(db, page, NULL, NULL, &iptr, NULL);
}

/*
 * The key needs to be stored indirectly.
 * Keys are striped but may not be provided by a get-function.
 */
static INLINE Dx_rc
item_add_indirect_key(Dx *db, Volume *volume, Page page, Dx_ui8 *hptr,
					  Dx_datum *key, Dx_hash hash, Dx_ui32 alignment)
{
  Dx_ui32 need;
  Dx_ui16 end_free, iptr, len;
  Physoff istripe_offset;
  Segment istripe_seg;
  Seglen istripe_seglen;
  Volume *v;
  Dx_ui8 *id;
  Dx_rc rc;

  /* Allocate and copy the new key. */
  v = Random_volume(db);
  rc = Item_stripe_copy(db, v, key, 0, &istripe_seg, &istripe_offset,
						&istripe_seglen, NULL);
  if (rc != DX_OK)
	return(rc);

  need = IDESC_SIZE + alignment;
  unpack_meta(db, page, NULL, &end_free, NULL, NULL);
  iptr = end_free - need;
  id = (Dx_ui8 *) page + iptr;
  pack_idescriptor_fast(db, id, &key->dlen, &istripe_offset,
						&istripe_seglen, &v->volnum);

  /* Fill in a new header entry. */
  pack_header_meta(db, hptr, NULL, NULL, &hash, NULL);
  len = (Dx_ui16) IDESC_SIZE;
  pack_header_kv(db, hptr, &iptr, &len, NULL, NULL);
  pack_header_or_flags(db, hptr,
					   ITEM_KEY_ALIGNMENT_FLAG_BITS(key->flags)
					   | ITEM_KEY_INDIRECT_FLAG_BITS);
  pack_meta_fast(db, page, NULL, NULL, &iptr, NULL);

  return(DX_OK);
}

static INLINE void
item_add_value(Dx *db, Page page, Dx_ui32 page_size, Dx_ui8 *hptr,
			   Dx_datum *value, Dx_ui32 alignment)
{
  Dx_ui32 need;
  Dx_ui16 end_free, iptr, len;
  Dx_ui8 *ptr;

  /* Copy the new value. */
  need = (Dx_ui32) value->dlen + alignment;
  unpack_meta(db, page, NULL, &end_free, NULL, NULL);
  iptr = end_free - need;
  ptr = (Dx_ui8 *) page + iptr;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(ptr, value->dptr, (size_t) value->dlen);

  /* Fill in the header entry. */
  len = (Dx_ui16) value->dlen;
  pack_header_kv(db, hptr, NULL, NULL, &iptr, &len);
  pack_header_or_flags(db, hptr, ITEM_VALUE_ALIGNMENT_FLAG_BITS(value->flags));

  pack_meta_fast(db, page, NULL, NULL, &iptr, NULL);
}

/*
 * Store the value indirectly.
 */
static INLINE Dx_rc
item_add_indirect_value(Dx *db, Volume *volume, Page page, Dx_ui8 *hptr,
						Dx_datum *value, Dx_ui32 alignment, Ehash_xarg *arg)
{
  Dx_ui32 need;
  Dx_ui16 end_free, iptr, len;
  Item_flags_t hflags;
  Physoff istripe_offset;
  Segment istripe_seg;
  Seglen istripe_seglen;
  Volume *v;
  Dx_ui8 *id;
  Dx_rc rc;

  /* Allocate and copy the new value. */
  v = Random_volume(db);
  rc = Item_stripe_copy(db, v, value, 0, &istripe_seg,
						&istripe_offset, &istripe_seglen, arg);
  if (rc != DX_OK)
	return(rc);

  hflags = ITEM_VALUE_ALIGNMENT_FLAG_BITS(value->flags)
	| ITEM_VALUE_INDIRECT_FLAG_BITS;

  need = IDESC_SIZE + alignment;
  unpack_meta(db, page, NULL, &end_free, NULL, NULL);
  iptr = end_free - need;
  id = (Dx_ui8 *) page + iptr;
  pack_idescriptor_fast(db, id, &value->dlen, &istripe_offset, &istripe_seglen,
						&v->volnum);

  /* Fill in a new header entry. */
  len = (Dx_ui16) IDESC_SIZE;
  pack_header_kv(db, hptr, NULL, NULL, &iptr, &len);
  pack_header_or_flags(db, hptr, hflags);

  pack_meta_fast(db, page, NULL, NULL, &iptr, NULL);

  return(DX_OK);
}

/*
 * Copy a key, using the given header, from PAGE to NEW_PAGE.
 */
static INLINE void
item_copy_key(Dx *db, Page page, Dx_ui32 page_size, Dx_ui8 *hptr,
			  Page new_page, Dx_ui16 new_h, Dx_ui32 alignment)
{
  Dx_ui32 need;
  Dx_ui16 keyp, keylen;
  Dx_ui16 end_free, iptr;
  Dx_ui8 *n_hptr, *ptr;
  Item_flags_t hflags;

  unpack_header_kv(db, hptr, &keyp, &keylen, NULL, NULL);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  n_hptr = (Dx_ui8 *) new_page + new_h;

  /*
   * Copy the key.
   * Note that it doesn't matter if the key is indirect.
   */
  unpack_meta(db, new_page, NULL, &end_free, NULL, NULL);
  need = keylen + alignment;
  iptr = end_free - need;
  ptr = (Dx_ui8 *) new_page + iptr;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(ptr, (char *) page + keyp, (size_t) keylen);

  /* Fill in a new header entry. */
  pack_header_kv(db, n_hptr, &iptr, NULL, NULL, NULL);
  pack_meta_fast(db, new_page, NULL, NULL, &iptr, NULL);
  pack_meta_add_avail(db, new_page, -need);

  if (ITEM_KEY_INDIRECT(hflags))
	hflags |= ITEM_KEY_INDIRECT_FLAG_BITS;
  hflags |= ITEM_KEY_ALIGNMENT_FLAG_BITS(hflags);
  pack_header_meta(db, n_hptr, NULL, &hflags, NULL, NULL);
  pack_header_kv(db, n_hptr, NULL, &keylen, NULL, NULL);
}

static INLINE void
item_copy_value(Dx *db, Page page, Dx_ui32 page_size, Dx_ui8 *hptr,
				Page new_page, Dx_ui16 new_h, Dx_ui32 alignment)
{
  Dx_ui32 need;
  Dx_ui16 valuep, valuelen;
  Dx_ui16 end_free, iptr;
  Dx_ui8 *n_hptr, *ptr;
  Item_flags_t hflags;

  n_hptr = (Dx_ui8 *) new_page + new_h;
  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  /*
   * Copy the value.
   * Note that it doesn't matter if the value is indirect.
   */
  unpack_meta(db, new_page, NULL, &end_free, NULL, NULL);
  need = valuelen + alignment;
  iptr = end_free - need;
  ptr = (Dx_ui8 *) new_page + iptr;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(ptr, (char *) page + valuep, (size_t) valuelen);

  /* Fill in a new header entry. */
  pack_header_kv(db, n_hptr, NULL, NULL, &iptr, NULL);
  pack_meta_fast(db, new_page, NULL, NULL, &iptr, NULL);
  pack_meta_add_avail(db, new_page, -need);

  if (ITEM_VALUE_INDIRECT(hflags))
	hflags |= ITEM_VALUE_INDIRECT_FLAG_BITS;
  hflags |= ITEM_VALUE_ALIGNMENT_FLAG_BITS(hflags);
  pack_header_meta(db, n_hptr, NULL, &hflags, NULL, NULL);
  pack_header_kv(db, n_hptr, NULL, NULL, NULL, &valuelen);
}

/*
 * Partially initialize an Item structure from a described item.
 * The resulting Item doesn't point to the key/value bytes and if
 * the described item has indirect Dx_datums, the resulting key/value lengths
 * are those of the meta data, not the true key/value.
 */
static INLINE void
item_make_partial(Dx *db, Dx_ui8 *hptr, Item *item)
{
  Dx_hash hval;
  Item_flags_t hflags;
  Dx_ui16 keyp, keylen, valuep, valuelen;

  unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, &hval, NULL);
  item->hash = hval;
  Dx_key(item->key, NULL, keylen);	/* XXX */
  item->key->flags = ITEM_KEY_ALIGNMENT_FLAG_BITS(hflags);

  Dx_value(item->value, NULL, valuelen);	/* XXX */
  item->value->flags = ITEM_VALUE_ALIGNMENT_FLAG_BITS(hflags);
}

/*
 * Move an item from one page to another.
 * The caller guarantees that there is sufficient space in the destination.
 */
static INLINE void
item_move_item(Dx *db, Page page, Page new_page, Dx_ui8 *hptr)
{
  int indirect_key, indirect_value;
  Dx_ui32 ind, key_alignment_bytes, need, value_alignment_bytes;
  Dx_hash hval;
  Dx_uid uid;
  Dx_ui16 keylen, valuelen;
  Dx_ui16 new_end_free, new_h;
  Dx_ui16 new_avail, new_mflags, new_nheaders;
  Dx_ui8 *n_hptr;
  Dx_datum key, value;
  Item item;
  Item_flags_t hflags;
  Dx_rc rc;

  unpack_header_meta(db, hptr, NULL, &hflags, &hval, &uid);
  unpack_header_kv(db, hptr, NULL, &keylen, NULL, &valuelen);

  unpack_meta(db, new_page, &new_avail, &new_end_free, &new_mflags,
			  &new_nheaders);

  item.key = &key;
  item.value = &value;
  item_make_partial(db, hptr, &item);

  rc = item_need(ITEM_INDIRECT_THRESHOLD(db->page_size), &item,
				 new_avail, new_end_free, hflags,
				 &indirect_key, &indirect_value,
				 &key_alignment_bytes, &value_alignment_bytes, &need);
  if (rc != DX_OK) {
	fprintf(stderr, "item_move_item: no space!\n");
	Fatal(rc);
	/*NOTREACHED*/
  }
				   
  new_h = ITEM_HEADER(new_page, new_nheaders, new_mflags);
  n_hptr = (Dx_ui8 *) new_page + new_h;
  pack_header_meta(db, n_hptr, &chain_zero, &flags_zero, &hval, &uid);

  if (!indirect_key && !indirect_value) {
	if (key_alignment_bytes <= value_alignment_bytes) {
	  item_copy_key(db, page, db->page_size, hptr, new_page, new_h,
					key_alignment_bytes);
	  pack_meta_add_avail(db, page, keylen);
	  item_copy_value(db, page, db->page_size, hptr, new_page, new_h,
					  value_alignment_bytes);
	  pack_meta_add_avail(db, page, valuelen);
	  pack_meta_add_avail(db, new_page, -ITEM_HEADER_SIZE);
	} else {
	  item_copy_value(db, page, db->page_size, hptr, new_page, new_h,
					  value_alignment_bytes);
	  pack_meta_add_avail(db, page, valuelen);
	  item_copy_key(db, page, db->page_size, hptr, new_page, new_h,
					key_alignment_bytes);
	  pack_meta_add_avail(db, page, keylen);
	  pack_meta_add_avail(db, new_page, -ITEM_HEADER_SIZE);
	}
  }
  else {
	item_copy_key(db, page, db->page_size, hptr, new_page, new_h,
				  key_alignment_bytes);
	pack_meta_add_avail(db, page, keylen);
	item_copy_value(db, page, db->page_size, hptr, new_page, new_h,
					value_alignment_bytes);
	pack_meta_add_avail(db, page, valuelen);
	pack_meta_add_avail(db, new_page, -ITEM_HEADER_SIZE);
  }

  /* Fill in the hash bucket. */
  ind = H(hval) & (ITEM_HASHTAB_NBUCKETS(new_mflags) - 1);

  if (unpack_hashtab_val(db, new_page, db->page_size, ind) == 0)
	pack_hashtab_val(db, new_page, db->page_size, ind, new_h);
  else {
	Chainp_t cp;

	cp = unpack_hashtab_val(db, new_page, db->page_size, ind);
	pack_header_meta(db, n_hptr, &cp, NULL, NULL, NULL);
	pack_hashtab_val(db, new_page, db->page_size, ind, new_h);
  }

  new_nheaders++;
  pack_meta_fast(db, new_page, &new_nheaders, NULL, NULL, NULL);
}

/*
 * If any items have been flagged for deletion or there is internal
 * fragmentation, reorganize the page to maximize the contiguous free
 * space, eliminate unused item headers, and correct the item count.
 */
int
Item_reorg(Dx *db, Page page)
{
  Dx_ui32 i, nmoves;
  Dx_ui16 iptr, mflags, nheaders;
  Dx_ui8 *hptr;
  Item_flags_t hflags;
  static void *tpage = NULL;
  static Dx_ui32 tpage_size = 0;

  if (tpage == NULL) {
	tpage = (void *) Xmalloc(db->page_size);
	tpage_size = db->page_size;
  } else if (tpage_size < db->page_size) {
	tpage = (void *) Xrealloc(page, db->page_size);
	tpage_size = db->page_size;
  }

  Item_init(db, tpage);

  /* Find each undeleted item and add it to the temp page */
  nmoves = 0;
  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);
  for (i = 0; i < nheaders; i++) {
	iptr = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
	if (! ITEM_DELETED(hflags)) {
	  item_move_item(db, page, tpage, hptr);
	  nmoves++;
	}
  }
 
  mflags &= ~(ITEM_REORG_FLAG_BITS);
  nheaders = nmoves;
  pack_meta_fast(db, tpage, &nheaders, &mflags, NULL, NULL);

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(page, tpage, (size_t) db->page_size);

  Stats_count(db, REORG);
  return(nmoves);
}

/*
 * Fetch an indirect datum (key or value) and its length.
 * When requested, handle a Dx_point and/or user I/O callback.
 */
static INLINE Dx_rc
item_fetch_indirect_datum(Dx *db, Dx_ui8 *id, Ehash_xarg *arg)
{
  Dx_rc rc;

  rc = item_stripe_contig(db, id, arg);

  return(rc);
}

typedef struct {
  Dx_hash (*hashfunc)(Dx_ui8 *str, Dx_len len, Dx_hash init);
  Dx_hash hash;
} Hash_callback_state;

static Dx_len
item_key_hash_callback(void *arg, void *dptr, Dx_len io_dlen,
					   Dx_desc *desc)
{
  Hash_callback_state *state;

  state = (Hash_callback_state *) arg;
  state->hash = state->hashfunc((Dx_ui8 *) dptr, io_dlen, state->hash);

  return(io_dlen);
}

/*
 * Compute the key hash for the given entry.
 */
static Dx_rc
item_key_hash(Dx *db, Page page, Dx_ui8 *hptr, Dx_hash *hash)
{
  Dx_ui8 *id;
  Dx_ui16 keyp, keylen;
  Hash_callback_state state;
  Ehash_xarg nullarg;
  Dx_hash hval;
  Item_flags_t hflags;
  Dx_user_io user_io;
  Dx_rc rc;

  unpack_header_kv(db, hptr, &keyp, &keylen, NULL, NULL);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  if (ITEM_DELETED(hflags)) {
	/* XXX Shouldn't happen! */
	return(DX_INTERNAL_ERROR);
  }

  if (ITEM_KEY_INDIRECT(hflags)) {
	id = (Dx_ui8 *) page + keyp;
	state.hash = 0;
	state.hashfunc = db->hashfunc;
	user_io.func = item_key_hash_callback;
	user_io.arg = (void *) &state;
	nullarg.io.uio = &user_io;
	nullarg.pos = NULL;
	nullarg.region = NULL;
	rc = item_fetch_indirect_datum(db, id, &nullarg);
	if (rc != DX_OK)
	  return(rc);
	hval = state.hash;
  }
  else
	hval = db->hashfunc((Dx_ui8 *) page + keyp, (Dx_len) keylen, 0);

  *hash  = hval;
  return(DX_OK);
}

/*
 * Test if a key for an item is identical to a given key.
 */
static INLINE Dx_rc
item_key_equal(Dx *db, Page page, Dx_ui8 *hptr, Item *item)
{
  void *dbkey;
  Dx_hash hval;
  Item_flags_t hflags;
  Ehash_xarg arg;
  Dx_ui16 keyp, keylen;
  Dx_ui8 *id;
  Dx_user_io user_io;
  Dx_rc rc;

  unpack_header_kv(db, hptr, &keyp, &keylen, NULL, NULL);
  unpack_header_meta(db, hptr, NULL, &hflags, &hval, NULL);
  if (hval != item->hash)
	return(ITEM_NOMATCH);

  if (item->ks != NULL && (hval ^ item->keyhash) != item->ks->hash)
	return(ITEM_NOMATCH);

  if (ITEM_KEY_INDIRECT(hflags)) {
	Dx_len dlen;

	/* They can't be equal if they're different lengths. */
	id = (Dx_ui8 *) page + keyp;
	unpack_idescriptor_fast(db, id, &dlen, NULL, NULL, NULL);
	if (dlen != item->key->dlen)
	  return(ITEM_NOMATCH);

	user_io.func = NULL;
	user_io.arg = (void *) &dbkey;
	arg.io.uio = &user_io;
	arg.pos = NULL;
	arg.region = NULL;
	if ((rc = item_fetch_indirect_datum(db, id, &arg)) != DX_OK)
	  return(rc);
  }
  else {
	if (keylen != item->key->dlen)
	  return(ITEM_NOMATCH);
	dbkey = (Dx_ui8 *) page + keyp;
  }

  if (Dx_memneq((const void *) dbkey, (const void *) item->key->dptr,
				(size_t) item->key->dlen))
	return(ITEM_NOMATCH);

  if (ITEM_KEY_INDIRECT(hflags))
	Xfree(dbkey);

  return(DX_OK);
}

/*
 * Return a pointer to the header for the given item, or NULL if it
 * wasn't found.
 * The key's ind field is used to differentiate the different instances
 * of the same key value.  Note that all identical keys will end up in the
 * same hash table chain.
 */
static INLINE Dx_ui8 *
item_locate(Dx *db, Page page, Item *item, Hashtab_bucket_t *piptr)
{
  Dx_ui32 ind, n;
  Hashtab_bucket_t iptr, next_iptr, prev_iptr;
  Dx_ui16 mflags, nheaders;
  Dx_ui8 *hptr, *prev_hptr;
  Item_flags_t hflags;

  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);

  if (nheaders == 0)
	return(NULL);

  ind = H(item->hash) & (ITEM_HASHTAB_NBUCKETS(mflags) - 1);
  n = 0;
  iptr = unpack_hashtab_val(db, page, db->page_size, ind);

  prev_iptr = iptr;		/* Shuts up compiler */
  while (iptr != 0) {
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_meta(db, hptr, &next_iptr, &hflags, NULL, NULL);
	if (item_key_equal(db, page, hptr, item) == DX_OK) {
	  /* XXX break if ind > n? */
	  if (! ITEM_DELETED(hflags) && item->key->ind == n) {
		if (piptr != NULL)
		  *piptr = prev_iptr;
		return(hptr);
	  }
	  n++;
	}
	prev_hptr = hptr;
	iptr = next_iptr;
  }
  
  return(NULL);
}

/*
 * Try to add ITEM to PAGE.
 * The item's key need not be unique.
 *
 * If the size of the aligned value (or key, or combination) doesn't exceed
 * ITEM_INDIRECT_THRESHOLD, then it will end up in a normal page; oth. its
 * meta-data will be stored in a normal page but the data will be stored
 * "indirectly" as one or more stripes.
 *
 * If there's not enough free space in this page (possibly after
 * reorganization), the caller is informed that this page must be split.
 */
Dx_rc
Item_add(Dx *db, Volume *v, Page page, Dx_opmode mode, Item *item,
		 Ehash_xarg *arg)
{
  Dx_ui8 *hptr;
  Dx_rc rc;

  if (DX_OPMODE(mode) == DX_INSERT_UNIQUE &&
	  (hptr = item_locate(db, page, item, NULL)) != NULL)
	return(DX_KEY_EXISTS);

  rc = item_add(db, v, page, item, arg);

  if (rc == DX_OK)
	item_advance_uid(db);

  return(rc);
}

static Dx_rc
item_add(Dx *db, Volume *v, Page page, Item *item, Ehash_xarg *arg)
{
  int indirect_key, indirect_value;
  Dx_ui32 ind, key_alignment_bytes, need, value_alignment_bytes;
  Dx_ui16 avail, end_free, mflags, new_h, nheaders;
  Dx_ui8 *hptr, *khptr;
  Hashtab_bucket_t hb;
  Chainp_t iptr, next_iptr, prev_iptr;
  Dx_rc rc;

  if (arg->io.uio != NULL) {
	if (arg->io.bytes_buffered == 0) {
	  Dx_len len, st;
	  Dx_desc *desc;

	  len = (Dx_len) ITEM_INDIRECT_THRESHOLD(db->page_size) + 1;
	  arg->io.uio_buf = (Dx_ui8 *) alloca((size_t) len);

	  /* Call the get-function. */
	  desc = &arg->io.desc;
	  desc->error_occurred = 0;
	  desc->last_errno = 0;
	  desc->dlen = 0;
	  desc->flags = DX_DATUM_ALIGN0;
	  st = (arg->io.uio->func)(arg->io.uio->arg, arg->io.uio_buf, len, desc);
	  if (desc->error_occurred) {
		db->last_errno = desc->last_errno;
		return(DX_USERIO_ERROR);
	  }

	  arg->io.bytes_buffered = st;
	  item->value->dlen = st;
	}
	else {
	  /*
	   * It's possible for some/all data to have already been read.
	   * If so, enough was read to differentiate between a direct and
	   * an indirect item.
	   */
	  item->value->dlen = arg->io.bytes_buffered;
	}

	if (item->value->dlen <= ITEM_INDIRECT_THRESHOLD(db->page_size))
	  item->value->dptr = arg->io.uio_buf;
  }

  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);

  rc = item_need(ITEM_INDIRECT_THRESHOLD(db->page_size), item, avail,
				 end_free, 0, &indirect_key, &indirect_value,
				 &key_alignment_bytes, &value_alignment_bytes, &need);
  if (rc != DX_OK)
	return(rc);

  /*
   * If there isn't enough contiguous space or the item count has reached the
   * maximum, see if a consolidation will help.
   */
  if (need > ITEM_CONTIG(db, page, nheaders, mflags)
	  || nheaders == ITEM_MAX_NHEADERS()) {
	if (ITEM_REORG_FLAG(mflags)) {
	  Item_reorg(db, page);
	  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);
	}

	/*
	 * Reorganization may mean less space needed due to better alignment,
	 * but it's only necessary to check if alignment is an issue.
	 * XXX Not calling item_need() will be faster but may result in poorer
	 * space utilization, at least until the page is reorganized.
	 */
	if (indirect_key || indirect_value
		|| DATUM_ALIGNMENT(item->key) != 1
		|| DATUM_ALIGNMENT(item->value) != 1) {
	  rc = item_need(ITEM_INDIRECT_THRESHOLD(db->page_size), item, avail,
					 end_free, 0, &indirect_key, &indirect_value,
					 &key_alignment_bytes, &value_alignment_bytes, &need);
	  if (rc != DX_OK
		  || need > ITEM_CONTIG(db, page, nheaders, mflags)
		  || nheaders == ITEM_MAX_NHEADERS())
		return(ITEM_NOSPACE);
	}
  }

  new_h = ITEM_HEADER(page, nheaders, mflags);
  hptr = (Dx_ui8 *) page + new_h;
  pack_header_meta(db, hptr, &chain_zero, &flags_zero, NULL, NULL);

  rc = DX_OK;
  if (!indirect_key && !indirect_value) {
	if (key_alignment_bytes <= value_alignment_bytes) {
	  item_add_key(db, page, db->page_size, hptr, item->key,
				   key_alignment_bytes);
	  item_add_value(db, page, db->page_size, hptr, item->value,
					 value_alignment_bytes);
	} else {
	  item_add_value(db, page, db->page_size, hptr, item->value,
					 value_alignment_bytes);
	  item_add_key(db, page, db->page_size, hptr, item->key,
				   key_alignment_bytes);
	}
  } else if (!indirect_key) {
	item_add_key(db, page, db->page_size, hptr, item->key,
				 key_alignment_bytes);
	rc = item_add_indirect_value(db, v, page, hptr, item->value,
								 value_alignment_bytes, arg);
  } else if (!indirect_value) {
	if ((rc = item_add_indirect_key(db, v, page, hptr, item->key,
									item->hash, key_alignment_bytes)) == DX_OK)
	  item_add_value(db, page, db->page_size, hptr, item->value,
					 value_alignment_bytes);
  } else {
	if ((rc = item_add_indirect_key(db, v, page, hptr, item->key,
									item->hash, key_alignment_bytes)) == DX_OK)
	  rc = item_add_indirect_value(db, v, page, hptr, item->value,
								   value_alignment_bytes, arg);
  }

  if (rc == DX_OK) {
	pack_header_meta(db, hptr, NULL, NULL, &item->hash, &db->current_uid);

	/* Fill in the hash bucket. */
	ind = H(item->hash) & (ITEM_HASHTAB_NBUCKETS(mflags) - 1);

	if ((hb = unpack_hashtab_val(db, page, db->page_size, ind)) == 0)
	  pack_hashtab_val(db, page, db->page_size, ind, new_h);
	else {
	  if (arg == NULL || (arg->flags & DX_ADD_FIRST)) {
		pack_header_meta(db, hptr, &hb, NULL, NULL, NULL);
		pack_hashtab_val(db, page, db->page_size, ind, new_h);
	  }
	  else if (arg->flags & DX_ADD_LAST) {
		iptr = hb;
		while (iptr != 0) {
		  hptr = (Dx_ui8 *) page + iptr;
		  unpack_header_meta(db, hptr, &next_iptr, NULL, NULL, NULL);
		  if (next_iptr == 0)
			break;
		  iptr = next_iptr;
		}
		pack_header_meta(db, hptr, &new_h, NULL, NULL, NULL);
	  }
	  else if (arg->flags & DX_ADD_INDEX) {
		khptr = item_locate(db, page, item, &prev_iptr);
		iptr = unpack_hashtab_val(db, page, db->page_size, ind);
		if (khptr == NULL || prev_iptr == iptr) {
		  pack_header_meta(db, hptr, &hb, NULL, NULL, NULL);
		  pack_hashtab_val(db, page, db->page_size, ind, new_h);
		}
		else {
		  khptr = (Dx_ui8 *) page + prev_iptr;
		  unpack_header_meta(db, khptr, &next_iptr, NULL, NULL, NULL);
		  pack_header_meta(db, khptr, &new_h, NULL, NULL, NULL);
		  hptr = (Dx_ui8 *) page + new_h;
		  pack_header_meta(db, hptr, &next_iptr, NULL, NULL, NULL);
		}
	  }
	  else
		return(DX_BADACCESSMODE);
	}

	unpack_meta(db, page, &avail, NULL, NULL, NULL);
	avail -= need;
	nheaders++;
	pack_meta_fast(db, page, &nheaders, NULL, NULL, &avail);
  }

  return(rc);
}

/*
 * Create a zero-length item.
 * This is useful when fielding a DX_EDIT/DX_STORE request for an item
 * that doesn't exist.
 */
static Dx_rc
item_force_create(Dx *db, Volume *v, Page page, Item *item, Ehash_xarg *arg)
{
  Item new_item;
  Dx_datum new_value;
  Ehash_xarg new_xarg;

  new_item = *item;
  new_value.dptr = NULL;
  new_value.dlen = 0;
  new_value.flags = DX_DATUM_ALIGN0;
  new_value.ind = 0;
  new_value.count = 0;
  new_value.keyspace = NULL;
  new_item.value = &new_value;
  
  new_xarg.pos = NULL;
  new_xarg.region = NULL;
  new_xarg.io.uio = NULL;
  new_xarg.io.bytes_buffered = 0;
  new_xarg.flags = DX_INSERT | DX_ADD_FIRST;

  return(Item_add(db, v, page, DX_INSERT, &new_item, &new_xarg));

}

/*
 * If ITEM exists, replace it with the new value, oth. just add the new value.
 *
 * XXX If you replace an existing item, its 'ind' position shouldn't change.
 * There should be a way to change its position without the user having
 * to do a full delete-and-add sequence.
 */
Dx_rc
Item_replace(Dx *db, Volume *v, Page page, Item *item, Ehash_xarg *arg)
{
  Dx_ui8 *hptr;
  Dx_ui16 valuelen;
  Item_flags_t hflags;

  if ((hptr = item_locate(db, page, item, NULL)) == NULL)
	return(Item_add(db, v, page, DX_REPLACE, item, arg));

  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  unpack_header_kv(db, hptr, NULL, NULL, NULL,  &valuelen);

  return(item_replace(db, v, page, hptr, hflags,
					  (Dx_len) valuelen, item, arg));
}

/*
 * Do an efficient replacement of the existing value with the new one.
 */
static Dx_rc
item_replace(Dx *db, Volume *v, Page page, Dx_ui8 *hptr,
			 Item_flags_t hflags, Dx_len valuelen, Item *item, Ehash_xarg *arg)
{
  Dx_rc rc;

  if (arg != NULL && (arg->io.uio != NULL || arg->io.bytes_buffered)) {
	if ((rc = Item_delete(db, page, item)) != DX_OK)
	  return(rc);

	return(Item_add(db, v, page, DX_REPLACE, item, arg));
  }

  /*
   * Check for an opportunity to do this quickly by not doing a delete.
   * This is possible if:
   * 1. the value is not indirect; and
   * 2. the alignment requirements for the new value remain satisfied;
   * 3. the length of the item's new value is not increasing
   */
  if (!ITEM_VALUE_INDIRECT(hflags)
	  && ITEM_VALUE_ALIGNMENT(hflags) <= DATUM_ALIGNMENT(item->value)
	  && valuelen > item->value->dlen)
	return(item_substitute(db, page, hptr, item));

  if ((rc = Item_delete(db, page, item)) != DX_OK)
	return(rc);

  return(Item_add(db, v, page, DX_REPLACE, item, arg));
}

/*
 * If new value data is the same size (or smaller), the value is not indirect,
 * and alignment is not more restrictive than the old, the old data can just
 * be overwritten.
 * The key's alignment must also not be more restrictive than previous.
 */
static Dx_rc
item_substitute(Dx *db, Page page, Dx_ui8 *hptr, Item *item)
{
  Dx_ui16 len, valuep, valuelen;
  Dx_ui8 *ptr;
  Item_flags_t hflags;

  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  hflags &= ~ITEM_KEY_ALIGNMENT_FLAG_BITS(0);
  hflags |= ITEM_KEY_ALIGNMENT_FLAG_BITS(item->key->flags);
  hflags &= ~ITEM_VALUE_ALIGNMENT_FLAG_BITS(0);
  hflags |= ITEM_VALUE_ALIGNMENT_FLAG_BITS(item->value->flags);

  ptr = (Dx_ui8 *) page + valuep;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(ptr, item->value->dptr, (size_t) item->value->dlen);

  pack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  len = (Dx_ui16) item->value->dlen;
  pack_header_kv(db, hptr, NULL, NULL, NULL, &len);

  /* If the new data is smaller, reorganization is possible. */
  if (item->value->dlen < valuelen)
	pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  return(DX_OK);
}

/*
 * Delete an item by simply marking it as deleted, although an
 * indirect datum is deallocated now.
 */
Dx_rc
Item_delete(Dx *db, Page page, Item *item)
{
  Dx_ui8 *hptr;

  if ((hptr = item_locate(db, page, item, NULL)) == NULL)
	return(ITEM_NOTFOUND);

  return(item_delete(db, page, hptr));
}

static Dx_rc
item_delete(Dx *db, Page page, Dx_ui8 *hptr)
{
  Dx_ui16 avail, keyp, keylen, valuep, valuelen;
  void *dptr;
  Item_flags_t hflags;
  Physvolnum initial_volume;
  Physoff off;
  Seglen seglen;
  Dx_rc rc;
  Dx_ui8 *id;

  unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  avail = ITEM_HEADER_SIZE;

  if (ITEM_KEY_INDIRECT(hflags)) {
	id = (Dx_ui8 *) page + keyp;
	unpack_idescriptor_fast(db, id, NULL, &off, &seglen, &initial_volume);
	rc = Stripe_free(db, initial_volume, off, seglen);
	if (rc != DX_OK)
	  return(rc);
  }
  else if (db->flags & DX_CONFIG_ERASE_ON_DELETE) {
	dptr = (void *) ((char *) page + keyp);
	Dx_memset(dptr, DX_ERASE_ON_DELETE_CHAR, (size_t) keylen);
  }
  avail += keylen;

  if (ITEM_VALUE_INDIRECT(hflags)) {
	id = (Dx_ui8 *) page + valuep;
	unpack_idescriptor_fast(db, id, NULL, &off, &seglen, &initial_volume);
	rc = Stripe_free(db, initial_volume, off, seglen);
	if (rc != DX_OK)
	  return(rc);
  }
  else if (db->flags & DX_CONFIG_ERASE_ON_DELETE) {
	dptr = (void *) ((char *) page + valuep);
	Dx_memset(dptr, DX_ERASE_ON_DELETE_CHAR, (size_t) valuelen);
  }
  avail += valuelen;

  /*
   * Increase the amount of free space available.
   * The item header is still in use, so don't count it in the total.
   */
  pack_meta_add_avail(db, page, avail);
  pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  pack_header_or_flags(db, hptr, ITEM_DELETED_BITS);

  return(DX_OK);
}

/*
 * Do the hard part of Dx_storex() here, deleting a region and/or
 * inserting/overwriting a region of an existing value Dx_datum.
 * The alignment of the target Dx_datum can't be changed unless a total
 * replacement operation is being done.
 *
 *         _____
 *      / |Head |
 *     |  |-----|
 *     |  |/ / /|
 * Dlen|  |Chunk|
 *     |  |/ / /|\  
 *     |  |-----| \_  ---
 *      \ |Tail |    |New|
 *         -----      ---
 * When this function is called,
 *   Head + Chunk + Tail == Dlen
 * and the size of any of the Head, Chunk, Tail, and New regions may be zero.
 *
 * For maximum efficiency, we want to avoid doing a full deletion followed
 * by an addition in the few cases where it makes sense (e.g., changing a
 * few bytes of a multi-megabyte datum).  There are a few things to consider:
 * o if there is a get-function, we don't immediately know the final
 *   size of the value and whether it will be direct or indirect.
 * o for indirect values, we want to reuse allocated segments as much
 *   as possible and make as few changes to the stripe descriptor pages
 *   as we can, while also trying to keep stripes balanced over volumes.
 *   A call to Dx_reorg() might also do some balancing chores one day.
 *
 * Note that deleting or replacing all of the existing value
 * (requested_delete_length == original_length) is not the same as deleting
 * the item.  It may result in a zero-length value or entirely new data.
 * The UID is not changed by a DX_EDIT operation.
 *
 * XXX edited out regions are not erased
 */
Dx_rc
Item_edit(Dx *db, Volume *volume, Page page, Item *item, Ehash_xarg *arg)
{
  Dx_ui8 *hptr, *id;
  Dx_ui16 valuep, valuelen;
  Dx_len end_pos, readahead_length, start_pos, original_length;
  Dx_len min_new_len, requested_delete_length;
  Item_flags_t hflags;
  Physoff off;
  Seglen seglen;
  Physvolnum initial_volume;
  Dx_len dlen;
  Dx_rc rc;

  /* Ordinarily, it must exist to be edited. */
  if ((hptr = item_locate(db, page, item, NULL)) == NULL) {
	if ((arg->flags & DX_MAY_CREATE) == 0)
	  return(ITEM_NOTFOUND);
	if (arg->region != NULL)
	  return(DX_BADARG);
	rc = item_force_create(db, volume, page, item, arg);
	if (rc != DX_OK)
	  return(rc);
	if ((hptr = item_locate(db, page, item, NULL)) == NULL)
	  return(DX_INTERNAL_ERROR);
  }

  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  /* Get the current size of the datum. */
  if (ITEM_VALUE_INDIRECT(hflags)) {
	id = (Dx_ui8 *) page + valuep;
	unpack_idescriptor_fast(db, id, &dlen, &off, &seglen, &initial_volume);
	original_length = (Dx_len) dlen;
  }
  else
	original_length = (Dx_len) valuelen;

  /* If there is a deletion to do, figure out how many bytes to remove. */
  requested_delete_length = 0;
  if (arg->region != NULL) {
	rc = item_verify_region(original_length, arg->region,
							&start_pos, &end_pos, &requested_delete_length);
	if (rc != DX_OK)
	  return(rc);
  }

  if (requested_delete_length > original_length)
	return(DX_BADEDITARG);

  /*
   * Determine the insertion point for inserting/overwriting, if there is any.
   * If an edit is being done, the insertion point will be the start of the
   * region being edited.
   */
  if (arg->region == NULL) {
	/* No deletion, only insertion. */
	if (arg->pos == NULL) {
	  /* The default is to append. */
	  start_pos = original_length;
	}
	else {
	  /* The start_pos may be beyond the current length. */
	  rc = item_verify_position(original_length, arg->pos, &start_pos);
	  if (rc != DX_OK)
		return(rc);
	}
  }

  /*
   * Will the resulting datum be direct or indirect?
   * If it's already indirect and any deletion is small, we know the result
   * must remain indirect.
   * If the insertion position is in the "indirect range" we may still do a
   * readahead in case zero bytes are read, to simplify this code.
   *
   * Some cases are best handled using a delete-and-add approach, while
   * others should use a custom approach for better efficiency.
   *
   * First, find out how much data will be inserted/overwritten.
   * If a get-function must be called, we don't know how much new data there
   * will be, but we can read and buffer enough to decide whether the final
   * value will be indirect.
   */
  arg->io.uio_buf = NULL;
  arg->io.bytes_buffered = 0;
  if (ITEM_VALUE_INDIRECT(hflags) && arg->io.uio != NULL
	  && ((original_length - requested_delete_length)
		  > ITEM_INDIRECT_THRESHOLD(db->page_size)))
	readahead_length = 0;
  else if (arg->io.uio != NULL) {
	Dx_len len, st;
	Dx_desc *desc;

	/* This will always be enough input to decide. */
	len = (Dx_len) ITEM_INDIRECT_THRESHOLD(db->page_size) + 1;
	arg->io.uio_buf = (Dx_ui8 *) alloca((size_t) len);

	/* Call the get-function. */
	desc = &arg->io.desc;
	desc->error_occurred = 0;
	desc->last_errno = 0;
	desc->dlen = 0;
	desc->flags = DX_DATUM_ALIGN0;
	st = (arg->io.uio->func)(arg->io.uio->arg, arg->io.uio_buf, len, desc);
	if (desc->error_occurred) {
	  db->last_errno = desc->last_errno;
	  return(DX_USERIO_ERROR);
	}
	arg->io.bytes_buffered = st;
	readahead_length = st;

	/* This may not be the final size... */
	item->value->dlen = st;
  }
  else if (item->value != NULL && item->value->dptr != NULL) {
	/* If a value datum was given, treat it like a readahead. */
	readahead_length = item->value->dlen;
  }
  else
	readahead_length = 0;

  /*
   * Compute the minimum length of the resulting datum, even in overwrite
   * mode or if there's a zero-fill area.
   */
  min_new_len = original_length - requested_delete_length + readahead_length;
  if (start_pos > original_length)
	min_new_len += (start_pos - original_length);

  if (ITEM_VALUE_INDIRECT(hflags)) {
	if (min_new_len <= ITEM_INDIRECT_THRESHOLD(db->page_size)) {
	  /*
	   * This is an indirect to direct change, but the value will remain
	   * indirect.  This may or may not turn out to be a good idea, but
	   * it simplifies things.
	   * There may be a (relatively small) Head and/or Tail to extract from the
	   * current datum.
	   * The end_pos must have been set because a delete must be happening.
	   */
	  void *hdptr, *tptr;
	  Contig_callback_state state;
	  Dx_ui8 *dst;
	  Dx_ui16 hlen, newlen, tlen;
	  Dx_datum new_value;
	  Dx_region head_region, tail_region;
	  Dx_desc *desc;
	  Dx_user_io uio;
	  Ehash_io io;

	  desc = &io.desc;
	  hlen = tlen = 0;
	  hdptr = tptr = NULL;
	  dst = NULL;

	  if (start_pos != 0) {
		/* Get the Head. */
		hlen = (Dx_ui16) start_pos;
		hdptr = alloca((size_t) start_pos);
		state.dptr = hdptr;
		state.offset = 0;
		uio.func = item_stripe_contig_callback;
		uio.arg = (void *) &state;
		io.uio = &uio;
		Dx_set_region_from_start(&head_region);
		Dx_set_region_len(&head_region, start_pos);

		rc = Stripe_contig(db, initial_volume, off, seglen, dlen,
						   &head_region, &io);
		if (rc != DX_OK)
		  return(rc);
	  }

	  if (end_pos != original_length) {
		/* Get the Tail. */
		tlen = (Dx_ui16) (original_length - end_pos);
		if (arg->flags & DX_OVERWRITE)
		  tlen -= readahead_length;
		if (tlen > 0) {
		  tptr = alloca((size_t) tlen);
		  state.dptr = tptr;
		  state.offset = 0;
		  uio.func = item_stripe_contig_callback;
		  uio.arg = &state;
		  io.uio = &uio;
		  if (arg->flags & DX_OVERWRITE)
			Dx_set_region_start(&tail_region, end_pos + readahead_length);
		  else
			Dx_set_region_start(&tail_region, end_pos);
		  Dx_set_region_len(&tail_region, tlen);

		  rc = Stripe_contig(db, initial_volume, off, seglen, dlen,
							 &tail_region, &io);
		  if (rc != DX_OK)
			return(rc);
		}
		else
		  tlen = 0;

		/* Now conflate the Head, new data, and Tail. */
		newlen = hlen + readahead_length + tlen;
		new_value.dptr = alloca((size_t) newlen);
		dst = (Dx_ui8 *) new_value.dptr;
		if (hlen) {
		  memmove(dst, hdptr, hlen);
		  dst += hlen;
		}
		if (arg->io.uio != NULL && readahead_length) {
		  memmove(dst, arg->io.uio_buf, (size_t) readahead_length);
		  dst += readahead_length;
		  arg->io.bytes_buffered = 0;
		} else if (item->value != NULL && item->value->dptr != NULL) {
		  memmove(dst, item->value->dptr, (size_t) readahead_length);
		  dst += readahead_length;
		}
		if (tlen) {
		  memmove(dst, tptr, tlen);
		  dst += tlen;
		}
	  }

	  new_value.dlen = dst - (Dx_ui8 *) new_value.dptr;
	  new_value.flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);

	  return(item_replace_indirect_value(db, volume, page, hptr, &new_value));
	}

	/* This is an indirect to indirect edit. */
	return(item_edit_indirect(db, volume, page, hptr, start_pos,
							  requested_delete_length, readahead_length,
							  item->value, arg));
  }

  if (min_new_len > ITEM_INDIRECT_THRESHOLD(db->page_size)) {
	/*
	 * This is a direct to indirect change.
	 * Simplify the code by converting it to indirect and then doing an
	 * indirect-indirect change.
	 */
	rc = item_force_indirect_value(db, volume, page, hptr);
	if (rc != DX_OK)
	  return(rc);

	/* This is an indirect to indirect edit. */
	return(item_edit_indirect(db, volume, page, hptr, start_pos,
							  requested_delete_length, readahead_length,
							  item->value, arg));
  }

  if (min_new_len > original_length) {
	Dx_ui8 *dst, *src;
	int tail_length;
	Dx_datum new_value;

	/*
	 * This is a direct to bigger size direct change.
	 * Piece together the new datum, then replace the old one with it.
	 */
	new_value.dptr = alloca((size_t) ITEM_INDIRECT_THRESHOLD(db->page_size));
	dst = (Dx_ui8 *) new_value.dptr;
	src = (Dx_ui8 *) page + valuep;

	if (start_pos > original_length) {
	  memmove(dst, src, (size_t) original_length);
	  /* Zero fill. */
	  Dx_memset(dst + original_length, 0, (size_t) (start_pos - original_length));
	  dst += start_pos;
	} else {
	  memmove(dst, src, (size_t) start_pos);
	  dst += start_pos;
	}
	if (arg->io.uio != NULL && readahead_length) {
	  memmove(dst, arg->io.uio_buf, (size_t) readahead_length);
	  dst += readahead_length;
	  arg->io.bytes_buffered = 0;
	} else if (item->value != NULL && item->value->dptr != NULL) {
	  memmove(dst, item->value->dptr, (size_t) readahead_length);
	  dst += readahead_length;
	}
	tail_length = original_length - start_pos - requested_delete_length;
	if (arg->flags & DX_OVERWRITE)
	  tail_length -= readahead_length;
	if (tail_length > 0) {
	  src = (Dx_ui8 *) page + valuep + start_pos + requested_delete_length;
	  if (arg->flags & DX_OVERWRITE)
		src += readahead_length;
	  memmove(dst, src, (size_t) tail_length);
	  dst += tail_length;
	}
	new_value.dlen = dst - (Dx_ui8 *) new_value.dptr;
	new_value.flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);
	return(item_replace_direct_value(db, volume, page, hptr, &new_value));
  }

  /* This is a direct to same-or-smaller size direct change. */
  return(item_edit_direct(db, page, hptr, valuep, valuelen,
                          start_pos, requested_delete_length,
                          readahead_length, item->value, arg));
}

/*
 * Perform an edit on a direct datum that results in a non-increasing length.
 * This is done "in place" to eliminate some copying.
 *
 * Note that there can't be any zero filling because it's shrinking.
 */
static Dx_rc
item_edit_direct(Dx *db, Page page, Dx_ui8 *hptr,
				 Dx_ui16 valuep, Dx_ui16 valuelen,
				 Dx_len start_pos, Dx_len requested_delete_length,
				 Dx_len readahead_length, Dx_datum *value, Ehash_xarg *arg)
{
  int tail_length;
  Dx_ui8 *dst, *src;
  Dx_ui16 new_valuelen;

  if (arg->io.bytes_buffered)
	readahead_length = arg->io.bytes_buffered;

  dst = (Dx_ui8 *) page + valuep + start_pos;

  /* Overwrite a range of the old value with new data. */
  if (readahead_length) {
	if (arg->io.bytes_buffered)
	  memmove(dst, (const void *) arg->io.uio_buf, (size_t) readahead_length);
	else
	  memmove(dst, (const void *) value->dptr, (size_t) readahead_length);
	dst += readahead_length;
  }

  /* If there's old data after the delete range, concatenate it. */
  tail_length = valuelen - start_pos - requested_delete_length;
  if (arg->flags & DX_OVERWRITE)
	tail_length -= readahead_length;
  if (tail_length > 0) {
	src = (Dx_ui8 *) page + valuep + start_pos + requested_delete_length;
	if (arg->flags & DX_OVERWRITE)
	  src += readahead_length;
	memmove(dst, (const void *) src, (size_t) tail_length);
  }

  new_valuelen = dst - ((Dx_ui8 *) page + valuep);
  pack_header_kv(db, hptr, NULL, NULL, NULL, &new_valuelen);

  /* Since the new data is smaller, reorganization is possible. */
  pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  return(DX_OK);
}

/*
 * Perform an edit on an indirect datum that results in another indirect datum.
 *
 * A simple-minded implementation will do for now:
 *   1) Read new data into stripes
 *   2) Delete old region
 *   3) Insert new stripes
 */
static Dx_rc
item_edit_indirect(Dx *db, Volume *volume, Page page,
				   Dx_ui8 *hptr, Dx_len start_pos,
				   Dx_len requested_delete_length, Dx_len readahead_length,
				   Dx_datum *value, Ehash_xarg *arg)
{
  Dx_ui8 *id;
  Dx_ui16 valuep, valuelen;
  Dx_len delete_length, new_dlen;
  Physoff off;
  Seglen seglen;
  Physvolnum initial_volume;
  Dx_len dlen;
  Dx_desc *desc;
  Dx_rc rc;

  if (arg->flags & DX_OVERWRITE)
	delete_length = requested_delete_length + value->dlen;
  else
	delete_length = requested_delete_length;

  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  id = (Dx_ui8 *) page + valuep;
  unpack_idescriptor_fast(db, id, &dlen, &off, &seglen, &initial_volume);
  
  desc = &arg->io.desc;
  desc->error_occurred = 0;
  desc->last_errno = 0;
  if (delete_length) {
	rc = Stripe_delete_data(db, volume->volnum, off, seglen, dlen, start_pos,
							delete_length, desc, &arg->io);
	if (rc != DX_OK)
	  return(rc);
  }

  new_dlen = dlen - requested_delete_length;
  value->dlen = new_dlen;
  pack_idescriptor_fast(db, id, &new_dlen, NULL, NULL, NULL);

  rc = Stripe_insert_data(db, volume->volnum, id, off, seglen, value, new_dlen,
						  start_pos, arg);
  if (rc == DX_OK)
	pack_idescriptor_fast(db, id, &value->dlen, NULL, NULL, NULL);

  return(rc);
}

static Dx_len
item_stripe_contig_callback(void *arg, void *dptr, Dx_len io_dlen,
							Dx_desc *desc)
{
  Dx_ui8 *ptr;
  Contig_callback_state *state;

  state = (Contig_callback_state *) arg;

  ptr = (Dx_ui8 *) state->dptr + state->offset;
  memmove(ptr, dptr, (size_t) io_dlen);

  state->offset += io_dlen;

  return(io_dlen);
}

/*
 * Extract a contiguous region from an indirect datum.
 * Allocate a chunk of memory (if necessary) and copy/output the region from
 * the stripes.
 * Used, for example, to flatten a datum for output or key comparison.
 */
static Dx_rc
item_stripe_contig(Dx *db, Dx_ui8 *id, Ehash_xarg *arg)
{
  int do_malloc;
  void *ptr;
  Physvolnum initial_volume;
  Dx_len dlen;
  Physoff offset;
  Seglen seglen;
  Ehash_io io;
  Dx_len chunk_dlen;
  Dx_user_io *uio, user_io;
  Contig_callback_state state;
  Dx_rc rc;

  uio = arg->io.uio;
  unpack_idescriptor_fast(db, id, &dlen, &offset, &seglen, &initial_volume);

  do_malloc = (uio == NULL || uio->func == NULL);

  chunk_dlen = dlen;
  if (arg->region != NULL) {
	Dx_len start_pos;

	rc = item_verify_region(dlen, arg->region, &start_pos, NULL, &chunk_dlen);
	if (rc != DX_OK)
	  return(rc);
  }

  arg->io.dlen = chunk_dlen;

  ptr = NULL;
  if (do_malloc) {
	if ((ptr = (Dx_ui8 *) Xmalloc(chunk_dlen)) == NULL) {
	  db->last_errno = errno;
	  return(DX_MALLOC_FAILED);
	}
	state.dptr = ptr;
	state.offset = 0;
	user_io.func = item_stripe_contig_callback;
	user_io.arg = &state;
  }
  else {
	user_io.func = uio->func;
	user_io.arg = uio->arg;
  }

  io.uio = &user_io;
  rc = Stripe_contig(db, initial_volume, offset, seglen, dlen,
					 arg->region, &io);
  if (rc != DX_OK) {
	if (ptr != NULL)
	  Xfree(ptr);
	return(rc);
  }

  if (do_malloc)
	*(void **) uio->arg = ptr;

  arg->io.dlen = io.dlen;

  return(DX_OK);
}

/*
 * Allocate stripes for a datum and copy it.
 * All that needs to be done within the entry is to point to the stripe
 * descriptor page.
 *
 * If a get-function is not being used, we can easily compute the sizes of
 * the needed data structures.
 * But if a get-function is being used, it's more complicated since we don't
 * know how big the Dx_datum is until we read all of it.
 * Also, there may be bytes buffered.
 */
Dx_rc
Item_stripe_copy(Dx *db, Volume *volume, Dx_datum *datum, Dx_len zero_fill,
				 Segment *istripe_seg, Physoff *istripe_offset,
				 Seglen *istripe_seglen, Ehash_xarg *arg)
{
  Dx_ui32 i, n, nbytes, nleft, npages;
  Istripe_count_t nstripes;
  Dx_ui8 *id, *ptr;
  Stripe *stripes;
  Dx_rc rc;

  if (arg != NULL && arg->io.uio != NULL)
	return(item_stripe_copy_uio(db, volume, datum, zero_fill, istripe_seg,
								istripe_offset, istripe_seglen, arg));

  /*
   * Allocate and initialize a segment to hold the header and a vector of
   * stripe descriptors.
   * Compute how many pages are needed to hold the datum, then the number
   * of stripes to spread those pages over.
   */
  npages = NBYTES_TO_NPAGES(db, datum->dlen);
  if (zero_fill)
	npages++;
  nstripes = ITEM_INDIRECT_NPAGES_TO_NSTRIPES(db, npages);

  /*
   * Compute the size of the segment to hold the header and the segment
   * descriptors.
   * XXX alignment?
   */
  nbytes = ISTRIPE_HDR_SIZE + (nstripes * ISTRIPE_SIZE);
  *istripe_seglen = NBYTES_TO_PAGE_BYTES(db, nbytes);

  if (nbytes == 0) {
	*istripe_seg = 0;
	*istripe_offset = 0;
	return(DX_OK);
  }

  rc = Segment_allocate_segment(db, volume, *istripe_seglen,
								0, istripe_seg, istripe_offset);
  if (rc != DX_OK)
	return(rc);

  /* Allocate the stripes and their meta data. */
  stripes = (Stripe *) alloca((size_t) (sizeof(Stripe) * nstripes));
  rc = Stripe_allocate(db, datum->dlen, nstripes, stripes);
  if (rc != DX_OK)
	return(rc);

  /* Initialize the header. */
  id = (Dx_ui8 *) *istripe_seg;
  pack_istripe_hdr(db, id, "cosv", nstripes, (Physoff) 0,
				   (Seglen) 0, (Physvolnum) 0);

  /* Copy the Dx_datum to the stripes and fill in the meta data. */
  nleft = datum->dlen;
  ptr = (Dx_ui8 *) datum->dptr;
  
  i = 0;
  if (zero_fill) {
	id = ISTRIPE_PTR(*istripe_seg, 0);
	stripes[0].volnum = 0;
	stripes[0].seglen = 0;
	stripes[0].offset = 0;
	stripes[0].dlen = zero_fill;
	pack_istripe(db, id, &stripes[i]);
	i = 1;
  }

  for (; i < nstripes; i++) {
	n = (nleft > ITEM_INDIRECT_STRIPE_NBYTES(db))
	  ? ITEM_INDIRECT_STRIPE_NBYTES(db) : nleft;
	/* Safe and faster not to use memmove(). */
	Dx_memcpy(stripes[i].segment, (void *) ptr, (size_t) n);
	nleft -= n;
	ptr += n;
	id = ISTRIPE_PTR(*istripe_seg, i);
	pack_istripe(db, id, &stripes[i]);
  }

  return(DX_OK);
}

enum { STRIPE_COPY_INITIAL_NSTRIPES = 4 };
/*  	((DB->page_size - ISTRIPE_HDR_SIZE) / ISTRIPE_SIZE) */

enum { STRIPE_COPY_INCREMENTAL_NSTRIPES = 4 };
/*STRIPE_COPY_INITIAL_NSTRIPES(DB) */

enum { ITEM_MAXIMUM_INITIAL_STRIPE_ALLOCATION = 16777216 };

/*
 * The Dx_datum size isn't known, so allocate meta data and
 * stripes, repeatedly extend them and allocate new ones as necessary, and
 * then when the Dx_datum size is finally known, truncate any excess.
 * It's assumed that any readahead data does not exceed
 * ITEM_INDIRECT_STRIPE_NBYTES bytes in length.
 *
 * If successful, ISTRIPE_OFFSET and ISTRIPE_SEGLEN describe the segment
 * (which will have an initialized header) that contains the stripe
 * descriptors.
 * The DATUM's dlen is set to the actual length.
 */
static Dx_rc
item_stripe_copy_uio(Dx *db, Volume *volume, Dx_datum *datum, Dx_len zero_fill,
					 Segment *istripe_seg, Physoff *istripe_offset,
					 Seglen *istripe_seglen, Ehash_xarg *arg)
{
  int did_copy, do_zero_fill;
  Dx_len stripe_copy_initial_nstripes, st, stripe_written, total_written;
  Dx_ui32 need, tail_nbytes, tail_nstripes, tail_alloc_nstripes, waste;
  Dx_ui8 *id, *ptr;
  Stripe *stripe, *tail_stripes;
  Physvolnum v;
  Istripe_count_t nstripes, nstripes_alloc;
  Dx_desc *desc;
  Dx_rc rc;

  v = volume->volnum;
  do_zero_fill = (zero_fill != 0);

  /*
   * Compute the number of stripes to allocate initially.
   * Stripe descriptors are put in this vector by Stripe_allocate().
   */
  if (arg->io.desc.dlen == 0)
	stripe_copy_initial_nstripes = STRIPE_COPY_INITIAL_NSTRIPES;
  else
	stripe_copy_initial_nstripes
	  = ITEM_INDIRECT_NBYTES_TO_NSTRIPES(db, arg->io.desc.dlen);
  if ((stripe_copy_initial_nstripes * ITEM_INDIRECT_STRIPE_NBYTES(db))
	  > ITEM_MAXIMUM_INITIAL_STRIPE_ALLOCATION)
	stripe_copy_initial_nstripes = ITEM_MAXIMUM_INITIAL_STRIPE_ALLOCATION
	  / ITEM_INDIRECT_STRIPE_NBYTES(db);

  need = Dx_maximum_ui32(stripe_copy_initial_nstripes,
						 STRIPE_COPY_INCREMENTAL_NSTRIPES);
  tail_stripes = (Stripe *) alloca((size_t) (sizeof(Stripe) * need));

  tail_nstripes = 0;

  /* The total number of stripes used. */
  nstripes = 0;
  /* The total number of stripes allocated. */
  nstripes_alloc = 0;

  tail_alloc_nstripes = 0;		/* Shuts up compiler. */
  tail_nbytes = 0;

  total_written = 0;

  *istripe_seglen = 0;

  /*
   * Allocate a set of stripes and fill them one by one using the get-function.
   * If the allocated set is used up, allocate more and continue.
   */
  while (1) {
	if (nstripes == nstripes_alloc) {
	  if (nstripes_alloc == 0) {
		/*
		 * Initialize by allocating enough for the stripe header and a set of
		 * stripes.
		 */
		need = NBYTES_TO_PAGE_BYTES(db, ISTRIPE_HDR_SIZE
									+ (stripe_copy_initial_nstripes
									   * ISTRIPE_SIZE));
		rc = Segment_allocate_segment(db, volume, need, SEGMENT_FIXED,
									  istripe_seg, istripe_offset);
		if (rc != DX_OK)
		  return(rc);
		*istripe_seglen = need;
		nstripes_alloc = tail_alloc_nstripes = stripe_copy_initial_nstripes;
	  }
	  else {
		/*
		 * We've used up the previous set, so expand the set.
		 */
		need =
		  NBYTES_TO_PAGE_BYTES(db, ISTRIPE_SIZE
							   * (nstripes_alloc
								  + STRIPE_COPY_INCREMENTAL_NSTRIPES));
		rc = Segment_reallocate(db, volume, *istripe_offset,
								*istripe_seglen, need,
								istripe_seg, istripe_offset);
		if (rc != DX_OK)
		  return(rc);
		*istripe_seglen = need;
		nstripes_alloc += STRIPE_COPY_INCREMENTAL_NSTRIPES;
		tail_alloc_nstripes = STRIPE_COPY_INCREMENTAL_NSTRIPES;
	  }

	  /*
	   * Allocate stripes for this newly allocated set.
	   */
	  tail_nbytes = tail_alloc_nstripes * ITEM_INDIRECT_STRIPE_NBYTES(db);
	  if (arg->io.desc.dlen != 0 && arg->io.desc.dlen < tail_nbytes)
		tail_nbytes = arg->io.desc.dlen;
	  rc = Stripe_allocate(db, tail_nbytes, tail_alloc_nstripes, tail_stripes);
	  if (rc != DX_OK) {
		/* XXX Release segments */
		return(rc);
	  }
	  tail_nstripes = 0;
	}

	if (do_zero_fill) {
	  Stripe s;

	  s.volnum = 0;
	  s.seglen = 0;
	  s.offset = 0;
	  s.dlen = zero_fill;
	  id = ISTRIPE_PTR(*istripe_seg, 0);
	  pack_istripe(db, id, &s);

	  datum->dlen += zero_fill;
	  nstripes++;
	  do_zero_fill = 0;
	}

	/*
	 * Fill the next stripe.
	 */
	did_copy = 0;
	stripe = &tail_stripes[tail_nstripes];
	ptr = (Dx_ui8 *) stripe->segment;
	need = Dx_minimum_len(tail_nbytes, ITEM_INDIRECT_STRIPE_NBYTES(db));
	if (arg->io.bytes_buffered != 0) {
	  /* Safe and faster not to use memmove(). */
	  Dx_memcpy(ptr, arg->io.uio_buf, (size_t) arg->io.bytes_buffered);
	  stripe_written = arg->io.bytes_buffered;
	  need -= stripe_written;
	  tail_nbytes -= stripe_written;
	  ptr += stripe_written;
	  datum->dlen = stripe_written;
	  total_written += stripe_written;
	  arg->io.bytes_buffered = 0;
	  did_copy = 1;
	}
	else
	  stripe_written = 0;

	/*
	 * Call the get-function to read the data into a stripe, then fill in the
	 * meta data (dlen and flags were previously initialized).
	 */
	desc = &arg->io.desc;
	desc->error_occurred = 0;
	desc->last_errno = 0;
	st = (arg->io.uio->func)(arg->io.uio->arg, ptr, need, desc);
	if (desc->error_occurred) {
	  /* XXX Release segments */
	  db->last_errno = desc->last_errno;
	  return(DX_USERIO_ERROR);
	}

	if (st != 0) {
	  total_written += st;
	  stripe_written += st;
	  stripe->dlen = stripe_written;

	  id = ISTRIPE_PTR(*istripe_seg, nstripes);
	  pack_istripe(db, id, stripe);

	  nstripes++;
	  tail_nstripes++;
	  datum->dlen += st;
	  tail_nbytes -= st;
	  did_copy = 1;
	}

	if (did_copy) {
	  /*
	   * If we've reached EOF by reading zero bytes, the segment should be
	   * deleted.  But if some bytes were read before EOF, the segment should
	   * be truncated.
	   */
	  if (stripe->dlen != stripe->seglen) {
		rc = Segment_reallocate(db, db->volume[stripe->volnum], stripe->offset,
								stripe->seglen, stripe->dlen,
								&stripe->segment, &stripe->offset);
		if (rc != DX_OK) {
		  fprintf(stderr, "item_stripe_copy_uio: Stripe_reallocate failed!\n");
		  Fatal(rc);
		}
	  }
	  Segment_clear_fixed_active(db, db->volume[stripe->volnum],
								 stripe->offset);
	  /*
	   * It is probably a good idea to uncache (unmap) some stripes since
	   * they tend to clog up the address space.
	   */
	  Segment_uncache(db, 1);
	}

	/* The EOF has been seen. */
	if (st == 0 || st != need || arg->io.desc.dlen == total_written)
	  break;

	v = (v + 1) % db->nvolumes;
  }

  if ((waste = tail_alloc_nstripes - tail_nstripes) != 0) {
	/* Free up unused stripes but leave excess stripe header space. */
	/*
	  fprintf(stderr, "Waste: %ld stripes (%ld)\n", waste,
	  nstripes_alloc - nstripes);
	*/
	rc = Stripe_free_set(db, &tail_stripes[tail_nstripes], waste);
	if (rc != DX_OK) {
	  fprintf(stderr, "item_stripe_copy_uio: Stripe_free_set failed!\n");
	  Fatal(rc);
	}
  }

  /* Initialize the header. */
  id = (Dx_ui8 *) *istripe_seg;
  pack_istripe_hdr(db, id, "cvos", nstripes, (Physvolnum) 0,
				   (Physoff) 0, (Seglen) 0);

  Segment_clear_fixed_active(db, volume, *istripe_offset);

  return(DX_OK);
}

/*
 * Given the index of a (non-deleted) item within a page, check if it exists.
 * If ITEM is NULL or MODE is DX_EXISTS, nothing else is returned.
 * If MODE is DX_PROBE, then the key and information about the value are
 * returned.
 * If MODE is DX_FETCH, then a full fetch is done.
 */
Dx_rc
Item_index(Dx *db, Page page, Dx_ui32 ind, Dx_opmode mode, Item *item,
		   Ehash_xarg *arg)
{
  Dx_ui32 i, keyind, n;
  Dx_hash hval, khval, prev_hval;
  Item_flags_t hflags;
  Dx_ui16 keyp, keylen, valuep, valuelen;
  Dx_ui16 iptr, mflags, nheaders;
  Dx_ui8 *hptr;
  Ehash_xarg nullarg;
  Dx_rc rc;

  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);
  if (nheaders == 0)
	return(ITEM_NOTFOUND);

  n = 0;
  prev_hval = 0;
  keyind = 0;

  rc = DX_OK;
  for (i = 0; i < nheaders; i++) {
	iptr = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
	unpack_header_meta(db, hptr, NULL, &hflags, &hval, NULL);

	if (!ITEM_DELETED(hflags)) {
	  if (item->ks != NULL && item->ks->name != NULL) {
		rc = item_key_hash(db, page, hptr, &khval);
		if (rc != DX_OK)
		  return(rc);
	  }
	  if (prev_hval != hval) {
		keyind = 0;
		prev_hval = hval;
	  }
	  else
		keyind++;

	  if ((item->ks == NULL || item->ks->name == NULL)
		  || (khval ^ item->ks->hash) == hval) {
		if (n == ind) {
		  if (item == NULL || DX_OPMODE(mode) == DX_EXISTS)
			return(DX_OK);

		  item->hash = hval;
		  item->key->dptr = NULL;
		  item->key->dlen = keylen;
		  item->key->flags = ITEM_KEY_ALIGNMENT_VALUE(hflags);
		  item->key->ind = keyind;
		  item->key->keyspace = NULL;

		  if (DX_OPMODE(mode) == DX_PROBE || DX_OPMODE(mode) == DX_FETCH) {
			if (ITEM_KEY_INDIRECT(hflags)) {
			  Dx_user_io user_io;
			  Dx_ui8 *id;

			  user_io.func = NULL;
			  user_io.arg = &item->key->dptr;
			  nullarg.io.uio = &user_io;
			  nullarg.pos = NULL;
			  nullarg.region = NULL;
			  id = (Dx_ui8 *) page + keyp;
			  rc = item_fetch_indirect_datum(db, id, &nullarg);
			  if (rc != DX_OK)
				return(rc);
			  item->key->dlen = nullarg.io.dlen;
			  item->key->flags |= DX_DATUM_MALLOC;
			}
			else {
			  if (arg != NULL && (arg->flags & DX_ALWAYS_COPY)) {
				item->key->dptr = (void *) Xmalloc((size_t) keylen);
				/* Safe and faster not to use memmove(). */
				Dx_memcpy(item->key->dptr, (char *) page + keyp, keylen);
				item->key->flags |= DX_DATUM_MALLOC;
			  }
			  else {
				item->key->dptr = (char *) page + keyp;
				item->key->flags &= ~DX_DATUM_MALLOC;
			  }
			}
		  }

		  item->value->flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);
		  if (DX_OPMODE(mode) == DX_PROBE) {
			Dx_ui8 *id;

			item->value->dptr = NULL;
			if (ITEM_VALUE_INDIRECT(hflags)) {
			  id = (Dx_ui8 *) page + valuep;
			  unpack_idescriptor_fast(db, id, &item->value->dlen, NULL, NULL,
									  NULL);
			}
			else
			  item->value->dlen = valuelen;
		  }
		  else if (mode == DX_FETCH) {
			if (ITEM_VALUE_INDIRECT(hflags)) {
			  Dx_user_io user_io;
			  Dx_ui8 *id;

			  id = (Dx_ui8 *) page + valuep;
			  if (arg == NULL || arg->io.uio == NULL) {
				user_io.func = NULL;
				user_io.arg = &item->value->dptr;
				nullarg.io.uio = &user_io;
				nullarg.pos = NULL;
				nullarg.region = NULL;
				rc = item_fetch_indirect_datum(db, id, &nullarg);
				if (rc != DX_OK)
				  return(rc);
				item->value->dlen = nullarg.io.dlen;
			  }
			  else {
				rc = item_fetch_indirect_datum(db, id, arg);
				if (rc != DX_OK)
				  return(rc);
				item->value->dlen = arg->io.dlen;
			  }
			  item->value->flags |= DX_DATUM_MALLOC;
			}
			else {
			  if (arg != NULL && (arg->flags & DX_ALWAYS_COPY)) {
				item->value->dptr = (void *) Xmalloc((size_t) valuelen);
				/* Safe and faster not to use memmove(). */
				Dx_memcpy(item->value->dptr, (char *) page + valuep, valuelen);
				item->value->flags |= DX_DATUM_MALLOC;
			  }
			  else {
				item->value->dptr = (char *) page + valuep;
				item->value->dlen = valuelen;
				item->value->flags &= ~DX_DATUM_MALLOC;
			  }
			}
		  }

		  return(rc);
		}
		n++;
	  }
	}
  }

  return(ITEM_NOTFOUND);
}

/*
 * This is used by ehash_split() to iterate through each undeleted item in
 * a page, regardless of the key.
 * It has nothing to do with Dx_first()/Dx_next(); see Item_index().
 */
Dx_rc
Item_iterate(Dx *db, Page page, int first, Dx_ui32 *ptr, Dx_hash *hashp)
{
  Dx_ui32 h, i, ind;
  Dx_hash hval;
  Item_flags_t hflags;
  Dx_ui16 mflags, nheaders;
  Dx_ui8 *hptr;

  if (first) {
	*ptr = 0;
	ind = 0;
  }
  else
	ind = (*ptr) + 1;

  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);
  for (i = *ptr; i < nheaders; i++) {
	h = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + h;
	unpack_header_meta(db, hptr, NULL, &hflags, &hval, NULL);
	if (ITEM_DELETED(hflags))
	  continue;

	if (i >= ind) {
	  *ptr = i;
	  *hashp = hval;
	  return(DX_OK);
	}
  }

  return(ITEM_NOTFOUND);
}

/*
 * Move the indicated item to a new home page.
 * Note that if indirect datums are involved, their data stays put.
 */
Dx_rc
Item_move(Dx *db, Page page, Dx_ui32 ind, Page new_page)
{
  int indirect_key, indirect_value;
  Dx_ui32 h, kab, need, vab;
  Dx_ui16 mflags, new_avail, new_end_free, new_mflags, new_nheaders, nheaders;
  Dx_ui8 *hptr;
  Item item;
  Dx_datum key, value;
  Item_flags_t hflags;
  Dx_rc rc;

  unpack_meta(db, page, NULL, NULL, &mflags, NULL);

  h = ITEM_HEADER(page, ind, mflags);
  hptr = (Dx_ui8 *) page + h;
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  unpack_meta(db, new_page, &new_avail, &new_end_free, &new_mflags,
			  &new_nheaders);

  item.key = &key;
  item.value = &value;
  item_make_partial(db, hptr, &item);
  rc = item_need(ITEM_INDIRECT_THRESHOLD(db->page_size), &item,
				 new_avail, new_end_free, hflags,
				 &indirect_key, &indirect_value, &kab, &vab, &need);

  if (rc != DX_OK)
	return(rc);

  /* Verify that the new home page has sufficient space. */
  if (need > ITEM_CONTIG(db, new_page, new_nheaders, new_mflags)
	  || new_nheaders == ITEM_MAX_NHEADERS()) {
	if (ITEM_REORG_FLAG(new_mflags))
	  Item_reorg(db, new_page);
	unpack_meta(db, new_page, NULL, NULL, &new_mflags, &new_nheaders);
	if (need > ITEM_CONTIG(db, new_page, new_nheaders, new_mflags)
		|| new_nheaders == ITEM_MAX_NHEADERS())
	  return(ITEM_NOSPACE);
  }

  item_move_item(db, page, new_page, hptr);

  pack_header_or_flags(db, hptr, ITEM_DELETED_BITS);
  pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  if (DX_CHECK) {
	unpack_meta(db, new_page, NULL, NULL, &new_mflags, &new_nheaders);
	unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);

	if (ITEM_CONTIG(db, page, nheaders, mflags) > db->page_size
		|| ITEM_CONTIG(db, new_page, new_nheaders, new_mflags) > db->page_size) {
	  fprintf(stderr, "Item_move: check failed\n");
	  exit(1);
	}
  }

  return(DX_OK);
}

/*
 * Extended fetch.
 */
Dx_rc
Item_fetchx(Dx *db, Page page, Item *item, Ehash_xarg *arg)
{
  int do_uio;
  void *dptr;
  Dx_ui16 valuep, valuelen;
  Dx_ui8 *hptr, *id;
  Dx_len rci, rlen;
  Dx_desc *desc;
  Ehash_xarg xarg;
  Item_flags_t hflags;
  Dx_rc rc;

  if ((hptr = item_locate(db, page, item, NULL)) == NULL)
	return(ITEM_NOTFOUND);

  rc = DX_OK;
  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

  desc = &arg->io.desc;
  do_uio = (arg->io.uio != NULL);

  if (!do_uio)
	item->value->flags = 0;

  if (ITEM_VALUE_INDIRECT(hflags)) {
	Dx_user_io user_io;

	if (do_uio) {
	  user_io.func = arg->io.uio->func;
	  user_io.arg = arg->io.uio->arg;
	} else {
	  user_io.func = NULL;
	  user_io.arg = &item->value->dptr;
	}
	xarg.io.uio = &user_io;
	xarg.pos = arg->pos;

	id = (Dx_ui8 *) page + valuep;

	unpack_idescriptor_fast(db, id, &desc->dlen, NULL, NULL, NULL);
	desc->flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);
	xarg.region = NULL;
	if ((rc = item_fetch_indirect_datum(db, id, &xarg)) != DX_OK)
	  return(rc);

	if (do_uio)
	  arg->io.dlen = xarg.io.dlen;
	else {
	  item->value->dlen = xarg.io.dlen;
	  item->value->flags = DX_DATUM_MALLOC;
	}
  }
  else {
	Dx_len start_pos;

	rc = item_verify_region((Dx_len) valuelen, arg->region,
							&start_pos, NULL, &rlen);
	if (rc != DX_OK)
	  return(rc);
	dptr = (void *) ((char *) page + valuep + start_pos);

	if (do_uio) {
	  /* Call the put-function. */
	  desc->dlen = rlen;
	  desc->flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);
	  desc->error_occurred = 0;
	  desc->last_errno = 0;
	  rci = (arg->io.uio->func)(arg->io.uio->arg, dptr, rlen, desc);
	  if (rci != rlen || desc->error_occurred) {
		db->last_errno = desc->last_errno;
		return(DX_USERIO_ERROR);
	  }
	  arg->io.dlen = rlen;
	} else {
	  if (arg != NULL && (arg->flags & DX_ALWAYS_COPY)) {
		item->value->dptr = (void *) Xmalloc((size_t) rlen);
		/* Safe and faster not to use memmove(). */
		Dx_memcpy(item->value->dptr, dptr, (size_t) rlen);
		item->value->flags = DX_DATUM_MALLOC;
	  }
	  else
		item->value->dptr = dptr;
	  item->value->dlen = (Dx_len) rlen;
	  item->value->flags |= ITEM_VALUE_ALIGNMENT_VALUE(hflags);
	}
  }

  return(rc);
}

/*
 * Check if an item exists and optionally return information about the key
 * and value without returning the value.
 * If a region was specified, check that it exists.
 * The key's ind field indicates which instance is of interest.
 * Also counts how many identical keys there are.
 */
Dx_rc
Item_probe(Dx *db, Page page, Dx_opmode mode, Item *item, Ehash_xarg *arg)
{
  int not_found;
  Dx_len len;
  Dx_ui32 c, ind;
  Dx_ui16 iptr, next_iptr, valuelen, valuep;
  Dx_ui16 mflags, nheaders;
  Dx_ui8 *hptr, *id;
  Item_flags_t hflags;
  Dx_rc rc;

  item->key->count = 0;

  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);
  if (nheaders == 0)
	return(ITEM_NOTFOUND);

  ind = H(item->hash) & (ITEM_HASHTAB_NBUCKETS(mflags) - 1);
  c = 0;
  not_found = 1;

  iptr = unpack_hashtab_val(db, page, db->page_size, ind);
  while (iptr != 0) {
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
	unpack_header_meta(db, hptr, &next_iptr, &hflags, NULL, NULL);
	if (item_key_equal(db, page, hptr, item) == DX_OK) {
	  if (!ITEM_DELETED(hflags)) {
		if (not_found && c++ == item->key->ind) {
		  if (ITEM_VALUE_INDIRECT(hflags)) {
			id = (Dx_ui8 *) page + valuep;
			unpack_idescriptor_fast(db, id, &len, NULL, NULL, NULL);
		  }
		  else
			len = (Dx_len) valuelen;
		  rc = item_verify_region(len, arg->region, NULL, NULL, NULL);
		  if (rc != DX_OK)
			return(rc);
		  if (DX_OPMODE(mode) == DX_EXISTS)
			return(DX_OK);
		  item->value->dptr = NULL;
		  item->value->dlen = len;
		  item->value->flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);
		  item->key->flags = ITEM_KEY_ALIGNMENT_VALUE(hflags);
		  not_found = 0;
		}
		item->key->count++;
	  }
	}
	iptr = next_iptr;
  }

  return(not_found ? ITEM_NOTFOUND : DX_OK);
}

/*
 * Convert a direct value to an indirect one, regardless of its size.
 * Note that the header is reused.
 */
static Dx_rc
item_force_indirect_value(Dx *db, Volume *volume, Page page, Dx_ui8 *hptr)
{
  Dx_ui32 need, value_alignment_bytes;
  Dx_ui16 avail, end_free, mflags, nheaders, valuelen, valuep;
  Dx_rc rc;
  Dx_datum value;
  Item_flags_t hflags;

  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  if (ITEM_VALUE_INDIRECT(hflags))
	return(DX_OK);

  value.dptr = (void *) ((Dx_ui8 *) page + valuep);
  value.dlen = (Dx_len) valuelen;
  value.flags = ITEM_VALUE_ALIGNMENT_VALUE(hflags);

  /*
   * XXX If there were atomic transactions, this could sometimes be done by
   * overwriting the old value rather than allocating new space.
   */
  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);
  if (end_free < IDESC_SIZE)
	return(ITEM_NOSPACE);
  value_alignment_bytes = (end_free - IDESC_SIZE) % ALIGNMENT32_CONSTRAINT;
  need = IDESC_SIZE + value_alignment_bytes;
  if (need > avail)
	return(ITEM_NOSPACE);
  if (need > ITEM_CONTIG(db, page, nheaders, mflags)) {
	if (ITEM_REORG_FLAG(mflags)) {
	  Item_reorg(db, page);
	  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);
	}
  }
  if (end_free < IDESC_SIZE)
	return(ITEM_NOSPACE);
  value_alignment_bytes = (end_free - IDESC_SIZE) % ALIGNMENT32_CONSTRAINT;
  need = IDESC_SIZE + value_alignment_bytes;
  if (need > ITEM_CONTIG(db, page, nheaders, mflags))
	  return(ITEM_NOSPACE);

  /* The old value will be garbage collected eventually. */
  rc = item_add_indirect_value(db, volume, page,
							   hptr, &value, value_alignment_bytes, NULL);

  if (rc != DX_OK)
	return(rc);

  avail = avail - need + valuelen;
  pack_meta_fast(db, page, NULL, NULL, NULL, &avail);
  pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  return(rc);
}

/*
 * Replace a direct value with a new one.
 * Note that the header is reused.
 */
static Dx_rc
item_replace_direct_value(Dx *db, Volume *volume, Page page,
						  Dx_ui8 *hptr, Dx_datum *value)
{
  Dx_ui32 need, value_alignment_bytes;
  Dx_ui16 avail, end_free, mflags, nheaders, valuelen, valuep;
  Item_flags_t hflags;

  unpack_header_kv(db, hptr, NULL, NULL, &valuep, &valuelen);
  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  if (ITEM_VALUE_INDIRECT(hflags))
	return(DX_INTERNAL_ERROR);

  /*
   * XXX If there were atomic transactions, this could sometimes be done by
   * overwriting the old value rather than allocating new space.
   */
  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);

  if (end_free < (Dx_ui16) value->dlen)
	return(ITEM_NOSPACE);
  value_alignment_bytes = (end_free - (Dx_ui16) value->dlen)
	% ALIGNMENT32_CONSTRAINT;
  need = (Dx_ui32) value->dlen + value_alignment_bytes;
  if (need > avail)
	return(ITEM_NOSPACE);
  if (need > ITEM_CONTIG(db, page, nheaders, mflags)) {
	if (ITEM_REORG_FLAG(mflags)) {
	  Item_reorg(db, page);
	  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);
	}
  }
  if (end_free < (Dx_ui16) value->dlen)
	return(ITEM_NOSPACE);
  value_alignment_bytes = (end_free - (Dx_ui16) value->dlen)
	% ALIGNMENT32_CONSTRAINT;
  need = (Dx_ui32) value->dlen + value_alignment_bytes;
  if (need > ITEM_CONTIG(db, page, nheaders, mflags))
	  return(ITEM_NOSPACE);

  /* The old value will be garbage collected eventually. */
  item_add_value(db, page, db->page_size, hptr, value, value_alignment_bytes);

  avail = avail - need + valuelen;
  pack_meta_fast(db, page, NULL, NULL, NULL, &avail);
  pack_meta_or_flags(db, page, ITEM_REORG_FLAG_BITS);

  return(DX_OK);
}

/*
 * Release the old stripes and force (regardless of its size) the VALUE into
 * new stripes.
 * Note that the header is reused.
 * XXX This should use Stripe_overwrite_data() to avoid releasing and
 * then reallocating stripes, if it can be done atomically.
 */
static Dx_rc
item_replace_indirect_value(Dx *db, Volume *volume, Page page,
							Dx_ui8 *hptr, Dx_datum *value)
{
  Dx_ui8 *id;
  Dx_ui16 valuep;
  Dx_len old_dlen;
  Item_flags_t hflags;
  Physoff off;
  Physvolnum initial_volume;
  Seglen istripe_seglen, seglen;
  Segment istripe_seg;
  Physoff istripe_offset;
  Dx_rc rc;

  unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
  unpack_header_kv(db, hptr, NULL, NULL, &valuep, NULL);
  if (!ITEM_VALUE_INDIRECT(hflags))
	return(DX_INTERNAL_ERROR);

  id = (Dx_ui8 *) page + valuep;
  unpack_idescriptor_fast(db, id, &old_dlen, &off, &seglen, &initial_volume);

  /* Copy to the same volume as the old data. */
  rc = Item_stripe_copy(db, db->volume[initial_volume], value, 0, &istripe_seg,
						&istripe_offset, &istripe_seglen, NULL);
  if (rc != DX_OK)
	return(rc);

  /* Free the old value. */
  rc = Stripe_free(db, initial_volume, off, seglen);
  if (rc != DX_OK)
	return(rc);

  pack_idescriptor_fast(db, id, &value->dlen, &istripe_offset, &istripe_seglen,
						&initial_volume);

  return(DX_OK);
}

Dx_ui32
Item_count(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui32 count, h, i;
  Dx_ui8 *hptr;
  Dx_ui16 mflags, nheaders;
  Item_flags_t hflags;

  count = 0;
  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);
  for (i = 0; i < nheaders; i++) {
	h = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + h;
	unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
	if (!ITEM_DELETED(hflags))
	  count++;
  }

  return(count);
}

/*
 * Return an indication of how full this page is.
 */
double
Item_page_usage(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui16 avail;
  Dx_ui16 mflags, nheaders;
  double val;

  unpack_meta(db, page, &avail, NULL, &mflags, &nheaders);

  val = (double) (page_size - avail) / (double) page_size;

  return(val);
}

void
Item_load(Dx *db, Page page, Dx_ui32 page_size, PageLoad *load, int full)
{
  Dx_ui32 h, i, indirect, ndeleted, nindirect, t, u;
  Dx_ui16 avail, keylen, keyp, valuelen, valuep;
  Dx_ui8 *hptr, *id;
  Dx_ui16 mflags, nheaders;
  Dx_len dlen;
  Item_flags_t hflags;

  unpack_meta(db, page, &avail, NULL, &mflags, &nheaders);

  t = ITEM_TOTAL_META_SIZE(db, page, mflags);
  u = 0;
  nindirect = 0;
  indirect = 0;
  ndeleted = 0;

  for (i = 0; i < nheaders; i++) {
	h = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + h;
	unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
	unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);

	if (ITEM_DELETED(hflags))
	  ndeleted++;
	else {
	  if (ITEM_KEY_INDIRECT(hflags)) {
		if (full) {
		  id = (Dx_ui8 *) page + keyp;
		  unpack_idescriptor_fast(db, id, &dlen, NULL, NULL, NULL);
		  indirect += dlen;
		}
		nindirect++;
		t += keylen;
	  }
	  else
		u += keylen;

	  if (ITEM_VALUE_INDIRECT(hflags)) {
		if (full) {
		  id = (Dx_ui8 *) page + valuep;
		  unpack_idescriptor_fast(db, id, &dlen, NULL, NULL, NULL);
		  indirect += dlen;
		}
		nindirect++;
		t += valuelen;
	  }
	  else
		u += valuelen;
	}
  }

  load->total_meta = t;
  load->avail = avail;
  load->used = u;
  load->actual_meta = t - (ndeleted * ITEM_HEADER_SIZE);
  load->nindirect = nindirect;
  load->indirect = indirect;
  load->ndeleted = ndeleted;
}

void
Item_init(Dx *db, Page page)
{
  unsigned int i;
  Avail_t avail;
  Metaflags_t flags;
  Endfree_t end_free;

  if (db->flags & DX_CONFIG_PERSISTENT)
	Dx_memset(page, 0, db->page_size);

  flags = HASHTAB_SIZE_OF_PAGE_SIZE(db->page_size) >> 4;

  end_free = (Endfree_t) db->page_size;
  pack_meta_fast(db, page, &nheaders_zero, &flags, &end_free, NULL);
  /* This must be done after the above values have been set. */
  avail = db->page_size - ITEM_TOTAL_META_SIZE(db, page, flags);
  pack_meta_fast(db, page, NULL, NULL, NULL, &avail);

  for (i = 0; i < HASHTAB_SIZE_OF_PAGE_SIZE(db->page_size); i++)
	pack_hashtab_val(db, page, db->page_size, i, 0);
}

void
page_check(Dx *db, PageNum b)
{
  Dx_ui32 i;
  Dx_ui16 keyp, valuep;
  Dx_ui16 avail, iptr;
  Dx_ui8 *hptr;
  Dx_ui16 mflags, nheaders;
  Page page;
  Item_flags_t hflags;
  Dx_rc rc;

  if ((rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page))
	  != DX_OK)
	return;

  unpack_meta(db, page, &avail, NULL, &mflags, &nheaders);
  if (ITEM_REORG_FLAG(mflags) == 0
	  && avail != ITEM_CONTIG(db, page, nheaders, mflags)) {
	fprintf(stderr, "page_check!\n");
	Fatal(DX_INTERNAL_ERROR);
	/*NOTREACHED*/
  }

  for (i = 0; i < nheaders; i++) {
	iptr = ITEM_HEADER(page, i, mflags);
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_meta(db, hptr, NULL, &hflags, NULL, NULL);
	unpack_header_kv(db, hptr, &keyp, NULL, &valuep, NULL);
	if (!ITEM_KEY_INDIRECT(hflags) && keyp >= db->page_size) {
	  fprintf(stderr, "page_check 2\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
	if (!ITEM_VALUE_INDIRECT(hflags) && valuep >= db->page_size) {
	  fprintf(stderr, "page_check 3\n");
	  Fatal(DX_INTERNAL_ERROR);
	  /*NOTREACHED*/
	}
  }

}

void
header_dump(Dx *db, Dx_ui8 *hptr)
{
  Item_flags_t hflags;
  Dx_ui16 chainp;
  Dx_hash hval;
  Dx_uid uid;
  Dx_ui16 keyp, keylen, valuep, valuelen;

  unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
  unpack_header_meta(db, hptr, &chainp, &hflags, &hval, &uid);

  fprintf(stderr, "(%4lu %4lu%s) (%4lu %4lu%s) %11lu/%-11lu %11llu %4u %u%s\n",
		  (Dx_ui32) keyp, (Dx_ui32) keylen,
		  ITEM_KEY_INDIRECT(hflags) ? " I" : "  ",
		  (Dx_ui32) valuep, (Dx_ui32) valuelen,
		  ITEM_VALUE_INDIRECT(hflags) ? " I" : "  ",
		  uid.tv_sec, uid.count,
		  (Hash64) hval, chainp, hflags,
		  ITEM_DELETED(hflags) ? " (D)" : "");
}

void
Item_dump(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui32 i;
  Dx_ui16 avail, end_free, iptr;
  Dx_ui8 *hptr;
  Dx_ui16 mflags, nheaders;
  PageLoad pl;

  if (page == NULL)
	return;

  unpack_meta(db, page, &avail, &end_free, &mflags, &nheaders);

  fprintf(stderr,
		  "nheaders=%u, avail=%u, flags=%u, free_end=%u, contig=%u reorg=%d\n",
		  nheaders,
		  avail,
		  mflags,
		  end_free,
		  ITEM_CONTIG(db, page, nheaders, mflags),
		  ITEM_REORG_FLAG(mflags) != 0);

  Item_load(db, page, page_size, &pl, 1);
  fprintf(stderr, "Load: meta=%.1f, actual_meta=%.1f, avail=%.1f, used=%.1f\n",
		  (double) pl.total_meta / (double) page_size,
		  (double) pl.actual_meta / (double) page_size,
		  (double) pl.avail / (double) page_size,
		  (double) pl.used / (double) page_size);

  for (i = 0; i < ITEM_HASHTAB_NBUCKETS(mflags); i++) {
	fprintf(stderr, "%5u ", unpack_hashtab_val(db, page, page_size, i));
	if (((i + 1) % 8) == 0)
	  fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n");

  for (i = 0; i < nheaders; i++) {
	iptr = ITEM_HEADER(page, i, mflags);
	fprintf(stderr, "\t%3lu +%4lu| ", i, (unsigned long) iptr);
	hptr = (Dx_ui8 *) page + iptr;
	header_dump(db, hptr);
  }
}

void
dump(Dx *db)
{
  Dx_ui32 i;
  Dx_ui16 nheaders;
  Page page;
  Dx_rc rc;

  for (i = 0; i < db->npages; i++) {
	if ((rc = Page_get(db, PAGENUM_TO_VOLUME(db, i), i, PAGE_MUSTEXIST, &page))
		== DX_OK) {
	  unpack_meta(db, page, NULL, NULL, NULL, &nheaders);
	  if (nheaders)
		Item_dump(db, page, db->page_size);
	}
  }
}

int
Item_show(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui32 i, n, nchains, total;
  Dx_ui16 iptr;
  Dx_ui16 mflags, nheaders;

  unpack_meta(db, page, NULL, NULL, &mflags, &nheaders);

  fprintf(stderr, "%u header%s\n",
		  nheaders, (nheaders != 1) ? "s" : "");
  fprintf(stderr, "contig=%d, reorg=%d\n",
		  ITEM_CONTIG(db, page, nheaders, mflags),
		  ITEM_REORG_FLAG(mflags) != 0);
  fprintf(stderr, "thresh=%lu\n",
		  (unsigned long) ITEM_INDIRECT_THRESHOLD(page_size));

  fprintf(stderr, "hash size=%lu\n", ITEM_HASHTAB_NBUCKETS(mflags));

  total = nchains = 0;
  for (i = 0; i < HASHTAB_SIZE_OF_PAGE_SIZE(page_size); i++) {
	if ((iptr = unpack_hashtab_val(db, page, page_size, i)) != 0) {
	  fprintf(stderr, "Bucket %lu:\n", (unsigned long) i);
	  if ((n = item_show_bucket(db, page, page_size, iptr)) != 0) {
		total += n;
		nchains++;
	  }
	}
  }

  if (nchains)
	fprintf(stderr, "Mean chain length: %.1f\n",
			(double) total / (double) nchains);

  return(nheaders);
}

static int
item_show_bucket(Dx *db, Page page, Dx_ui32 page_size, Dx_ui32 hindex)
{
  Dx_ui32 n;
  Dx_len len;
  Dx_ui16 keyp, keylen, valuep, valuelen;
  Dx_ui16 iptr;
  Dx_ui8 *hptr, *id;
  Dx_uid uid;
  Dx_hash h;
  Item_flags_t hflags;

  n = 0;
  iptr = hindex;
  do {
	hptr = (Dx_ui8 *) page + iptr;
	unpack_header_kv(db, hptr, &keyp, &keylen, &valuep, &valuelen);
	unpack_header_meta(db, hptr, NULL, &hflags, &h, &uid);

	if (ITEM_DELETED(hflags))
	  fprintf(stderr, "(deleted) ");
	if (ITEM_KEY_INDIRECT(hflags)) {
	  id = (Dx_ui8 *) page + keyp;
	  unpack_idescriptor_fast(db, id, &len, NULL, NULL, NULL);
	  fprintf(stderr, "key=indirect (len=%lu) ",
			  (unsigned long) len);
	}
	else
	  fprintf(stderr, "key=\"%s\" (len=%lu, align=%lu) ",
			  (char *) page + keyp,
			  (unsigned long) keylen,
			  ITEM_KEY_ALIGNMENT(hflags));
	if (ITEM_VALUE_INDIRECT(hflags)) {
	  id = (Dx_ui8 *) page + valuep;
	  unpack_idescriptor_fast(db, id, &len, NULL, NULL, NULL);
	  fprintf(stderr, "value=indirect (len=%lu) ",
			  (unsigned long) len);
	}
	else
	  fprintf(stderr, "value=\"%s\" (len=%lu, align=%lu)\n",
			  (char *) page + valuep,
			  (unsigned long) valuelen,
			  ITEM_VALUE_ALIGNMENT(hflags));
	fprintf(stderr, "\n");
	n++;
	unpack_header_meta(db, hptr, &iptr, NULL, NULL, NULL);
  } while (iptr != 0);

  return(n);
}

void
Item_info(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui32 n;
  Dx_ui16 iptr;
  Dx_ui8 *hptr;
  Dx_ui16 avail, contig, mflags, nheaders;

  unpack_meta(db, page, &avail, NULL, &mflags, &nheaders);
  contig = ITEM_CONTIG(db, page, nheaders, mflags);

  iptr = unpack_hashtab_val(db, page, page_size, 0);
  hptr = (Dx_ui8 *) page + iptr;

  n = ITEM_TOTAL_META_SIZE(db, page, mflags);

  fprintf(stderr, "fixed meta size=%lu bytes, ",
		  META_HEADER_SIZE
		  + ITEM_HASHTAB_NBUCKETS(mflags) * sizeof(Dx_ui16));
  fprintf(stderr, "total meta size=%lu bytes\n", (unsigned long) n);
  fprintf(stderr, "header size=%u bytes, hashtab size=%lu buckets\n",
		  ITEM_HEADER_SIZE, ITEM_HASHTAB_NBUCKETS(mflags));
  fprintf(stderr, "page overhead=%.1f%%\n", (100.0 *  n) / (double) page_size);
  fprintf(stderr, "%u headers (%lu deleted)\n",
		  nheaders,
		  (unsigned long) (nheaders - Item_count(db, page, page_size)));
  fprintf(stderr, "avail=%lu, contig=%lu, flags=%lx\n",
		  (unsigned long) avail, (unsigned long) contig,
		  (unsigned long) mflags);
}

void
Item_layout(Dx *db, Page page, Dx_ui32 page_size)
{
  Dx_ui16 mflags;

  fprintf(stderr, "Meta header:\n");
  fprintf(stderr, "%2d: %6d (Nheaders)\n",
		  META_NHEADERS_OFFSET, ITEM_GET_NHEADERS(db, page));
  fprintf(stderr, "%2d: %06x (Metaflags)\n",
		  META_METAFLAGS_OFFSET, ITEM_GET_METAFLAGS(db, page));
  fprintf(stderr, "%2d: %6d (Endfree)\n",
		  META_ENDFREE_OFFSET, ITEM_GET_ENDFREE(db, page));
  fprintf(stderr, "%2d: %6d (Avail)\n",
		  META_AVAIL_OFFSET, ITEM_GET_AVAIL(db, page));
  fprintf(stderr, "META_HEADER_SIZE = %d bytes\n", META_HEADER_SIZE);

  fprintf(stderr, "\n");
  unpack_meta(db, page, NULL, NULL, &mflags, NULL);
  fprintf(stderr, "ITEM_HASHTAB_NBUCKETS = %ld (%ld bytes)\n",
		  ITEM_HASHTAB_NBUCKETS(mflags), ITEM_HASHTAB_SIZE(mflags));

  fprintf(stderr, "\nItem header:\n");
  fprintf(stderr, "%2d: (Keyp)\n", ITEM_KEYP_OFFSET);
  fprintf(stderr, "%2d: (Valuep)\n", ITEM_VALUEP_OFFSET);
  fprintf(stderr, "%2d: (Keylen)\n", ITEM_KEYLEN_OFFSET);
  fprintf(stderr, "%2d: (Valuelen)\n", ITEM_VALUELEN_OFFSET);
  fprintf(stderr, "%2d: (UIDHI)\n", ITEM_UIDHI_OFFSET);
  fprintf(stderr, "%2d: (Hash)\n", ITEM_HASH_OFFSET);
  fprintf(stderr, "%2d: (Chainp)\n", ITEM_CHAINP_OFFSET);
  fprintf(stderr, "%2d: (Flags)\n", ITEM_FLAGS_OFFSET);
  fprintf(stderr, "ITEM_HEADER_SIZE = %d\n", ITEM_HEADER_SIZE);
}

/*
 * Verify that the given region lies completely within a Dx_datum that's
 * DLEN bytes long.
 * A region that's zero bytes long is considered valid.
 * A region that goes beyond the end of the Dx_datum is invalid.
 * Optionally return pointers to the starting offset, ending offset,
 * and total length of the region.
 * Then ending offset is one past the last byte in the region.
 */
Dx_rc
item_verify_region(Dx_len dlen, Dx_region *r,
				   Dx_len *spos, Dx_len *epos, Dx_len *lenp)
{
  Dx_len end_pos, start_pos;

  if (r->mode == DX_REGION_ALL) {
	if (spos != NULL)
	  *spos = 0;
	if (epos != NULL)
	  *epos = dlen - 1;
	if (lenp != NULL)
	  *lenp = dlen;
	return(DX_OK);
  }

  if (r->start.mode == DX_POINT_ABSOLUTE)
	start_pos = r->start.u.pos;
  else if (r->start.mode == DX_POINT_RELATIVE) {
	if (r->start.u.offset_from_end < 0
		&& (Dx_len) (0 - r->start.u.offset_from_end) > dlen)
	  return(DX_BAD_REGION);
	start_pos = dlen + r->start.u.offset_from_end;
  }
  else
	return(DX_BAD_REGION);

  if (r->mode == DX_REGION_LEN) {
	end_pos = start_pos + r->u.dlen;
	if (end_pos < start_pos)
	  return(DX_BAD_REGION);
  }
  else if (r->mode == DX_REGION_PTOP) {
	if (r->u.end.mode == DX_POINT_ABSOLUTE)
	  end_pos = r->u.end.u.pos;
	else if (r->u.end.mode == DX_POINT_RELATIVE) {
	  if (r->u.end.u.offset_from_end > 0
		  || (r->u.end.u.offset_from_end < 0
			  && (Dx_len) (0 - r->u.end.u.offset_from_end) > dlen))
		return(DX_BAD_REGION);
	  end_pos = dlen + r->u.end.u.offset_from_end;
	}
	else
	  return(DX_BAD_REGION);
  }
  else
	return(DX_BAD_REGION);

  if (start_pos >= dlen || end_pos < start_pos || end_pos > dlen)
	return(DX_BAD_REGION);

  if (spos != NULL)
	*spos = start_pos;

  if (epos != NULL)
	*epos = end_pos;

  if (lenp != NULL)
	*lenp = end_pos - start_pos;

  return(DX_OK);
}

/*
 * Verify that the given position lies either completely within a Dx_datum that's
 * DLEN bytes long or past its end.
 */
Dx_rc
item_verify_position(Dx_len dlen, Dx_point *r, Dx_len *pos)
{
  Dx_len start_pos;

  if (r->mode == DX_POINT_ABSOLUTE)
	start_pos = r->u.pos;
  else if (r->mode == DX_POINT_RELATIVE) {
	if (r->u.offset_from_end < 0
		&& (Dx_len) (0 - r->u.offset_from_end) > dlen)
	  return(DX_BAD_REGION);
	start_pos = dlen + r->u.offset_from_end;
  }
  else
	return(DX_BAD_REGION);

  if (pos != NULL)
	*pos = start_pos;

  return(DX_OK);
}

