/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _DX_LOCAL_H_
#define _DX_LOCAL_H_

/*
 * "Simplicity, Clarity, Generality"
 * Style influenced in part by "The Practice of Programming",
 * Brian W. Kernighan and Rob Pike, Addison-Wesley Professional Computer
 * Series, ISBN 020161586X, February 1999.
 */

#include <sys/types.h>
#ifdef LINUX
#endif
#include <netinet/in.h>
#ifdef BSD
#include <sys/param.h>
#endif

#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "dx.h"

#undef INLINE
#if defined(G_HAVE___INLINE)
#define INLINE __inline
#elif defined(G_HAVE___INLINE__)
#define INLINE __inline__
#elif defined(G_HAVE_INLINE)
#define INLINE inline
#else
#define INLINE
#endif

#ifndef DO_CHECKS
#define DO_CHECKS				0
#endif

#ifndef DO_DEBUG
#define DO_DEBUG				0
#endif

#ifndef DO_STATS
#define DO_STATS				0
#endif

#ifndef DO_VERBOSE
#define DO_VERBOSE              0
#endif

enum {
  DX_DBMCOMPAT = 0,
  DX_DEBUG     = DO_DEBUG,
  DX_CHECK     = DO_CHECKS,
  DX_STATS     = DO_STATS,
  DX_VERBOSE   = DO_VERBOSE
};

#if DX_BITS == 64
typedef Dx_ui64 Physoff;
typedef Dx_ui64 Seglen;

#define Physoff64
#define Seglen64

#elif DX_BITS == 32
typedef Dx_ui32 Physoff;
typedef Dx_ui32 Seglen;

#define Physoff32
#define Seglen32

#else
#error "DX_BITS must be defined to be 32 or 64!"
#endif

typedef void *Segment;
typedef void *Page;
typedef Dx_ui32 Physvolnum;		/* 32 bits ought to always be enough */
typedef Dx_ui32 Logvolnum;		/* 32 bits ought to always be enough */
#define Physvolnum32
#define Logvolnum32
typedef Dx_ui64 Hash64;

enum {
  Physvolnum_s  = sizeof(Physvolnum),
  Logvolnum_s  = sizeof(Logvolnum),
  Dx_len_s      = sizeof(Dx_len),
  Dx_hash_s     = sizeof(Dx_hash),
  Physoff_s     = sizeof(Physoff),
  Seglen_s      = sizeof(Seglen)
};

/* Dx_hash, PageNum, and DX_MAX_UINT must be the same size. */
typedef Dx_hash			PageNum;

#define Dx_hash_nbits	(Dx_hash_s * BITS_PER_BYTE)

#define MAX_LEVELS		(Dx_ui64) (sizeof(DX_MAX_UINT) * (BITS_PER_BYTE))

enum {
  /* A (hopefully) unique value. */
  META_MAGIC = 0x626a622e,					/* "bjb." */

  BITS_PER_BYTE = 8,

  /*
   * All of these must be a power of two.
   * A recommended minimum page size is 4096, but because mmap() wants to be
   * given a page-aligned area, the minimum actually has to be the system's
   * page size as indicated by getpagesize().  The hard maximum limit is
   * 65536 since offsets within a page are 16 bits.
   *
   * XXX There are database portability issues here, since the system page
   * size may be smaller on the platform where the database was created.
   * Incidentally, there's a similar problem with alignment constraints.
   */
  META_HEADER_NBYTES = 8192,
  PHYSMAP_MAP_NBYTES = 4096,
  LOGMAP_MAP_NBYTES  = ((PHYSMAP_MAP_NBYTES) * 8 * Physoff_s),
  DEFAULT_PAGE_SIZE  = 8192,

  /* This is the default maximum volume "stripe size", in pages. */
  DEFAULT_MAX_STRIPE_NPAGES = 16,

  /* This is the maximum allowed stripe size.  See dx.c. */
  MAXIMUM_MAX_STRIPE_NBYTES = 8388608,		/* 2^23 */

  /* Configuration flags for Dx_open(). */
  DX_DEFAULT_OPEN_MODE = ((mode_t) 0644),

  /* The byte to write for DX_CONFIG_ERASE_ON_DELETE. */
  DX_ERASE_ON_DELETE_CHAR = 0
};

typedef struct Physmap {
  Segment map;
  Dx_ui32 avail;		/* Number of pages available. */
} Physmap;

typedef struct Area {
  Dx_ui32 areanum;
  Segment logmap;
  Physmap physmap;
  struct Area *next;
} Area;

typedef struct Volume {
  Physvolnum volnum;
  char *path;
  int fd;
  unsigned long st_blksize;
  Segment meta;
  Dx_ui64 mmap_size;
  Area *area;
} Volume;

/* State information for iteration through all keys in the db. */
typedef struct Dx_iter {
  PageNum b;		/* Current page number in the iteration. */
  Dx_ui32 ind;		/* Current item number within that page. */
} Dx_iter;

/*
 * XXX This probably shouldn't make use of the time at all.
 */
typedef struct Dx_uid {
  Dx_ui32 tv_sec;
  Dx_ui32 count;
} Dx_uid;

typedef struct Dx_stamp {
  Dx_ui32 high;
  Dx_ui32 low;
} Dx_stamp;

enum {
  /* Number of hash buckets; should be a prime. */
  SEGMENT_MAP_NENTRIES      = 127,

  /* Soft limit on the number of cached mmapped segments. */
  SEGMENT_CACHE_ENTRY_LIMIT = 256,

  /* Soft limit on the total bytes in cached mmapped segments. */
  SEGMENT_CACHE_BYTE_LIMIT  = 2000000
};

/* These are bits in SegmentMap flags. */
enum {
  SEGMENT_IS_PAGE  =  1,	/* A fixed-length page (bucket) */
  SEGMENT_RW       =  2,	/* Read-write access rather than read-only */
  SEGMENT_FIXED    =  4,	/* Fixed in memory - don't uncache the mapping */
  SEGMENT_META     =  8,	/* This is a meta data segment */
  SEGMENT_MODIFIED = 16
};

typedef struct SegmentMap {
  Dx_ui32 flags;
  PageNum b;
  Physoff offset;
  Volume *volume;
  Segment seg;
  Seglen length;		/* This is the allocated length in bytes */
  Dx_ui32 count;
  struct SegmentMap *next;
  struct SegmentMap *active_list_next;
} SegmentMap;

typedef struct VolumeWeight {
  Dx_ui32 *weights;
  Dx_ui32 *cw;			/* Cumulative weights */
} VolumeWeight;

struct Dx {
  Dx_ui32 flags;
  mode_t open_mode;
  Dx_ui32 dx_bits;

  Dx_ui32 max_stripe_npages;	/* Max number of pages in a stripe. */

  Dx_endian endian;			/* Either DX_BIG_ENDIAN or DX_LITTLE_ENDIAN. */

  size_t system_page_size;
  Dx_ui32 page_size;

  Dx_ui64 npages;
  Dx_ui64 nexpansions;		/* Number of full bucket expansions so far. */
  Dx_ui64 boundary;			/* The next bucket to have a controlled split.
							   Actually, it's a PageNum. */
  Dx_ui32 controlled_split_check;

  Dx_iter iter;				/* Iteration state. */
  Dx_uid current_uid;		/* The unique id for the next added item. */
  Dx_stamp creation_stamp;	/* Creation date of this database. */

  Volume **volume;
  Physvolnum nvolumes;		/* Number of physical volumes. */
  VolumeWeight volume_weights;
  Logvolnum nlogical_volumes;

  int mmap_flags;
  int mmap_prot;

  Dx_keyspace *keyspace;

  int last_errno;
  Dx_hash (*hashfunc)(Dx_ui8 *str, Dx_len len, Dx_hash init);

  SegmentMap *segment_map[SEGMENT_MAP_NENTRIES];
  /*
   * The same page is often searched for several times consecutively.
   * Cache a pointer to the result of the last lookup, per volume.
   */
  SegmentMap **last_segment_lookup;
  SegmentMap *active_list;
  Dx_ui32 nmapped;
  Dx_ui64 nbytesmapped;
  Dx_ui32 nmmap_calls;

  Dx_stats_counters stats;
};

enum stats_mode { SPLIT, ADD, MOVE, LOCATE, PROBE, REORG };

static INLINE Dx_ui32
Dx_maximum_ui32(Dx_ui32 x, Dx_ui32 y)
{

  if (x > y)
	return(x);
  return(y);
}

static INLINE Dx_len
Dx_maximum_len(Dx_len x, Dx_len y)
{

  if (x > y)
	return(x);
  return(y);
}

static INLINE Dx_ui32
Dx_minimum_ui32(Dx_ui32 x, Dx_ui32 y)
{

  if (x < y)
	return(x);
  return(y);
}

static INLINE Dx_len
Dx_minimum_len(Dx_len x, Dx_len y)
{

  if (x < y)
	return(x);
  return(y);
}

static INLINE Dx_optype
DX_OPMODE(Dx_opmode mode)
{
  Dx_ui32 m;
  Dx_optype type;

  m = (Dx_ui32) mode;
  type = (Dx_optype) (m & DX_OPMODE_MASK);
  return(type);
}

/* The number of full and partial pages needed for LEN bytes. */
static INLINE Dx_len
NBYTES_TO_NPAGES(Dx *db, Dx_len len)
{
  Dx_len val;

  val = (len + (db->page_size - 1)) / db->page_size;
  return(val);
}

/* The number of bytes in NPAGES. */
static INLINE Dx_len
NPAGES_TO_NBYTES(Dx *db, Dx_ui32 npages)
{
  Dx_len val;

  val = npages * db->page_size;
  return(val);
}

/* The rounded-up segment length needed for LEN bytes. */
static INLINE Dx_len
NBYTES_TO_PAGE_BYTES(Dx *db, Dx_len len)
{
  Dx_len val;

  val = NBYTES_TO_NPAGES(db, len) * db->page_size;
  return(val);
}

static INLINE Dx_ui64
PHYSMAP_NPAGES_MAPPED(void)
{
  Dx_ui64 val;

  val = (Dx_ui64) ((PHYSMAP_MAP_NBYTES) << 3);
  return(val);
}

static INLINE Dx_len
AREA_SIZE_USERPAGES_NBYTES(Dx_len map_nbytes, Dx_ui32 page_size)
{
  Dx_len val;

  val = PHYSMAP_NPAGES_MAPPED() * (Dx_ui64) page_size;
  return(val);
}

static INLINE Dx_ui64
AREA_SIZE(Dx *db)
{
  Dx_ui64 val;

  val = (Dx_ui64) (LOGMAP_MAP_NBYTES
				  + PHYSMAP_MAP_NBYTES
				  + AREA_SIZE_USERPAGES_NBYTES(PHYSMAP_MAP_NBYTES,
											   db->page_size));
  return(val);
}

static INLINE Dx_len
AREA_OFFSET(Dx *db, Dx_ui32 rel_areanum)
{
  Dx_len val;

  val = META_HEADER_NBYTES + rel_areanum * AREA_SIZE(db);
  return(val);
}

typedef struct PhysVolnumIter {
  Physvolnum physvolnum;
  Logvolnum logvolnum;
} PhysVolnumIter;

static INLINE Physvolnum
SET_PHYSICAL_VOLNUM(Dx *db, Logvolnum logical_volnum, PhysVolnumIter *pvi)
{
  Physvolnum v;
  Logvolnum lv;

  lv = logical_volnum % db->nlogical_volumes;
  for (v = 0; v < db->nvolumes; v++) {
	if (lv < db->volume_weights.cw[v]) {
	  if (pvi != NULL) {
		pvi->physvolnum = v;
		pvi->logvolnum = lv;
	  }
	  return(v);
	}
  }

  /* Shouldn't happen... */
  fprintf(stderr, "SET_PHYSICAL_VOLNUM!?!\n");
  abort();
}

#ifdef NOTDEF
static INLINE Physvolnum
NEXT_PHYSICAL_VOLNUM(Dx *db, PhysVolnumIter *pvi)
{

  if (++pvi->logvolnum > db->volume_weights.cw[pvi->physvolnum]) {
	if (++pvi->physvolnum == db->nvolumes) {
	  pvi->physvolnum = 0;
	  pvi->logvolnum = 0;
	}
  }

  return(pvi->physvolnum);
}
#endif

static INLINE Volume *
AREANUM_TO_VOLUME(Dx *db, Dx_ui32 areanum)
{

  return(db->volume[areanum % db->nvolumes]);
}

enum { META_HEADER_OFFSET = 0 };

static INLINE Dx_ui32
REL_AREANUM(Dx *db, Dx_ui32 area_num)
{
  Dx_ui32 val;

  val = area_num / db->nvolumes;
  return(val);
}

static INLINE Dx_len
LOGMAP_OFFSET(Dx *db, Dx_ui32 rel_areanum)
{
  Dx_len val;

  val = AREA_OFFSET(db, rel_areanum);
  return(val);
}

static INLINE Dx_len
PHYSMAP_OFFSET(Dx *db, Dx_ui32 rel_areanum)
{
  Dx_len val;

  val = AREA_OFFSET(db, rel_areanum) + LOGMAP_MAP_NBYTES;
  return(val);
}

static INLINE Dx_len
LOGMAP_PAGES_MAPPED(void)
{
  Dx_len val;

  val = LOGMAP_MAP_NBYTES / Physoff_s;
  return(val);
}

/*
 * Map a page number, B, to its area number and volume.
 */
static INLINE void
PAGENUM_TO_ANV(Dx *db, PageNum b, Volume **volumep, Dx_ui32 *areanump)
{
  Dx_ui32 an;

  an = b / LOGMAP_PAGES_MAPPED();

  if (areanump != NULL)
	*areanump = an;
  if (volumep != NULL)
	*volumep = AREANUM_TO_VOLUME(db, an);
}

static INLINE Dx_len
USERPAGE_OFFSET_START(Dx *db, Dx_ui32 rel_areanum)
{
  Dx_len val;

  val = AREA_OFFSET(db, rel_areanum) + LOGMAP_MAP_NBYTES + PHYSMAP_MAP_NBYTES;
  return(val);
}

static INLINE Volume *
PAGENUM_TO_VOLUME(Dx *db, PageNum b)
{
  Physvolnum v;

  v = SET_PHYSICAL_VOLNUM(db, (Logvolnum) b, NULL);
  return(db->volume[v]);
}

static INLINE Logvolnum
PAGENUM_TO_LOGICAL_VOLNUM(Dx *db, PageNum b)
{

  return(b % db->nlogical_volumes);
}

static INLINE Dx_ui32
AREA_PHYSOFFSET_TO_AREANUM(Dx *db, Volume *volume, Physoff offset)
{
  Dx_ui32 areanum, volrelnum;

  if (offset < META_HEADER_NBYTES)
	abort();

  volrelnum = (offset - META_HEADER_NBYTES) / AREA_SIZE(db);
  areanum = (volrelnum * db->nvolumes) + volume->volnum;

  return(areanum);
}

static INLINE Area *
AREA_PHYSOFFSET_TO_AREA(Dx *db, Physvolnum volnum, Physoff offset)
{
  Area *area;
  Dx_ui32 areanum;

  areanum = AREA_PHYSOFFSET_TO_AREANUM(db, db->volume[volnum], offset);
  for (area = db->volume[volnum]->area; area != NULL; area = area->next) {
	if (area->areanum == areanum)
	  return(area);
  }

  return(NULL);
}

static INLINE Area *
PAGENUM_TO_AREA(Dx *db, PageNum b)
{
  Dx_ui32 areanum;
  Area *area;
  Volume *volume;

  PAGENUM_TO_ANV(db, b, &volume, &areanum);

  for (area = volume->area; area != NULL; area = area->next) {
	if (area->areanum == areanum)
	  return(area);
  }

  return(NULL);
}

static INLINE Dx_ui32
Compute_expansion_buffers(Dx *db, Dx_ui32 exp)
{

  return(1 << exp);
}

static INLINE void
Stats_count(Dx *db, enum stats_mode type)
{
  Dx_stats_counters *s;

  if (DX_STATS) {
	s = &db->stats;
	switch (type) {
	case SPLIT:
	  s->nsplits++; break;
	case ADD:
	  s->nadds++; break;
	case MOVE:
	  s->nmoves++; break;
	case LOCATE:
	  s->nsearches++; break;
	case PROBE:
	  s->nprobes++; break;
	case REORG:
	  s->nreorgs++; break;
	}
  }
}

/*
 * If the load factor of a page meets or exceeds this threshold, it is
 * eligible to be split (in a round-robin scheduling).
 * To disable controlled splitting, set the factor to something greater than 1.
 * Increasing this value will tend to increase the average number of probes
 * required when searching for a key because chaining is more likely.  It
 * may also lead to a bigger db file.
 * Decreasing this value will tend to minimize the average number of probes
 * (1 probe for any key), but making it "too small" will tend to make the
 * db file bigger than necessary and increase overhead.
 */
#define EHASH_CONTROLLED_SPLIT_FACTOR  0.83

/*
 * If a key or value is bigger than this fraction of the page size
 * (not counting the meta data size and alignment requirements), it will
 * be stored indirectly.
 */
#define ITEM_INDIRECT_THRESHOLD_FACTOR  0.75

/* Byte-wise alignment, in bytes.  Use 1 if no alignment is necessary.  */
enum {
  ALIGNMENT16_CONSTRAINT = 2,
  ALIGNMENT32_CONSTRAINT = 4
};

typedef struct Item {
  Dx_datum *key;
  Dx_datum *value;
  Dx_hash keyhash;		/* The key's hash value. */
  Dx_hash hash;			/* The item's hash value in the keyspace. */
  Dx_keyspace *ks;		/* The keyspace of this key, or NULL. */
} Item;

typedef struct Stripe {
  Physvolnum volnum;
  Physoff offset;
  Seglen seglen;
  Seglen dlen;
  Segment segment;
} Stripe;

typedef struct Ehash_io {
  Dx_user_io *uio;
  Dx_len dlen;				/* Byte count up from lower-level functions. */
  Dx_ui8 *uio_buf;			/* Read-ahead buffer... */
  Dx_len bytes_buffered;	/* and its byte count. */
  Dx_desc desc;				/* Descriptive info for put/get-functions. */
} Ehash_io;

typedef struct Ehash_xarg {
  Dx_point *pos;		/* A position within a datum */
  Dx_region *region;	/* A region to edit or fetch within a datum */
  Ehash_io io;
  Dx_opmode flags;
} Ehash_xarg;

/* Page_get() modes. */
enum {
  PAGE_MAYCREATE = 0,
  PAGE_MUSTEXIST = 1
};

typedef struct PageLoad {
  Dx_ui32 total_meta;
  Dx_ui32 actual_meta;
  Dx_ui32 avail;
  Dx_ui32 used;
  Dx_ui32 indirect;
  Dx_ui32 nindirect;
  Dx_ui32 ndeleted;
} PageLoad;

#if BYTE_ORDER == BIG_ENDIAN
enum { DX_HOST_BIG_ENDIAN = 1 };
#else
enum { DX_HOST_BIG_ENDIAN = 0 };
#endif

/* Swap the two bytes of a 16 bit unsigned integer. */
static INLINE Dx_ui16
dx_swap16(Dx_ui16 *val)
{
  Dx_ui8 *ptr;

  ptr = (Dx_ui8 *) val;
  return(((Dx_ui16) *(ptr + 1)) | (((Dx_ui16) (*ptr)) << 8));
}

static INLINE Dx_ui16
Dx_export_ui16(Dx *db, Dx_ui16 *val)
{
  Dx_ui8 *ptr;
  
  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Big endian to little endian. */
	  return(dx_swap16(val));
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Little endian to big endian. */
	return(dx_swap16(val));
  }
}

static INLINE Dx_ui16
Dx_import_ui16(Dx *db, Dx_ui16 *val)
{
  Dx_ui8 *ptr;
  Dx_ui16 *hi, *lo;

  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Little endian to big endian. */
	  return(dx_swap16(val));
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Big endian to little endian. */
	return(dx_swap16(val));
  }
}

/* Convert VAL to the proper 32 bit external form. */
static INLINE Dx_ui32
Dx_export_ui32(Dx *db, Dx_ui32 *val)
{
  Dx_ui8 *ptr;
  Dx_ui16 *hi, *lo;
  
  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Big endian to little endian. */
	  hi = (Dx_ui16 *) ptr;
	  lo = (Dx_ui16 *) (ptr + 2);
	  return((Dx_ui32) dx_swap16(lo) | ((Dx_ui32) dx_swap16(hi)) << 16);
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Little endian to big endian. */
	return(htonl(*val));
  }
}

/* Convert VAL to the proper 32 bit internal form. */
static INLINE Dx_ui32
Dx_import_ui32(Dx *db, Dx_ui32 *val)
{
  Dx_ui8 *ptr;
  Dx_ui16 *hi, *lo;

  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Little endian to big endian. */
	  hi = (Dx_ui16 *) (ptr + 2);
	  lo = (Dx_ui16 *) ptr;
	  return((Dx_ui32) dx_swap16(hi) | ((Dx_ui32) dx_swap16(lo)) << 16);
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Big endian to little endian. */
	return(ntohl(*val));
  }
}

/* Convert VAL to its 64 bit external form. */
static INLINE Dx_ui64
Dx_export_ui64(Dx *db, Dx_ui64 *val)
{
  Dx_ui8 *ptr;
  Dx_ui32 *hi, *lo;

  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Big endian to little endian. */
	  hi = (Dx_ui32 *) ptr;
	  lo = (Dx_ui32 *) (ptr + 4);
	  return((Dx_ui64) Dx_export_ui32(db, lo)
			 | ((Dx_ui64) Dx_export_ui32(db, hi) << 32));
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Little endian to big endian. */
	hi = (Dx_ui32 *) (ptr + 4);
	lo = (Dx_ui32 *) ptr;
	return((Dx_ui64) htonl(*hi) | ((Dx_ui64) htonl(*lo) << 32));
  }
}

/* Convert VAL to its 64 bit internal form. */
static INLINE Dx_ui64
Dx_import_ui64(Dx *db, Dx_ui64 *val)
{
  Dx_ui8 *ptr;
  Dx_ui32 *hi, *lo;

  ptr = (Dx_ui8 *) val;
  if (DX_HOST_BIG_ENDIAN) {
	if (db->endian == DX_BIG_ENDIAN)
	  return(*val);
	else {
	  /* Little endian to big endian. */
	  lo = (Dx_ui32 *) ptr;
	  hi = (Dx_ui32 *) (ptr + 4);
	  return((Dx_ui64) Dx_import_ui32(db, hi)
			 | ((Dx_ui64) Dx_import_ui32(db, lo) << 32));
	}
  }
  else {
	if (db->endian == DX_LITTLE_ENDIAN)
	  return(*val);
	/* Big endian to little endian. */
	hi = (Dx_ui32 *) ptr;
	lo = (Dx_ui32 *) (ptr + 4);
	return((Dx_ui64) ntohl(*lo) | ((Dx_ui64) ntohl(*hi) << 32));
  }
}

#ifdef INLINE_MEMFUNCS
/*
 * It's important that these functions be as efficient as possible, so it
 * may be worthwhile to inline them.  At present, they are far from
 * optimal and on Linux at least, the library versions are faster so don't
 * use these.
 */

static INLINE void
Dx_memcpy(void *dest, const void *src, size_t n)
{
  Dx_ui8 *d, *s;

  d = (Dx_ui8 *) dest;
  s = (Dx_ui8 *) src;
  while (n-- > 0)
	*d++ = *s++;
}

static INLINE void
Dx_memset(void *dest, int c, size_t n)
{
  Dx_ui8 *d;

  d = (Dx_ui8 *) dest;
  while (n-- > 0)
	*d++ = (Dx_ui8) c;
}

/* Return 0 if the first N bytes of SRC1 and SRC2 are equal, 1 otherwise. */
static INLINE int
Dx_memneq(const void *src1, const void *src2, size_t n)
{
  Dx_ui8 *p, *q;

  p = (Dx_ui8 *) src1;
  q = (Dx_ui8 *) src2;
  while (n > 0 && *p++ == *q++)
	n--;
  if (n == 0)
	return(0);	/* Equal */
  return(1);
}
#else
#define Dx_memcpy(D, S, L)		memcpy(D, S, L)
#define Dx_memset(D, C, L)		memset(D, C, L)
#define Dx_memneq(P, Q, L)		memcmp(P, Q, L)
#endif

extern Dx_rc Area_create(Dx *db);
extern Dx_rc Area_create_volume(Dx *db, Physvolnum volnum,
								Dx_ui32 areanum, Area **);
extern Dx_rc Area_init(Dx *db);
extern Area *Area_Physoffo_area(Dx *db, Physvolnum volnum,
									 Physoff offset);
extern Dx_rc Area_write(Dx *db);
void Area_show(Dx *, Dx_ui32);

extern Dx_rc Physmap_allocate_page(Dx *, Volume *volume, Seglen seg_len,
								   Physoff *offset);
extern Dx_rc Physmap_allocate_segment(Dx *, Volume *volume, Seglen seg_len,
									  Dx_ui32 npages, Physoff *offset);
extern Dx_rc Physmap_create(Dx *db, Volume *volume, Dx_ui32 areanum, Physmap *);
extern Dx_rc Physmap_block_allocate(Dx *db, Volume *volume,
									Segment *new_seg, Physoff *pa);
extern Dx_rc Physmap_extend_segment(Dx *db, Volume *volume, Seglen seg_len,
									Physoff offset, Dx_ui32 old_npages,
									Dx_ui32 new_npages);
extern Dx_rc Physmap_free_segment(Dx *db, Physvolnum volnum,
								  Physoff offset, Dx_ui32 npages);
extern Dx_rc Physmap_truncate_segment(Dx *db, Physvolnum volnum,
									  Physoff offset,
									  Dx_ui32 old_npages, Dx_ui32 new_npages);
extern Dx_rc Physmap_init(Dx *, Dx_ui32 areanum, Physmap *);

extern Dx_rc Logmap_create(Dx *db, Volume *volume, int areanum, Segment *);
extern Dx_rc Logmap_block_allocate(Dx *, Volume *, Segment *, Physoff *);
extern Dx_rc Logmap_init(Dx *, Area *area);
extern Dx_rc Logmap_page_add(Dx *db, PageNum b, Physoff pa);
extern Dx_rc Logmap_page_remove(Dx *db, PageNum b);
extern void Logmap_block_init(Segment seg);
extern Dx_rc Logmap_page_lookup(Dx *db, PageNum b, Dx_ui32 mode, Physoff *pa);

extern Dx_rc Item_add(Dx *, Volume *, Page page, Dx_opmode mode,
					  Item *item, Ehash_xarg *arg);
extern Dx_ui32 Item_count(Dx *db, Page, Dx_ui32 page_size);
extern Dx_rc Item_edit(Dx *db, Volume *v, Page page, Item *item,
					   Ehash_xarg *arg);
extern Dx_rc Item_iterate(Dx *db, void *page, int first, Dx_ui32 *ptr, Dx_hash *hashp);
extern void Item_load(Dx *db, Page, Dx_ui32 page_size, PageLoad *, int full);
extern Dx_rc Item_move(Dx *, void *page, Dx_ui32 ind, void *new_page);
extern Dx_rc Item_delete(Dx *, void *page, Item *item);
extern Dx_rc Item_fetchx(Dx *db, Page page, Item *item, Ehash_xarg *arg);
extern Dx_rc Item_exists(void *page, Dx_ui32 page_size, Item *item);
extern void Item_init(Dx *, Page);
extern Dx_rc Item_index(Dx *, void *page, Dx_ui32 ind,
						Dx_opmode mode, Item *item, Ehash_xarg *arg);
extern double Item_page_usage(Dx *db, Page page, Dx_ui32 page_size);
extern Dx_rc Item_probe(Dx *, void *page, Dx_opmode mode,
						Item *item, Ehash_xarg *arg);
extern Dx_rc Item_region(Dx_len dlen, Dx_region *r, Dx_len *pos);
extern int Item_reorg(Dx *, Page page);
extern Dx_rc Item_replace(Dx *, Volume *v, Page page, Item *item,
						  Ehash_xarg *arg);
extern int Item_show(Dx *db, void *page, Dx_ui32 page_size);

extern Dx_rc Segment_allocate_page(Dx *, Volume *, PageNum b,
								   Segment *new_seg, Physoff *pa);
extern Dx_rc Segment_allocate_segment(Dx *, Volume *, Seglen nbytes,
									  Dx_ui32 flags,
									  Segment *new_seg, Physoff *pa);
extern Dx_rc Segment_unmap(Dx *, Volume *, Physoff first_block,
						   Segment seg, Seglen bytes, int do_remove);
extern Dx_rc Segment_unmap_zeroes(Dx *db, Seglen nbytes);
extern Dx_rc Segment_init(Dx *db);
extern Dx_rc Segment_map(Dx *db, Volume *, Physoff pb, Seglen nbytes,
						 PageNum b, int is_page, Dx_ui32 flags,
						 Segment *new_seg);
extern Dx_rc Segment_map_zeroes(Dx *db, Seglen nbytes, int prot,
								Segment *new_seg);
extern Dx_rc Segment_reallocate(Dx *, Volume *, Physoff offset,
								Seglen old_length, Seglen new_length,
								Segment *new_seg, Physoff *pb);
extern Dx_rc Segment_truncate(Dx *db, Volume *volume, Physoff first_block,
							  Dx_ui32 old_npages, Dx_ui32 new_npages);
extern void Segment_mapshow(Dx *db);
extern Dx_rc Segment_uncache(Dx *db, int segments_only);
extern SegmentMap *Segment_lookup(Dx *db, Volume *volume, Physoff offset);
extern Dx_rc Segment_free(Dx *db, Volume *volume, Physoff offset,
						  Dx_ui32 npages);
extern Dx_rc Segment_free_using_map(Dx *db, SegmentMap *sm);
extern Dx_rc Segment_release(Dx *db);
extern Dx_rc Segment_set_flags(Dx *db, Volume *v, Physoff offset,
							   Dx_ui32 new_flags);
extern Dx_rc Segment_clear_flags(Dx *db, Volume *volume, Physoff offset,
								 Dx_ui32 new_flags);
extern Dx_rc Segment_get_flags(Dx *db, Volume *volume, Physoff offset,
							   Dx_ui32 *flags);
extern Dx_rc Segment_clear_fixed_active(Dx *db, Volume *volume, Physoff offset);
extern Dx_rc Segment_clear_fixed_page_active(Dx *db, PageNum b);
extern Dx_rc Segment_duplicate(Dx *new_db, Dx *old_db);

extern Page Page_alloc(Dx *db, PageNum b);
extern Dx_rc Page_duplicate(Dx *new_db, Dx *old_db);
extern Dx_rc Page_exists(Dx *db, PageNum b);
extern Dx_rc Page_get(Dx *db, Volume *, PageNum b, Dx_ui32 mode, Page *page);
extern void Page_init(Dx *);
extern Dx_rc Page_release(Dx *);
extern Dx_rc Page_reorg(Dx *db);
extern Dx_rc Page_sync(Dx *db);

extern Dx_rc Ehash_add(Dx *db, Item *item, Dx_ui32 mode, Ehash_xarg *arg);
extern Dx_rc Ehash_complete_expansion(Dx *db);
extern Dx_rc Ehash_create(Dx *db);
extern Dx_rc Ehash_get(Dx *db, Item *item, Dx_opmode mode, Ehash_xarg *arg);
extern Dx_rc Ehash_iter(Dx *db, Item *item, Dx_opmode mode, Ehash_xarg *arg);
extern void Ehash_iter_init(Dx *);
extern int Ehash_bucket_exists(Dx *db, PageNum b);

extern Dx_rc Meta_create(Dx *db);
extern Dx_rc Meta_init(Dx *db, off_t);
extern Dx_rc Meta_read(Dx *db);
extern Dx_rc Meta_write(Dx *db);
extern Dx_rc Meta_duplicate(Dx *new_db, Dx *old_db);
extern void meta_show(Dx *db);

extern void Fatal(Dx_rc rc);

extern Hash64 hash_bj64(Dx_ui8 *k, Dx_ui32 length, Dx_ui64 init);
extern Dx_hash hash_bj(Dx_ui8 *str, Dx_len len, Dx_hash init);
extern Dx_hash hash_sdbm(Dx_ui8 *str, Dx_len len);
extern Dx_hash hash_sdbm32(Dx_ui8 *str, Dx_len len, Dx_ui32 init);

extern void page_check(Dx *db, PageNum);
extern void db_check(Dx *db);
extern void db_dump(Dx *db);
extern void dump(Dx *db);
extern void Item_info(Dx *db, Page page, Dx_ui32 page_size);
extern void dx_show(Dx *db);

extern void Stats_count(Dx *, enum stats_mode type);
extern void Stats_show(Dx *, FILE *);
extern void Stats_init(Dx *);

extern int Xrand(int low, int high);
extern void Xsrand(unsigned int seed);
extern Volume *Random_volume(Dx *db);

extern void *Xmalloc(size_t);
extern void *Xrealloc(void *, size_t);
extern void Xfree(void *);
extern char *Xstrdup(const char *s);

#endif
