/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Map a logical page to a physical address within the db.
 * Each logical block is a vector that maps a set of logical blocks to their
 * corresponding physical addresses.
 */

#include "local.h"

#if defined(Physoff32)
#define logmap_get_physoff(DB, V)		(Physoff) Dx_import_ui32(DB, V)
#define logmap_put_physoff(DB, V)		(Physoff) Dx_export_ui32(DB, V)
#elif defined(Physoff64)
#define logmap_get_physoff(DB, V)		(Physoff) Dx_import_ui64(DB, V)
#define logmap_put_physoff(DB, V)		(Physoff) Dx_export_ui64(DB, V)
#else
#error "Unsupported length for Physoff"
#endif

Dx_rc
Logmap_create(Dx *db, Volume *volume, int rel_areanum, Segment *logmap)
{
  Physoff offset;
  Dx_rc rc;

  offset = LOGMAP_OFFSET(db, rel_areanum);
  rc = Segment_map(db, volume, offset, LOGMAP_MAP_NBYTES, 0, 0,
				   SEGMENT_META | SEGMENT_FIXED, logmap);

  if (rc == DX_OK)
	Dx_memset((void *) *logmap, 0, (size_t) LOGMAP_MAP_NBYTES);

  return(rc);
}

Dx_rc
Logmap_init(Dx *db, Area *area)
{
  Dx_rc rc;

  rc = Segment_map(db, AREANUM_TO_VOLUME(db, area->areanum),
				   LOGMAP_OFFSET(db, REL_AREANUM(db, area->areanum)),
				   LOGMAP_MAP_NBYTES, 0, 0,
				   SEGMENT_META | SEGMENT_FIXED, &area->logmap);

  return(rc);
}

/*
 * Create the mapping page for the given page number.
 */
static INLINE Dx_rc
create_pagenum_logmap(Dx *db, PageNum b, Segment *seg)
{
  Dx_ui32 areanum;
  Area *area;
  Volume *volume;
  Dx_rc rc;

  /* The area doesn't exist yet, so create it. */
  PAGENUM_TO_ANV(db, b, &volume, &areanum);
  if ((rc = Area_create_volume(db, volume->volnum, areanum, &area)) != DX_OK)
	return(rc);

  *seg = area->logmap;
  return(DX_OK);
}

/*
 * Locate the mapping page for the given page number.
 *
 * XXX Should support mapping on demand instead of assuming
 * areas are all mapped at startup time.
 */
static INLINE Dx_rc
pagenum_to_logmap(Dx *db, PageNum b, Dx_ui32 mode, Segment *seg)
{
  Area *area;
  Dx_rc rc;

  if ((area = PAGENUM_TO_AREA(db, b)) != NULL) {
	if (area->logmap == NULL) {
	  if ((rc = Logmap_init(db, area)) != DX_OK)
		return(rc);
	}
	*seg = area->logmap;
	return(DX_OK);
  }

  if (mode != PAGE_MAYCREATE)
	return(PAGE_NOTFOUND);

  return(create_pagenum_logmap(db, b, seg));
}

/*
 * Add a mapping from the given page to the given db address.
 */
Dx_rc
Logmap_page_add(Dx *db, PageNum b, Physoff pa)
{
  Dx_ui32 areanum, map_offset;
  Area *area;
  Physoff *map;
  Segment logmap;
  Volume *volume;
  Dx_rc rc;

  if ((rc = pagenum_to_logmap(db, b, PAGE_MAYCREATE, &logmap)) != DX_OK)
	return(rc);

  if (logmap == NULL) {
	if ((area = PAGENUM_TO_AREA(db, b)) == NULL) {
	  PAGENUM_TO_ANV(db, b, &volume, &areanum);
	  rc = Area_create_volume(db, volume->volnum, areanum, &area);
	  if (rc != DX_OK)
		return(rc);
	}
	if ((rc = Logmap_init(db, area)) != DX_OK)
	  return(rc);
	logmap = area->logmap;
  }

  map = (Physoff *) logmap;
  map_offset = b % LOGMAP_PAGES_MAPPED();

  if (map[map_offset] != 0) {
	fprintf(stderr, "Logmap_page_add: remap!\n");
	return(DX_INTERNAL_ERROR);
  }

  map[map_offset] = logmap_put_physoff(db, &pa);

  return(DX_OK);
}

Dx_rc
Logmap_page_lookup(Dx *db, PageNum b, Dx_ui32 mode, Physoff *pa)
{
  Dx_ui32 map_offset;
  Segment logmap;
  Physoff *map;
  Dx_rc rc;

  rc = pagenum_to_logmap(db, b, mode, &logmap);
  if (rc != DX_OK)
	return(rc);

  map = (Physoff *) logmap;

  map_offset = b % LOGMAP_PAGES_MAPPED();

  if (pa != NULL) {
	*pa = logmap_get_physoff(db, &map[map_offset]);
  }

  return(map[map_offset] == 0 ? PAGE_NOTFOUND : DX_OK);
}

Volume *
Random_volume(Dx *db)
{
  int r;
  Logvolnum lv;
  Physvolnum v;

  r = Xrand(0, db->nlogical_volumes - 1);
  lv = (Logvolnum) r;

  v = SET_PHYSICAL_VOLNUM(db, lv, NULL);
printf("%u\n", (unsigned int) v);
  return(db->volume[v]);
}

