/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Interface for user-level raw segment management and frames-based
 * memory management.
 */

/*
 * Frames     Memory Chunk Allocation List
 *  ----         -----           -----
 * | F0 | ----> | MC0 |  ---->  | MC1 | ---->
 *  ----         -----           ----- 
 * | F1 |        /  /            \  \   \
 *  ----        /  /              \  \   \
 * | F2 |      /  /                \  \   \
 *  ----     MCA MCA               MCA MCA MCA    
 *           /   /                 /   /   /
 *         mem mem               mem mem mem
 */ 

/*
 * XXX should be able to ask for a particular alignment
 */

#include "local.h"

#define MEM_ALIGNMENT_TYPE	int

enum {
  MEM_MINIMUM_CHUNK_SIZE = 4096,
  MEM_ROUND_CHUNK_SIZE   = 4096
};

typedef struct MemSeg MemSeg;
typedef struct MemChunk MemChunk;
typedef struct MemChunkAlloc MemChunkAlloc;
typedef struct Frame Frame;
typedef struct PSeg PSeg;

/*
 * If db is NULL, ignore the other elements and use malloc.
 * Otherwise, this describes the source in db for this allocation.
 */
struct  MemSeg {
  Dx *db;
  Physvolnum volnum;
  Physoff offset;
  Segment seg;
  Dx_ui32 npages;
};

struct PSeg {
  Physvolnum volnum;
  Physoff offset;
  Dx_ui32 npages;
  Dx_ui32 size;
  Dx_ui32 avail;
};

/*
 * A prototype structure for what ultimately gets passed back after a
 * successful allocation.
 * This is currently only necessary to support Dx_realloc() - a convenient
 * way of knowing the size of an allocation and where the allocation came
 * from is needed.
 * The mem_mem_to_mca() function returns a pointer to this structure given
 * a pointer passed back to a user.
 *
 * XXX may want to put a magic number (a checksum?) in this header to
 * help detect if it's been clobbered.
 */
struct MemChunkAlloc {
  size_t size;					/* This must be the first element */
  MemChunk *mc;
  MEM_ALIGNMENT_TYPE bogus;		/* The user is given a pointer to this */
};

/*
 * A MemChunk corresponds to one big memory allocation.  One or more
 * allocations are made from it.
 */
struct MemChunk {
  MemSeg ms;			/* Source for this allocation */
  void *mem;			/* A big block from which allocations are made */
  Dx_ui32 size;			/* The size, in bytes, of this big block */
  Dx_ui32 avail;			/* The amount available, in bytes */
  MemChunk *next;		/* Next allocated block */
};

/*
 * Each MemFrame represents one "frame" of memory, each having zero or more
 * allocations.  The Frames are linked together into a "stack" (an ordered
 * list).  The current or top Frame is pointed to by top_frame.
 * All allocations for a frame come from the same source (malloc, volatile db,
 * persistent db).
 */
struct Frame {
  char *name;			/* Currently, only used for debugging */
  int popped;			/* Non-zero if the frame isn't on the stack */
  MemSeg ms;			/* The default is for allocations to come from here */
  MemChunk *chunks;		/* The list of allocations for this frame */
  Frame *next;
};

Dx_rc Dx_last_rc;

#ifdef NOTDEF
/*
 * XXX Instead of freeing chunks and MemChunk and Frame structures, cache
 * some of them so they can be reused more efficiently.
 */
static MemChunk *available_chunks = NULL;
static Frame *available_frames = NULL;
#endif

/* A pointer to the top of the frame stack. */
static Frame *top_frame = NULL;

static Frame *popped_frames = NULL;

/*
 * Compute the number of bytes needed for an object of SIZE bytes and space
 * following it sufficient to result in an ending address that will be
 * aligned to an address evenly divisible by ALIGNMENT.
 * ALIGNMENT is a power of 2.
 */
static INLINE size_t
mem_align(size_t size, size_t alignment)
{
  size_t aligned_size;

  aligned_size = (size + alignment - 1) & (~ (alignment - 1));
  return(aligned_size);
}

/*
 * If the user requests SIZE bytes, compute the total amount needed, including
 * bookkeeping overhead and alignment constraints.
 * XXX May be based on constraints and actual chunk.mem value one day so
 * it is currently called from places where that pointer is accessible.
 */
static INLINE size_t
mem_need(size_t size, size_t sizeof_mem_alignment_type)
{
  size_t need;

  need = mem_align(sizeof(MemChunkAlloc) - sizeof_mem_alignment_type + size,
				   sizeof_mem_alignment_type);
  return(need);
}

/*
 * Given a pointer to allocated memory that was previously returned to the
 * user, return a pointer to that memory's header.
 * XXX Might guard against a clobbered header.
 */
static INLINE MemChunkAlloc *
mem_mem_to_mca(void *ptr)
{
  size_t diff;
  MemChunkAlloc *mca;

  diff = (size_t) ((Dx_ui8 *) &mca->bogus - (Dx_ui8 *) &mca->size);
  mca = (MemChunkAlloc *) ((Dx_ui8 *) ptr - diff);

  return(mca);
}

/*
 * A new MemChunk is about to be allocated in response to a user request
 * for SIZE bytes.  Determine how big a MemChunk to actually allocate.
 */
static INLINE size_t
mem_compute_chunk_size(Frame *f, size_t size)
{
  size_t actual_size;

  if (f == NULL || f->ms.db == NULL) {
	if (size < MEM_MINIMUM_CHUNK_SIZE)
	  return(MEM_MINIMUM_CHUNK_SIZE);

	actual_size = (size + MEM_ROUND_CHUNK_SIZE - 1)
	  & (~ (MEM_ROUND_CHUNK_SIZE - 1));
  }
  else {
	actual_size = NBYTES_TO_PAGE_BYTES(f->ms.db, size);
  }
  return(actual_size);
}

/*
 * Create a new MemChunk on FRAME of NEED bytes.
 * The start of this new chunk must be maximally aligned.
 */
static INLINE MemChunk *
mem_new_chunk(Frame *f, Dx *dbsrc, size_t need)
{
  void *ptr;
  MemChunk *mc;
  Dx_rc rc;

  mc = (MemChunk *) malloc(sizeof(MemChunk));
  if (mc == NULL)
	return(NULL);

  if (dbsrc == NULL) {
	ptr = (void *) malloc(need);
	if (ptr == NULL) {
	  free(mc);
	  return(NULL);
	}
	mc->ms.db = NULL;
  }
  else {
	mc->ms.db = dbsrc;
	mc->ms.volnum = 0;
	rc = Segment_allocate_segment(dbsrc, dbsrc->volume[mc->ms.volnum], need,
								  0, &mc->ms.seg, &mc->ms.offset);
	if (rc != DX_OK) {
	  Dx_last_rc = rc;
	  free(mc);
	  return(NULL);
	}
	mc->ms.npages = (Dx_ui32) NBYTES_TO_NPAGES(dbsrc, need);
	ptr = (void *) mc->ms.seg;
  }

  mc->mem = ptr;
  mc->size = need;
  mc->avail = need;
  mc->next = f->chunks;

  f->chunks = mc;

  return(mc);
}

/*
 * Search for an existing chunk with sufficient space.
 * First fit.
 */
static INLINE MemChunk *
mem_find_chunk(Frame *f, Dx *dbsrc, size_t size, size_t alignment)
{
  MemChunk *mc;
  size_t need;

  need = mem_need(size, alignment);

  for (mc = f->chunks; mc != NULL; mc = mc->next) {
	if (mc->ms.db == dbsrc && mc->avail >= need)
	  return(mc);
  }
  return(NULL);
}

static INLINE void *
mem_pointer(MemChunkAlloc *mca)
{

  return((void *) &mca->bogus);
}

/*
 * Return a new allocation of SIZE bytes from MC, which the caller guarantees
 * to be big enough.
 */
static INLINE void *
mem_alloc_from_chunk(Frame *f, MemChunk *mc, size_t size, size_t alignment)
{
  MemChunkAlloc *mca;
  size_t need;
  void *mem, *ptr;

  need = mem_need(size, alignment);

  ptr = (void *) ((Dx_ui8 *) mc->mem + (mc->size - mc->avail));
  mc->avail -= need;

  mca = (MemChunkAlloc *) ptr;
  mca->size = need;
  mca->mc = mc;
  mem = mem_pointer(mca);
  return(mem);
}

/*
 * Make an allocation of SIZE bytes on the given frame.
 */
static INLINE void *
mem_alloc(Frame *f, Dx *dbsrc, size_t size)
{
  void *ptr;
  size_t need;
  MemChunk *mc;

#ifdef NOTDEF
  if (dbsrc != NULL && (dbsrc->flags & DX_CONFIG_PERSISTENT)) {
	Dx_last_rc = DX_NOTIMPLEMENTED;
	return(NULL);
  }
#endif

  /* Look for something already available. */
  mc = mem_find_chunk(f, dbsrc, size, sizeof(MEM_ALIGNMENT_TYPE));
  if (mc == NULL) {
	/* Round up the size to satisfy alignment constraints. */
	need = mem_need(size, sizeof(MEM_ALIGNMENT_TYPE));
	mc = mem_new_chunk(f, dbsrc, mem_compute_chunk_size(f, need));
	if (mc == NULL)
	  return(NULL);
  }
  ptr = mem_alloc_from_chunk(f, mc, size, sizeof(MEM_ALIGNMENT_TYPE));

  return(ptr);
}

static INLINE void *
mem_calloc(Frame *f, Dx *dbsrc, size_t nmemb, size_t size)
{
  void *ptr;
  size_t need;

#ifdef NOTDEF
  if (dbsrc != NULL && (dbsrc->flags & DX_CONFIG_PERSISTENT) == 0) {
	Dx_last_rc = DX_NOTIMPLEMENTED;
	return(NULL);
  }
#endif

  need = nmemb * size;		/* XXX overflow */
  ptr = mem_alloc(f, dbsrc, need);
  if (ptr != NULL)
	Dx_memset(ptr, 0, need);
  return(ptr);
}

static INLINE int
frame_push(Frame *f)
{
  Frame *ff;

  f->popped = 0;

  /* Remove the frame from the popped list. */
  ff = popped_frames;
  if (ff != NULL) {
	if (ff == f)
	  popped_frames = popped_frames->next;
	else {
	  while (ff->next != f && ff->next != NULL) {
		ff = ff->next;
	  }
	  if (ff != NULL)
		ff->next = (ff->next)->next;
	}
  }

  if (ff == NULL) {
	Dx_last_rc = DX_INTERNAL_ERROR;
	return(-1);
  }

  f->next = top_frame;
  top_frame = f;
  return(0);
}

/*
 * Create a new, empty frame.
 * The source of allocations for the frame is DB; if DB is NULL, the
 * malloc(3) functions are used.
 */
static INLINE Frame *
frame_new_frame(Dx *db, char *name)
{
  Frame *f;

  f = (Frame *) malloc(sizeof(Frame));
  if (f == NULL) {
	Dx_last_rc = DX_MALLOC_FAILED;
	return(NULL);
  }

  if (name == NULL || *name == '\0')
	f->name = NULL;
  else
	f->name = Xstrdup(name);

  f->ms.db = db;
  if (db != NULL) {
	f->ms.volnum = 0;
  }
  f->popped = 1;
  f->chunks = NULL;
  f->next = popped_frames;
  popped_frames = f;

  return(f);
}

/*
 * Make an allocation of SIZE bytes on FRAME.
 * If FRAME is NULL, use the top frame, creating it if necessary.
 * The source for the allocation is obtained from FRAME.
 */
static INLINE void *
frame_alloc_on_frame(Frame *f, Dx *dbsrc, size_t size)
{
  Frame *ff;
  void *p;

  if (f == NULL) {
	/* Use the top frame. */
	if (top_frame == NULL) {
	  ff = frame_new_frame(NULL, NULL);
	  if (ff == NULL)
		return(NULL);
	  if (frame_push(ff) == -1)
		return(NULL);		/* XXX leak */
	}
	else
	  ff = top_frame;
  }
  else
	ff = f;

  p = mem_alloc(ff, dbsrc, size);

  return(p);
}

static INLINE void *
frame_calloc_on_frame(void *frame, Dx *dbsrc, size_t nmemb, size_t size)
{
  void *ptr;
  size_t need;
  Frame *f;

  f = (Frame *) frame;
  need = nmemb * size;		/* XXX overflow */
  ptr = frame_alloc_on_frame(f, dbsrc, need);
  if (ptr != NULL)
	Dx_memset(ptr, 0, need);
  return(ptr);
}

/*
 * Reallocation is currently just making a new allocation and copying the
 * old allocation.
 * XXX Trying to grow the old allocation is a possibility.
 */
static INLINE void *
mem_realloc(Frame *f, Dx *dbsrc, void *ptr, size_t new_size)
{
  void *new_ptr;
  MemChunkAlloc *mca;

  mca = mem_mem_to_mca(ptr);

  if (mca->size >= new_size)
	return(ptr);

  if ((new_ptr = frame_alloc_on_frame(f, dbsrc, new_size)) == NULL)
	return(NULL);

  /* Safe and faster not to use memmove(). */
  memcpy(new_ptr, ptr, mca->size);

  return(new_ptr);
}

/*
 * Free all allocations associated with FRAME.
 */
static INLINE int
frame_free_frame(Frame *f)
{
  MemChunk *mc, *mc_next;
  Dx_rc rc;

  if (f == NULL)
	return(0);

  for (mc = f->chunks; mc != NULL; mc = mc_next) {
	mc_next = mc->next;
	if (mc->ms.db == NULL) {
	  free(mc->mem);
	  mc->mem = NULL;
	}
	else {
	  Dx *db;

	  db = mc->ms.db;
	  rc = Segment_free(db, db->volume[mc->ms.volnum],
						mc->ms.offset, mc->ms.npages);
	  if (rc != DX_OK) {
		Dx_last_rc = rc;
		return(-1);
	  }
	}
	free(mc);
  }
  if (f->name != NULL) {
	free(f->name);
	f->name = NULL;
  }
  free(f);
  return(0);
}

static INLINE int
mem_swizzle_mc(MemChunk *mc, Dx_ui32 mc_offset, Dx_ui8 *ptr, Dx_ui32 *swizzled)
{
  Dx_ui32 ptr_offset;

  if ((char *) mc->mem <= (char *) ptr
	  && (char *) ptr < ((char *) mc->mem + mc->size)) {
	ptr_offset = mc_offset + (Dx_ui32) ((char *) ptr - (char *) mc->mem);
	*swizzled = ptr_offset;
	return(1);
  }
  return(-1);
}
  
static INLINE int
mem_swizzle(Dx *db, Frame *f, void *frame_mem_ptr, Dx_ui32 *swizzled)
{
  int i;
  Dx_ui32 mc_offset;
  Dx_ui8 *ptr;
  MemChunk *mc;

#define SWIZZLED_NULL_POINTER	0xffffffff

  if (frame_mem_ptr == NULL) {
	*swizzled = SWIZZLED_NULL_POINTER;
	return(1);
  }

  ptr = (Dx_ui8 *) frame_mem_ptr;
  i = 0;
  mc_offset = 0;
  for (mc = f->chunks; mc != NULL; mc = mc->next) {
	if (mc->ms.db != db)
	  continue;
	if (mem_swizzle_mc(mc, mc_offset, ptr, swizzled) != -1)
	  return(i);
	mc_offset += mc->size;
	i++;
  }
  return(-1);
}

Dx_rc
Dx_swizzle(Dx *db, void **frame_mem_ptr)
{
  Dx_ui32 swizzled;

  if (top_frame == NULL)
	return(DX_FRAME_NO_FRAME);

  if (mem_swizzle(db, top_frame, *frame_mem_ptr, &swizzled) == -1)
	return(DX_FRAME_BAD_POINTER);
  *frame_mem_ptr = (void *) swizzled;
  return(DX_OK);
}

Dx_rc
Dx_frame_swizzle(Dx *db, void *frame, void **frame_mem_ptr)
{
  Dx_ui32 swizzled;
  Frame *f;

  if (frame == NULL) {
	if (top_frame == NULL)
	  return(DX_FRAME_NO_FRAME);
	f = top_frame;
  }
  else
	f = (Frame *) frame;

  if (mem_swizzle(db, f, *frame_mem_ptr, &swizzled) == -1)
	return(DX_FRAME_BAD_POINTER);
  *frame_mem_ptr = (void *) swizzled;
  return(DX_OK);
}

int
mem_unswizzle(Dx *db, Frame *f, Dx_ui32 swizzled_ptr, void **unswizzled)
{
  Dx_ui32 mc_offset;
  MemChunk *mc;

  if (swizzled_ptr == SWIZZLED_NULL_POINTER) {
	*unswizzled = NULL;
	return(1);
  }

  mc_offset = 0;
  for (mc = f->chunks; mc != NULL; mc = mc->next) {
	if (mc->ms.db == db) {
	  if (swizzled_ptr >= mc_offset
		  && swizzled_ptr < (mc_offset + mc->size)) {
		*unswizzled
		  = (void *) ((Dx_ui8 *) mc->mem + (swizzled_ptr - mc_offset));
		return(1);
	  }
	  mc_offset += mc->size;
	}
  }

  return(-1);
}

Dx_rc
Dx_unswizzle(Dx *db, void **swizzledp)
{
  Dx_ui32 swizzled;

  if (top_frame == NULL)
	return(DX_FRAME_NO_FRAME);

  swizzled = *(Dx_ui32 *) swizzledp;
  if (mem_unswizzle(db, top_frame, swizzled, swizzledp) == -1)
	return(DX_FRAME_BAD_POINTER);
  return(DX_OK);
}

Dx_rc
Dx_frame_unswizzle(Dx *db, void *frame, void **swizzledp)
{
  Dx_ui32 swizzled;
  Frame *f;

  if (frame == NULL) {
	if (top_frame == NULL)
	  return(DX_FRAME_NO_FRAME);
	f = top_frame;
  }
  else
	f = (Frame *) frame;

  swizzled = *(Dx_ui32 *) swizzledp;
  if (mem_unswizzle(db, top_frame, swizzled, swizzledp) == -1)
	return(DX_FRAME_BAD_POINTER);
  return(DX_OK);
}

/*
 * XXX Need to externalize binary meta-data
 */
static INLINE int
frame_export(Dx *db, Frame *f, void *ptr, void **buf, size_t *buflen)
{
  int i, n;
  Dx_ui32 ptr_offset;
  MemChunk *mc;
  PSeg *pseg;

  n = 0;
  for (mc = f->chunks; mc != NULL; mc = mc->next) {
	if (mc->ms.db == db)
	  n++;
  }
	
  if (n == 0) {
	*buflen = 0;
	*buf = NULL;
	return(0);
  }

  pseg = (PSeg *) malloc((n + 1) * sizeof(PSeg));
  if (pseg == NULL) {
	Dx_last_rc = DX_MALLOC_FAILED;
	return(-1);
  }

  i = 0;
  for (mc = f->chunks; mc != NULL; mc = mc->next) {
	if (mc->ms.db == db) {
	  pseg[i].volnum = mc->ms.volnum;
	  pseg[i].offset = mc->ms.offset;
	  pseg[i].npages = mc->ms.npages;
	  pseg[i].size = mc->size;
	  pseg[i].avail = mc->avail;
	  i++;
	}
  }

  if (ptr != NULL && mem_swizzle(db, f, ptr, &ptr_offset) == -1) {
	fprintf(stderr, "Didn't find pointer!\n");
	pseg[i].offset = SWIZZLED_NULL_POINTER;
	pseg[i].volnum = 0;
  }
  else {
	pseg[i].offset = ptr_offset;
	pseg[i].volnum = 1;
  }

  *buf = (void *) pseg;
  *buflen = (n + 1) * sizeof(PSeg);

  return(0);
}

int
Dx_frame_export(Dx *db, void *ptr, void **buf, size_t *buflen)
{
  int rc;

  if (db == NULL && (db->flags & DX_CONFIG_PERSISTENT) == 0) {
	Dx_last_rc = DX_BADARG;
	return(-1);
  }

  Dx_last_rc = DX_OK;
  rc = frame_export(db, top_frame, ptr, buf, buflen);

  return(rc);
}

int
Dx_frame_export_frame(Dx *db, void *frame, void *ptr,
					  void **buf, size_t *buflen)
{
  int rc;
  Frame *f;

  if (db == NULL && (db->flags & DX_CONFIG_PERSISTENT) == 0) {
	Dx_last_rc = DX_BADARG;
	return(-1);
  }

  Dx_last_rc = DX_OK;
  f = (Frame *) frame;
  rc = frame_export(db, f, ptr, buf, buflen);

  return(rc);
}

/*
 * XXX Need to internalize binary info
 */
static INLINE Frame *
frame_import(Dx *db, void *buf, size_t buflen, void **ptr)
{
  int i, is_ptr, n;
  Dx_ui32 mc_offset, ptr_offset;
  Frame *f;
  MemChunk *mc;
  PSeg *pseg;
  Dx_rc rc;

  f = frame_new_frame(db, "");
  if (f == NULL)
	return(NULL);

  n = (buflen / sizeof(PSeg)) - 1;
  pseg = (PSeg *) buf;

  is_ptr = pseg[n].volnum;
  ptr_offset = pseg[n].offset;

  mc_offset = 0;
  for (i = 0; i < n; i++) {
	mc = (MemChunk *) malloc(sizeof(MemChunk));
	if (mc == NULL) {
	  /* XXX free */
	  Dx_last_rc = DX_MALLOC_FAILED;
	  return(NULL);
	}
	mc->ms.db = db;
	mc->ms.volnum = pseg[i].volnum;
	mc->ms.offset = pseg[i].offset;
	mc->ms.npages = pseg[i].npages;
	rc = Segment_map(db, db->volume[pseg[i].volnum], pseg[i].offset,
					 db->page_size * pseg[i].npages, 0, 0, 0, &mc->ms.seg);
	if (rc != DX_OK) {
	  Dx_last_rc = rc;
	  /* XXX free */
	  return(NULL);
	}
	mc->mem = (void *) mc->ms.seg;
	mc->size = pseg[i].size;
	mc->avail = pseg[i].avail;
	mc->next = f->chunks;
	f->chunks = mc;

	if (is_ptr && ptr != NULL && ptr_offset >= mc_offset
		&& ptr_offset < (mc_offset + mc->size))
	  *ptr = (void *) ((Dx_ui8 *) mc->mem + (ptr_offset - mc_offset));

	mc_offset += mc->size;
  }

  return(f);
}

int
Dx_frame_import(Dx *db, void *buf, size_t buflen, void **ptr)
{
  int rc;
  Frame *f;

  if (db == NULL && (db->flags & DX_CONFIG_PERSISTENT) == 0) {
	Dx_last_rc = DX_BADARG;
	return(-1);
  }

  Dx_last_rc = DX_OK;
  f = frame_import(db, buf, buflen, ptr);
  rc = frame_push(f);
  return(rc);
}

void *
Dx_frame_import_frame(Dx *db, void *buf, size_t buflen, void **ptr)
{
  Frame *f;

  if (db == NULL && (db->flags & DX_CONFIG_PERSISTENT) == 0) {
	Dx_last_rc = DX_BADARG;
	return(NULL);
  }

  Dx_last_rc = DX_OK;
  f = frame_import(db, buf, buflen, ptr);

  return((void *) f);
}

/*
 * Create a new frame, of a type determined by DB, and make it the top frame.
 */
int
Dx_new_frame(Dx *db, char *name)
{
  int rc;
  Frame *f;

  Dx_last_rc = DX_OK;

  f = frame_new_frame(db, name);
  if (f == NULL)
	return(-1);

  rc = frame_push(f);

  return(rc);
}

/*
 * Make FRAME, a previously popped frame, the top frame.
 */
int
Dx_push_frame(void *frame)
{
  int rc;
  Frame *f;

  Dx_last_rc = DX_OK;
  if (frame == NULL) {
	Dx_last_rc = DX_BADARG;
	return(-1);
  }

  f = (Frame *) frame;
  rc = frame_push(f);

  return(rc);
}

/*
 * Remove the top frame and return it.
 */
void *
Dx_pop_frame(void)
{
  Frame *f;
  
  Dx_last_rc = DX_OK;
  if (top_frame == NULL)
	return(NULL);

  f = top_frame;

  top_frame = top_frame->next;

  f->popped = 1;
  f->next = popped_frames;
  popped_frames = f;

  return((void *) f);
}

/*
 * Exchange the top-most two frames.
 */
void
Dx_exch_frame(void)
{
  Frame *new_top, *old_top;

  Dx_last_rc = DX_OK;
  if (top_frame != NULL && top_frame->next != NULL) {
	new_top = top_frame->next;
	old_top = top_frame;

	old_top->next = (top_frame->next)->next;
	new_top->next = old_top;
	top_frame = new_top;
  }

}

/*
 * Combine allocations on FRAME1 with allocations on FRAME2, leaving
 * FRAME1 and destroying FRAME2.
 * If either is NULL, it refers to the top frame.
 */
int
Dx_unite_frames(void *frame1, void *frame2)
{
  Frame *f1, *f2;
  MemChunk *mc, *mc_next;

  Dx_last_rc = DX_OK;

  if (frame1 == NULL) {
	if (top_frame == NULL) {
	  Dx_last_rc = DX_FRAME_NO_FRAME;
	  return(-1);
	}
	f1 = top_frame;
  }
  else
	f1 = (Frame *) frame1;

  if (frame2 == NULL) {
	if (top_frame == NULL) {
	  Dx_last_rc = DX_FRAME_NO_FRAME;
	  return(-1);
	}
	f2 = top_frame;
  }
  else
	f2 = (Frame *) frame2;

  if (frame1 == frame2) {
	Dx_last_rc = DX_FRAME_NO_FRAME;
	return(-1);
  }

  for (mc = f2->chunks; mc != NULL; mc = mc_next) {
	mc_next = mc->next;
	mc->next = f1->chunks;
	f1->chunks = mc;
  }
  if (f2->name != NULL) {
	free(f2->name);
	f2->name = NULL;
  }

  if (f2 == top_frame)
	top_frame = f2->next;

  free(f2);

  return(0);
}

/*
 * Free the top frame's allocations and destroy it.
 */
int
Dx_free(void)
{
  int rc;
  Frame *f_next;

  Dx_last_rc = DX_OK;
  if (top_frame == NULL)
	return(0);

  f_next = top_frame->next;

  rc = frame_free_frame(top_frame);

  top_frame = f_next;

  return(rc);
}

/*
 * Free FRAME's allocations and destroy it.
 * FRAME must have already been popped.
 */
int
Dx_free_frame(void *frame)
{
  int rc;
  Frame *f;

  Dx_last_rc = DX_OK;

  if (frame == NULL)
	return(Dx_free());

  f = (Frame *) frame;
  if (f->popped == 0) {
	Dx_last_rc = DX_FRAME_NOT_POPPED;
	return(-1);
  }

  rc = frame_free_frame(f);
  return(rc);
}

/*
 * Free all pushed frames.
 */
int
Dx_free_frames(void)
{
  int rc;
  Frame *f, *f_next;

  Dx_last_rc = DX_OK;
  if (top_frame == NULL)
	return(0);

  rc = 0;
  for (f = top_frame; f != NULL; f = f_next) {
	f_next = f->next;
	if (frame_free_frame(f) == -1)
	  rc = -1;
	free(f);
  }

  top_frame = NULL;
  return(rc);
}

/*
 * Free all frames, even popped ones.
 */
int
Dx_free_all_frames(void)
{
  int rc;
  Frame *f, *f_next;

  Dx_last_rc = DX_OK;
  rc = 0;

  for (f = top_frame; f != NULL; f = f_next) {
	f_next = f->next;
	if (frame_free_frame(f) == -1)
	  rc = -1;
	free(f);
  }
  top_frame = NULL;

  for (f = popped_frames; f != NULL; f = f_next) {
	f_next = f->next;
	if (frame_free_frame(f) == -1)
		rc = -1;
	free(f);
  }
  popped_frames = NULL;
  return(0);
}

/*
 * Allocate from the current frame, creating a new one if necessary.
 * If DB is NULL, use the system malloc().
 */
void *
Dx_malloc(size_t size)
{
  Dx *dbsrc;

  Dx_last_rc = DX_OK;
  if (top_frame == NULL)
	dbsrc = NULL;
  else
	dbsrc = top_frame->ms.db;

  return(frame_alloc_on_frame(NULL, dbsrc, size));
}

void *
Dx_malloc_src(Dx *dbsrc, size_t size)
{

  return(frame_alloc_on_frame(NULL, dbsrc, size));
}

/*
 * Allocate from FRAME.
 * If FRAME is NULL, use (or create) the top frame.
 */
void *
Dx_frame_malloc(void *frame, size_t size)
{
  Dx *dbsrc;
  Frame *f;

  Dx_last_rc = DX_OK;
  f = (Frame *) frame;
  if (f == NULL)
	dbsrc = NULL;
  else
	dbsrc = f->ms.db;

  return(frame_alloc_on_frame(f, dbsrc, size));
}

void *
Dx_frame_malloc_src(void *frame, Dx *dbsrc, size_t size)
{
  Frame *f;

  Dx_last_rc = DX_OK;
  f = (Frame *) frame;

  return(frame_alloc_on_frame(f, dbsrc, size));
}

/*
 * Allocate memory from the current frame for an array of NMEMB elements,
 * each of SIZE bytes.  Zero the memory.
 */
void *
Dx_calloc(size_t nmemb, size_t size)
{
  Dx *dbsrc;

  Dx_last_rc = DX_OK;
  if (top_frame == NULL)
	dbsrc = NULL;
  else
	dbsrc = top_frame->ms.db;

  return(frame_calloc_on_frame(NULL, dbsrc, nmemb, size));
}

void *
Dx_calloc_src(Dx *dbsrc, size_t nmemb, size_t size)
{

  Dx_last_rc = DX_OK;

  return(frame_calloc_on_frame(NULL, dbsrc, nmemb, size));
}

/*
 * Allocate, via calloc(), from FRAME.
 * If FRAME is NULL, use (or create) the top frame.
 */
void *
Dx_frame_calloc(void *frame, size_t nmemb, size_t size)
{
  Dx *dbsrc;
  Frame *f;

  Dx_last_rc = DX_OK;
  f = (Frame *) frame;
  if (f == NULL)
	dbsrc = NULL;
  else
	dbsrc = f->ms.db;

  return(frame_calloc_on_frame(f, dbsrc, nmemb, size));
}

void *
Dx_frame_calloc_src(void *frame, Dx *dbsrc, size_t nmemb, size_t size)
{
  Frame *f;

  Dx_last_rc = DX_OK;
  f = (Frame *) frame;

  return(frame_calloc_on_frame(f, dbsrc, nmemb, size));
}

/*
 * Reallocate memory.
 * If there is no current frame or allocation on that frame, this is
 * equivalent to Dx_malloc().
 */
void *
Dx_realloc(void *ptr, size_t size)
{
  Dx *dbsrc;

  Dx_last_rc = DX_OK;
  if (ptr == NULL)
	return(Dx_malloc(size));

  /* This is equivalent to free(3). */
  if (size == 0)
	return(NULL);

  if (top_frame == NULL)
	dbsrc = NULL;
  else
	dbsrc = top_frame->ms.db;

  return(mem_realloc(NULL, dbsrc, ptr, size));
}

void *
Dx_realloc_src(Dx *dbsrc, void *ptr, size_t size)
{

  Dx_last_rc = DX_OK;
  if (ptr == NULL)
	return(Dx_malloc_src(dbsrc, size));

  /* This is equivalent to free(3). */
  if (size == 0)
	return(NULL);

  return(mem_realloc(NULL, dbsrc, ptr, size));
}

void *
Dx_frame_realloc(void *frame, void *ptr, size_t size)
{
  Dx *dbsrc;
  Frame *f;

  Dx_last_rc = DX_OK;
  if (ptr == NULL)
	return(Dx_frame_malloc(frame, size));

  /* This is equivalent to free(3). */
  if (size == 0)
	return(NULL);

  f = (Frame *) frame;
  if (f == NULL)
	dbsrc = NULL;
  else
	dbsrc = f->ms.db;
  return(mem_realloc(f, dbsrc, ptr, size));
}

void *
Dx_frame_realloc_src(void *frame, Dx *dbsrc, void *ptr, size_t size)
{
  Frame *f;

  Dx_last_rc = DX_OK;
  if (ptr == NULL)
	return(Dx_frame_malloc_src(frame, dbsrc, size));

  /* This is equivalent to free(3). */
  if (size == 0)
	return(NULL);

  f = (Frame *) frame;
  return(mem_realloc(f, dbsrc, ptr, size));
}
