/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Each page is db->page_size bytes in length and contains zero or more items.
 * The extensible hashing algorithm generates logical page numbers,
 * numbered 0, 1, 2, ...
 * A logical page number is mapped onto a physical page, which resides in
 * some contiguous region of the database file.
 * A bitmap allocation scheme is used to allocate space for pages within the
 * database file.
 */

#include "local.h"

Page
Page_alloc(Dx *db, PageNum b)
{
  Physoff newb;
  Segment seg;
  Volume *volume;
  Dx_rc rc;

  volume = PAGENUM_TO_VOLUME(db, b);

  if ((rc = Segment_allocate_page(db, volume, b, &seg, &newb)) != DX_OK) {
	fprintf(stderr, "Page_alloc: Segment_allocate_page() failed\n");
	exit(1);
  }

  Item_init(db, (Page) seg);

  if (Logmap_page_add(db, b, newb) != DX_OK) {
	fprintf(stderr, "Page_alloc: Logmap_page_add() failed.\n");
	exit(1);
  }

  return((Page) seg);
}

Dx_rc
Page_exists(Dx *db, PageNum b)
{
  Dx_rc rc;

  rc = Logmap_page_lookup(db, b, PAGE_MAYCREATE, NULL);
  return(rc);
}

Dx_rc
Page_get(Dx *db, Volume *volume, PageNum b, Dx_ui32 mode, Page *page)
{
  Physoff pb;
  SegmentMap *sm;
  Dx_rc rc;

  if ((rc = Logmap_page_lookup(db, b, mode, &pb)) != DX_OK)
	return(rc);

  sm = Segment_lookup(db, PAGENUM_TO_VOLUME(db, b), pb);

  /*
   * If the page doesn't exist yet, create it.
   * XXX sm->flags are not set to SEGMENT_FIXED
   */
  if (sm == NULL) {
	Dx_ui32 need;
	Segment seg;

	need = pb + db->page_size;

	if (mode != PAGE_MAYCREATE && need > volume->mmap_size)
	  return(PAGE_NOTFOUND);

	if (need <= volume->mmap_size) {
	  rc = Segment_map(db, volume, pb, db->page_size, b, 1, 0, &seg);
	  if (rc != DX_OK)
		return(rc);
	} else {
	  if ((rc = Segment_allocate_page(db, volume, b, &seg, &pb)) != DX_OK)
		return(rc);
	}

	*page = (Page) seg;
  }
  else
	*page = (Page) sm->seg;

  return(DX_OK);
}

Dx_rc
Page_reorg(Dx *db)
{
  PageNum b;
  Page page;
  Dx_rc rc;

  if (db->flags & DX_CONFIG_READ_ONLY)
	return(DX_READONLY);

  for (b = 0; b < db->npages; b++) {
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc == DX_OK) {
	  if (Item_reorg(db, page))
		Stats_count(db, REORG);
	  Segment_clear_fixed_page_active(db, b);
	}
  }

  return(DX_OK);
}

Dx_rc
Page_release(Dx *db)
{
  Dx_rc rc;

  rc = DX_OK;

#ifdef NOTDEF
	/* This doesn't seem to be necessary. */
  for (i = 0; i < db->nvolumes; i++) {
	if (munmap((caddr_t) db->volume[i]->meta,
			   (size_t) META_HEADER_NBYTES) == -1) {
	  db->last_errno = errno;
	  rc = DX_OS_ERROR;
	  fprintf(stderr, "Page_release: munmap error\n");
	}
  }

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = db->segment_map[i]) != NULL) {
	  while (sm != NULL) {
		if (munmap((caddr_t) sm->seg, db->page_size) == -1) {
		  db->last_errno = errno;
		  rc = DX_OS_ERROR;
		  fprintf(stderr, "Page_release: munmap error\n");
		}
		sm->seg = NULL;
		sm = sm->next;
	  }
	}
  }
#endif

  rc = Segment_release(db);

  return(rc);
}

Dx_rc
Page_sync(Dx *db)
{
  Dx_ui32 i;
  SegmentMap *sm;
  Dx_rc rc;

  if ((db->flags & DX_CONFIG_PERSISTENT) == 0)
	return(DX_OK);

  rc = DX_OK;

#ifdef __FreeBSD__
  for (i = 0; i < db->nvolumes; i++) {
	if (fsync(db->volume[i]->fd) == -1) {
	  db->last_errno = errno;
	  rc = DX_OS_ERROR;
	  fprintf(stderr, "Page_sync: fsync error\n");
	}
  }
#else

  for (i = 0; i < db->nvolumes; i++) {
	if (msync((caddr_t) db->volume[i]->meta,
			  META_HEADER_NBYTES, MS_SYNC) == -1) {
	  db->last_errno = errno;
	  rc = DX_OS_ERROR;
	  fprintf(stderr, "Page_sync: msync error\n");
	}
  }

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = db->segment_map[i]) != NULL) {
	  while (sm != NULL) {
		if (msync((char *) sm->seg, sm->length, MS_SYNC) == -1) {
		  db->last_errno = errno;
		  rc = DX_OS_ERROR;
		  fprintf(stderr, "Page_sync: msync error\n");
		  fprintf(stderr, "addr=%lx\n", (Dx_ui32) sm->seg);
		}
		sm = sm->next;
	  }
	}
  }
#endif

  return(rc);
}

Dx_rc
Page_duplicate(Dx *new_db, Dx *old_db)
{

  return(Segment_duplicate(new_db, old_db));
}

void
Page_init(Dx *db)
{

  db->controlled_split_check = 0;
  Segment_init(db);
}

void
page_show(Dx *db)
{
  Dx_ui32 i;
  SegmentMap *sm;

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = db->segment_map[i]) != NULL) {
	  fprintf(stderr, "%lu: ", (unsigned long) i);
	  while (sm != NULL) {
		fprintf(stderr, "%lu ", (unsigned long) sm->b);
		sm = sm->next;
	  }
	  fprintf(stderr, "\n");
	}
  }
}

void
page_stats(Dx *db)
{
  int npages, nused;
  Dx_ui32 i;
  SegmentMap *sm;

  npages = nused = 0;
  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = db->segment_map[i]) != NULL) {
	  nused++;
	  while (sm != NULL) {
		npages++;
		sm = sm->next;
	  }
	}
  }
  fprintf(stderr, "npages=%d, nused=%d\n", npages, nused);

}

void
db_check(Dx *db)
{
  unsigned int *dist, i, j, k, nempty, nitems, nprobes, nused;
  Dx_ui32 ind, max, n;
  Dx_rc rc;
  Area *area;
  PageNum b;
  Page page;
  void *bp;
  char **key_vec;
  int *probe_vec;
  Item item;
  Dx_datum key, value;
  SegmentMap *next, *sm;
  unsigned long t_actual_meta, t_meta, t_avail, t_used;
  unsigned long t_indirect, t_nindirect;
  PageLoad pl;
#define DD_SIZE  30
  static unsigned int dd[DD_SIZE] = {
	   0,   1,   5,  10,  15,  20,  30,  40,  50,  60,
	  70,  80,  90, 100, 110, 120, 130, 140, 150, 160,
	 170, 180, 190, 200, 220, 240, 260, 280, 300, 500 };

  dist = (unsigned int *) Xmalloc((DD_SIZE + 1) * sizeof(int));
  for (i = 0; i < (DD_SIZE + 1); i++)
	dist[i] = 0;

  nitems = 0;
  nused = 0;
  nempty = 0;
  t_meta = t_actual_meta = t_avail = t_used = t_indirect = t_nindirect = 0;
  max = 0;

  for (b = 0; Ehash_bucket_exists(db, b); b++) {
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc == DX_OK) {
	  page_check(db, b);
	  n = Item_count(db, page, db->page_size);

	  nitems += n;
	  if (n == 0)
		nempty++;
	  nused++;
	  if (n > max)
		max = n;
	  for (i = 0; i < DD_SIZE; i++) {
		if (n <= dd[i]) {
		  dist[i]++;
		  break;
		}
	  }
	  if (n > dd[DD_SIZE - 1])
		dist[DD_SIZE]++;

	  Item_load(db, page, db->page_size, &pl, 1);
	  t_meta += pl.total_meta;
	  t_actual_meta += pl.actual_meta;
	  t_avail += pl.avail;
	  t_used += pl.used;
	  t_indirect += pl.indirect;
	  t_nindirect += pl.nindirect;
	}
  }

  fprintf(stderr, "system_page_size=%lu, st_blksize=%lu (vol0)\n",
		  (unsigned long) db->system_page_size,
		  (unsigned long) db->volume[0]->st_blksize);
  fprintf(stderr, "area=%llu (header=%d, physmap=%d, logmap=%d)\n",
		  AREA_SIZE(db), META_HEADER_NBYTES, PHYSMAP_MAP_NBYTES,
		  LOGMAP_MAP_NBYTES);
  fprintf(stderr, "indirect_threshold=%lu bytes\n",
		  (Dx_ui32) (ITEM_INDIRECT_THRESHOLD_FACTOR * db->page_size));
  fprintf(stderr, "\n");

  fprintf(stderr,
		  "current_size=%lu, npages=%lu, page_size=%lu, persistent=%ld\n",
		  (unsigned long) Compute_expansion_buffers(db, db->nexpansions),
		  (unsigned long) db->npages,
		  (unsigned long) db->page_size,
		  (unsigned long) db->flags & DX_CONFIG_PERSISTENT);
  fprintf(stderr, "nexpansions=%lu, boundary=%lu\n",
		  (unsigned long) db->nexpansions, (unsigned long) db->boundary);
  fprintf(stderr, "pages inuse=%d, empty=%d\n", nused, nempty);
  fprintf(stderr, "nmmap_calls=%lu\n", db->nmmap_calls);
  fprintf(stderr, "\n");

  fprintf(stderr, "%d item%s, mean items/page=%.2f, max=%lu\n",
		  nitems, nitems == 1 ? "" : "s",
		  (double) nitems / (double) nused, max);
  fprintf(stderr, "\n");

  for (i = 0; i < db->nvolumes; i++) {
	for (area = db->volume[i]->area; area != NULL; area = area->next) {
	  Area_show(db, area->areanum);
	}
  }

  fprintf(stderr, "Mean page load (bytes):\n");
  fprintf(stderr, "meta=%.1f (%.2f), actual_meta=%.1f (%.2f)\n",
		  (double) t_meta / (double) nused,
		  ((double) t_meta / (double) nused) / db->page_size,
		  (double) t_actual_meta / (double) nused,
		  ((double) t_actual_meta / (double) nused) / db->page_size);
  fprintf(stderr, "data=%.1f (%.2f), avail=%.1f (%.2f)\n",
		  (double) t_used / (double) nused,
		  ((double) t_used / (double) nused) / db->page_size,
		  (double) t_avail / (double) nused,
		  ((double) t_avail / (double) nused) / db->page_size);

  fprintf(stderr, "Mean db load: data=%.1f, overall=%.1f\n",
		  (double) t_used / (((double) nused) * db->page_size),
		  (double) (t_meta + t_used) / (((double) nused) * db->page_size));
  fprintf(stderr, "Indirect: count=%lu, bytes=%lu\n",
		  t_nindirect, t_indirect);
  fprintf(stderr, "\n");

  /* Pick any old page to report on. */
  for (b = 0; b < SEGMENT_MAP_NENTRIES; b++) {
	sm = db->segment_map[b];
	if (sm != NULL && (sm->flags & SEGMENT_IS_PAGE)) {
	  fprintf(stderr, "Page %llu:\n", (Dx_ui64) sm->b);
	  Item_info(db, (Page) sm->seg, db->page_size);
	  fprintf(stderr, "\n");
	  break;
	}
  }

  if (DX_STATS)
	Stats_show(db, stderr);

  for (i = 0; i < (DD_SIZE + 1); i++) {
	if (i == DD_SIZE)
	  fprintf(stderr, "%4s: %4d  ", ">", dist[i]);
	else
	  fprintf(stderr, "%4d: %4d  ", dd[i], dist[i]);
	if (((i + 1) % 5) == 0)
	  fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n");
  Xfree(dist);

  if ((key_vec = (char **) Xmalloc(nitems * sizeof(char *))) == NULL) {
	db->last_errno = errno;
	perror("Out of memory");
	exit(1);
  }

  item.key = &key;
  item.value = &value;
  item.ks = NULL;
  k = 0;
  for (b = 0; Ehash_bucket_exists(db, b); b++) {
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc == DX_OK) {
	  for (ind = 0; Item_index(db, page, ind, DX_PROBE, &item, NULL)
			 == DX_OK;
		   ind++)
		key_vec[k++] = Xstrdup((char *) item.key->dptr);
	}
  }

  if (DX_STATS) {
	probe_vec = (int *) Xmalloc(8 * sizeof(int));
	for (i = 0; i < 8; i++)
	  probe_vec[i] = 0;

	nprobes = 0;
	for (i = 0; i < k; i++) {
	  Dx_key(&key, key_vec[i], strlen(key_vec[i]) + 1);
	  Stats_init(db);
	  if (Dx_fetchx(db, &key, DX_PROBE, &value) != DX_OK)
		fprintf(stderr, "Not found: %s\n", key_vec[i]);
	  Xfree(key_vec[i]);
	  nprobes += db->stats.nprobes;
	  if ((j = db->stats.nprobes - 1) >= 8)
		j = 7;
	  probe_vec[j]++;
	}

	Xfree(key_vec);

	fprintf(stderr, "nprobes = %d, mean = %.2f\n",
			nprobes, k == 0 ? 0.0 : (double) nprobes / (double) k);
	for (i = 0; i < 8; i++)
	  fprintf(stderr, "%d:%3d ", i + 1, probe_vec[i]);
	fprintf(stderr, "\n");

	Xfree(probe_vec);
  }
}

void
db_dump(Dx *db)
{
  unsigned int i, nitems, nused;
  Dx_ui32 ind;
  void *bp;
  Item item;
  Dx_datum key, value;
  Page page;
  PageNum b;
  Dx_rc rc;

  nitems = 0;
  nused = 0;
  item.key = &key;
  item.value = &value;

  for (b = 0; b < db->npages; b++) {
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	if (rc != DX_OK)
	  continue;
	nused++;
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &page);
	nitems += Item_count(db, page, db->page_size);
	if (b == db->boundary)
	  fprintf(stderr, "->");
	else
	  fprintf(stderr, "  ");
	fprintf(stderr, "%4llu:\n", (Dx_ui64) b);
	rc = Page_get(db, PAGENUM_TO_VOLUME(db, b), b, PAGE_MUSTEXIST, &bp);
	if (rc != DX_OK)
	  continue;
	for (ind = 0; Item_index(db, bp, ind, DX_FETCH, &item, NULL)
		   != ITEM_NOTFOUND; ind++) {
	  fprintf(stderr, "    %3lu: %s = %s\n",
			  ind, (char *) item.key->dptr, (char *) item.value->dptr);
	}
  }

  fprintf(stderr, "nitems=%d, nused=%d\n", nitems, nused);
}
