/*
 * Copyright (c) 2000 DSS Distributed Systems Software, Inc.
 * All rights reserved.
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * Segment allocation and mapping.
 * A segment is an aligned area of the database that is memory mapped.
 * Segment sizes are multiples of the system page size.
 * A database page is an instance of a segment, but other kinds of segments
 * can be bigger.  Segments the size of a page are managed using the
 * physmap data structures, but bigger segments are managed here.
 *
 * XXX Note that segments for meta data can't be resized and they are
 * usually mapped permanently (SEGMENT_FIXED).
 * Also, a database page (bucket) can't be resized.
 */

#include <stdlib.h>
#include <stdio.h>

#include "local.h"

/*
 * This internal version doesn't increment the usage count.
 * Also, it doesn't cache the lookup; this is called in situations where
 * it will probably be a while before this segment is accessed.
 * XXX This assumption may be wrong in practice...
 */
static INLINE SegmentMap *
segment_lookup(Dx *db, Volume *volume, Physoff offset)
{
  SegmentMap *sm;

  sm = db->last_segment_lookup[volume->volnum];

  if (sm != NULL && sm->offset == offset)
	return(sm);

  sm = db->segment_map[offset % SEGMENT_MAP_NENTRIES];

  while (sm != NULL) {
	if (sm->offset == offset && sm->volume == volume)
	  return(sm);
	sm = sm->next;
  }

  return(NULL);
}

SegmentMap *
Segment_lookup(Dx *db, Volume *volume, Physoff offset)
{
  SegmentMap *sm;

  if ((sm = segment_lookup(db, volume, offset)) != NULL) {
	sm->count++;
	db->last_segment_lookup[volume->volnum] = sm;
  }

  return(sm);
}

static INLINE SegmentMap *
segment_lookup_active(Dx *db, Volume *volume, Physoff offset)
{
  SegmentMap *sm;

  sm = db->active_list;

  while (sm != NULL) {
	if (sm->offset == offset && sm->volume == volume)
	  return(sm);
	sm = sm->active_list_next;
  }

  return(NULL);
}

static INLINE SegmentMap *
segment_lookup_page_active(Dx *db, PageNum b)
{
  SegmentMap *sm;

  sm = db->active_list;

  while (sm != NULL) {
	if (sm->b == b)
	  break;
	sm = sm->active_list_next;
  }

  return(sm);
}

/*
 * Add a segment description to the table.
 * The argument has been malloc-ed.
 */
static INLINE void
segment_mapadd(Dx *db, SegmentMap *nsm)
{
  SegmentMap *sm;

  sm = db->segment_map[nsm->offset % SEGMENT_MAP_NENTRIES];

  if (sm == NULL) {
	nsm->next = NULL;
	db->segment_map[nsm->offset % SEGMENT_MAP_NENTRIES] = nsm;
  }
  else {
	if (DX_DEBUG) {
	  SegmentMap *osm;
	  if ((osm = segment_lookup(db, nsm->volume, nsm->offset)) != NULL)
		fprintf(stderr, "segment_mapadd: duplicate entry!\n");
	}
	nsm->next = db->segment_map[nsm->offset % SEGMENT_MAP_NENTRIES];
	db->segment_map[nsm->offset % SEGMENT_MAP_NENTRIES] = nsm;
  }

  nsm->active_list_next = db->active_list;
  db->active_list = nsm;

  db->nmapped++;
  db->nbytesmapped += nsm->length;
}

static INLINE void
segment_not_active(Dx *db, SegmentMap *osm)
{
  SegmentMap *prev, *sm;

  sm = db->active_list;
  prev = NULL;

  while (sm != NULL) {
	if (sm == osm) {
	  if (prev == NULL)
		db->active_list = sm->active_list_next;
	  else
		prev->active_list_next = sm->active_list_next;
	  sm->active_list_next = NULL;
	  return;
	}
	prev = sm;
	sm = sm->active_list_next;
  }

  fprintf(stderr, "segment_not_active: not found!\n");
}

static INLINE void
segment_mapfree(Dx *db, SegmentMap *sm)
{

  db->nbytesmapped -= sm->length;
  db->nmapped--;

  if (sm == db->last_segment_lookup[sm->volume->volnum])
	db->last_segment_lookup[sm->volume->volnum] = NULL;

  segment_not_active(db, sm);

  if (DX_DEBUG)
	Dx_memset((void *) sm, 0, sizeof(SegmentMap));

  Xfree(sm);
}

/*
 * Remove a segment description from the table.
 * The structure can be freed.
 */
static INLINE void
segment_mapremove(Dx *db, Volume *volume, Physoff offset)
{
  SegmentMap *prev, *sm;

  sm = db->segment_map[offset % SEGMENT_MAP_NENTRIES];
  prev = NULL;

  while (sm != NULL) {
	if (sm->volume == volume && sm->offset == offset) {
	  if (prev == NULL)
		db->segment_map[offset % SEGMENT_MAP_NENTRIES] = sm->next;
	  else
		prev->next = sm->next;

	  segment_mapfree(db, sm);
	  return;
	}
	prev = sm;
	sm = sm->next;
  }

  /*
   * At the moment, don't panic if it's not found.
   */
#ifdef NOTDEF
  fprintf(stderr, "segment_mapremove: not found!\n");
  return(DX_INTERNAL_ERROR);
#endif
}

/*
 * Clean up and free all segments for DB.
 * XXX Only unmap modified segments?  or leave that up to the OS?
 * XXX Only needs to be partially implemented at present.
 */
Dx_rc
Segment_release(Dx *db)
{
  Dx_ui32 i;
  SegmentMap *next, *sm;
  Dx_rc rc;

  /* There's no point doing any work for a volatile db. */
  if ((db->flags & DX_CONFIG_PERSISTENT) == 0)
	return(DX_OK);

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = db->segment_map[i]) != NULL) {
	  while (sm != NULL) {
		next = sm->next;
		if (sm->flags & SEGMENT_RW) {
		  if ((rc = Segment_unmap(db, sm->volume, sm->offset, sm->seg,
								  sm->length, 0)) != DX_OK)
		  return(rc);
		}
		segment_mapfree(db, sm);
		sm = next;
	  }
	  db->segment_map[i] = NULL;
	}
  }

  return(DX_OK);
}

/*
 * Allocate and initialize a segment.
 * The SM structure is from malloc().
 * Returns a pointer to the initialized segment in memory and the physical
 * address of the segment in the db.
 */
static INLINE Dx_rc
segment_allocate(Dx *db, Volume *volume, PageNum b, int is_page,
				 Seglen length, Dx_ui32 flags, Segment *new_seg, Physoff *pb)
{
  Dx_rc rc;

  if (length <= db->page_size)
	rc = Physmap_allocate_page(db, volume, PHYSMAP_MAP_NBYTES, pb);
  else
	rc = Physmap_allocate_segment(db, volume, PHYSMAP_MAP_NBYTES,
								  length / db->page_size, pb);

  if (rc != DX_OK)
	return(rc);

  rc = Segment_map(db, volume, *pb, length, b, is_page, flags, new_seg);

  return(rc);
}

static INLINE SegmentMap *
segment_mapalloc(void)
{
  SegmentMap *sm;

  sm = (SegmentMap *) Xmalloc(sizeof(SegmentMap));
  sm->b = 0;
  sm->length = 0;
  sm->flags = 0;
  sm->offset = 0;
  sm->volume = NULL;
  sm->seg = NULL;
  sm->count = 1;
  sm->next = NULL;
  sm->active_list_next = NULL;

  return(sm);
}

Dx_rc
Segment_allocate_page(Dx *db, Volume *volume, PageNum b,
					  Segment *new_seg, Physoff *pb)
{
  Dx_rc rc;

  rc = segment_allocate(db, volume, b, 1, db->page_size, SEGMENT_FIXED,
						new_seg, pb);

  return(rc);
}

Dx_rc
Segment_allocate_segment(Dx *db, Volume *volume, Seglen nbytes, Dx_ui32 flags,
						 Segment *new_seg, Physoff *pb)
{
  Dx_rc rc;

  rc = segment_allocate(db, volume, 0, 0, NBYTES_TO_PAGE_BYTES(db, nbytes),
						flags, new_seg, pb);

  return(rc);
}

/*
 * If the new size is greater than the old size, the existing segment may be
 * extended or a new segment may be allocated.
 * If the new size is less than the old size, the existing segment will be
 * truncated.
 * The sizes are given in bytes.
 * The current length of the segment is provided so that the segment needn't
 * be already mapped.
 * The returned segment is always mapped.
 * Note that a regular page can't be resized.
 */
Dx_rc
Segment_reallocate(Dx *db, Volume *volume, Physoff old_offset,
				   Seglen old_length, Seglen new_length,
				   Segment *new_seg, Physoff *pb)
{
  Dx_ui32 flags, new_npages, old_npages;
  Segment old_seg;
  SegmentMap *sm;
  Dx_rc rc;

  rc = DX_OK;
  if ((sm = segment_lookup(db, volume, old_offset)) != NULL) {
	if (old_length != sm->length) {
	  fprintf(stderr, "Segment_reallocate: bad initial length!\n");
	  return(DX_INTERNAL_ERROR);
	}
  }

  old_npages = NBYTES_TO_NPAGES(db, old_length);
  new_npages = NBYTES_TO_NPAGES(db, new_length);

  if (old_npages >= new_npages) {
	/* Just truncate the old segment. */
	if (old_npages > new_npages)
	  rc = Segment_truncate(db, volume, old_offset, old_npages, new_npages);

	if (rc == DX_OK) {
	  if (sm == NULL)
		/* XXX flags? */
		rc = Segment_map(db, volume, old_offset, new_length, 0, 0, 0, new_seg);
	  if (rc == DX_OK)
		*pb = old_offset;
	}

	return(rc);
  }

  /* See if the existing segment can simply be extended. */
  rc = Physmap_extend_segment(db, volume, PHYSMAP_MAP_NBYTES, old_offset,
							  old_npages, new_npages);
  if (rc == DX_OK) {
	int need_to_map;

	/* It was either unmapped or mapped at a smaller size. */
	need_to_map = (sm == NULL);
	rc = Segment_map(db, volume, old_offset, new_length, 0, 0, 0, new_seg);
	if (rc != DX_OK)
	  return(rc);
	*pb = old_offset;
	return(DX_OK);
  }
  else if (rc != DX_PHYSMAP_FAILED)
	return(rc);

  /* Allocate a new segment, copy the old segment, free the old segment. */
  if (sm == NULL)
	flags = 0;
  else
	flags = sm->flags;
  rc = Segment_allocate_segment(db, volume, new_length, flags, new_seg, pb);
  if (rc != DX_OK)
	return(rc);

  /* The old segment must be mapped in order to copy it. */
  if (sm == NULL) {
	rc = Segment_map(db, volume, old_offset, old_length, 0, 0, 0, &old_seg);
	if (rc != DX_OK) {
	  /* XXX New segment should be deallocated. */
	  Segment_free(db, volume, *pb, new_npages);
	  return(rc);
	}
  }
  else
	old_seg = sm->seg;

  /* Safe and faster not to use memmove(). */
  Dx_memcpy(*new_seg, old_seg, old_length);
  rc = Segment_free(db, volume, old_offset, old_npages);
  
  return(rc);
}

/*
 * Memory map the contiguous byte range that starts at offset PB within
 * the given volume and is NBYTES long.
 * Return a pointer to the start of the range.
 * The range may already be mapped.
 */
Dx_rc
Segment_map(Dx *db, Volume *volume, Physoff pb, Seglen nbytes,
			PageNum b, int is_page, Dx_ui32 flags, Segment *new_seg)
{
  int is_new;
  Dx_ui32 need_len;
  SegmentMap *sm;

  if ((sm = Segment_lookup(db, volume, pb)) == NULL) {
	sm = segment_mapalloc();
	sm->volume = volume;
	sm->length = nbytes;
	sm->b = b;
	if (is_page)
	  sm->flags |= SEGMENT_IS_PAGE;
	sm->flags |= (flags & (SEGMENT_FIXED | SEGMENT_META));
	/* XXX The FLAGS should specify this */
	if ((db->mmap_prot & PROT_WRITE))
	  sm->flags |= SEGMENT_RW;
	is_new = 1;
  }
  else {
	if (sm->length == nbytes) {
	  *new_seg = sm->seg;
	  return(DX_OK);
	}
	is_new = 0;
  }

  /* The file must be at least this big.  If not, extend it.  */
  need_len = pb + nbytes;
  if (need_len > volume->mmap_size) {
	if (ftruncate(volume->fd, need_len) == -1) {
	  db->last_errno = errno;
	  perror("ftruncate");
	  return(DX_OS_ERROR);
	}
	volume->mmap_size = need_len;
  }

  /*
   * The region may be shrinking or expanding.
   * "A successful mmap deletes any previous mapping in the allocated address
   * range."
   * -- mmap(2), FreeBSD 2.2.2-RELEASE/2.2.5-RELEASE/3.1-RELEASE
   */
  if (nbytes < sm->length) {
	if (munmap((caddr_t) sm->seg, (size_t) sm->length) == -1) {
	  db->last_errno = errno;
	  perror("munmap");
	  return(DX_OS_ERROR);
	}
  }

  *new_seg = mmap(NULL, (size_t) nbytes, db->mmap_prot,
				  db->mmap_flags, volume->fd, (off_t) pb);
  if (*new_seg == (caddr_t) -1) {
	db->last_errno = errno;
	perror("mmap");
	return(DX_OS_ERROR);
  }

  sm->length = nbytes;
  sm->offset = pb;
  sm->seg = *new_seg;

  db->nmmap_calls++;

  if (is_new)
	segment_mapadd(db, sm);

  return(DX_OK);
}

Dx_rc
Segment_unmap(Dx *db, Volume *volume, Physoff first_block,
			  Segment seg, Seglen nbytes, int do_remove)
{

  if (munmap((caddr_t) seg, (size_t) nbytes) == -1) {
	db->last_errno = errno;
	perror("munmap");
	return(DX_OS_ERROR);
  }

  if (do_remove)
	segment_mapremove(db, volume, first_block);

  return(DX_OK);
}

static int fd_zero = -1;
static Seglen current_size = 0;

/*
 * This is probably a dumb idea.
 * It's used to map a region of zeroes that is copied into a buffer by
 * Stripe_contig().  This region corresponds to a hole in a db.
 */
Dx_rc
Segment_map_zeroes(Dx *db, Seglen nbytes, int prot, Segment *new_seg)
{

  if (fd_zero == -1) {
	if ((fd_zero = open("/dev/zero", O_RDWR)) == -1) {
	  db->last_errno = errno;
	  perror("open /dev/zero");
	  return(DX_OS_ERROR);
	}
  }

  if (current_size < nbytes) {
	if (ftruncate(fd_zero, nbytes) == -1) {
	  db->last_errno = errno;
	  perror("ftruncate");
	  return(DX_OS_ERROR);
	}
	current_size = nbytes;
  }

  *new_seg = mmap(NULL, (size_t) nbytes, prot, MAP_PRIVATE, fd_zero,
				  (off_t) 0);
  if (*new_seg == (caddr_t) -1) {
	db->last_errno = errno;
	perror("mmap");
	return(DX_OS_ERROR);
  }

  return(DX_OK);
}

Dx_rc
Segment_unmap_zeroes(Dx *db, Seglen nbytes)
{

  if (nbytes > NPAGES_TO_NBYTES(db, db->max_stripe_npages)) {
	/* If the mapped area is very big, avoid clogging the address space. */
	if (munmap((caddr_t) 0, (size_t) nbytes) == -1) {
	  db->last_errno = errno;
	  perror("munmap");
	  return(DX_OS_ERROR);
	}
	current_size = 0;
  }

  return(DX_OK);
}

Dx_rc
Segment_get_flags(Dx *db, Volume *volume, Physoff offset, Dx_ui32 *flags)
{
  SegmentMap *sm;

  if ((sm = segment_lookup(db, volume, offset)) == NULL)
	return(SEGMENT_NOTFOUND);

  *flags = sm->flags;
  return(DX_OK);
}

Dx_rc
Segment_clear_fixed_active(Dx *db, Volume *volume, Physoff offset)
{
  SegmentMap *sm;

  if ((sm = segment_lookup_active(db, volume, offset)) == NULL)
	return(SEGMENT_NOTFOUND);

  sm->flags &= (~SEGMENT_FIXED);

  return(DX_OK);
}

Dx_rc
Segment_clear_fixed_page_active(Dx *db, PageNum b)
{
  SegmentMap *sm;

  if ((sm = segment_lookup_page_active(db, b)) == NULL)
	return(SEGMENT_NOTFOUND);

  sm->flags &= (~SEGMENT_FIXED);

  return(DX_OK);
}

/*
 * Set the given flag bits (logically OR them with existing flags).
 */
Dx_rc
Segment_set_flags(Dx *db, Volume *volume, Physoff offset, Dx_ui32 set_flags)
{
  SegmentMap *sm;

  if ((sm = segment_lookup(db, volume, offset)) == NULL)
	return(SEGMENT_NOTFOUND);

  if ((set_flags & SEGMENT_RW) && (sm->flags & SEGMENT_RW) == 0) {
	/* Change from read-only to read-write. */
	if (mprotect((caddr_t) sm->seg, (size_t) sm->length, PROT_READ|PROT_WRITE)
		== -1) {
	  db->last_errno = errno;
	  perror("mprotect");
	  return(DX_OS_ERROR);
	}
  }

  sm->flags |= set_flags;
  return(DX_OK);
}

/*
 * Clear the given flag bits.
 */
Dx_rc
Segment_clear_flags(Dx *db, Volume *volume, Physoff offset, Dx_ui32 clr_flags)
{
  SegmentMap *sm;

  if ((sm = segment_lookup(db, volume, offset)) == NULL)
	return(SEGMENT_NOTFOUND);

  if ((clr_flags & SEGMENT_RW) && (sm->flags & SEGMENT_RW)) {
	/* Change from read-write to read-only. */
	if (mprotect((caddr_t) sm->seg, (size_t) sm->length, PROT_READ) == -1) {
	  db->last_errno = errno;
	  perror("mprotect");
	  return(DX_OS_ERROR);
	}
  }

  sm->flags &= (~clr_flags);
  return(DX_OK);
}

Dx_rc
Segment_free(Dx *db, Volume *volume, Physoff offset, Dx_ui32 npages)
{

  return(Segment_truncate(db, volume, offset, npages, 0));
}

/*
 * Truncate the segment starting at OFFSET of length OLD_NPAGES to
 * NEW_NPAGES.
 * Do nothing if OLD_NPAGES and NEW_NPAGES are the same.
 * The OLD_NPAGES is needed because the segment may not be mapped (therefore
 * there is no SegmentMap entry for it) and then we don't know its length.
 */
Dx_rc
Segment_truncate(Dx *db, Volume *volume, Physoff offset,
				 Dx_ui32 old_npages, Dx_ui32 new_npages)
{
  Dx_len truncate_bytes;
  Physoff start_offset;
  Segment truncate_ptr;
  SegmentMap *sm;
  Dx_rc rc;

  if (old_npages < new_npages) {
	fprintf(stderr, "Segment_truncate: old_npages < new_npages!\n");
	return(DX_INTERNAL_ERROR);
  }

  if (old_npages == new_npages)
	return(DX_OK);

  truncate_bytes = NPAGES_TO_NBYTES(db, old_npages - new_npages);

  rc = DX_OK;
  if ((sm = Segment_lookup(db, volume, offset)) != NULL) {
	if (DX_DEBUG && sm->length != NPAGES_TO_NBYTES(db, old_npages)) {
	  fprintf(stderr,
			  "Segment_truncate: sm->length (%llu) != old_npages (%llu)!\n",
			  (unsigned long long) sm->length,
			  (unsigned long long) NPAGES_TO_NBYTES(db, old_npages));
	  return(DX_INTERNAL_ERROR);
	}
	sm->length = NPAGES_TO_NBYTES(db, new_npages);
	start_offset = offset + NPAGES_TO_NBYTES(db, new_npages);
	truncate_ptr =
	  (Segment) ((Dx_ui8 *) sm->seg + NPAGES_TO_NBYTES(db, new_npages));

	if (db->flags & DX_CONFIG_ERASE_ON_DELETE)
	  Dx_memset((void *) truncate_ptr, DX_ERASE_ON_DELETE_CHAR,
			 (size_t) truncate_bytes);

	rc = Segment_unmap(db, volume, start_offset, truncate_ptr, truncate_bytes,
					   (new_npages == 0));
  }

  if (rc == DX_OK) {
	if (sm == NULL && db->flags & DX_CONFIG_ERASE_ON_DELETE) {
	  Segment seg;

	  /*
	   * It must be mapped to get zapped.
	   * XXX Don't always need to map the whole thing, although it may not
	   * matter, efficiency wise.
	   */
	  rc = Segment_map(db, volume, offset, NPAGES_TO_NBYTES(db, old_npages),
					   0, 0, 0, &seg);
	  if (rc == DX_OK) {
		truncate_ptr =
		  (Segment) ((Dx_ui8 *) seg + NPAGES_TO_NBYTES(db, new_npages));
		Dx_memset((void *) truncate_ptr, DX_ERASE_ON_DELETE_CHAR,
			   (size_t) truncate_bytes);
		rc = Segment_unmap(db, volume, offset, truncate_ptr,
						   NPAGES_TO_NBYTES(db, old_npages), 1);
	  }
	}
	rc = Physmap_truncate_segment(db, volume->volnum, offset,
								  old_npages, new_npages);
  }

  return(rc);
}

Dx_rc
Segment_free_using_map(Dx *db, SegmentMap *sm)
{
  Physoff offset;
  Seglen seglen;
  Volume *v;
  Dx_rc rc;

  v = sm->volume;
  offset = sm->offset;
  seglen = sm->length;

  rc = Segment_unmap(db, v, offset, sm->seg, seglen, 1);

  if (rc == DX_OK)
	rc = Physmap_free_segment(db, v->volnum, offset,
							  NBYTES_TO_NPAGES(db, seglen));

  return(rc);
}

Dx_rc
Segment_init(Dx *db)
{
  Dx_ui32 i;

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++)
	db->segment_map[i] = NULL;

  for (i = 0; i < db->nvolumes; i++)
	db->last_segment_lookup[i] = NULL;

  db->nmapped = 0;
  db->nbytesmapped = 0;
  db->nmmap_calls = 0;

  return(DX_OK);
}

/*
 * XXX This is a temporary solution.
 * Guard against mapping too many segments by reducing the number to at
 * most SEGMENT_CACHE_LIMIT.
 * Try to keep the most heavily used mappings.
 * Segments flagged SEGMENT_FIXED must not be unmapped here.
 * Only invoke this when the segments aren't in use, specifically at the
 * start of executing a top-level function.
 */
Dx_rc
Segment_uncache(Dx *db, int segments_only)
{
  int changed, i;
  Dx_ui32 min, segment_cache_entry_limit;
  Dx_ui64 new_total, original_length, total_length;
  SegmentMap *prev, *sm, *sm_prev;
  Dx_rc rc;

  if (!segments_only)
	return(DX_OK);

  /*
   * Don't ever uncache volatile stuff - it'd be gone forever.
   */
  if ((db->flags & DX_CONFIG_PERSISTENT) == 0)
	return(DX_OK);

 if (db->nmapped <= SEGMENT_CACHE_ENTRY_LIMIT
	 && db->nbytesmapped <= SEGMENT_CACHE_BYTE_LIMIT)
   return(DX_OK);

  if (DX_DEBUG && DX_VERBOSE)
	fprintf(stderr, "Total mapped: %lu entries, %llu bytes\n",
			db->nmapped, db->nbytesmapped);

  segment_cache_entry_limit = SEGMENT_CACHE_ENTRY_LIMIT;
  new_total = (Dx_ui64) (SEGMENT_CACHE_BYTE_LIMIT * 0.25);
  original_length = db->nbytesmapped;

  changed = 1;
  while (changed && (db->nmapped > segment_cache_entry_limit
					 || db->nbytesmapped > new_total)) {
	changed = 0;
	for (i = 0;
		 (db->nmapped > segment_cache_entry_limit
		  || db->nbytesmapped > new_total) && i < SEGMENT_MAP_NENTRIES;
		 i++) {
	  min = DX_MAX_UI32;
	  sm_prev = prev = NULL;
	  for (sm = db->segment_map[i]; sm != NULL; sm = sm->next) {
		if ((sm->flags & SEGMENT_FIXED) == 0
			&& (!segments_only || (sm->flags & SEGMENT_IS_PAGE) == 0)
			&& sm->count < min) {
		  min = sm->count;
		  prev = sm_prev;
		}
		sm_prev = sm;
	  }

	  if (min == DX_MAX_UI32)
		continue;

	  if (prev == NULL) {
		sm = db->segment_map[i];
		db->segment_map[i] = sm->next;
	  }
	  else {
		sm = prev->next;
		prev->next = sm->next;
	  }

	  rc = Segment_unmap(db, sm->volume, sm->offset, sm->seg, sm->length, 1);
	  if (rc != DX_OK)
		return(rc);

	  changed = 1;
	}
  }

  if (DX_DEBUG && DX_VERBOSE)
	fprintf(stderr, "Total mapped: %lu entries, %llu bytes\n",
			db->nmapped, db->nbytesmapped);
  return(DX_OK);
}

/*
 * Duplicate each segment of OLD_DB in NEW_DB.
 */
Dx_rc
Segment_duplicate(Dx *new_db, Dx *old_db)
{
  int i;
  Segment new_seg;
  SegmentMap *sm;
  Dx_rc rc;

  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	if ((sm = old_db->segment_map[i]) != NULL) {
	  while (sm != NULL) {
		rc = Segment_map(new_db, new_db->volume[sm->volume->volnum],
						 sm->offset, sm->length, sm->b,
						 sm->flags & SEGMENT_IS_PAGE, 0, &new_seg);
		if (rc != DX_OK)
		  return(rc);		/* XXX */

		Dx_memcpy(new_seg, sm->seg, sm->length);

		/* Don't unmap meta data, it's still needed. */
		if ((sm->flags & SEGMENT_META) == 0) {
		  rc = Segment_unmap(new_db, new_db->volume[sm->volume->volnum],
							 sm->offset, new_seg, sm->length, 1);
		  if (rc != DX_OK)
			return(rc);
		}

		sm = sm->next;
	  }
	}
  }

  return(DX_OK);
}

static void
segment_mapshow(SegmentMap *sm)
{
  char fbuf[20];

  if (sm->flags & SEGMENT_RW)
	strcpy(fbuf, "rw");
  else
	strcpy(fbuf, "ro");
  if (sm->flags & SEGMENT_FIXED)
	strcat(fbuf, ",f");
  if (sm->flags & SEGMENT_META)
	strcat(fbuf, ",M");
  if (sm->flags & SEGMENT_MODIFIED)
	strcat(fbuf, ",*");

  fprintf(stderr, "%4s %2ld %7llu %9llu %8lx  %7llu %5lu %s\n",
		  (sm->flags & SEGMENT_IS_PAGE) ? "P" : "S",
		  sm->volume->volnum,
		  (unsigned long long) sm->b,
		  (unsigned long long) sm->offset, (Dx_ui32) sm->seg,
		  (unsigned long long) sm->length,
		  sm->count, fbuf);
}

void
Segment_mapshow(Dx *db)
{
  int first, i;
  SegmentMap *sm;
  Dx_ui64 total_length;

  first = 1;
  total_length = 0;
  for (i = 0; i < SEGMENT_MAP_NENTRIES; i++) {
	sm = db->segment_map[i];

	while (sm != NULL) {
	  if (first) {
		fprintf(stderr,
				"BKT Type  V PageNum   Offset   SegAddr   Length     # Flags\n");
		first = 0;
	  }

	  fprintf(stderr, "%3d ", i);
	  segment_mapshow(sm);

	  total_length += sm->length;
	  sm = sm->next;
	}
  }

  fprintf(stderr, "nmapped=%lu, total bytes=%llu\n",
		  db->nmapped, total_length);
}

void
Segment_mapshow_active(Dx *db)
{
  int first, i;
  SegmentMap *sm;
  Dx_ui64 total_length;

  first = 1;
  total_length = 0;
  for (sm = db->active_list; sm != NULL; sm = sm->active_list_next) {
	if (first) {
	  fprintf(stderr,
			  "BKT Type  V PageNum   Offset   SegAddr   Length     # Flags\n");
	  first = 0;
	}

	fprintf(stderr, "    ");
	segment_mapshow(sm);

	total_length += sm->length;
  }

  fprintf(stderr, "nmapped=%lu, total bytes=%llu\n",
		  db->nmapped, total_length);
}

#ifdef NOTDEF
SegmentMap *
Segment_lookup_page(Dx *db, PageNum b)
{
  SegmentMap *sm;

  if (db->last_segment_lookup != NULL && (db->last_segment_lookup)->b == b) {
	(db->last_segment_lookup)->count++;
	return(db->last_segment_lookup);
  }

  sm = db->segment_map[b % SEGMENT_MAP_NENTRIES];
  while (sm != NULL && (sm->b != b || (sm->flags & SEGMENT_IS_PAGE) == 0))
	sm = sm->next;

  if (sm != NULL) {
	sm->count++;
	db->last_segment_lookup = sm;
  }

  return(sm);
}
#endif
